"""
This module handles the data files for MC and recorded data which shall
be processed for an analysis.

When executed directly, the modules performs a file scan for Belle files
creating file dictionaries with information about the location and number
of events of the file.

Felix Metzner 2018
"""

import os
import sys
import glob
import json
import copy
import click
import logging
import argparse
import subprocess
import numpy as np
import multiprocessing

from collections import OrderedDict
from functools import partial, lru_cache
from typing import Any, Union, Optional, Tuple, Dict, List, Generator, Sequence

from b2biiConversion import parse_process_url

from analysistools.utilities import PathType


__all__ = [
    "Belle_url_mc_types",
    "Belle_list_types",
    "Belle_data_types",
    "off_resonance_types",
    "get_file_set",
    "load_json_config",
    "store_json_config",
    "get_n_events_from_file_dict",
]


fmt_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
logger = logging.getLogger(__name__)
console = logging.StreamHandler()
console.setFormatter(logging.Formatter(fmt_str))
console.setLevel(logging.INFO)

fh = logging.FileHandler("file_handler_log.log")
fh.setFormatter(logging.Formatter(fmt_str))
fh.setLevel(logging.INFO)

logger.addHandler(console)
logger.addHandler(fh)


Belle_exp_list = [
    7,
    9,
    11,
    13,
    15,
    17,
    19,
    21,
    23,
    25,
    27,
    31,
    33,
    35,
    37,
    39,
    41,
    43,
    45,
    47,
    49,
    51,
    55,
    61,
    63,
    65,
]  # type: List[int]


Belle_url_mc_types = {
    "Charged": "evtgen-charged",
    "Mixed": "evtgen-mixed",
    "Charm": "evtgen-charm",
    "UDS": "evtgen-uds",
    "OffResonanceCharm": "evtgen-charm",
    "OffResonanceUDS": "evtgen-uds",
}  # type: Dict[str, str]

Belle_mc_streams = {
    "Charged": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    "Mixed": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
    "Charm": [0, 1, 2, 3, 4, 5],
    "UDS": [0, 1, 2, 3, 4, 5],
    "OffResonanceCharm": [0, 1, 2, 3, 4, 5],
    "OffResonanceUDS": [0, 1, 2, 3, 4, 5],
}  # type: Dict[str, List[int]]

Belle_list_types = {
    "ulnuCharged": "/group/belle/bdata-files/fileloc/caseB/specialmc2010/caseB-chargedulnu.list",
    "ulnuMixed": "/group/belle/bdata-files/fileloc/caseB/specialmc2010/caseB-mixedulnu.list",
    "rareCharged": "/group/belle/bdata-files/fileloc/caseB/specialmc2010/caseB-chargedrare.list",
    "rareMixed": "/group/belle/bdata-files/fileloc/caseB/specialmc2010/caseB-mixedrare.list",
    "jpsiCharged": "/group/belle/bdata-files/fileloc/caseB/specialmc2010/caseB-chargedjpsi.list",
    "jpsiMixed": "/group/belle/bdata-files/fileloc/caseB/specialmc2010/caseB-mixedjpsi.list",
    "psi2sCharged": "/group/belle/bdata-files/fileloc/caseB/specialmc2010/caseB-chargedpsi2s.list",
    "psi2sMixed": "/group/belle/bdata-files/fileloc/caseB/specialmc2010/caseB-mixedpsi2s.list",
}  # type: Dict[str, PathType]

Belle_data_types = {
    "DataY4S": "on_resonance",
    "DataContinuum": "continuum",
}  # type: Dict[str, str]

off_resonance_types = [
    "DataContinuum",
    "OffResonanceCharm",
    "OffResonanceUDS",
]  # type: List[str]


# TODO: Still to be considered are
#  * QED MC under /ghi/fs01/belle/bdata1/mcprod/...
#  * tautau continuum under /hsm/belle/bdata1/mcprod/tautau/...


def get_file_set(
    type_list: List[str],
    events_per_job: int,
    dict_directory: PathType,
    stream: Optional[int] = None,
    fraction: float = 1.0,
    path_to_signal_mc_config: Optional[PathType] = None,
    only_use_files_from_to: Tuple[Optional[int], Optional[int]] = (None, None),
) -> Generator[Dict[str, Any], None, None]:
    """
    Getter for file set based on event type and stream number.
    Files are put into chunks based on the number of events set via events_per_job.
    Returns lists of tuples containing the file location and the entry sequence for the file:
    [([paths to files of chunk], ["start:end" entry sequence for each file in chunk]), ...]
    For each type in type_list a separate list is returned.
    The requested types must be in one of the dictionaries Belle_url_mc_tyoes, Belle_list_types,
    Belle_data_types, ... TODO: Add other type dicts here...
    ... or in the provided signal_mc_config.
    :param type_list: List of types to get files for.
    :param events_per_job: Number of events per chunk.
    :param dict_directory: Base path to look for file dictionaries.
    :param stream: Stream to get for evtgen event type (in Belle_url_mc_types)
    :param fraction: Allows to only process a fraction of a given sample
    :param path_to_signal_mc_config: Path to json file containing a dictionary with the analysis specific channels.
    :param only_use_files_from_to: For testing: Only use this slice of the data files
        data_files[only_use_files_from_to[0]:only_use_files_from_to[1]]
    :return: Lists of Tuples of lists of strings for file locations and entry sequences.
    """
    if set(type_list) & set(Belle_data_types):
        if np.setdiff1d(type_list, list(Belle_data_types.keys())):
            raise ValueError("You are trying to process MC and recorded data in the same process.")

    if not os.path.exists(dict_directory):
        raise ValueError(f"Path to dictionary directory is not valid. \nPath was: {dict_directory}")

    if path_to_signal_mc_config:
        signal_mc_config = load_json_config(
            path_to_file=path_to_signal_mc_config,
            hint="Loading signal mc config in get_file_set",
        )  # type: Dict[str, Dict[str, PathType]]
        belle_signal_types = signal_mc_config["Belle_signal_types"]  # type: Dict[str, PathType]
    else:
        belle_signal_types = {}

    for event_type in type_list:
        if event_type in Belle_url_mc_types:
            prefix = "evtgen"
        elif event_type in Belle_list_types:
            prefix = "special"
        elif event_type in Belle_data_types:
            prefix = "data"
        elif event_type in belle_signal_types:
            prefix = "signal"
        else:
            raise KeyError(f"Encountered unknown event type {event_type}. Aborting...")

        # Building path to relevant file dictionary
        type_dir = os.path.join(dict_directory, f"{prefix}_{event_type}")
        if stream is None or event_type not in Belle_url_mc_types:
            dict_path = os.path.join(type_dir, f"{event_type}.json")
        elif stream in Belle_mc_streams[event_type]:
            dict_path = os.path.join(type_dir, f"{event_type}_stream{stream}", f"{event_type}_stream{stream}.json")
        else:
            raise ValueError(
                f"Trying to gather files for the invalid combination " f"of event type {event_type} and stream {stream}"
            )
        if not os.path.exists(dict_path):
            raise ValueError(f"The file dictionary which is supposed to be at\n{dict_path}\ndoes not exist...")

        # Loading file dictionary
        ordered_file_dict = read_file_dict(dict_path, sort=True)
        if not ordered_file_dict:
            logger.error("No data files. This will probably cause errors.")
        if only_use_files_from_to != (None, None):
            logger.warning(
                "Only using slice of data files! This is not a dataset you should use for science, "
                "only for quick first checks."
            )
            ordered_file_dict = OrderedDict(
                [
                    (k, ordered_file_dict[k])
                    for k in list(ordered_file_dict.keys())[only_use_files_from_to[0] : only_use_files_from_to[1]]
                ]
            )
            if not ordered_file_dict:
                logger.error("No data files left over after slicing. This will probably cause errors.")
        for file in ordered_file_dict:
            if not os.path.exists(file):
                raise FileNotFoundError(f"Input file {file} from dictionary {dict_path} for {event_type} does not exist.")
        file_chunks = slice_data_files(
            ordered_file_dict=ordered_file_dict,
            max_events=events_per_job,
            fraction=fraction,
        )

        for chunk_num, chunk in file_chunks.items():
            yield {"files": chunk["files"], "parts": chunk["parts"]}


def scan_for_belle_files(
    output_dir: PathType,
    processes: int,
    path_to_signal_mc_config: Optional[PathType] = None,
    perform_consistency_check: bool = False,
) -> None:
    """
    Scanning the known file locations to collect a dictionary comprising information on
    the respective files.
    Running this method a second time with perform_consistency_check=True allows to validate the produced file
    dictionaries to some extend. With this option turned on the already stored file dictionaries will be compared
    to newly produced ones. If inconsistencies are found they will be reported. The respective file dictionaries
    must be deleted to produce new ones in another execution of this function.
    :param output_dir: Where to put the resulting file dictionaries.
    :param processes: Number of processes to be used for multiprocessing.
    :param path_to_signal_mc_config: Path to json file containing a dictionary with the analysis specific channels.
    :param perform_consistency_check: Whether to perform a check of the consistency of the already available Belle file
                dictionaries. This will evaluate the files a second time and compare the result to the stored
                information. Default is False.
    :return: None
    """
    logger.info(
        f"Performing file scan for Belle files using {processes} processes.\n"
        f"The output is written to {output_dir}.\n"
        f"This can take a while..."
    )

    # Generic Monte Carlo
    all_mc_collection = {}  # type: Dict[PathType, Dict[str, Union[str, int, None]]]

    logger.info("Scanning Generic MC Files...")
    scan_belle_event_type(
        output_dir=output_dir,
        processes=processes,
        types_identifier="evtgen",
        types=Belle_url_mc_types,
        collection=all_mc_collection,
        check=perform_consistency_check,
    )

    # Special Monte Carlo
    logger.info("Scanning Special MC Files...")
    scan_belle_event_type(
        output_dir=output_dir,
        processes=processes,
        types_identifier="special",
        types=Belle_list_types,
        collection=all_mc_collection,
        check=perform_consistency_check,
    )

    # Signal Monte Carlo from config file
    if path_to_signal_mc_config:
        signal_config = load_json_config(path_to_signal_mc_config, hint="Loading signal mc in scan_for_belle_files")
        belle_signal_types = signal_config["Belle_signal_types"]  # type: Dict[str, PathType]

        logger.info("Scanning Private MC Files...")
        scan_belle_event_type(
            output_dir=output_dir,
            processes=processes,
            types_identifier="signal",
            types=belle_signal_types,
            collection=all_mc_collection,
            signal_types=belle_signal_types,
            check=perform_consistency_check,
        )

    # TODO: Handle missing types
    #           - QED MC?
    #           - tau tau continuum MC?

    store_file_dict(
        file_dict=all_mc_collection,
        path_to_file=os.path.join(output_dir, "all_mc_files.json"),
        overwrite=True,
    )

    # Data
    all_data_collection = {}  # type: Dict[PathType, Dict[str, Union[str, int, None]]]

    logger.info("Scanning Data Files...")
    scan_belle_event_type(
        output_dir=output_dir,
        processes=processes,
        types_identifier="data",
        types=Belle_data_types,
        collection=all_data_collection,
        check=perform_consistency_check,
    )

    store_file_dict(
        file_dict=all_data_collection,
        path_to_file=os.path.join(output_dir, "all_data_files.json"),
        overwrite=True,
    )

    logger.info(f"File dictionary creation completed. \nStored to: {output_dir}")


def scan_belle_event_type(
    output_dir: PathType,
    processes: int,
    types_identifier: str,
    types: Union[Dict[str, PathType], Dict[str, str]],
    collection: Optional[Dict[PathType, Dict[str, Union[str, int, None]]]] = None,
    signal_types: Optional[Dict[str, PathType]] = None,
    check: bool = False,
) -> None:
    """
    Convenience function to collect the file information for all Belle files of a
    specific data type.
    :param output_dir: Dictionary into which the output will be written.
    :param processes: Number of processes to use for event number evaluation via multi processing.
    :param types_identifier: Identifier string of the scanned event types. E.g. Signal, Data, Generic, Special...
    :param types: Belle event types to be scanned.
    :param collection: Dictionary into which all collected information will be written. Default is None.
    :param signal_types: Dictionary of signal types if private signal types shall be scanned. Default is None.
    :param check: Whether to perform a check of the consistency of the already available Belle file dictionaries.
                  This will evaluate the files a second time and compare the result to the stored information.
                  Default is False.
    :return: None
    """
    stream_collection = {}  # type: Dict[PathType, Dict[str, Union[str, int, None]]]

    for evt_type in types:
        type_collection = {}
        output_type_dir = os.path.join(output_dir, f"{types_identifier}_{evt_type}")
        os.makedirs(os.path.dirname(output_type_dir), exist_ok=True)

        type_streams = [None]  # type: Sequence[Optional[int]]
        stream_tag = ""  # type: str
        if evt_type in Belle_mc_streams.keys():
            type_streams = Belle_mc_streams[evt_type]

        for stream in type_streams:
            logger.info(f"Event Type: {evt_type}" + (f" (Stream {stream})" if stream is not None else ""))
            if stream is not None:
                stream_tag = f"_stream{stream}"
                stream_collection = {}
                out_target_dir = os.path.join(output_type_dir, f"{evt_type}_stream{stream}")
                os.makedirs(os.path.dirname(out_target_dir), exist_ok=True)
            else:
                out_target_dir = output_type_dir

            for exp in Belle_exp_list:
                stream_num = stream  # type: Optional[int]
                if stream is not None and exp < 30:
                    if evt_type.startswith("OffResonance"):
                        stream_num = stream + 20
                    else:
                        stream_num = stream + 10
                output_name = f"{evt_type}{stream_tag}_exp{exp}.json"
                file_dict = evaluate_belle_files(
                    target_file=os.path.join(out_target_dir, output_name),
                    processes=processes,
                    event_type=evt_type,
                    exp_number=exp,
                    stream_number=stream_num,
                    signal_types=signal_types,
                    check=check,
                )

                if stream is not None:
                    stream_collection.update(file_dict)
                type_collection.update(file_dict)
                if collection is not None:
                    collection.update(file_dict)
            if stream is not None:
                stream_output = os.path.join(out_target_dir, f"{evt_type}_stream{stream}.json")
                store_file_dict(
                    file_dict=stream_collection,
                    path_to_file=stream_output,
                    overwrite=True,
                )

            finished_tag = "." if stream is None else f" Stream {stream}"
            logger.info(f"Finished {types_identifier} {evt_type}{finished_tag}")
        type_output = os.path.join(output_type_dir, f"{evt_type}.json")
        store_file_dict(
            file_dict=type_collection,
            path_to_file=type_output,
            overwrite=True,
        )


def evaluate_belle_files(
    target_file: PathType,
    processes: int,
    event_type: str,
    exp_number: int,
    stream_number: Optional[int],
    signal_types: Optional[Dict[str, PathType]] = None,
    check: bool = False,
) -> Dict[PathType, Dict[str, Union[str, int, None]]]:
    """
    Convenience function to read out number of events in Belle files or from an already existing
    file dictionary for a given type/experiment/stream combination.
    :param target_file: Path to file into which the file dictionary shall be stored.
    :param processes: Number of processes to use for event number evaluation via multi processing.
    :param event_type: Event type to evaluate.
    :param exp_number: Experiment number to evaluate.
    :param stream_number: Stream number to evaluate.
    :param signal_types: Dictionary of signal types if private signal types shall be evaluated. Default is None.
    :param check: Whether to perform consistency check if file is already present. Default is False.
    :return: Dictionary containing information about the evaluated Belle files.
    """
    if os.path.exists(target_file):
        logger.warning(f"Dictionary for file {target_file} already exists. {'Checking' if check else 'Skipping'} it...")
        file_dict = read_file_dict(target_file, sort=False)

        if check:
            file_list = get_belle_files(
                event_type=event_type,
                exp_number=exp_number,
                stream_number=stream_number,
                belle_signal_types=signal_types,
            )

            file_dict_compare = create_file_dict(file_list, processes, use_b2bii=True)
            problematic_files = compare_dicts(file_dict, file_dict_compare)
            logger.info(
                f"Check result: \nFound {len(problematic_files)} inconsistent files in the "
                f"checked dictionary json files."
            )
            if len(problematic_files) > 0:
                logger.warning("Please check the dictionaries containing the following files:")
                for p_file in problematic_files:
                    print(p_file)

        return file_dict
    else:
        file_list = get_belle_files(
            event_type=event_type,
            exp_number=exp_number,
            stream_number=stream_number,
            belle_signal_types=signal_types,
        )

        if len(file_list) == 0:
            logger.warning(f"WARNING: No file for {event_type} experiment {exp_number}!")
            return dict()

        meta_data = {"ext_num": exp_number, "stream": stream_number}  # type: Dict[str, Union[int, str, None]]
        file_dict = create_file_dict(
            file_list=file_list,
            processes=processes,
            use_b2bii=True,
            additional_vals=meta_data,
        )
        store_file_dict(
            file_dict=file_dict,
            path_to_file=target_file,
            overwrite=False,
        )

        return file_dict


def get_belle_files(
    event_type: str,
    exp_number: Optional[int] = None,
    stream_number: Optional[int] = None,
    belle_signal_types: Optional[Dict[str, PathType]] = None,
) -> List[PathType]:
    """
    Wrapper for functions to retrieve belle files of different event types.
    :param event_type: Event type for which the files shall be gathered
    :param exp_number: Experiment number to evaluate.
    :param stream_number: Stream number to evaluate.
    :param belle_signal_types: Dictionary of signal types if private signal types shall be evaluated. Default is None.
    :return: List of files.
    """
    if event_type in Belle_url_mc_types or event_type in Belle_data_types:
        return get_belle_files_via_url(event_type=event_type, exp_number=exp_number, stream_number=stream_number)
    elif event_type in Belle_list_types:
        return get_belle_files_from_list(event_type=event_type, exp_number=exp_number)
    elif belle_signal_types is not None and event_type in belle_signal_types:
        return list(
            get_belle_private_mc(
                event_type=event_type,
                exp_number=exp_number,
                belle_signal_types=belle_signal_types,
            )
        )
    else:
        raise ValueError(f"No method available to retrieve information about Belle event type {event_type}")


def get_belle_files_via_url(
    event_type: str,
    exp_number: Optional[int] = None,
    stream_number: Optional[int] = None,
) -> List[PathType]:
    """
    Gets list of Belle files via url request.
    Possible for Charged, Mixed, Charm and UDS MC as well as data.
    You can request the data for certain experiment or stream numbers.
    If not further specified, all possible values will be considered.
    :param event_type: Event type for which the files shall be gathered
    :param exp_number: Specifies the experiment to get. Default: all
    :param stream_number: Specifies the stream to get (only on MC). Default: all
    :return: List of files.
    """
    if stream_number is None and event_type not in Belle_data_types:
        _stream_number = "Any"  # type: str
    else:
        _stream_number = str(stream_number)

    experiments = []  # type: List[int]
    if exp_number is not None:
        if exp_number not in Belle_exp_list:
            sys.exit(f"Requested experiment number {exp_number} is not valid.")
        experiments.append(exp_number)
    else:
        experiments.extend(Belle_exp_list)

    if event_type in Belle_data_types and stream_number is None:
        file_list = []
        for exp in experiments:
            url = "http://bweb3.cc.kek.jp/mdst.php?ex={exp}&rs=1&re=9999&skm=HadronBorJ&dt={type}&bl=caseB".format(
                exp=exp,
                type=Belle_data_types[event_type],
            )
            file_list.extend(parse_process_url(url))
        return file_list
    elif event_type in Belle_url_mc_types and stream_number is not None:
        file_list = []
        for exp in experiments:
            url = "http://bweb3/montecarlo.php?ex={exp}&rs=1&re=9999&ty={ty}&dt={res}&bl=caseB&st={s}".format(
                exp=exp,
                ty=Belle_url_mc_types[event_type],
                res="on_resonance" if not event_type.startswith("OffResonance") else "continuum",
                s=stream_number,
            )
            file_list.extend(parse_process_url(url))
        return file_list
    else:
        sys.exit(
            f"The file request you made with {event_type}, exp = {exp_number}, stream = {_stream_number} is not valid!"
        )


def get_belle_files_from_list(
    event_type: str,
    exp_number: Optional[int] = None,
) -> List[PathType]:
    """
    Returns list of files for requested event type and experiment number.
    Valid for MC event types in Belle_list_types.
    If no experiment number is provided, all files are gathered.
    A distinction based on stream numbers is not possible for special MC.
    :param event_type: Get files for this event type
    :param exp_number: Specifies experiment number to get. Default: all
    :return: List of files.
    """
    path_to_file = Belle_list_types[event_type]
    if not os.path.isfile(path_to_file):
        raise ValueError(f"The file you are looking for does not exist:\n{path_to_file}")
    with open(path_to_file, "r") as in_file:
        line_list = in_file.readlines()
    file_list = [line.strip().split()[1] for line in line_list]  # type: List[PathType]
    if exp_number is not None:
        if exp_number not in Belle_exp_list:
            raise ValueError(f"Requested experiment number {exp_number} is not valid.")
        file_list = list(filter(lambda x: f".e{exp_number}" in str(x), file_list))
    return file_list


def get_belle_private_mc(
    event_type: str,
    belle_signal_types: Dict[str, PathType],
    exp_number: Optional[int] = None,
) -> Sequence[PathType]:
    """
    Gather files from belle private mc directory.
    If experiment number is specified, only files matching this requirement are gathered.
    :param event_type: Event type for which the files shall be collected. Must be in belle_signal_types.
    :param exp_number: Specifies experiment number. Default: all.
    :param belle_signal_types: Dictionary containing the analysis specific decay channels.
    :return: List of files.
    """
    try:
        path_to_files = belle_signal_types[event_type]  # type: PathType
    except KeyError:
        raise ValueError(f"The event type {event_type} is invalid, as it is not available in belle_signal_types!")

    if not os.path.isdir(path_to_files):
        raise ValueError(
            f"The directory from which you requested private mc files does not exist!\n" f"Path: {path_to_files}"
        )
    if exp_number is not None:
        if exp_number not in Belle_exp_list:
            raise ValueError(f"Requested experiment number {exp_number} is not valid.")
        return glob.glob(os.path.join(path_to_files, f"evtgen_exp_{exp_number:02d}*.mdst"))

    else:
        return glob.glob(os.path.join(path_to_files, "evtgen_exp_*.mdst"))


def create_file_dict(
    file_list: List[PathType],
    processes: int,
    use_b2bii: bool,
    additional_vals: Optional[Dict[str, Union[str, int, None]]] = None,
) -> Dict[PathType, Dict[str, Union[str, int, None]]]:
    """
    Loop over all files provided via a list and check for the number of events
    they contain.
    A dictionary mapping the files path to the number of events is returned.
    As the check of the number of events takes some time, this task is done
    using multiprocessing. The number of processes used can be specified.
    Optional additional dictionary entries such es experiment or stream number
    can be added via the parameter additional_vals.
    E.g.: additional_vals={"exp_num": 55, "stream": 2}
    :param file_list: List of file paths to consider
    :param processes: Number of processes to use for this task
    :param additional_vals: Optional additional dictionary entries.
    :param use_b2bii: Whether converted Belle data or Belle II data is used.
    :return: dict {str "path/to/file": int(number of events)}
    """
    file_dict = {}  # type: Dict[PathType, Dict[str, Union[str, int, None]]]

    with multiprocessing.Pool(processes=processes) as process_pool:
        for file, events in process_pool.map(partial(count_events_in_file, use_b2bii=use_b2bii), file_list):
            file_dict[file] = {"nevents": events}
            if additional_vals is not None:
                file_dict[file].update(additional_vals)

    return file_dict


def compare_dicts(
    file_dict: Dict[PathType, Dict[str, Union[str, int, None]]],
    file_dict_compare: Dict[PathType, Dict[str, Union[str, int, None]]],
) -> List[PathType]:
    """
    Compare two file dictionaries to detect differences in the number of events stored for each file.
    :param file_dict: Original file dictionary.
    :param file_dict_compare: File dictionary used to compare to original file dictionary.
    :return: List of problematic files.
    """
    problematic_files = []  # type: List[PathType]
    for file, file_info in file_dict.items():
        try:
            comparison_file_info = file_dict_compare[file]
        except KeyError:
            logger.warning(
                f"WARNING: The file\n"
                f"{file}\n"
                f"from the original file dictionary was not found in the second file dictionary."
            )
            problematic_files.append(file)
            continue

        if file_info["nevents"] != comparison_file_info["nevents"]:
            logger.warning(
                f"For the file\n"
                f"{file}\n"
                f"different event numbers have been found:\n"
                f"Original file dictionary: {file_info['nevents']}\n"
                f"Comparison file dictionary: {comparison_file_info['nevents']}."
            )
            problematic_files.append(file)

    for file in file_dict_compare.keys():
        if file in file_dict.keys():
            continue
        else:
            logger.warning(
                f"The file\n{file}\nwas found in the comparison file dictionary, " f"but not in the original one."
            )
            problematic_files.append(file)

    return problematic_files


def store_file_dict(
    file_dict: Dict[PathType, Dict[str, Union[str, int, None]]],
    path_to_file: PathType,
    overwrite: bool = False,
) -> None:
    """
    Stores the provided file dictionary into a json file.
    :param file_dict: Dictionary containing information for a set of files.
    :param path_to_file: Target location for the generated json file.
    :param overwrite: Whether to allow to overwrite an existing file.
    :return: None
    """
    if os.path.isfile(path_to_file) and not overwrite:
        raise ValueError(
            f"The data dictionary file you specified to write the dictionary to, "
            f"which was \n {path_to_file} \n already exists and will not be overwritten.\n"
            f"You can either choose another path to write to, or allow to overwrite."
        )
    else:
        os.makedirs(os.path.dirname(path_to_file), exist_ok=True)
        with open(path_to_file, "w") as out_file:
            json.dump(file_dict, out_file, indent=2)


def read_file_dict(
    path_to_file: PathType,
    sort: bool = True,
) -> Dict[PathType, Dict[str, Union[str, int, None]]]:
    """
    Reads a json file containing a file dictionary to return (a sorted version of) the dict.
    The sorting is done based on the key of the dictionary.
    :param path_to_file: Location of the file to read.
    :param sort: Whether to return a sorted OrderedDict version of the read dict.
    :return: Sorted Ordered dictionary containing the information about the files.
        {path to file: {nevents: .., ext_num: .., stream: .., ... }}
    """
    if not os.path.isfile(path_to_file):
        raise ValueError(f"The file you are looking for does not exist:\n{path_to_file}")
    with open(path_to_file, "r") as in_file:
        file_dict = json.load(in_file)

    to_drop = list()
    for key, value in file_dict.items():
        if value["nevents"] < 0:
            to_drop.append(key)
            logger.warning(f"Dropping mdst file\n{key}\nfrom file dict\n{path_to_file}.\n" f"Dropped entry:\n{value}")

    for key in to_drop:
        file_dict.pop(key)

    if sort:
        ordered_file_dict = OrderedDict(sorted(file_dict.items()))
        return ordered_file_dict
    else:
        return file_dict


def make_sorted_dict(input_dict: Dict[str, Any]) -> Dict[str, Any]:
    """
    Returns a sorted OrderedDict version of the input dictionary.
    :param input_dict: Input dictionary
    :return: Sorted OrderedDict version of input
    """
    return OrderedDict([(k, input_dict[k]) for k in sorted(input_dict.keys())])


def count_events_in_file(
    file_path: PathType,
    use_b2bii: bool,
) -> Tuple[PathType, int]:
    """
    Checks first if file specified by file_path is present and
    if so, will count the number of events in the file using either
    numberOfEventsInBelle2Mdst of numberOfEventsInBelleMdst.
    :param file_path: path to file of interest
    :param use_b2bii: Whether converted Belle data or Belle II data is used.
    :return: (str, int) tuple of file_path and int number of events in the file
    """
    if not os.path.isfile(file_path):
        raise ValueError(f"The file you are looking for does not exist:\n{file_path}")

    if use_b2bii:
        command = "numberOfEventsInBelleMdst"
    else:
        command = "numberOfEventsInBelle2Mdst"

    try:
        events = int(subprocess.check_output([command, file_path]))
    except ValueError:
        logger.warning(f"The file: {file_path}\ndid not yield a parsable event number!")
        events = -999

    return file_path, events


def slice_data_files(
    ordered_file_dict: Dict[PathType, Dict[str, Union[str, int, None]]],
    max_events: int,
    fraction: float = 1.0,
    wiggle: float = 0.0,
) -> Dict[int, Dict[str, Union[List[PathType], List[str]]]]:
    """
    Combine files to chunks with specified number of events.
    If files are too large, be it as last element of a chunk or by
    themselves they will be split and put into multiple chunks.
    The created OrderedDict has the form
    { slice_number: {"files": [list of files], "parts": [list of parts]}}
    e.g.:
    {
        1: {"files": [file1, file2], "parts": ["0:1000", "0:500"]}
        2: {"files": [file2], "parts": ["501:1000"]}
    }
    :param ordered_file_dict: Ordered dictionaries containing info about the files which will be sliced.
    :param max_events: Maximal number of events in a chunk.
    :param fraction: set x to only get the x-th fraction of events of a stream to be processed. The events are still
           equally distributed over all experiments, etc.
    :param wiggle: Tolerance for lager or smaller chunk sizes in percentage of the max_events. Must be in [0., 0.1]!
    :return: Dictionary containing the chunks specifying the files and their respective entry sequence.
    """

    if wiggle > 0.1 or wiggle < 0.0:
        raise ValueError(
            f"The chosen value for the wiggle room of the slice sizes is not valid\n"
            f"Should be between 0. and 0.1; was {wiggle}"
        )
    if not 1.0 >= fraction > 0.0:
        raise ValueError(
            f"The chosen fraction-size is greater than 1 or zero or negative which is invalid.\n"
            f"0.5 means, half the stream is computed. "
            f"So the value should be between 0 and 1, but was {fraction}"
        )

    sliced_dict = OrderedDict({})  # type: Dict[int, Dict[str, Union[List[PathType], List[str]]]]
    current_events = 0  # type: int
    my_slice = 1  # type: int
    max_scaled = int(max_events / fraction)
    for file, file_info in ordered_file_dict.items():
        _n_events_in_file = copy.copy(file_info["nevents"])
        assert isinstance(_n_events_in_file, int), type(_n_events_in_file).__name__
        n_events_in_file = _n_events_in_file  # type: int

        remaining_events = copy.copy(n_events_in_file)  # type: int

        file_split = {"slices": [], "parts": []}  # type: Dict[str, List[Union[int, str]]]
        start = 0  # type: int
        while remaining_events > 0:
            if remaining_events + current_events <= max_scaled * (1.0 + wiggle):
                end = start + int((n_events_in_file - start) * fraction) - 1
                file_split["slices"].append(my_slice)
                file_split["parts"].append(f"{start}:{int(end)}")
                current_events += remaining_events
                remaining_events = 0
            else:  # remaining_events + current_events > max_scaled :
                end = start + max_events - int(current_events * fraction) - 1
                file_split["slices"].append(my_slice)
                file_split["parts"].append(f"{start}:{int(end)}")
                start = int(start + max_scaled - current_events)
                remaining_events -= max_scaled - current_events
                my_slice += 1
                current_events = 0
        # Add new file parts to sliced_dict:
        for s, p in zip(file_split["slices"], file_split["parts"]):
            assert isinstance(s, int), type(s).__name__
            if s in sliced_dict:
                sliced_dict[s]["files"].append(file)  # type: ignore
                sliced_dict[s]["parts"].append(p)  # type: ignore
            else:
                sliced_dict[s] = {"files": [file], "parts": [p]}  # type: ignore

        # Create new slice if current is almost full:
        if current_events > (1.0 - wiggle) * max_scaled:
            my_slice += 1
            current_events = 0

    return sliced_dict


def valid_file_extension(
    file_list: List[PathType],
    use_b2bii: bool,
) -> List[PathType]:
    """
    Returns subset of given file list, in which all files have a valid extension.
    :param file_list: list of file paths to be checked for valid extension.
    :param use_b2bii: Whether converted Belle data or Belle II data is used.
    :return: filtered file list
    """
    if use_b2bii:
        identifier = ".mdst"
    else:
        raise NotImplementedError("No Belle II case implemented, yet.")

    return list(filter(lambda x: str(x).endswith(identifier), file_list))


@lru_cache()
def _load_json_config(
    path_to_file: PathType,
) -> Dict[str, Any]:
    with open(path_to_file, "r") as in_file:
        config = json.load(in_file)
    return config


def load_json_config(
    path_to_file: PathType,
    hint: Optional[str] = None,
) -> Dict[str, Any]:
    """
    Convenience function to load json config file.
    :param path_to_file: Path to the json config file.
    :param hint: String which can be set as hint for where reading the configuration file from
                 the provided path fails, if it fails.
    :return: config dictionary
    """
    if not os.path.exists(path_to_file):
        raise ValueError(f"{hint + ': ' if hint else ''}The provided config file path {path_to_file} does not exits!")
    with open(path_to_file, "r") as in_file:
        config = json.load(in_file)

    return config


def store_json_config(
    config: Dict[str, Any],
    path_to_file: PathType,
    overwrite: bool = False,
) -> None:
    """
    Stores the provided config into a json file.
    :param config: Dictionary containing configuration information.
    :param path_to_file: Target location for the generated json file.
    :param overwrite: Whether to allow to overwrite an existing file.
    :return: None
    """
    if os.path.isfile(path_to_file) and not overwrite:
        raise ValueError(
            f"The data dictionary file you specified to write the dictionary to, "
            f"which was \n {path_to_file} \n already exists and will not be overwritten.\n"
            f"You can either choose another path to write to, or allow to overwrite."
        )

    os.makedirs(os.path.dirname(path_to_file), exist_ok=True)
    with open(path_to_file, "w") as out_file:
        json.dump(config, out_file, indent=2)


def get_commandline_args() -> Tuple[int, PathType, PathType, bool]:
    """
    Convenience function to get commandline arguments for
        - number of processes
        - and the configuration file for the analysis for the target directory for the dictionaries and
          the list of analysis specific decay channels.
        - whether to perform a consistency check on the already present file dictionaries by reevaluating them in
          a second run of the function scan_for_belle_files.
    :return: number_of_processes, target_directory, signal_specific_decays_config, perform_check_flag
    """
    parser = argparse.ArgumentParser(
        description="Usage: python3 file_handler.py "
        "-- --config <path/to/config>  [--processes <number_of_processes>] [--check]\n"
        "File handler scanning for Belle files required for the analysis. "
        "The json config file of the analysis must be provided via the commandline parameter --config.\n"
        "The scan is done using multiple processes. "
        "The maximum number of processes can be set via the option --processes.\n"
        "In a second execution of this method a consistency check of the "
        "already stored file dictionaries can be performed by setting the flag --check. "
        "This will again retrieve the information by evaluating all files."
    )

    parser.add_argument(
        "--config",
        type=str,
        required=True,
        help="If set MC info is used.",
    )

    parser.add_argument(
        "--processes",
        type=int,
        required=False,
        default=20,
        help="Number of processes to use.",
    )

    parser.add_argument(
        "--check",
        action="store_true",
        help="Perform consistency check on already stored file dictionaries in a second execution of this command. ",
    )

    args = vars(parser.parse_args())
    config_file = args["config"]
    processes = args["processes"]
    perform_check_flag = args["check"]

    config = load_json_config(config_file, hint="Loading json config in get_commandline_args")

    logger.info(f"Running Belle File Scan with {processes} processes.")
    logger.info(f"Loading analysis specific decay channels from configuration file:\n{config['signal_MC_config']}")
    logger.info(f"The file dictionaries will be written to:\n{config['file_dict_location']}")

    return processes, config["file_dict_location"], config["signal_MC_config"], perform_check_flag


def get_n_events_from_file_dict(file_dict_path: PathType) -> int:
    if not os.path.isfile(file_dict_path):
        raise RuntimeWarning(f"Path to file_dict is not valid:\n\t{file_dict_path}")

    with open(file_dict_path) as infile:
        file_dict_content = json.load(infile)

    total = 0  # type: int
    for name, fd in file_dict_content.items():
        n_events = fd["nevents"]  # type: int
        assert isinstance(n_events, int), (n_events, type(n_events).__name__)
        if n_events < 0:
            logging.warning(f"Mdst file {name} is skipped since its could not pre processed.")
            continue
        else:
            total += n_events

    return total


def count_events_in_file_dict(file_dict_path: PathType) -> None:
    n_events = get_n_events_from_file_dict(file_dict_path=file_dict_path)  # type: int
    event_type_name = os.path.splitext(os.path.basename(file_dict_path))[0]  # type: str
    print(f"{event_type_name}: Total number of events: {n_events}")


@click.command()
@click.option(
    "--file_dict",
    type=click.Path(),
    help="Path to file dictionary.",
)
def nevents_in_filedict(file_dict: PathType) -> None:
    """Counts the total number of events in the given file dict."""
    count_events_in_file_dict(file_dict_path=file_dict)


@click.command()
@click.option(
    "--path_to_file_dict_dir",
    type=click.Path(),
    help="Path to the directory with the file dictionaries.",
)
def nevents_in_all_signal_filedicts(path_to_file_dict_dir: PathType) -> None:
    if not os.path.exists(path_to_file_dict_dir):
        raise RuntimeError(f"The provided directory path\n\t{path_to_file_dict_dir}\ndoes not exist!")
    logging.info(
        f"Processing all main signal file dictionaries in the sub-directories of the directory\n\t"
        f"{path_to_file_dict_dir}"
    )

    prefix = "signal_"
    for _, dirs, _ in os.walk(top=path_to_file_dict_dir):
        for subdir_name in dirs:
            if subdir_name.startswith(prefix):
                subdir = os.path.join(path_to_file_dict_dir, subdir_name)
                event_type_name = subdir_name.split(prefix, 1)[1]
                file_dict_path = os.path.join(subdir, f"{event_type_name}.json")
                count_events_in_file_dict(file_dict_path=file_dict_path)
        break


if __name__ == "__main__":
    num_processes, target_dir, specific_config, perform_check = get_commandline_args()

    scan_for_belle_files(
        output_dir=target_dir,
        processes=num_processes,
        path_to_signal_mc_config=specific_config,
        perform_consistency_check=perform_check,
    )
