"""
Job submission tools.

Contains example of how to submit jobs to LSF batch system in the main function.

Furthermore useful tasks for the sampling of the run times of an analysis' steering
scripts for different event types.

Felix Metzner 2018
"""

import os
import time
import random
import resource
import subprocess
import b2luigi as luigi

from typing import Union, List, Dict, Generator

import analysistools.batch_tools.tasks as analysis_tasks

from analysistools.utilities import PathType
from analysistools.batch_tools import file_handler
from analysistools.batch_tools.file_handler import get_file_set, load_json_config, store_json_config


KEK_LSF_queue_limits = {
    "l": 24 * 60 * 60,
    "s": 3 * 60 * 60,
}


class RunTimeSamplingMaster(luigi.WrapperTask):
    batch_config_file = luigi.Parameter(positional=False)  # type: str
    sample_chunks = luigi.IntParameter(default=10, positional=False)  # type: int
    chunk_size = luigi.IntParameter(default=5000, positional=False)  # type: int

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.batch_config = load_json_config(
            path_to_file=self.batch_config_file,
            hint="Loading batch config for run time sampling.",
        )

        self.global_config = load_json_config(
            path_to_file=self.batch_config["global_config"],
            hint="Loading global config for run time sampling.",
        )

        self.work_dir = os.path.join(self.global_config["target_directory"], "run_time_sampling")
        if not os.path.exists(self.work_dir):
            os.makedirs(self.work_dir)

        self.mc_config = self.batch_config.copy()
        self.mc_config.update(self.global_config)
        self.mc_config["batch_config"] = str(os.path.join(self.work_dir, "mc_config.json"))
        store_json_config(self.mc_config, os.path.join(self.work_dir, "mc_config.json"), True)

        self.data_config = self.mc_config.copy()
        self.data_config["use_mc_info"] = False
        self.data_config["batch_config"] = str(os.path.join(self.work_dir, "data_config.json"))
        store_json_config(self.data_config, os.path.join(self.work_dir, "data_config.json"), True)

    def get_task_file_dir(self) -> PathType:
        os.makedirs(self.work_dir, exist_ok=True)
        return os.path.join(self.work_dir, "{task_name}_".format(task_name=self.get_task_family()))

    def get_log_file_dir(self) -> PathType:
        return self.get_task_file_dir()

    def output(self) -> luigi.LocalTarget:
        return luigi.LocalTarget(os.path.join(self.work_dir, "run_time_sampling_result.json"))

    def requires(self) -> Generator[luigi.Task, None, None]:
        file_dict_dir = self.global_config["file_dict_location"]
        signal_mc_config = self.global_config["signal_MC_config"]
        signal_dict = load_json_config(signal_mc_config, hint="Loading Signal MC config for run time sampling")

        if not os.path.exists(file_dict_dir):
            raise ValueError(
                f"The provided file dictionary directory path is not valid!\n" f"The provided path is {file_dict_dir}"
            )

        # Determine event types for which the sampling will be performed.
        generic_types = list(file_handler.Belle_url_mc_types.keys())
        data_types = list(file_handler.Belle_data_types.keys())

        evt_types = generic_types
        evt_types.extend(list(file_handler.Belle_list_types.keys()))
        evt_types.extend(data_types)
        evt_types.extend(list(signal_dict["Belle_signal_types"].keys()))

        for evt_type in evt_types:
            # Select a sample of 10 chunks a 10000 events at random for the run time
            # sampling of each event type. If a event type provides less than 10
            # samples all available ones are used...
            file_chunks = get_file_set(
                type_list=[evt_type],
                events_per_job=self.chunk_size,
                dict_directory=file_dict_dir,
                stream=0 if evt_type in generic_types else None,
                path_to_signal_mc_config=signal_mc_config,
            )

            chunk_list = list(file_chunks)
            number_of_chunks = self.sample_chunks if len(chunk_list) >= self.sample_chunks else len(chunk_list)
            random.seed(8847)
            random_chunks = random.sample(chunk_list, number_of_chunks)
            sampling_chunk = {"files": [], "parts": []}  # type: Dict[str, List[Union[PathType, str]]]
            for chunk in random_chunks:
                sampling_chunk["files"].extend(chunk["files"])
                sampling_chunk["parts"].extend(chunk["parts"])

            yield RunTimeSamplingTask(
                file_chunk=sampling_chunk,
                stream=0 if evt_type in generic_types else None,
                evt_type=evt_type,
                config=self.mc_config if evt_type not in data_types else self.data_config,
                job_dir=os.path.join(self.work_dir, evt_type),
                log_file=os.path.join(self.work_dir, evt_type, f"{evt_type}_runtime_sampling.log"),
                error_file=os.path.join(self.work_dir, evt_type, f"{evt_type}_runtime_sampling.err"),
                stats_file=os.path.join(self.work_dir, evt_type, f"{evt_type}_runtime_sampling.json"),
            )

    def run(self) -> None:
        result_dict = {}
        files = [f["stats_file"].path for f in self.input()]
        for file in files:
            result = load_json_config(file, hint="Loading timing result for result collection")
            result_dict.update({result["event_type"]: result})

        store_json_config(result_dict, self.output().path, True)


class RunTimeSamplingTask(luigi.Task):
    file_chunk = luigi.DictParameter(positional=False)  # type: dict
    evt_type = luigi.Parameter(positional=False)  # type: str
    stream = luigi.IntParameter(positional=False)  # type: int
    config = luigi.DictParameter(positional=False)  # type: dict

    job_dir = luigi.Parameter(positional=False)  # type: str
    log_file = luigi.Parameter(positional=False)  # type: str
    error_file = luigi.Parameter(positional=False)  # type: str
    stats_file = luigi.Parameter(positional=False)  # type: str

    def get_task_file_dir(self) -> PathType:
        os.makedirs(self.job_dir, exist_ok=True)
        identifier = "{task_name}_{evt_type}".format(task_name=self.get_task_family(), evt_type=self.evt_type)
        return os.path.join(self.job_dir, f"{identifier}_")

    def get_log_file_dir(self) -> PathType:
        return self.get_task_file_dir()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()
        for output_name in self.config["output"]:
            ntuple = os.path.join(self.job_dir, f"chunk_{self.evt_type}_{output_name}.root")
            output_dict.update({output_name: luigi.LocalTarget(ntuple)})
        output_dict.update({"stats_file": luigi.LocalTarget(self.stats_file)})
        return output_dict

    def run(self) -> None:
        # Prepare target directory
        os.makedirs(self.job_dir, exist_ok=True)

        analysis_tasks.document_conditions(self.job_dir, self.config)

        # Prepare input for basf2 execution command
        steering = self.config["steering_file"]
        batch_config = self.config["batch_config"]
        file_list = self.file_chunk["files"]
        sequence_list = self.file_chunk["parts"]

        n_events = 0
        for sequence in sequence_list:
            n_events += int(sequence.split(":")[1]) - int(sequence.split(":")[0]) + 1

        output_dict_list = []
        for output_key, output_path in self.output().items():
            output_dict_list.append(output_key)
            output_dict_list.append(output_path.path)

        # Prepare basf2 command
        cmd = ["basf2", steering, "--", "--config", batch_config, "--input_files"]
        cmd.extend(file_list)
        cmd.append("--input_sequences")
        cmd.extend(sequence_list)
        cmd.append("--output_dict")
        cmd.extend(output_dict_list)

        start_time = time.time()

        env = os.environ.copy()
        with open(self.log_file, "w") as log_file_target, open(self.error_file, "w") as error_file_target:
            subprocess.check_call(cmd, env=env, stdout=log_file_target, stderr=error_file_target)

        run_time = time.time() - start_time
        user_time = resource.getrusage(resource.RUSAGE_CHILDREN).ru_utime
        sys_time = resource.getrusage(resource.RUSAGE_CHILDREN).ru_stime

        # Saving results into json file. Events are calculated using the run_time as
        # this time measurements reflects the values in the LSF overview of a job closest.
        stats_dict = {
            "event_type": self.evt_type,
            "events": n_events,
            "run_time": run_time,
            "user_time": user_time,
            "sys_time": sys_time,
            "events_on_l_queue": int(n_events / run_time * KEK_LSF_queue_limits["l"]),
            "events_on_s_queue": int(n_events / run_time * KEK_LSF_queue_limits["s"]),
        }

        store_json_config(stats_dict, self.output()["stats_file"].path, True)


if __name__ == "__main__":
    num_workers = 20  # type: int
    config = "/home/belle2/metzner/RDStar/rdstar/configs/batch_config.json"  # type: PathType
    print(f"Starting to process events with\n\t\tworkers = {num_workers}\n\t\tconfiguration file = {config}")
    luigi.process(
        analysis_tasks.Master(batch_config_file=config),
        workers=num_workers,
    )
