"""
Task set for processing Belle data sets on large scale at KEKcc with luigi.

The following tasks make up an acyclic dependency graph which will process
a set of event types specified by user provided configuration files with a
provided basf2 steering file.
The configuration files necessary are described in the Master task's documentation.

The log output of the jobs will be checked for errors before the n-tuple files
will be merged for each event type.

Felix Metzner 2018
"""

import os
import sys
import git
import shutil
import inspect
import tarfile
import subprocess
import b2luigi as luigi

from pathlib import Path
from typing import Optional, Tuple, Dict, Any, Generator

from analysistools.utilities import PathType
from analysistools.batch_tools.file_handler import get_file_set, load_json_config


__all__ = [
    "Master",
    "FileDictArchiveTask",
]


class Master(luigi.WrapperTask):
    """
    Receives the batch-specific config file containing instructions on how to process the data types.
    Furthermore, it loads the global config file, linked in the batch-specific config, which contains
    information about database locations and general parameters that will probably be the same for a
    series of executions of this program.
    Both json config files are explained below.

    In the end the required ([files], [entry-sequences], config) tuples are generated to start tasks to
    build n-tuples for requested files.

    ----------------------------------
    Batch Config File:
    The batch config file contains the settings used for a set of batch jobs. This allows to handle
    different data types differently, as e.g. continuum events will be processed faster than MC files
    containing only signal events. Furthermore, the settings for data and MC should also be different and
    hence different batch config files allow for this.
    Thus, one specifies the data types to be processed, as well as the number of events per chunk and
    the steering file to be used along with some other relevant settings in the batch config file.
    The general settings relevant for every batch job are contained in the global config the path to
    which is also specified in the batch config.

    The format of the batch config is as follows:

    {
      "global_config": "/home/belle2/metzner/RDStar/rdstar/configs/global_config.json",
      "steering_file": "/home/belle2/metzner/RDStar/rdstar/processing/reconstruction.py",
      "events_per_chunk": 2000,
      "queue": "l",
      "tag": null,
      "data": {
        "Test": {
          "streams": null,
          "fraction": 1.0,
          "only_use_files_from_to": [null, null]
        }
      },
      "output": [
        "Y4S_ntuple"
      ],
      "use_mc_info": true
      "roe_mask": "all"
    }

    Here the entries have the following meaning:

    global_config: Path to the global config. Its purpose and structure is explained below.
    steering_file: The path to the basf2 steering file to be used to process the data.
    events_per_chunk: The maximal number of events per chunk which is processed in one job.
    queue: LSF queue the jobs will be submitted to. Most likely 'l' for long or 's' for short.
    tag: Additional optional identifier for this batch of jobs which will be used in the name of
         the sub-directory produced for this batch in the global target directory.
    data: A dictionary containing all the data types and the streams to be processed.
          This dictionary has the keys:
          evt_type: Event type to be processed. Must be one of the ones defined in
                    file_handler.py or one defined in the analysis specific signal_config json file:
                    E.g.: Charged, Mixed, Charm, UDS, ulnuCharged, ulnuMixed, rareCharged,... ,
                    one of the private signal types or actual data such as DataY4S,
                    DataContinuum.
                    The event type holds again a dictionary with the keys:
                    streams: List of streams to be processed for this event type.
                    fraction: Fraction of events to process
                    only_use_files_from_to (default [null, null] = use all files): For testing:
                        Only use these data files and ignore the rest. The fraction keyword still applies.
                        Possible values, e.g. [0, 1] or [null, 1] (only use first file), [-1, null] (only use last file),
                        [0, 100] use files 0 to 99, etc.
    output: A list of key words which correspond to a output created in the reconstruction script.
            For each keyword the corresponding root file will be created for each reconstruction task
            and then merged by the merger task.
    use_mc_info: Whether to use MC Info for MC Matching and to retrieve generator level information.
    roe_mask: Tag of the RestOfEvent-mask that will be used.
    merge_ntuples: Bool, defaults to True. If set, the single chunks fill be merged into a single ROOT
        file.

    ----------------------------------
    Global Config File:
    The global config file contains settings and information necessary and equal for all jobs of
    an analysis.
    This config file is a json file containing a nested dictionary with the following structure:

    {
      "target_directory": "/home/belle2/metzner/test/rdstar_tests",
      "file_dict_location": "/home/belle2/metzner/file_dicts/",
      "known_error_list": "/home/belle2/metzner/test/rdstar_tests/test_known_error_list.json",
      "signal_MC_config": "/home/belle2/metzner/RDStar/rdstar/batch_tools/signal_config.json",
      "own_local_mc_db_path": "/home/belle2/metzner/databases/RDStar_b2bii/mc_beamparams/localdb/",
      "FEI_options": {
        "FEI_tag": "FEIv4_2017_MCConverted_Track14_2",
        "hadronic": true,
        "semileptonic": false
      },
      "use_b2bii": false
    }

    The meaning of the different entries is:

    target_directory: Path to the directory where the output of this processing shall be put.
    file_dict_location: Path to json file containing the file dictionary with information about the
                        events in all available files.
    known_error_list: Path to json file containing a list with known errors which indicate a malfunction
                      of the n-tuple production. If a known error occurred, the job has to be fixed and redone.
    signal_MC_config: Path to the json file that contains the Belle_signal_types dictionary.
    own_local_mc_db_path: Path to the MC beam parameters.
    FEI_options: A dictionary of all options concerning the FEI in the reconstruction process.
        FEI_tag: Tag of an already trained configuration of the FEI
        hadronic: If hadronic FEI is used.
        semileptonic: If semileptonic FEI is used.
    use_b2bii: If converted Belle data/MC is used.
    ----------------------------------

    To run this task, call luigi.build(Master(batch_config_file='path_to_batch_config.json'))
    """

    batch_config_file = luigi.Parameter(positional=False)  # type: PathType

    # TODO Function to check config!

    def get_batch_config_and_tag(self) -> Tuple[Dict[str, Any], str]:
        batch_config = load_json_config(self.batch_config_file, hint="Loading batch config file")
        batch_tag = "" if not batch_config["tag"] else batch_config["tag"] + "_"
        batch_tag += "_".join(batch_config["data"].keys())
        return batch_config, batch_tag

    def get_task_file_dir(self) -> PathType:
        batch_config, batch_tag = self.get_batch_config_and_tag()
        global_config = load_json_config(batch_config["global_config"], hint="Loading global config in Master")
        os.makedirs(global_config["target_directory"], exist_ok=True)
        return os.path.join(global_config["target_directory"], batch_tag, f"{self.get_task_family()}")

    def get_log_file_dir(self) -> PathType:
        return self.get_task_file_dir()

    def requires(self) -> Generator[luigi.Task, None, None]:
        batch_config, batch_tag = self.get_batch_config_and_tag()

        global_config_file = batch_config["global_config"]
        global_config = load_json_config(global_config_file, hint="Loading global config file")

        config = batch_config.copy()
        config.update(global_config)
        config["batch_config"] = str(self.batch_config_file)

        config["target_directory"] = os.path.join(config["target_directory"], batch_tag)

        for evt_type, params in config["data"].items():
            if "fraction" in params:
                fraction = params["fraction"]
            else:
                fraction = 1.0
            if "only_use_files_from_to" in params:
                assert len(params["only_use_files_from_to"]) == 2
                start, stop = params["only_use_files_from_to"]
                if start is not None:
                    start = int(start)
                if stop is not None:
                    stop = int(stop)
                only_use_files_from_to = (start, stop)
            else:
                only_use_files_from_to = (None, None)
            if params["streams"] is not None and len(params["streams"]) > 0:
                for stream in params["streams"]:
                    yield ArchiverTask(
                        config=config,
                        evt_type=evt_type,
                        stream=stream,
                        fraction=fraction,
                        only_use_files_from_to=only_use_files_from_to,
                    )
            else:
                yield ArchiverTask(
                    config=config,
                    evt_type=evt_type,
                    stream="None",
                    fraction=fraction,
                    only_use_files_from_to=only_use_files_from_to,
                )

        yield FileDictArchiveTask(config=config)


class FileDictArchiveTask(luigi.Task):
    config = luigi.DictParameter(positional=False)  # type: Dict[str, Any]

    archive_format = "gztar"  # type: str
    archive_extension = "tar.gz"  # type: str
    archive_name = "file_dict_archive"  # type: str
    archive_dir_name = "file_dict_archive"  # type: str

    # The archive task is run directly on the worker nodes.
    batch_system = "local"  # type: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        batch_config_file_name = os.path.basename(os.path.splitext(self.config["batch_config"])[0])  # type: str

        self.archive_dir = self.get_archive_dir_path(batch_config=self.config)  # type: PathType

        self.output_key = self.get_output_key()  # type: str
        self.output_path = self.get_output_path(batch_config=self.config)  # type: PathType

        _output_base_path = os.path.join(self.archive_dir, self.archive_name)  # type: PathType
        assert self.output_path == f"{_output_base_path}.{self.archive_extension}", (
            self.output_path,
            _output_base_path,
            self.archive_extension,
        )

        self.temp_output_base_path = os.path.join(
            self.archive_dir,
            f"temp_{batch_config_file_name}_{self.archive_name}",
        )  # type: PathType

    @staticmethod
    def get_output_key() -> str:
        return FileDictArchiveTask.archive_name

    @staticmethod
    def get_archive_dir_path(
        batch_config: Dict[str, Any],
        local_config_dir_path: Optional[PathType] = None,
    ) -> PathType:
        global_config_file = batch_config["global_config"]  # type: str

        if local_config_dir_path is not None:
            _split_str = "/rdstar/configs/"  # type: str
            assert os.path.isdir(local_config_dir_path), local_config_dir_path
            assert str(local_config_dir_path).endswith(_split_str) or str(local_config_dir_path).endswith(
                _split_str[:-1]
            ), (local_config_dir_path, _split_str)
            assert global_config_file.count(_split_str) == 1, (global_config_file, _split_str)
            _end_of_path = global_config_file.split(sep=_split_str, maxsplit=1)[1]  # type: str
            global_config_file = os.path.join(local_config_dir_path, _end_of_path)

        global_config = load_json_config(global_config_file, hint="Loading global config file")

        main_target_dir_path = global_config["target_directory"]
        production_tag = batch_config["tag"]

        archive_dir = os.path.join(
            main_target_dir_path,
            f"{production_tag}_{FileDictArchiveTask.archive_dir_name}",
        )  # type: PathType
        return archive_dir

    @staticmethod
    def get_output_path(
        batch_config: Dict[str, Any],
        local_config_dir_path: Optional[PathType] = None,
    ) -> PathType:
        archive_dir = FileDictArchiveTask.get_archive_dir_path(
            batch_config=batch_config,
            local_config_dir_path=local_config_dir_path,
        )  # type: PathType
        archive_path = os.path.join(
            archive_dir,
            f"{FileDictArchiveTask.archive_name}.{FileDictArchiveTask.archive_extension}",
        )  # type: PathType
        return archive_path

    def get_task_file_dir(self) -> PathType:
        os.makedirs(self.archive_dir, exist_ok=True)
        return os.path.join(self.archive_dir, f"{self.get_task_family()}_")

    def get_log_file_dir(self) -> PathType:
        return self.get_task_file_dir()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        return {self.output_key: luigi.LocalTarget(self.output_path)}

    def run(self) -> None:
        __file_dict_dir = self.config["file_dict_location"]
        assert isinstance(__file_dict_dir, str), __file_dict_dir
        assert os.path.isdir(__file_dict_dir), __file_dict_dir
        file_dict_dir_path = __file_dict_dir  # type: PathType

        temp_result_path = shutil.make_archive(
            base_name=str(self.temp_output_base_path),
            format=self.archive_format,
            root_dir=file_dict_dir_path,
        )
        assert temp_result_path == f"{self.temp_output_base_path}.{self.archive_extension}", (
            temp_result_path,
            f"{self.temp_output_base_path}.{self.archive_extension}",
        )

        if not os.path.isfile(self.output_path):
            os.rename(src=temp_result_path, dst=self.output_path)


class ArchiverTask(luigi.Task):

    config = luigi.DictParameter(positional=False)  # type: Dict[str, Any]
    evt_type = luigi.Parameter(positional=False)  # type: str
    stream = luigi.IntParameter(positional=False)  # type: int
    fraction = luigi.FloatParameter(positional=False)  # type: float
    only_use_files_from_to = luigi.TupleParameter(
        positional=False,
        default=(None, None),
    )  # type: Tuple[Optional[int], Optional[int]]

    # The merger task is run directly on the worker nodes, as it might require more memory.
    batch_system = "local"  # type: str

    log_archive_name = "log_files_archive"  # type: str
    log_archive_extension = "tar.gz"  # type: str
    file_types_to_archive = ("", ".log", ".info", ".json", ".py", ".sh")  # type: Tuple[str, ...]

    def get_log_archive_dir_path(self) -> PathType:
        target_dir_path = self.config["target_directory"]  # type: PathType
        return os.path.join(target_dir_path, f"{self.evt_type}_{self.log_archive_name}")

    def get_task_file_dir(self) -> PathType:
        log_archive_dir_path = self.get_log_archive_dir_path()  # type: PathType
        os.makedirs(log_archive_dir_path, exist_ok=True)
        identifier = "{task}_{evt}".format(task=self.get_task_family(), evt=self.evt_type)
        return os.path.join(log_archive_dir_path, f"{identifier}_")

    def get_log_file_dir(self) -> PathType:
        return self.get_task_file_dir()

    def create_log_files_archive(self, dry_run: bool = False) -> Tuple[str, PathType]:
        log_file_dir_path = self.get_log_archive_dir_path()  # type: PathType

        target_dir_path = self.config["target_directory"]  # type: PathType
        root_dir_path = os.path.join(target_dir_path, self.evt_type)  # type: PathType

        archive_output_key = self.log_archive_name
        archive_file_name = f"{self.log_archive_name}.{self.log_archive_extension}"  # type: str
        archive_path = os.path.join(log_file_dir_path, archive_file_name)  # type: PathType
        temp_archive_path = os.path.join(log_file_dir_path, f"temp_{archive_file_name}")  # type: PathType

        if not dry_run:
            os.makedirs(log_file_dir_path, exist_ok=True)

            assert not os.path.exists(archive_path), archive_path

            with tarfile.open(temp_archive_path, "w:gz") as tar_archive:
                for root_path, _, files in os.walk(root_dir_path):
                    for a_file in files:
                        if os.path.splitext(a_file) in self.file_types_to_archive:
                            tar_archive.add(os.path.join(root_path.split(root_dir_path, 1)[1], a_file))

            os.rename(src=temp_archive_path, dst=archive_path)

        return archive_output_key, archive_path

    def output(self) -> Dict[str, luigi.LocalTarget]:
        archive_key, archive_path = self.create_log_files_archive(dry_run=True)
        return {archive_key: luigi.LocalTarget(archive_path)}

    def requires(self) -> Generator[luigi.Task, None, None]:
        yield self.clone(MergerTask)

    def run(self) -> None:
        self.create_log_files_archive(dry_run=False)


class MergerTask(luigi.Task):

    config = luigi.DictParameter(positional=False)  # type: Dict[str, Any]
    evt_type = luigi.Parameter(positional=False)  # type: str
    stream = luigi.IntParameter(positional=False)  # type: int
    fraction = luigi.FloatParameter(positional=False)  # type: float
    only_use_files_from_to = luigi.TupleParameter(
        positional=False,
        default=(None, None),
    )  # type: Tuple[Optional[int], Optional[int]]

    # The merger task is run directly on the worker nodes, as it might require more memory.
    batch_system = "local"  # type: str

    @property
    def fake_merge(self) -> bool:
        """If this is set, do not actually merge files."""
        return not self.config.get("merge_ntuples", True)

    def get_task_file_dir(self) -> PathType:
        os.makedirs(self.config["target_directory"], exist_ok=True)
        identifier = "{task}_{evt}".format(task=self.get_task_family(), evt=self.evt_type)
        return os.path.join(self.config["target_directory"], f"{identifier}_")

    def get_log_file_dir(self) -> PathType:
        return self.get_task_file_dir()

    def get_output_file_paths(self) -> Generator[Tuple[str, PathType], None, None]:
        target_dir = self.config["target_directory"]
        output_names = self.config["output"]

        stream_tag = ""  # type: str
        if self.stream != "None":
            stream_tag = f"_{self.stream}"
        for name in output_names:
            out_name = f"{self.evt_type}{stream_tag}_{name}.root"
            yield name, os.path.join(target_dir, out_name)

    def output(self) -> Dict[str, luigi.LocalTarget]:
        output_dict = dict()  # type: Dict[str, luigi.LocalTarget]
        for output_name, output in self.get_output_file_paths():
            if self.fake_merge:
                p = Path(output)
                output = str(p.parent / ("." + p.name.replace(".root", ".complete")))
            output_dict.update({output_name: luigi.LocalTarget(output)})

        return output_dict

    def requires(self) -> Generator[luigi.Task, None, None]:
        max_events = self.config["events_per_chunk"]
        file_dict_location = self.config["file_dict_location"]
        signal_mc_config = self.config["signal_MC_config"]

        for chunk_num, chunk in enumerate(
            get_file_set(
                [self.evt_type],
                events_per_job=max_events,
                path_to_signal_mc_config=signal_mc_config,
                stream=self.stream,
                fraction=self.fraction,
                dict_directory=file_dict_location,
                only_use_files_from_to=self.only_use_files_from_to,
            )
        ):
            yield JobCheckerTask(
                chunk_number=chunk_num,
                file_chunk=chunk,
                stream=self.stream,
                evt_type=self.evt_type,
                config=self.config,
            )

    def run(self) -> None:
        if self.fake_merge:
            for output_target in self.output().values():
                info = (
                    "This file marks the completion of the corresponding MergerTask. "
                    "As merge_ntuples == False option was set in the config file, "
                    "no actual files were merged, so no more output files but this one"
                    " were generated."
                )
                Path(output_target.path).write_text(info)
            return

        for output_key, output_target in self.output().items():
            tmp_path = os.path.join(os.path.dirname(output_target.path), f"temp_{os.path.basename(output_target.path)}")
            files = [f[output_key].path for f in self.input()]
            cmd = ["fei_merge_files", tmp_path] + files
            subprocess.check_call(cmd)
            os.rename(src=tmp_path, dst=output_target.path)


class JobCheckerTask(luigi.Task):
    """
    Scans the log files of a basf2 task to decide whether the produced nTuple is sane.
    If it is sane, give the nTuple file free for merging.
    """

    chunk_number = luigi.IntParameter(positional=False)  # type: int
    file_chunk = luigi.DictParameter(positional=False)  # type: Dict[str, Any]
    evt_type = luigi.Parameter(positional=False)  # type: str
    stream = luigi.IntParameter(positional=False)  # type: int
    config = luigi.DictParameter(positional=False)  # type: Dict[str, Any]

    def get_output_path(self) -> PathType:
        target_dir = self.config["target_directory"]

        chunk_dir = f"chunk_{self.chunk_number}"  # type: str
        stream_suffix = ""  # type: str
        if self.stream != "None":
            stream_suffix = f"_{self.stream}"
        type_dir = f"{self.evt_type}{stream_suffix}"
        return os.path.join(target_dir, type_dir, chunk_dir)

    def get_task_file_dir(self) -> PathType:
        os.makedirs(self.get_output_path(), exist_ok=True)
        identifier = "{task}_{evt}".format(task=self.get_task_family(), evt=self.evt_type)
        return os.path.join(self.get_output_path(), f"{identifier}_")

    def get_log_file_dir(self) -> PathType:
        return self.get_task_file_dir()

    def get_log_file_name(self) -> PathType:
        return os.path.join(self.get_output_path(), f"chunk_{self.chunk_number}_log.log")

    def get_error_file_name(self) -> PathType:
        return os.path.join(self.get_output_path(), f"chunk_{self.chunk_number}_error.log")

    def output(self) -> Dict[str, luigi.LocalTarget]:
        """
        Path to moved n-tuple file for this chunk.
        :return: String containing path to n-tuple produced by job.
        """
        output_dict = dict()
        for output_name in self.config["output"]:
            checked_ntuple = os.path.join(
                self.get_output_path(),
                f"checked_chunk_{self.chunk_number}_{output_name}.root",
            )
            output_dict.update({output_name: luigi.LocalTarget(checked_ntuple)})
        return output_dict

    def requires(self) -> luigi.Task:
        """
        Requires reconstruction job yielding the relevant n-tuple file to be done and finished successfully.
        :return: Reconstruction task object for the desired n-tuple.
        """
        return ReconstructionTask(
            chunk_number=self.chunk_number,
            file_chunk=self.file_chunk,
            job_dir=self.get_output_path(),
            log_file=self.get_log_file_name(),
            error_file=self.get_error_file_name(),
            stream=self.stream,
            evt_type=self.evt_type,
            config=self.config,
        )

    def complete(self) -> bool:
        is_complete = super().complete()
        if not is_complete:
            return False

        required = luigi.task.flatten(self.requires())
        if not required:
            return True

        if not all([task.complete() for task in required]):
            return False

        return True

    def run(self) -> None:
        """
        Performs check on log files produced by the basf2 call executed in the
        ReconstructionTask this task depends on.
        If the log files are sane, the n-tuple file produced by the ReconstructionTask
        is approved for merging by moving it.
        :return:
        """
        if not os.path.exists(self.get_log_file_name()):
            stream_info = f" stream {self.stream}" if self.stream != "None" else ""
            sys.exit(
                f"Log file of task {self.evt_type}{stream_info} chunk {self.chunk_number}\n"
                f"which should be {self.get_log_file_name()}\n"
                f"does not exist!"
            )

        known_errors = load_json_config(self.config["known_error_list"], hint="Loading known errors list")

        log_file_sane = True
        with open(self.get_log_file_name(), "r") as log_file:
            for line in log_file.readline():
                for known_error in known_errors:
                    if known_error in line:
                        log_file_sane = False

        if log_file_sane:
            for output_key, output_path in self.output().items():
                os.symlink(self.input()[output_key].path, output_path.path)
        else:
            # If log was not sane, move input n-tuple file, so that requirement is not fulfilled.
            for output_key in self.output():
                failed_out = os.path.join(self.get_output_path(), f"failed_chunk_{self.chunk_number}_{output_key}.root")
                os.rename(self.input()[output_key].path, failed_out)
            # Move also log the files.
            failed_log = str(self.get_log_file_name()) + "_failed.log"
            os.rename(self.get_log_file_name(), failed_log)
            failed_error_log = str(self.get_error_file_name()) + "_failed.log"
            os.rename(self.get_error_file_name(), failed_error_log)


class ReconstructionTask(luigi.Task):
    """
    Runs basf2 steering file on LSF batch system to produce nTuple file for given file chunk.
    """

    chunk_number = luigi.IntParameter(positional=False)  # type: int
    file_chunk = luigi.DictParameter(positional=False)  # type: Dict[str, Any]
    job_dir = luigi.Parameter(positional=False)  # type: str
    log_file = luigi.Parameter(positional=False)  # type: str
    error_file = luigi.Parameter(positional=False)  # type: str
    stream = luigi.IntParameter(positional=False)  # type: int
    evt_type = luigi.Parameter(positional=False)  # type: str
    config = luigi.DictParameter(positional=False)  # type: Dict[str, Any]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.queue = "l"
        if self.config["queue"]:
            self.queue = self.config["queue"]

    def get_task_file_dir(self) -> PathType:
        os.makedirs(self.job_dir, exist_ok=True)
        identifier = "{task}_{evt}".format(task=self.get_task_family(), evt=self.evt_type)
        return os.path.join(self.job_dir, f"{identifier}_")

    def get_log_file_dir(self) -> PathType:
        return self.get_task_file_dir()

    def output(self) -> Dict[str, luigi.LocalTarget]:
        """
        The path to the n-tuple file.
        The n-tuple file is moved to this location after the job is finished successfully.
        :return: String containing path to n-tuple file.
        """
        output_dict = dict()
        for output_name in self.config["output"]:
            unchecked_ntuple = os.path.join(self.job_dir, f"unchecked_chunk_{self.chunk_number}_{output_name}.root")
            output_dict.update({output_name: luigi.LocalTarget(unchecked_ntuple)})
        return output_dict

    def complete(self) -> bool:
        is_complete = super().complete()
        if not is_complete:
            return False

        if os.path.exists(self.log_file) and os.path.exists(self.error_file):
            return True
        else:
            return False

    def run(self) -> None:
        """
        Start processing of events with the given parameters and configuration
        using the steering file defined in the provided configuration files.
        :return: None
        """
        # Prepare working directory for this task
        os.makedirs(self.job_dir, exist_ok=True)

        document_conditions(self.job_dir, self.config)

        # Check if output files or log files are already present end abort if so
        for output in luigi.task.flatten(self.output()):
            if output.exists():
                raise FileExistsError("Output file for task already exists.")
        if os.path.exists(self.log_file):
            raise FileExistsError("Log file for task already exists.")
        if os.path.exists(self.error_file):
            raise FileExistsError("Error file for task already exists.")

        # Prepare input for basf2 execution command
        steering = self.config["steering_file"]
        batch_config = self.config["batch_config"]
        file_list = self.file_chunk["files"]
        sequence_list = self.file_chunk["parts"]

        output_dict_list = []
        temp_output_dict = {}
        for output_key, output_path in self.output().items():
            output_dict_list.append(output_key)
            temp_output_name = "temp" + os.path.basename(output_path.path)
            temp_output_path = os.path.join(os.path.dirname(output_path.path), temp_output_name)
            output_dict_list.append(temp_output_path)
            temp_output_dict.update({temp_output_path: output_path.path})

        # Prepare basf2 command
        cmd = ["basf2", steering, "--", "--config", batch_config, "--input_files"]
        cmd.extend(file_list)
        cmd.append("--input_sequences")
        cmd.extend(sequence_list)
        cmd.append("--output_dict")
        cmd.extend(output_dict_list)

        # Call basf2 command with copy of current env and log files to write to
        env = os.environ.copy()
        try:
            with open(self.log_file, "w") as log_file_target, open(self.error_file, "w") as error_file_target:
                subprocess.check_call(cmd, env=env, stdout=log_file_target, stderr=error_file_target)
        except subprocess.CalledProcessError as process_failed_exception:
            failed_error_log = self.error_file + "_failed.log"
            os.rename(self.error_file, failed_error_log)
            failed_log = self.log_file + "_failed.log"
            os.rename(self.log_file, failed_log)
            raise process_failed_exception

        for temp_path, output_path in temp_output_dict.items():
            os.rename(temp_path, output_path)


def document_conditions(
    target_dir: PathType,
    config: Dict[str, Any],
) -> None:
    """
    Function to document
        - git hashes of the used basf2, analysistools and specific analysis package,
        - the global config and batch config which was used
        - the steering file that was used
    in the working directory of each job.
    :param target_dir: The working directory of the job which shall be documented and where this function will write to.
    :param config: The configuration used in the job to be documented.
    :return: None
    """
    try:
        import basf2_version
    except ImportError:
        from basf2 import version as basf2_version
    basf2_repo = basf2_version.version

    import analysistools

    target = os.path.join(target_dir, "git_versions_used.info")

    try:
        at_repo = git.Repo(inspect.getfile(analysistools), search_parent_directories=True).head.commit
    except git.exc.InvalidGitRepositoryError:
        at_repo = "not available"
    try:
        own_repo = git.Repo(config["steering_file"], search_parent_directories=True).head.commit
    except git.exc.InvalidGitRepositoryError:
        own_repo = "not available"

    with open(target, "w") as output_file:
        output_file.write(f"basf2 = {basf2_repo}\n")
        output_file.write(f"analysistools = {at_repo}\n")
        output_file.write(f"analysis repo = {own_repo}\n")

    batch_config_target = os.path.join(target_dir, "used_batch_config.json")
    global_config_target = os.path.join(target_dir, "used_global_config.json")
    steering_file_target = os.path.join(target_dir, "used_steering_file.py")
    shutil.copy(config["batch_config"], batch_config_target)
    shutil.copy(config["global_config"], global_config_target)
    shutil.copy(config["steering_file"], steering_file_target)
