import glob
import pandas as pd

from concurrent.futures import ThreadPoolExecutor


class DataFrameManager:
    """
    Reads in pandas.DataFrames from hdf files using multiple
    threads and saveds these dataframes in an internal dictionary.
    The file structure is supposed to be like
    `source_dir/data_dir/hdf_files`.

    :Example:

    >>> from analysistools.dataframe_tools.dataframe_manager import DataFrameManager
    >>> dfm = DataFrameManager(4)
    >>> dfm.load_hdf("stream_0", "/storage/9/fmetzner/rdstar/data_set_prod_27thFeb2019_stream0", ["Charged", "Charm", "UDS", "Mixed"], "_Y4S_ntuple_with_CS_CS1.h5")
    >>> dfm.load_hdf("stream_1", "/storage/9/fmetzner/rdstar/data_set_prod_07thMar2019_stream1", ["Charged", "Charm", "UDS", "Mixed"], "_Y4S_ntuple_with_CS_CS1.h5")
    >>> dfm.load_hdf("dss_sig", "/storage/9/fmetzner/rdstar/data_set_prod_07thMar2019_stream0", ["BMD1E", "BMD1Mu", "BZD1E", "BZD1Mu"], "_Y4S_ntuple_with_CS_CS1.h5")
    >>> stream_0_dataframes = dfm.by_store_key("dss_sig")  # returns {"BMD1E": pd.DataFrame, "BMD1Mu": pd.DataFrame, "BZD1E": pd.DataFrame,"BZD1Mu": pd.DataFrame}
    >>> mixed_dataframes = dfm.by_data_key("Mixed")  # returns {"stream_0": pd.DataFrame, "stream_1": pd.DataFrame}
    """

    def __init__(self, max_workers):
        """Constructor.

        :param max_workers: Maximum number of threads to spawn for
        reading in data.
        """
        self._max_workers = max_workers
        self._df_storage = dict()

    def load_hdf(self, store_tag, source_dir, data_dirs, file_tag):
        """
        Loads pandas.DataFrames from the hdf format and stores them internally.

        :param store_tag: Key for the dictionary which stores these
        dataframes.
        :param source_dir: Path to the source directory which holds
        several subdirectories with the dataframes.
        :param data_dirs: List of names of the subdirectories to load
        the dataframes from.
        :param file_tag: Tag which specifies the hdf file to load. This
        is expected to be equal for all subdirectories specified by
        the `data_dir` parameter.
        """

        file_list = []

        if store_tag not in self._df_storage.keys():
            self._df_storage[store_tag] = dict()

        for data_dir in data_dirs:
            for f in glob.glob(f"{source_dir}/{data_dir}/*.h5"):
                if file_tag in f:
                    file_list.append((data_dir, f))
                else:
                    continue

        print(f"Loading files with {self._max_workers} workers")
        with ThreadPoolExecutor(max_workers=self._max_workers) as executor:
            for ftuple in file_list:
                executor.submit(self.load_hdf_into_dict, self._df_storage[store_tag], *ftuple)

    @staticmethod
    def load_hdf_into_dict(dfdict, tag, path):
        print(f"Loading file {path} into storage dict with key {tag}...")
        dfdict[tag] = pd.read_hdf(path)
        print(f"Finished loading file {path}")

    @property
    def storage_keys(self):
        return list(self._df_storage.keys())

    @property
    def storage(self):
        """Returns the internal storage dictionary."""
        return self._df_storage

    def by_data_key(self, data_tag):
        """
        Returns a dictionary with pandas.DataFrames holding data of
        the type specified by `data_tag` in the internal storage.

        :param data_tag: Key of the desired data types in the internal storage.

        :returns: Dictionary with keys equal to the storage_tags and
        values equal to pandas.DataFrames of type specified by
        `data_tag`.
        """
        data_dict = dict()
        for store_key, df_dict in self._df_storage.items():
            if data_tag in df_dict.keys():
                data_dict[store_key] = df_dict[data_tag]

        return data_dict

    def by_store_key(self, store_key):
        """
        Returns a dictionary with pandas.DataFrames holding data of the types loaded with store_key.

        :param store_key: Key of the desired storage type in the internal storage.

        :returns: Dictionary of pandas.DataFrames.
        """
        return self._df_storage[store_key]
