"""
This module contains convenience functions for handling pd.DataFrames.

Max Welsch 2018
"""

import os
import hashlib
import shutil
import subprocess
import numpy as np
import pandas as pd
import multiprocessing

from functools import partial
from collections import OrderedDict
from typing import Dict, Tuple, Set, List, Iterable, Any, NamedTuple, Callable, Optional

from analysistools.utilities import PathType
from analysistools.statistics.limits import binomial_limit


class CutStatistics(NamedTuple):
    evts_before_cut: float
    evts_after_cut: float
    eff: float
    eff_cl_interval: Tuple[float, float]


def apply_cut(
    df: pd.DataFrame,
    cut_string: str,
    weight: str = "__weight__",
) -> Tuple[pd.DataFrame, CutStatistics]:
    """
    Applies cuts on a pandas dataframe using a cut string like in basf2.

    :param df: pd.Dataframe on which the cut will be applied to.
    :param cut_string: basf2 cut string to be applied. E.g. `1.1 < InvMass < 1.5 and pi0_momentum > 200`.
    :param weight: Column name of the event weight. If the data frame has no
                   column named weight, the length of the index is uses as number of events.
    :return: DataFrames with cuts applied and statistics.
    """

    if weight in df.columns:
        evts_before_cut = float(np.sum(df[weight].values))  # type: float
    else:
        evts_before_cut = len(df.index)

    cut_df = df.query(cut_string)

    if weight in df.columns:
        evts_after_cut = float(np.sum(cut_df[weight].values))  # type: float
    else:
        evts_after_cut = len(cut_df.index)

    eff = evts_after_cut / evts_before_cut  # type: float
    eff_cl_interval = binomial_limit(evts_before_cut, evts_after_cut)  # type: Tuple[float, float]

    stats = CutStatistics(
        evts_before_cut=evts_before_cut,
        evts_after_cut=evts_after_cut,
        eff=eff,
        eff_cl_interval=eff_cl_interval,
    )

    return cut_df, stats


def apply_cut_to_df(df: pd.DataFrame, cut_string: str) -> pd.DataFrame:
    """
    Apply cuts to a single data frame. Mostly a convenience function for other functions.
    :param df: Pandas data frame.
    :param cut_string: String with arbitrary cuts combined by `and` and `or`,
                       e.g. `1.1 < InvMass < 1.5 and pi0_momentum > 200`.
    :return: Pandas data frame with applied cuts.
    """
    cut_df = df.query(cut_string)
    return cut_df


def get_max(full_dfs: Dict[str, pd.DataFrame], variable: str) -> np.array:
    """
    Calculates the maximum of a variable within each data frame given in the data frame dictionary.
    Then lists them in a numpy array.
    :param full_dfs: Data frame dictionary
    :param variable: Column name
    :return: Array of maximum of each data frame given by the data frame dictionary
    """
    maxima = [df[variable].max() for df in full_dfs.values()]
    return np.max(maxima)


def get_min(full_dfs: Dict[str, pd.DataFrame], variable: str) -> np.array:
    """
    Calculates the minimum of a variable within each data frame given in the data frame dictionary.
    Then lists them in a numpy array.
    :param full_dfs: Data frame dictionary
    :param variable: Column name
    :return: Array of minimum of each data frame given by the data frame dictionary
    """
    minima = [df[variable].min() for df in full_dfs.values()]
    return np.min(minima)


def remove_columns_from_df_except(df: pd.DataFrame, columns: Iterable[Any]) -> pd.DataFrame:
    """
    removes all columns from a dataframe except the ones you hand over
    :param df: Pandas data frame.
    :param columns: Iterable container of column-names you want to keep.
    :return: Skimmed Pandas data frame.
    """
    cols_total: Set[Any] = set(df.columns)
    diff: Set[Any] = cols_total - set(columns)
    df.drop(diff, axis=1, inplace=True)
    return df


def skim_dfs_by_column(dfs: Dict[str, pd.DataFrame], columns: List[Any]):
    """
    Removes all columns from the data frames in the provided dictionary except the ones handed over in an iterable.
    :param dfs: Dictionary of Pandas data frames.
    :param columns: List of column-names you want to keep.
    :return: Dictionary of skimmed Pandas data frames.
    """
    columns.append("__weight__")
    skimmed_dfs = OrderedDict()
    for key in dfs.keys():
        skimmed_dfs[key] = remove_columns_from_df_except(dfs[key], columns)
    return skimmed_dfs


def add_column_to_dfs(dfs: Dict[str, pd.DataFrame], func: Callable, new_column_name: str, variables: List[str]) -> None:
    """
    Adds a column to each pandas data frame in a dictionary of data frames.

    The values will be calculated via the given function, with the specified variables as arguments.
    The function has to be able to take the full data from one variable as an array and should return
    the full result as array.

    :param dfs: Dictionary of pandas data frames.
    :param func: Function to calculate the value you want to add as new column to the data frame.
                 Only takes other columns (= variables) as arguments.
    :param new_column_name: New name of the added column.
    :param variables: List of column names that are arguments of the function.
                      Order has to be the same as the expected by the function.
    :return: None
    """
    for df in dfs.values():
        df[new_column_name] = func(df[variables].values)


def concatenate_dfs(dfs: Dict[str, pd.DataFrame], new_key: str):
    """
    Concatenates multiple data frames stored in a dictionary to one data frame.
    :param dfs: Dictionary of pandas data frames.
    :param new_key: Key of the combined data frame.
    :return: Ordered data frame
    """
    return OrderedDict([(new_key, pd.concat(dfs.values(), keys=dfs.keys()))], sort=False)


def get_unique_event_id(
    ntuple_df: pd.DataFrame,
    data_type: str,
    optional_columns: Optional[List[str]] = None,
    column_name: str = "unique_event_id",
    hash_id: bool = False,
) -> None:
    """
    Generates and appends unique event ID build from several columns to the provided data frame.

    The unique event ID is build using
        - the name of the data type which has to be provided via the respective argument
        - experiment-, run- and event-number which are taken from the respective columns of the data frame
        - additional columns of the data frame, which can be provided as list of column names.
    which are combined to a string and then hashed using sha512.

    The created hash will be added to the data frame as new column, the name of which can be provided
    via the argument column_name and is set to "unique_event_id" as default name.

    :param ntuple_df: The data frame to which the unique event id column will be added.
    :param data_type: Name of the data type with stream if existing (e.g. Charged_0, Charm_3, UDS_5, signalA...).
    :param optional_columns: Optional additional columns to be used for the unique event id.
    :param column_name: Name to be used for the added column. Default: unique_event_id.
    :param hash_id: Whether to store a hashed version of the id instead of the plain string. Default = False
    :return: None
    """
    if optional_columns is None:
        optional_columns = []

    if len(ntuple_df.index) == 0:
        # Make sure that the column exists nonetheless. If we only get a
        # copy/view, the column won't be set on the original dataframe,
        # but we also won't get a SettingWithCopyWarning
        ntuple_df[column_name] = np.nan
        return

    assert column_name not in ntuple_df.columns, column_name

    def get_id_string(row: pd.Series) -> str:
        """Generate an ID string from one row of the dataframe"""
        return "{data_type}_{exp}_{run}_{evt}".format(
            data_type=data_type,
            exp=row.expNum,
            run=row.runNum,
            evt=row.evtNum,
        ) + \
        ("_" + "_".join(["{}".format(row[x]) for x in optional_columns]) if optional_columns else "")

    if not hash_id:
        ntuple_df.loc[:, column_name] = ntuple_df.apply(
            lambda row: str(get_id_string(row)),
            axis=1,
        )
    else:
        ntuple_df[column_name] = ntuple_df.apply(
            lambda row: hashlib.sha512(get_id_string(row).encode("utf-8")),
            axis=1,
        )

    assert not pd.isnull(ntuple_df[column_name]).any()


def copy_file_dict_archive_from_remote(
    remote_host: str,
    source_path: PathType,
    target_dir: PathType,
    file_dict_dir_name: str,
    overwrite: bool = False,
    dry_run: bool = False,
) -> Tuple[PathType, PathType]:
    archive_file_name = os.path.basename(source_path)  # type: str
    temp_archive_target_path = os.path.join(target_dir, f"temp_{archive_file_name}")  # type: PathType
    archive_target_path = os.path.join(target_dir, archive_file_name)  # type: PathType

    file_dict_target_path = os.path.join(target_dir, file_dict_dir_name)  # type: PathType

    if not dry_run:
        get_file_from_remote(
            source_target_tuple=(source_path, temp_archive_target_path),
            remote_host=remote_host,
            overwrite=overwrite,
        )
        assert os.path.isfile(temp_archive_target_path), temp_archive_target_path
        shutil.unpack_archive(
            filename=str(temp_archive_target_path),
            extract_dir=file_dict_target_path,
        )

        assert os.path.isdir(file_dict_target_path), file_dict_target_path
        os.rename(src=temp_archive_target_path, dst=archive_target_path)
        assert os.path.isfile(archive_target_path), archive_target_path

    return source_path, archive_target_path


def copy_online_analysis_output_from_remote(
    remote_host: str,
    source_dir: PathType,
    target_dir: PathType,
    configs: Iterable[Dict],
    processes: int = 4,
    overwrite: bool = False,
    dry_run: bool = False,
) -> List[Tuple[PathType, PathType]]:
    """
    Convenience function to copy the output of online analysis batch tasks from a remote host to the
    local host. The files to be copied are gathered by evaluating the batch configs of the tasks
    for which the output shall be copied.
    The copy process is done by spawning one rsync process for each file using multiprocessing.
    The number of processes to be used can be specified via the processes parameter.
    The source directory as defined in the global config of the tasks and the target directory on
    the local host must be specified.
    By default local files will not be overwritten if they already exist, and the copy process will be
    aborted. To overwrite local files set the overwrite parameter to True.

    The dry_run option allows to produce only the list of source and target file tuples, without doing
    the actual copying.

    :param remote_host: Identifier for the remote host. Can be a ssh config entry or a user@remote_host
                        combination.
    :param source_dir: Source directory on the remote host which is defined by the target_directory entry
                       in the global config of an analysis.
    :param target_dir: Target directory on the local host.
    :param configs: Batch configurations for which the output files shall be copied.
    :param processes: Number of processes to be used. Default is 4.
    :param overwrite: Whether to overwrite existing files. Default is False.
    :param dry_run: If True, only return the list of tuples of file sources and targets that would be
                    copied, but do not perform the actual copy process. Default is False.
    :return: List of tuples of source and target files of the copy process.
    """
    source_target_list = []  # type: List[Tuple[PathType, PathType]]
    for config in configs:
        batch_tag = "" if not config["tag"] else config["tag"] + "_"
        batch_tag += "_".join(config["data"].keys())
        base_path = os.path.join(source_dir, batch_tag)

        for evt_type in config["data"].keys():
            if config["data"][evt_type]["streams"] is not None:
                stream_tags = [f"_{stream}" for stream in config["data"][evt_type]["streams"]]
            else:
                stream_tags = [""]

            for stream_tag in stream_tags:
                for output in config["output"]:
                    file_name = f"{evt_type}{stream_tag}_{output}.root"

                    source_path = os.path.join(base_path, file_name)  # type: PathType
                    target_path = os.path.join(target_dir, evt_type, file_name)  # type: PathType

                    source_target_list.append((source_path, target_path))

    if not dry_run:
        get_files_from_remote(
            remote_host=remote_host,
            source_target_list=source_target_list,
            processes=processes,
            overwrite=overwrite,
        )

    return source_target_list


def get_files_from_remote(
    remote_host: str,
    source_target_list: List[Tuple[PathType, PathType]],
    processes: int = 4,
    overwrite: bool = False,
) -> None:
    """
    Copies multiple files from a remote host to the local machine using multiprocessing.
    The number of processes spawned can be set via the processes parameter (default = 4).
    Uses rsync in one process per file to copy it from remote_host : source_target_tuple[0]
    to the local target source_target_tuple[1] for all given source file - target combinations
    in source_target_list.
    :param remote_host: Identifier for the remote host. Can be a ssh config entry or a user@remote_host
                        combination.
    :param source_target_list: List of tuples containing the source on the remote_host as first (=0)
                               entry and the full target path as the second (=1). This also allows
                               to give the copied file a new name on the local machine.
    :param processes: Number of processes to be used. Default is 4.
    :param overwrite: Whether to overwrite existing files. Default is False.
    :return: None
    """
    with multiprocessing.Pool(processes=processes) as process_pool:
        process_pool.map(
            partial(get_file_from_remote, remote_host=remote_host, overwrite=overwrite),
            source_target_list,
        )


def get_file_from_remote(
    source_target_tuple: Tuple[PathType, PathType],
    remote_host: str,
    overwrite: bool = False,
) -> None:
    """
    Executes rsync in a subprocess check_call to copy a file from <remote_host>:<source_target_tuple[0]>
    to the local target source_target_tuple[1].
    :param source_target_tuple: Tuple containing the source on the remote_host as first (=0) entry and
                                the full target path as the second (=1). This also allows to give the
                                copied file a new name on the local machine.
    :param remote_host: Identifier for the remote host. Can be a ssh config entry or a user@remote_host
                        combination.
    :param overwrite: Whether to overwrite an existing file. Default is False.
    :return: None
    """
    source, target = source_target_tuple  # type: PathType, PathType

    if os.path.isfile(target) and not overwrite:
        print(f"File target\n{target}\nalready exists and will not be overwritten!\n")
        return

    os.makedirs(os.path.dirname(target), exist_ok=True)

    target_dir, target_file = os.path.split(target)
    temp_target = os.path.join(target_dir, f"temp_rsync_{target_file}")  # type: PathType

    rsync_command = ["rsync", "-az", "--checksum", remote_host + ":" + str(source), temp_target]
    subprocess.check_call(rsync_command)

    if os.path.isfile(temp_target):
        os.rename(src=temp_target, dst=target)

    if not os.path.isfile(target):
        raise RuntimeError(
            f"File {target}\n" f"does not exist after it should have been copied from\n" f"f{remote_host}:f{source}"
        )
