"""
Convenience functions to load the data from n-tuple ROOT files and store them to hdf5.
Provides also functionality to gather the data in a given directory.

Gabriel Seitz
2018
"""

import os
import glob
import uproot
import decayHash
import root_pandas
import pandas as pd

from typing import List, Tuple, Dict, Optional, Callable

from analysistools.utilities import PathType
from analysistools.dataframe_tools.uproot_compatibility import uproottree2df
from analysistools.dataframe_tools.dataframe_tools import get_unique_event_id


def convert_data_from_root(
    source_dir: PathType,
    keywords: List[Tuple[str, Optional[str], Optional[bool]]],
    columns: Optional[List[str]] = None,
    drop_list: Optional[List[str]] = None,
    overwrite: bool = False,
    reduce_size: bool = True,
    cut_function: Optional[Callable[[pd.DataFrame], None]] = None,
    file_tag: Optional[str] = None,
    dry_run: bool = False,
) -> List[Tuple[PathType, PathType]]:
    """
    Loads specified n-tuples from ROOT files in a given directory and stores them as hdf5 files.

    For a given source_dir path, all ROOT files with matching key words in its subdirectories
    are loaded. The names of the sub directories must match the data types.
    The loaded data is then stored into an hdf5 store with the file name matching the one of the
    original ROOT file. E.g. the data of
        <source_dir>/<data_type>/<data_type>_<stream_num>_<ntuple_key>.root
    will be stored under
        <source_dir>/<data_type>/<data_type>_<stream_num>_<ntuple_key>.h5

    In addition to the conversion, further information will be added to the data frame:
        - A unique event ID, which is generated from several columns of the data frame
          and allows to match events between several data frames.
        - The decay string will be added if a decay tree ROOT file is provided as second
          entry of the keywords tuple.

    The data frame will be stored under the key "<data_type>_<stream_num>_<ntuple_key>"
    in the generated hdf5 store in the .h5 file.

    The stream number will be omitted, if not relevant.

    If the respective hdf5 file is already present, the file will be skipped, unless the
    boolean parameter overwrite is set to True, in which case the data will be loaded anew
    and the existing .h5 file will be overwritten.

    Via the function parameter drop_list, a list of column names can be specified, which shall
    be dropped from the loaded data frame before it is written to hdf5.

    The dry_run option allows to produce the list of tuples of root and hdf5 files which are the
    source and the target of the conversion operation.

    :param source_dir: Path to the dictionary containing the ROOT n-tuple files.
    :param keywords: Tuple of strings specifying the n-tuple file names,
                     the respective decay tree file names and a boolean
                     indicating whether to apply the cut_function if specified.
    :param columns: Columns (=Branches) to be read from the ROOT tree.
                    Default: None (= all columns).
    :param drop_list: List of basf2 variable names to be dropped from
                      the data frames. Default: None.
    :param overwrite: Whether to reload from ROOT and overwrite present
                      hdf5 files. Default: False.
    :param reduce_size: Flag whether to reduce size of the DataFrame by
                        casting float64 to float32 if possible and python
                        objects (like str) to pd dtype categorical.
                        Default: True.
    :param cut_function: Function to be applied on the DataFrame loaded
                         from ROOT. The function must have one parameter,
                         which is a DataFrame and no return value. Default = None.
    :param file_tag: Additional tag appended to the keyword in the output filename.
    :param dry_run: If True, only return a list of tuples of the source (root file)
                    and target (h5 file) of the conversion process, without doing
                    the actual conversion. Default is False.
    :return: List of tuples of source (.root) and target (.h5) files of the conversion process.
    """
    file_list = [str(os.path.basename(f)) for f in glob.glob(str(source_dir) + "/*/*.root")]  # type: List[str]
    data_types = list(set([file.split("_")[0] for file in file_list]))  # type: List[str]

    keywords = [k if isinstance(k, tuple) else (k,) for k in keywords]
    keyword_tuples = [
        (
            kt[0],
            kt[1] if len(kt) >= 2 else None,
            kt[2] if len(kt) == 3 and isinstance(kt[2], bool) else False,
        )
        for kt in keywords
    ]  # type: List[Tuple[str, Optional[str], bool]]

    source_target_tuple_list = []  # type: List[Tuple[PathType, PathType]]

    if not dry_run:
        print(
            "Loading data for\nDatatypes:\n  "
            + "\n  ".join([f"{d}" for d in data_types])
            + "\nKeywords:\n  "
            + "\n  ".join([f"{k}" for k in keyword_tuples])
        )

    tag = ""
    if file_tag is not None:
        tag = f"_{file_tag}"

    for data_type in data_types:
        streams = [f.split("_")[1] for f in file_list if f.split("_")[1].isdigit() and f.startswith(data_type)]
        streams = list(set(streams))

        if len(streams) == 0:
            streams.append("")

        for stream in streams:
            if stream.isdigit():
                stream = "_" + stream

            for ntuple_key, decay_tree_key, cut_bool in keyword_tuples:
                hdf5_key = f"{data_type}{stream}_{ntuple_key}"
                root_file = os.path.join(source_dir, data_type, f"{hdf5_key}.root")
                hdf5_file = os.path.join(source_dir, data_type, f"{hdf5_key}{tag}.h5")

                source_target_tuple_list.append((root_file, hdf5_file))

                if dry_run:
                    continue

                print(f"Converting from {os.path.join(source_dir, data_type)}/{hdf5_key}.root")

                if os.path.isfile(hdf5_file) and not overwrite:
                    print(f"  Skipping, as already converted and overwrite is {overwrite}...")
                    continue

                ntuple_df = load_ntuple_decay_tree_pair(
                    source_directory=os.path.join(source_dir, data_type),
                    data_type=f"{data_type}{stream}",
                    ntuple_key=ntuple_key,
                    decay_tree_key=decay_tree_key,
                    columns=columns,
                    drop_list=drop_list,
                    reduce_size=reduce_size,
                    cut_function=cut_function if cut_bool else None,
                )

                print("  Adding unique event ID hash column to n-tuple data frame.")
                get_unique_event_id(ntuple_df=ntuple_df, data_type=f"{data_type}{stream}")

                print(f"  Storing n-tuple data frame to {hdf5_file}")
                save_to_hdf5(ntuple_df, hdf5_file, hdf5_key, decay_tree_key=decay_tree_key)

    return source_target_tuple_list


def convert_root_to_hdf5(
    root_file_path: PathType,
    hdf5_file_path: PathType,
    hdf5_key: str,
    data_type: str,
    stream_tag: str,
    overwrite: bool = False,
    reduce_size: bool = False,
    columns: Optional[List[str]] = None,
    drop_list: Optional[List[str]] = None,
    cut_function: Optional[Callable[[pd.DataFrame], None]] = None,
    add_unique_event_id: bool = True,
) -> bool:
    print(f"Converting from {root_file_path}")

    if os.path.isfile(hdf5_file_path) and not overwrite:
        print(f"  Skipping, as already converted and overwrite is {overwrite}...")
        return False

    ntuple_df = load_ntuple_from_root(
        file_path=root_file_path,
        reduce_size=reduce_size,
        columns=columns,
        drop_list=drop_list,
        cut_function=cut_function,
    )

    if add_unique_event_id:
        print("  Adding unique event ID hash column to n-tuple data frame.")
        get_unique_event_id(ntuple_df=ntuple_df, data_type=f"{data_type}{stream_tag}")

    print(f"  Storing n-tuple data frame to {hdf5_file_path}")
    ntuple_df.to_hdf(path_or_buf=hdf5_file_path, key=hdf5_key)

    if os.path.isfile(hdf5_file_path):
        return True
    return False


def load_data_from_hdf5(
    source_dir: PathType,
    keywords: List[Tuple[str, Optional[str]]],
    drop_list: Optional[List[str]] = None,
    file_tag: Optional[str] = None,
) -> Dict[str, Dict[str, pd.DataFrame]]:
    """
    Loads specified n-tuples from hdf5 files in a given directory.

    In the given directory, all hdf5 files matching the given keywords are considered.
    The hdf5 files must have a name with the structure
        <data_type>_[<stream_number>_]<ntuple/decay_tree_name>.h5

    The keywords specify the n-tuples or decay_trees which shall be loaded.
    If a decay tree shall be loaded for a given n-tuple, it can be specified as the second
    entry in the keywords tuple. If specified this decay tree will be loaded from the
    respective ROOT file and the information it contains will be added to the n-tuple data frame.

    All data types, such as mixed, charged, charm, uds, signal_mc are considered.
    Different streams will be loaded into different data frames.

    The resulting data frame dictionary will have the structure
    {"data_type_a": {"ntuple_name_x": ntuple_df_x,
                     "ntuple_name_y": ntuple_df_y,
                     "ntuple_name_z": ntuple_df_z},
     "data_type_b_1": {"ntuple_name_x": ntuple_df_x,
                     "ntuple_name_y": ntuple_df_y,
                     "ntuple_name_z": ntuple_df_z},
     "data_type_b_2": {"ntuple_name_x": ntuple_df_x,
                     "ntuple_name_y": ntuple_df_y,
                     "ntuple_name_z": ntuple_df_z}
    }
    where the level 1 key is the data type with contingently a stream number appended.
    The level 2 key is the n-tuple name and the respective item is the actual data frame.

    If drop_list is not an empty list, the columns matching the variable names it contains
    will be removed from the loaded data frames.

    :param source_dir: Path to the dictionary containing the hdf5 n-tuple files.
    :param keywords: Tuple of strings specifying the n-tuple file names and respective decay tree file names.
    :param drop_list: List of basf2 variable names to be dropped from the data frames.
    :param file_tag: Additional tag appended to the keyword in the filename.
    :return: Nested dictionary of Pandas data frames.
    """
    if drop_list is None:
        drop_list = []

    file_list = [
        str(os.path.basename(f))
        for f in glob.glob(str(source_dir) + "*/*.h5")
        if str(os.path.basename(f)).startswith(os.path.basename(os.path.dirname(f)))
    ]
    data_types = list(set([file.split("_")[0] for file in file_list]))

    tag = ""
    if file_tag is not None:
        tag = f"_{file_tag}"

    print(
        "Loading data for\nDatatypes:\n  "
        + "\n  ".join([f"{d}" for d in data_types])
        + "\nKeywords:\n  "
        + "\n  ".join([f"{k}" for k in keywords])
    )

    df_dict = dict()

    for data_type in data_types:
        streams = [f.split("_")[1] for f in file_list if f.split("_")[1].isdigit() and f.startswith(data_type)]
        streams = list(set(streams))

        if len(streams) == 0:
            streams.append("")

        for stream in streams:
            data_type_dict = dict()
            if stream.isdigit():
                stream = "_" + stream

            for ntuple_key, decay_tree_key in keywords:
                hdf5_key = f"{data_type}{stream}_{ntuple_key}"
                hdf5_file = os.path.join(source_dir, data_type, f"{hdf5_key}{tag}.h5")

                print(f"Loading {hdf5_file}")

                ntuple_df = load_hdf5(hdf5_file, hdf5_key)
                ntuple_df.drop(columns=drop_list, inplace=True)

                if decay_tree_key is not None:
                    decay_tree_file = os.path.join(source_dir, data_type, f"{data_type}{stream}_{decay_tree_key}.h5")
                    decay_tree = load_decay_tree_from_root(decay_tree_file)
                    add_decay_tree_to_ntuple(ntuple_df, decay_tree, full=True)

                data_type_dict.update({ntuple_key: ntuple_df})

            df_dict.update({data_type + stream: data_type_dict})

    return df_dict


def load_ntuple(
    source_directory: PathType,
    data_type: str,
    ntuple_key: str,
    reduce_size: bool = True,
    columns: Optional[List[str]] = None,
    drop_list: Optional[List[str]] = None,
    cut_function: Optional[Callable[[pd.DataFrame], None]] = None,
) -> pd.DataFrame:
    """
    Loads n-tuple from ROOT file or from hdf5 file, if already converted.

    The n-tuple for the given data type and n-tuple key combination is loaded from the
    respective ROOT file.
    The function will look for files with the signature <data_type>_<ntuple_key>.root
    in the specified source directory.

    If a drop list is given, the respective columns matching the given variable names will be
    dropped from the loaded n-tuple.

    :param source_directory: Source dictionary of the data and where the hdf5 files will be written to.
    :param data_type: Data type for which the data shall be loaded.
    :param ntuple_key: Key of the n-tuple to be loaded.
    :param reduce_size: Flag whether to reduce size of the DataFrame by casting float64
                        to float32 if possible. Default: True.
    :param columns: Columns (=Branches) to be read from the ROOT tree.
                    Default: None (= all columns).
    :param drop_list: List of variable names to drop. Default: None.
    :param cut_function: Function to be applied on the DataFrame loaded
                         from ROOT. The function must have one parameter,
                         which is a DataFrame and no return value.
    :return: Pandas data frame containing the n-tuple data.
    """
    if columns is not None and drop_list is not None:
        raise ValueError("The parameters columns and drop_list can not be used at the same time!")

    return load_ntuple_decay_tree_pair(
        source_directory=source_directory,
        data_type=data_type,
        ntuple_key=ntuple_key,
        reduce_size=reduce_size,
        columns=columns,
        drop_list=drop_list,
        cut_function=cut_function,
    )


def load_ntuple_decay_tree_pair(
    source_directory: PathType,
    data_type: str,
    ntuple_key: str,
    decay_tree_key: Optional[str] = None,
    column_name: str = "decay_string",
    reduce_size: bool = True,
    columns: Optional[List[str]] = None,
    drop_list: Optional[List[str]] = None,
    cut_function: Optional[Callable[[pd.DataFrame], None]] = None,
) -> pd.DataFrame:
    """
    Loads pair of n-tuple and optional decay tree for given data type and key from source_directory.

    The n-tuple for the given data_type ntuple_key combination and the respective decay tree are loaded
    and decay tree information is added to the n-tuple data frame in a new column.
    If no decay tree key is provided no decay tree is added to the n-tuple.
    The new column's name can be specified via the column_name parameter, which defaults to 'decay_string'.

    :param source_directory: Source dictionary of the data and where the
    hdf5 files will be written to.
    :param data_type: Data type for which the data shall be loaded.
    :param ntuple_key: Key of the n-tuple to be loaded.
    :param decay_tree_key: Optional key for decay tree to be loaded and attached to the n-tuple. Default: None.
    :param column_name: Optional name for the new column for the decay
                        tree information. Default: 'decay_string'.
    :param reduce_size: Flag whether to reduce size of the DataFrame by casting float64
                        to float32 if possible. Default: True
    :param columns: Columns (=Branches) to be read from the ROOT tree.
                    Default: None (= all columns).
    :param drop_list: List of variable names to drop. Default: None.
    :param cut_function: Function to be applied on the DataFrame loaded
                         from ROOT. The function must have one parameter,
                         which is a DataFrame and no return value. Default = None.
    :return: Pandas data frame containing the n-tuple with the decay tree information added, if available.
    """
    if columns is not None and drop_list is not None:
        raise ValueError("The parameters columns and drop_list can not be used at the same time!")

    file_name = f"{data_type}_{ntuple_key}"
    root_ntuple_file_path = os.path.join(source_directory, f"{file_name}.root")

    root_decay_tree_file_path = None
    if decay_tree_key is not None:
        root_decay_tree_file_path = os.path.join(source_directory, f"{data_type}_{decay_tree_key}.root")

    ntuple_df = load_ntuple_decay_tree_pair_from_root(
        ntuple_file_path=root_ntuple_file_path,
        decay_tree_file_path=root_decay_tree_file_path,
        column_name=column_name,
        reduce_size=reduce_size,
        columns=columns,
        drop_list=drop_list,
        cut_function=cut_function,
    )

    return ntuple_df


def load_ntuple_decay_tree_pair_from_root(
    ntuple_file_path: PathType,
    decay_tree_file_path: Optional[PathType] = None,
    column_name: str = "decay_string",
    reduce_size: bool = True,
    columns: Optional[List[str]] = None,
    drop_list: Optional[List[str]] = None,
    cut_function: Optional[Callable[[pd.DataFrame], None]] = None,
) -> pd.DataFrame:
    """
    Loads pair of n-tuple and optional decay tree from ROOT file into pandas data frame and drops given columns.

    The n-tuple and the decay tree are loaded from the given file paths. The decay tree information is
    added to the n-tuple data frame in a new column.
    If no decay tree path is provided no decay tree is added to the n-tuple.
    The new column's name can be specified via the column_name parameter, which defaults to 'decay_string'.

    :param ntuple_file_path: Path of n-tuple ROOT file to be loaded.
    :param decay_tree_file_path: Path of decay tree ROOT file to be loaded.
                                 Can be None in which case no decay tree information will be added to
                                 the n-tuple. Default: None.
    :param column_name: Optional name for the new column for the decay
                        tree information. Default: 'decay_string'.
    :param reduce_size: Flag whether to reduce size of the DataFrame by casting float64
                        to float32 if possible. Default: True.
    :param columns: Columns (=Branches) to be read from the ROOT tree.
                    Default: None (= all columns).
    :param drop_list: List of variable names to drop. Default: None.
    :param cut_function: Function to be applied on the DataFrame loaded
                         from ROOT. The function must have one parameter,
                         which is a DataFrame and no return value. Default = None.
    :return: Pandas data frame containing the n-tuple with the decay tree
    information added, if available.
    """
    if columns is not None and drop_list is not None:
        raise ValueError("The parameters columns and drop_list can not be used at the same time!")

    ntuple_df = load_ntuple_from_root(ntuple_file_path, reduce_size, columns, drop_list, cut_function)

    if decay_tree_file_path is not None:
        decay_tree = load_decay_tree_from_root(decay_tree_file_path)
        add_decay_tree_to_ntuple(ntuple_df, decay_tree, column_name=column_name)

    return ntuple_df


def load_ntuple_from_root(
    file_path: PathType,
    reduce_size: bool,
    columns: Optional[List[str]] = None,
    drop_list: Optional[List[str]] = None,
    cut_function: Optional[Callable[[pd.DataFrame], None]] = None,
) -> pd.DataFrame:
    """
    Loads n-tuple from ROOT file into pandas data frame and drops given
    columns.

    :param file_path: Path to n-tuple ROOT file.
    :param reduce_size: Whether to cast float64 to float32 (if possible,
                        decided by pandas) in the DataFrames.
    :param columns: Columns (=Branches) to be read from the ROOT tree.
                    Default: None (= all columns).
    :param drop_list: List of variable names to drop. Default: None.
    :param cut_function: Function to be applied on the DataFrame loaded
                         from ROOT. The function must have one parameter,
                         which is a DataFrame and no return value. Default = None.
    :return: Pandas data frame containing the n-tuple information.
    """
    if columns is not None and drop_list is not None:
        raise ValueError("The parameters columns and drop_list can not be used at the same time!")

    print(f"  Loading {file_path}")

    df = load_root_file(file_path, columns=columns, drop_list=drop_list, cut_function=cut_function)

    if reduce_size:
        df = df.apply(pd.to_numeric, downcast="float")

    return df


def load_decay_tree_from_root(file_path: PathType):
    """
    Loads a decay basf2 tree object from a given ROOT file.

    :param file_path: Path to the decay tree ROOT file.
    :return: The decay tree object.
    """
    if not os.path.isfile(file_path):
        raise ValueError(f"The file\n{file_path}\ndoes not exist!")

    print(f"  Loading decay tree file {file_path}")

    return decayHash.DecayHashMap(rootfile=file_path, removeRadiativeGammaFlag=True)


def add_decay_tree_to_ntuple(
    ntuple_df: pd.DataFrame,
    decay_tree,
    column_name: str = "decay_string",
    full: bool = False,
) -> None:
    """
    Adds the decay tree information to the given n-tuple data frame.

    By matching the decay tree hashes the information stored into the separate decay tree root file
    will be added to a given n-tuple data frame as a new column.
    The data frame must contain the respective decay hashes.
    The name of the new columns for the decay tree can be specified. By default these are 'decay_string' and
    then 'decay_string_mc' and 'decay_string_reco' for the decay string containing only the original MC decay
    information and the actually reconstructed decay, respectively.
    The whole decay string will only be added to the data frame if the argument full is set to True. This is by
    default set to False, as a data frame containing the whole decay string can sometimes not be converted to hdf5.

    :param ntuple_df: The data frame to which the decay tree column will
                      be added.
    :param decay_tree: The decay tree object to add to the data frame.
    :param column_name: Optional name for the new columns for the decay
                        tree information. Default: 'decay_string'.
    :param full: Whether to add the full decay string or only the MC decay info
                 and what was reconstructed to the data frame. Default: False.
    :return: None
    """
    decay_hash1 = "extraInfo__boDecayHash__bc"
    decay_hash2 = "extraInfo__boDecayHashExtended__bc"

    def _get_full_decay_string(x):
        return decay_tree.get_string(x[0], x[1])

    def _get_mc_decay_string(x):
        return decay_tree.get_original_decay(x[0], x[1]).to_string()

    def _get_reco_decay_string(x):
        return decay_tree.get_reconstructed_decay(x[0], x[1]).to_string()

    print("  Adding " + (" full " if full else "") + "decay tree to n-tuple.")

    if full:
        ntuple_df[column_name] = ntuple_df[[decay_hash1, decay_hash2]].apply(_get_full_decay_string, axis=1)

    ntuple_df[column_name + "_mc"] = ntuple_df[[decay_hash1, decay_hash2]].apply(_get_mc_decay_string, axis=1)
    ntuple_df[column_name + "_reco"] = ntuple_df[[decay_hash1, decay_hash2]].apply(_get_reco_decay_string, axis=1)


def is_key_with_stream(key) -> bool:
    """
    Checks if a df_dict key is of a data type that can be available in multiple streams.

    :param key: Key of a data type in a data frame dictionary.
    :return: Boolean indicating if streams are available.
    """
    split_key = key.rsplit("_", 1)
    if len(split_key) > 1:
        return key.rsplit("_", 1)[1].isdigit()
    return False


def remove_stream_from_key(key: str) -> str:
    """
    Function to remove the stream number suffix from a data frame dictionary
    data type key.

    :param key: The data frame dict key to be modified.
    :return: The new data frame dict key string.
    """
    if is_key_with_stream(key):
        return key.rsplit("_", 1)[0]
    return key


def get_stream_from_key(key: str) -> Optional[int]:
    """
    Returns the stream number encoded in a data frame dictionary data type key
    as integer, and None, if no stream number is encoded.

    :param key: The data frame dictionary data type key to be evaluated.
    :return: Stream number as int, or None
    """
    if is_key_with_stream(key):
        return int(key.rsplit("_", 1)[1])
    return None


def load_root_file(
    file_path: PathType,
    columns: Optional[List[str]] = None,
    drop_list: Optional[List[str]] = None,
    cut_function: Optional[Callable[[pd.DataFrame], None]] = None,
    use_uproot: bool = False,
) -> pd.DataFrame:
    """
    Loads the first tree in the given ROOT file and returns the data contained in this tree
    as Pandas data frame.
    If the data frame contains a column name multiple time, the duplicates are removed.
    :param file_path: Path to the file which shall be loaded.
    :param columns: Columns (=Branches) to be read from the ROOT tree. Default: None (= all columns).
    :param drop_list: List of variable names to drop. Default: None.
    :param cut_function: Function to be applied on the DataFrame loaded
                         from ROOT. The function must have one parameter,
                         which is a DataFrame and no return value. Default = None.
    :param use_uproot: Whether to use uproot instead of root_pandas to read root file. Default = False.
    :return: The data of the loaded ROOT file as Pandas data frame.
    """
    if columns is not None and drop_list is not None:
        raise ValueError("The parameters columns and drop_list can not be used at the same time!")
    if not os.path.isfile(file_path):
        raise ValueError(f"The file\n{file_path}\ndoes not exist!")

    tree = None
    if columns is not None or drop_list is not None or use_uproot:
        afile = uproot.open(file_path)
        tree = afile[afile.keys()[0]]
        available_branches = list(set(tree.keys()))  # type: List[str]

        if drop_list is None and columns is None:
            branches = None  # type: Optional[List[str]]
        elif drop_list is not None and columns is None:
            branches = [c for c in list(set(drop_list)) if c not in available_branches]
        elif columns is not None:
            branches = [c for c in list(set(columns)) if c in available_branches]
        else:
            raise RuntimeError("The parameters columns and drop_list can not be used at the same time!")
    else:
        branches = None

    if use_uproot:
        assert tree is not None
        original_df = uproottree2df(tree=tree, branches=branches)
    else:
        original_df = root_pandas.read_root(paths=file_path, columns=branches)

    if cut_function is not None:
        n_before = len(original_df.index)
        cut_function(original_df)
        n_after = len(original_df.index)
        quotient = n_after / n_before
        print(f"  Applied cut: events after/before = {n_after} / {n_before} = {quotient:.4f}")

    if any(original_df.columns.duplicated()):
        reduced_df = original_df.loc[:, ~original_df.columns.duplicated()]
        del original_df
        return reduced_df
    return original_df


def save_to_hdf5(
    ntuple_df: pd.DataFrame,
    target_hdf5_file: PathType,
    hdf5_key: str,
    decay_tree_key: Optional[str] = None,
) -> None:
    """
    Stores data frames into hdf5 store. If necessary, the decay strings are stored
    under a separate key in the store.

    The given n-tuple data frame is stored via the to_hdf method to a hdf5 store
    with the specified path target_hdf5_file and under the provided key hdf5_key.

    If the to be stored data frame contains decay tree information, the optional
    argument decay_tree_key should be provided. In this case, the data frame will
    be split into two parts by trying to store the decay tree information under
    a separate key and in the table format, if it is not possible, to store the
    whole information as one part in the fixed format. The decay tree information
    will then be stored under the additional key <hdf5_key>_decay_trees.

    Use the function load_hdf5 to retrieve the data frame again as a whole.

    :param ntuple_df: Pandas data frame to be stored.
    :param target_hdf5_file: Target path for the hdf5 store.
    :param hdf5_key: Key under which the data frame will be stored in the hdf5 store.
    :param decay_tree_key: A possible decay tree key for the decay info included
                           in the data frame. Default: None.
    :return: None
    """
    if decay_tree_key is None:
        ntuple_df.to_hdf(path_or_buf=target_hdf5_file, key=hdf5_key)
    else:
        temp_target = os.path.join(os.path.dirname(target_hdf5_file), f"temp_{os.path.basename(target_hdf5_file)}")
        try:
            ntuple_df.to_hdf(path_or_buf=temp_target, key=hdf5_key)
        except OverflowError:
            print("OverflowError exception occurred. Trying to split n-tuple data frame before saving to hdf5...")
            if os.path.isfile(temp_target):
                os.remove(temp_target)
            ntuple_df_base = ntuple_df.drop(columns=["decay_string_mc", "decay_string_reco"])
            ntuple_df_base.to_hdf(path_or_buf=temp_target, key=hdf5_key)
            # Note: One could also try to only save the whole file in the table format, if exception occurs...
            ntuple_df[["decay_string_mc", "decay_string_reco", "unique_event_id"]].to_hdf(
                path_or_buf=temp_target,
                key=f"{hdf5_key}_decay_trees",
                format="table",
            )

        os.rename(temp_target, target_hdf5_file)


def load_hdf5(hdf5_file_path: PathType, hdf5_key: str) -> pd.DataFrame:
    """
    Loads data frames from hdf5 stores under consideration of a possible split into two parts
    due to the decay strings.

    This function loads data frames from hdf5 stores as read_hdf would do.
    Additionally, this function checks if the decay tree information is stored separately
    in the hdf5 store and retrieved it as well, if this is the case.
    The data frames will then be joined based on the unique_event_id column. The unique_event_id
    column of the decay_tree part will be added again with the suffix '_decay_tree' to allow
    for checks.

    :param hdf5_file_path: Path to the hdf5 store.
    :param hdf5_key: Key of the data to be retrieved from the hdf5 store.
    :return: Pandas Dataframe.
    """
    with pd.HDFStore(hdf5_file_path, mode="r") as hdf_store:
        hdf5_keys = hdf_store.keys()
        if len(hdf5_keys) == 1 and hdf5_key in hdf_store:
            return hdf_store[hdf5_key]
        elif len(hdf5_keys) == 2 and hdf5_key in hdf_store and f"{hdf5_key}_decay_trees" in hdf_store:
            df_base = hdf_store[hdf5_key]
            df_decay_trees = hdf_store[f"{hdf5_key}_decay_trees"]
            return df_base.join(df_decay_trees, how="left", rsuffix="_decay_tree")
        else:
            raise LookupError(
                f"An error occurred when trying to load the data from\n"
                f"{hdf5_file_path}\n"
                f"The hdf5 store contained the following keys:\n{hdf5_keys}\n"
                f"and you were looking for the key {hdf5_key}"
            )
