"""
K Short correction for Belle

Creates table with variations for the K Short correction weights from the original table.

Felix Metzner 2019
"""

import os
import copy
import logging
import numpy as np
import pandas as pd

from typing import Tuple, NamedTuple

from analysistools.utilities import PathType

logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = [
    "KShortVariationsCreator",
]


class KShortColumns(NamedTuple):
    mom_lower: str = "mom_lower"
    mom_upper: str = "mom_upper"
    ratio: str = "ratio"
    err_stat: str = "err_stat"
    err_sys: str = "err_sys"

    @staticmethod
    def defaults() -> Tuple[str, ...]:
        default_k_short_columns = KShortColumns()
        defaults = [getattr(default_k_short_columns, name) for name in default_k_short_columns.__annotations__.keys()]
        assert all(isinstance(d, str) for d in defaults)
        return tuple(defaults)


class KShortVariationsCreator:
    own_base_path = os.path.dirname(os.path.abspath(__file__))  # type: PathType
    original_table_file_name = "k_short_mom_table.csv"  # type: str
    original_table_path = os.path.join(own_base_path, original_table_file_name)  # type: PathType

    new_table_file_name = "varied_k_short_mom_table.csv"  # type: str
    new_table_file_path = os.path.join(own_base_path, new_table_file_name)  # type: PathType

    original_columns = KShortColumns.defaults()  # type: Tuple[str, ...]
    momentum_bins = (0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 100.0)  # type: Tuple[float, ...]

    var_col_prefix = "variation_"  # type: str
    var_precision = 4  # type: int

    def __init__(self) -> None:
        assert os.path.isfile(self.original_table_path), f"Path {self.original_table_path} does not exist..."
        self.column_names = KShortColumns()

    def produce_variation_tables(self, overwrite: bool = False) -> Tuple[pd.DataFrame, pd.DataFrame]:
        original_table = pd.read_csv(filepath_or_buffer=self.original_table_path, header=1, delimiter=",")
        assert len(original_table.columns) == len(self.original_columns), (original_table.columns, self.original_columns)
        assert all(ec in c for ec, c in zip(self.original_columns, original_table.columns)), (
            self.original_columns,
            original_table.columns,
        )

        new_column_names_dict = {oc: nc for oc, nc in zip(original_table.columns, self.original_columns)}
        original_table.rename(columns=new_column_names_dict, inplace=True)
        self._check_momentum_bins(table=original_table)

        new_table = copy.copy(original_table)

        cov = self._creat_covariance_matrix(table=original_table)
        eigen_values, eigen_vectors = np.linalg.eig(cov)
        fluctuation = np.sqrt(eigen_values)
        p_inv = np.linalg.inv(eigen_vectors)
        rho_tilde = np.matmul(original_table[self.ratio_col].values, p_inv)

        table_size = len(original_table.index)
        for n in range(2 * table_size):
            rho_tilde_temp = copy.copy(rho_tilde)
            weight_name = self._get_variation_col_name(index=n)

            if n < table_size:
                # Calculate upward fluctuation
                rho_tilde_temp[n] += fluctuation[n]
            else:
                # Calculate downward fluctuation
                rho_tilde_temp[n - table_size] -= fluctuation[n - table_size]

            # Transform to original basis and append to new table
            rho_prime = np.matmul(rho_tilde_temp, eigen_vectors)
            new_table.loc[:, weight_name] = rho_prime

        if os.path.isfile(self.new_table_file_path) and not overwrite:
            logging.warning(
                f"The csv table containing the K-Short variations already exists under\n"
                f"\t{self.new_table_file_path}\n"
                f"and will not be overwritten, but compared to the newly created one."
            )
            self._compare_to_existing_table(table=new_table)
        else:
            new_table.to_csv(
                filepath_or_buffer=self.new_table_file_path,
                sep=",",
                header=False,
                index=False,
                float_format=f"%.{self.var_precision}f",
            )

        return original_table, new_table

    def _creat_covariance_matrix(self, table: pd.DataFrame) -> np.ndarray:
        dim = len(table.index)
        cov = np.zeros(shape=(dim, dim))
        for row in range(dim):
            for col in range(dim):
                if row == col:
                    cov[row][col] = self._get_total_error_squared(table=table, index=row)
                else:
                    cov[row][col] = table.iloc[row][self.sys_err_col] * table.iloc[col][self.sys_err_col]

        assert self._check_symmetric(matrix=cov)
        return cov

    def _get_total_error_squared(self, table: pd.DataFrame, index: int) -> float:
        return table.iloc[index][self.stat_err_col] ** 2 + table.iloc[index][self.sys_err_col] ** 2

    @staticmethod
    def _check_symmetric(
        matrix: np.ndarray,
        rtol: float = 1e-05,
        atol: float = 1e-08,
    ) -> bool:
        return np.allclose(matrix, matrix.T, rtol=rtol, atol=atol)

    def _check_momentum_bins(self, table: pd.DataFrame) -> None:
        for row_index, (lower, upper) in enumerate(zip(self.momentum_bins[:-1], self.momentum_bins[1:])):
            row = table.iloc[row_index]
            assert row[self.l_mom_col] == lower, (row[self.l_mom_col], lower)
            assert row[self.u_mom_col] == upper, (row[self.u_mom_col], upper)

    @staticmethod
    def _get_variation_col_name(index: int) -> str:
        return f"{KShortVariationsCreator.var_col_prefix}_{index + 1}"

    def _compare_to_existing_table(self, table: pd.DataFrame) -> None:
        existing_table = pd.read_csv(filepath_or_buffer=self.new_table_file_path, header=None, delimiter=",")
        assert len(existing_table.columns) == len(table.columns), (existing_table.columns, table.columns)
        existing_table.columns = table.columns
        col_precision_mapping = {c: self.var_precision for c in table.columns if c.startswith(self.var_col_prefix)}
        new_rounded_table = table.round(decimals=col_precision_mapping)
        if not existing_table.equals(new_rounded_table):
            logging.warning(
                f"The existing table and the newly created table are not equal:\n\n"
                f"Existing table\n{existing_table}\n\nNewly created table {table}\n"
            )
        else:
            logging.warning("=> The comparison did not show any differences, so the tables are equal")

    @property
    def ratio_col(self) -> str:
        return self.column_names.ratio

    @property
    def sys_err_col(self) -> str:
        return self.column_names.err_sys

    @property
    def stat_err_col(self) -> str:
        return self.column_names.err_stat

    @property
    def u_mom_col(self) -> str:
        return self.column_names.mom_upper

    @property
    def l_mom_col(self) -> str:
        return self.column_names.mom_lower


if __name__ == "__main__":
    k_short_var_creator = KShortVariationsCreator()
    k_short_var_creator.produce_variation_tables()
