"""
K Short correction for Belle

Calculates correction weights as well as variation of such for K shorts.

Felix Metzner 2019
"""

import os
import numpy as np
import pandas as pd

from typing import List, Dict

from analysistools.utilities import PathType
from analysistools.pid_corrections.table_manager import TableManager


__all__ = [
    "KShortCorrection",
]


_k_short_table_manager = TableManager.instance()  # type: ignore
k_short_table_manager = _k_short_table_manager  # type: TableManager


class KShortCorrection:
    number_of_variations = 14  # type: int

    def __init__(self, exp_column_name: str) -> None:
        self.exp_column_name = exp_column_name  # type: str

        self.name = "name"  # type: str
        self.pdg = "pdg_col"  # type: str
        self.mom = "mom_col"  # type: str

        self.k_short_pdg = 310  # type: int

        self.weight_name = "k_short"  # type: str

        self.table_name = "varied_k_short_mom_table"  # type: str
        self.base_path = os.path.dirname(os.path.abspath(__file__))  # type: PathType

        self.columns = self.get_k_short_table_columns()  # type: List[str]

        self.momentum_column = "momentum_lower"  # type: str
        self.momentum_bins = self.get_momentum_bins()  # type: List[float]

        self.correction_info = {}  # type: Dict[int, Dict[str, str]]

    def add_track(
        self,
        name: str,
        pdg_column: str,
        momentum_column: str,
        check_names: bool = True,
    ) -> None:
        is_new = self.check_track(
            name=name,
            pdg_column=pdg_column,
            momentum_column=momentum_column,
            check_names=check_names,
        )
        if not is_new:
            raise RuntimeError("Track info was already added. Was this intended?")
        index = len(self.correction_info) + 1  # type: int
        self.correction_info.update(
            {
                index: {
                    self.name: name,
                    self.pdg: pdg_column,
                    self.mom: momentum_column,
                }
            }
        )

    def get_corrections(self, df: pd.DataFrame) -> pd.DataFrame:
        self.check_columns(columns=df.columns)
        sdf = df.loc[:, self.get_columns()]

        k_short_corr = pd.DataFrame({c: np.ones(len(sdf.index)) for c in self.get_correction_cols()}, index=sdf.index)

        for track_info in self.correction_info.values():
            pdg_cond = abs(sdf[track_info[self.pdg]]) == self.k_short_pdg

            df_with_ks = sdf.loc[pdg_cond]
            if not len(df_with_ks) > 0:
                continue
            corr_res = self._calculate_corr_for_track(df=df_with_ks, df_mom_col=track_info[self.mom])
            assert len(df_with_ks.index) == len(corr_res.index)

            for col in k_short_corr.columns:
                k_short_corr.loc[corr_res.index, col] *= corr_res[col]

        assert not pd.isnull(k_short_corr).values.any()
        assert len(k_short_corr) == len(df.index)

        k_short_corr.rename(columns={c: self.get_weight_column_name(c) for c in k_short_corr.columns}, inplace=True)
        return k_short_corr

    def _calculate_corr_for_track(
        self,
        df: pd.DataFrame,
        df_mom_col: str,
    ) -> pd.DataFrame:
        if not len(df.index) > 0:
            raise RuntimeError("The dataframe you are passing to get_slow_pion_correction is empty. " "Is this intended?")

        assert df[df_mom_col].min() >= 0.0, df[df_mom_col].min()
        assert df[df_mom_col].max() <= self.momentum_bins[-1], df[df_mom_col].max()

        res_df = pd.DataFrame(df[df_mom_col])
        res_column_names = []  # type: List[str]

        for correction_col in self.get_correction_cols():
            corr_res_column_name = correction_col
            res_column_names.append(corr_res_column_name)
            res_df.loc[:, corr_res_column_name] = np.nan

            corr_matrix = self._get_correction_matrix(correction_col=correction_col)

            corr_for_exps = self._get_k_short_correction(
                df=df[[df_mom_col]], df_mom_col=df_mom_col, correction_matrix=corr_matrix
            )

            res_df.loc[:, corr_res_column_name] = corr_for_exps

            assert not res_df.isnull().values.any()

        return res_df[res_column_names]

    def _get_k_short_correction(
        self,
        df: pd.DataFrame,
        df_mom_col: str,
        correction_matrix: np.ndarray,
    ) -> np.ndarray:
        mom_indices = np.digitize(df[df_mom_col], self.momentum_bins, right=True)
        assert (0 not in mom_indices) and (len(self.momentum_bins) not in mom_indices)
        mom_indices -= 1

        return correction_matrix[mom_indices]

    def _get_correction_matrix(self, correction_col: str) -> np.ndarray:
        matrix_name = f"{self.table_name}_{correction_col}"  # type: str
        matrix = k_short_table_manager.get_correction_matrix(name=matrix_name)  # type: np.ndarray

        if matrix is None:
            table = self._get_k_short_table()
            matrix = self._build_correction_matrix(table=table, correction_col=correction_col)
            k_short_table_manager.add_correction_matrix(name=matrix_name, new_matrix=matrix)

        return matrix

    def _get_k_short_table(self) -> pd.DataFrame:
        table_path = os.path.join(self.base_path, f"{self.table_name}.csv")  # type: PathType

        return k_short_table_manager.get_table(name=self.table_name, path=table_path, column_names=self.columns)

    def _build_correction_matrix(
        self,
        table: pd.DataFrame,
        correction_col: str,
    ) -> np.ndarray:
        assert self.momentum_column in table.columns
        assert correction_col in table.columns

        # Ensure that momentum is unique in table:
        assert not (table.groupby(self.momentum_column).size() != 1).any()

        matrix = np.full(len(self.momentum_bins[:-1]), np.nan)  # type: np.ndarray

        for mom_index, mom_bin in enumerate(self.momentum_bins[:-1]):
            value_rows = table.loc[(table[self.momentum_column] == mom_bin)]
            assert len(value_rows) == 1
            assert len(value_rows[correction_col]) == 1

            matrix[mom_index] = value_rows[correction_col].values[0]

        assert not np.isnan(matrix).any()
        return matrix

    def get_columns(self) -> List[str]:
        columns = [self.exp_column_name]
        for track_info in self.correction_info.values():
            columns.append(track_info[self.pdg])
            columns.append(track_info[self.mom])
        return columns

    def check_columns(self, columns: List[str]) -> None:
        for column in self.get_columns():
            assert column in columns, f"{column}"

    def check_track(
        self,
        name: str,
        pdg_column: str,
        momentum_column: str,
        check_names: bool,
    ) -> bool:
        if check_names:
            self.check_column_names(pdg=pdg_column, mom=momentum_column)

        for track_info in self.correction_info.values():
            if track_info[self.name] == name:
                return False
            if track_info[self.pdg] == pdg_column:
                return False
            if track_info[self.mom] == momentum_column:
                return False

        return True

    @staticmethod
    def check_column_names(pdg: str, mom: str) -> None:
        pdg_parts = pdg.split("daughter__bo")
        mom_parts = mom.split("daughter__bo")

        if not len(pdg_parts) == len(mom_parts):
            raise RuntimeError(f"Column names do not match:\n {pdg}\n {mom}")

        pdg_daughters = [int(p[0]) for p in pdg_parts[1:] if p[0].isdigit()]
        mom_daughters = [int(p[0]) for p in mom_parts[1:] if p[0].isdigit()]

        if not len(pdg_daughters) == len(mom_daughters):
            raise RuntimeError(f"Column names do not match:\n {pdg}\n {mom}")

        for a, b in zip(pdg_daughters, mom_daughters):
            if not a == b:
                raise RuntimeError(f"Column names do not match:\n {pdg}\n {mom}")

    @staticmethod
    def get_k_short_table_columns() -> List[str]:
        cols = [
            "momentum_lower",
            "momentum_upper",
            "ratio",
            "stat",
            "sys",
        ]  # type: List[str]

        cols += [f"var_{i + 1}" for i in range(KShortCorrection.number_of_variations)]

        return cols

    @staticmethod
    def get_correction_cols() -> List[str]:
        cols = ["ratio"]
        cols += [f"var_{i + 1}" for i in range(KShortCorrection.number_of_variations)]
        return cols

    def get_correction_col_names(self) -> List[str]:
        return [self.get_weight_column_name(c) for c in self.get_correction_cols()]

    @staticmethod
    def get_momentum_bins() -> List[float]:
        return [
            0.0,
            0.5,
            1.0,
            1.5,
            2.0,
            2.5,
            3.0,
            100.0,
        ]

    def get_weight_column_name(self, column_name: str) -> str:
        return f"__weight_{self.weight_name}_{column_name}__"
