"""
This module allows for a quick overview of the properties of variable.
First, you can quickly check the correlation of all your observables/variables with any other
variable (mostly your final fitting variable, e.g. m2RecoilSignalSide) with the correlation_analysis function

Second, the function cut_analysis scans a given variable and calculates the optimal right- and left-sided cut
and returns the resulting sqrt(signal / signal+bkg) in a given window in your fitting variable.
If you pass a filename, it will save the plot there, otherwise it will show the plot.



All functions need the "signal" and "background" to be handed over in a certain format: (dataframe-set)

data_dict = OrderedDict([
    'dataframe_name(e.g. decay mode)': pd.DataFrame,
    ...
])



MY USECASE:
# define a list of all the variables you want to do the cut analysis for
variable_list = [
    ('daughter__bo0__cm__spextraInfo__boSignalProbability__bc__bc',[0,0.1]),
]

# loop over them and do the analysis
for variable,plot_window in variable_list:
    cut_analysis_utils.cut_analysis(signal_dfs, bkg_dfs, variable, n_steps=1000.,
                              plot_window=plot_window,
                              mass_window = [-0.2,0.2],
                              # filename = variable + '.pdf',
                              )

    # then in addition plot the fitting_variable with and without the cut to check shape,etc.

Gabriel Seitz 2018
"""

import os
import numpy as np
import pandas as pd
from typing import Dict, Tuple, List, Optional

from ROOT import gSystem
from ROOT import Belle2

import matplotlib.pyplot as plt
from analysistools.dataframe_tools import dataframe_tools as dft


def correlation_analysis(
    signal_dfs: Dict[str, pd.DataFrame],
    bkg_dfs: Dict[str, pd.DataFrame] = None,
    filename: str = None,
    comp_variable: str = "m2RecoilSignalSide",
    exception_list: Optional[List[str]] = None,
    comp_list: Optional[List[str]] = None,
) -> None:
    """
    calculates the correlation between all variables and the given comparison variable.
    if you have strings stored in your dataframe (like the decaystring) then you should add those variables
    to the exception list.

    if a filename is given it will try and store the values there, otherwise it will print it out
    :param signal_dfs: signal dataframe-set
    :param bkg_dfs: bkg dataframe-set (optional)
    :param filename: full filename to store the values in e.g. './correlation.txt'
    :param comp_variable: default 'm2RecoilSignalSide', but you can give any variable if you are interested in
                          other correlations
    :param exception_list: list of variables that will not be taken into account
    :param comp_list: list of variables you want to do the calculation with.
                      NONE means all variables in the df will be taken
    :return: None
    """
    if exception_list is None:
        _exception_list = []  # type: List[str]
    else:
        _exception_list = exception_list

    full_dfs_all_data = {}
    full_dfs_all_data.update(signal_dfs)
    if bkg_dfs:
        full_dfs_all_data.update(bkg_dfs)
    table, format_str = correlation_to_variable(
        full_dfs_all_data,
        comp_variable=comp_variable,
        exception_list=_exception_list,
        comp_list=comp_list,
    )

    if filename:
        if not os.path.isfile(filename):
            raise ValueError(
                f"The filename you provided for the output is not valid!" f"You provided the path: {filename}"
            )
        else:
            f = open(filename, "w")
            for row in table:
                f.write(format_str.format(*row))
            f.close()
        return
    for row in table:
        print(format_str.format(*row))
    return


def correlation_to_variable(
    full_dfs: Dict[str, pd.DataFrame],
    comp_variable: str = "m2RecoilSignalSide",
    exception_list: Optional[List] = None,
    comp_list: Optional[List[str]] = None,
) -> Tuple[List, str]:
    """
    calculates the correlation of each variable in a dataframe compared to the given variable.
    :param full_dfs: dataframe-set
    :param comp_variable: variable to compare with
    :param exception_list: list of variables that will not be taken into account
    :param comp_list: list of variables you want to do the calculation with.
                      NONE means all variables in the df will be taken
    :return: table of correlation coefficients and the format_string to allow for pretty printing
    """
    if exception_list is None:
        _exception_list = []  # type: List[str]
    else:
        _exception_list = exception_list

    column_headers = ["variable"]
    column_headers += [df_name for df_name in full_dfs.keys()]
    table_as_list = [column_headers]

    if not comp_list:
        _comp_list = list(full_dfs[column_headers[1]])  # type: List[str]
    else:
        _comp_list = comp_list

    for variable in _comp_list:
        if variable in _exception_list:
            continue
        correlation_row = [variable]
        for df in full_dfs.values():
            if type(df[comp_variable]) == str:
                raise TypeError(
                    f"values in the column {comp_variable} have type string! "
                    f"Can not calculate correlations of non-numeric values."
                )
            correlation_row.append("{:.2f}".format(df[variable].corr(df[comp_variable])))
        table_as_list.append(correlation_row)

    column_size = [max(map(len, map(str, col))) for col in zip(*table_as_list)]
    format_str = " | ".join("{{:<{i}}}".format(i=i) for i in column_size)
    table_as_list.insert(1, ["-" * i for i in column_size])

    return table_as_list, format_str


def cut_analysis(
    signal_dfs: Dict[str, pd.DataFrame],
    bkg_dfs: Dict[str, pd.DataFrame],
    variable: str,
    fitting_variable: str = "m2RecoilSignalSide",
    mass_window=(-0.5, 0.5),
    n_steps: int = 100,
    plot_window: Tuple[float, float] = None,
    filename: str = None,
    only_plot: str = None,
    logscale: bool = False,
) -> None:
    """
    main tool function

    scans the give variable and calculates the sqrt(sig/sig+bkg) at each step, if you would apply a one-sided cut there.
    then collects these information in a table
    also plots the signal efficiency & purity for the whole range of possible cuts on that variable
    :param signal_dfs: dataframe-set with signal
    :param bkg_dfs: dataframe-set with bkg
    :param variable: variable of interest
    :param fitting_variable: variable, in which the signal to bkg ratio will be calculated
    :param mass_window: window in the missing mass where the s/s+b will be calculated
    :param n_steps: resolution of the analysis
    :param plot_window: [min,max] of where to plot the information, none means full variable-space
    :param filename: optional: if a filename is given, the plots are saved under that path and not shown on screen
    :param only_plot: give 'left' or 'right' if only a one-sided cut should be analysed
    :param logscale: True if x-axis should be log-scaled
    :return:
    """
    if not plot_window:
        min_val = dft.get_min(signal_dfs, variable)
        max_val = dft.get_max(signal_dfs, variable)
    else:
        min_val, max_val = plot_window
    stepsize = 1.0 * (max_val - min_val) / n_steps

    if stepsize == 0:
        print(f"error in variable: {variable}")
        print(f"unreasonable binning: {min_val}, {max_val}, {stepsize}")
        print("skipped analysis for this variable")
        return

    if logscale is True:
        if min_val == 0:
            raise ValueError(f"pls provide a plot-window if you use logscaleing for x-axis. variable = {variable}")
        x_array = np.logspace(np.log10(min_val), np.log10(max_val), n_steps)
    else:
        x_array = np.arange(min_val, max_val, stepsize)

    left, left_signal = onesided_sig_bkg_ratio(
        x_array, signal_dfs, bkg_dfs, variable, fitting_variable, mass_window, side="left"
    )
    right, right_signal = onesided_sig_bkg_ratio(
        x_array, signal_dfs, bkg_dfs, variable, fitting_variable, mass_window, side="right"
    )

    gSystem.Load("libanalysis.so")
    Belle2.Variable.Manager.Instance()
    label = str(Belle2.invertMakeROOTCompatible(variable))

    column_headers = [label]
    column_headers += ["left"]
    column_headers += ["right"]
    table_as_list = [column_headers]

    optimal_cut = ["optimal cut"]
    optimal_cut.append("{:.4f}".format(x_array[np.argmax(left)]))
    optimal_cut.append("{:.4f}".format(x_array[np.argmax(right)]))

    signal_bkg_ratio = ["sqrt s/s+b"]
    signal_bkg_ratio.append("{:.3f}".format(np.max(left)))
    signal_bkg_ratio.append("{:.3f}".format(np.max(right)))

    remaining_signal = ["remaining signal %"]
    remaining_signal.append("{:.4f}".format(left_signal[np.argmax(left)]))
    remaining_signal.append("{:.4f}".format(right_signal[np.argmax(right)]))

    table_as_list.append(optimal_cut)
    table_as_list.append(signal_bkg_ratio)
    table_as_list.append(remaining_signal)

    column_size = [max(map(len, map(str, col))) for col in zip(*table_as_list)]
    format_str = " | ".join("{{:<{i}}}".format(i=i) for i in column_size)
    table_as_list.insert(1, ["-" * i for i in column_size])

    for row in table_as_list:
        print(format_str.format(*row))

    cut_analysis_plot(
        x_array,
        left,
        right,
        left_signal,
        right_signal,
        plot_window=plot_window,
        x_label=label,
        filename=filename,
        only_plot=only_plot,
        logscale=logscale,
    )


def onesided_sig_bkg_ratio(
    x_array: np.array,
    signal_dfs: Dict[str, pd.DataFrame],
    bkg_dfs: Dict[str, pd.DataFrame],
    variable: str,
    fitting_variable: str = "m2RecoilSignalSide",
    mass_window=(-0.5, 0.5),
    side: str = "left",
) -> Tuple[np.array, np.array]:
    """
    scans the give variable and calculates the sqrt(sig/sig+bkg) at each step, if you would apply a one-sided cut there.

    :param x_array: value-space for which the efficiencies should be calculated.
                    just hand over the same as you would use to plot the results afterwards
    :param signal_dfs: dataframe-set with signal channels
    :param bkg_dfs: dataframe-set with bkg channels
    :param variable: variable for which the cut efficiencies should be calculated
    :param fitting_variable: variable, in which the signal to bkg ratio is calculated
    :param mass_window: window of MissingMass in which the signal to background ratio will be calculated.
                        since this is you fit variable in the end you are not actually interested in the removal of
                        events outside the signal-relevant area.
    :param side: from which side the cut should be calculated. 'left' means that events whose values are smaller
                 than the cut (left of the cut) are removed.
    :return: two arrays
                first one contains signal to background ratio within the M_miss window depending on the cut
                second one contains the percentage of remaining signal events after the cut
    """
    signal = np.zeros(len(x_array))
    background = np.zeros(len(x_array))
    full_sig = 0

    for signal_df in signal_dfs.values():
        signal_array = np.array(signal_df[variable].values)
        signal_mass_array = np.array(signal_df[fitting_variable].values)
        signal_weight_array = np.array(signal_df["__weight__"].values)

        full_sig += np.sum(
            signal_weight_array[(signal_mass_array > mass_window[0]) & (signal_mass_array < mass_window[1])]
        )

        for index, x in enumerate(x_array):
            if side == "left":
                sig = np.sum(
                    signal_weight_array[
                        (signal_array > x) & (signal_mass_array > mass_window[0]) & (signal_mass_array < mass_window[1])
                    ]
                )
            else:
                sig = np.sum(
                    signal_weight_array[
                        (signal_array < x) & (signal_mass_array > mass_window[0]) & (signal_mass_array < mass_window[1])
                    ]
                )
            signal[index] += sig

    for bkg_df in bkg_dfs.values():
        bkg_array = np.array(bkg_df[variable].values)
        bkg_mass_array = np.array(bkg_df[fitting_variable].values)
        bkg_weight_array = np.array(bkg_df["__weight__"].values)

        for index, x in enumerate(x_array):
            if (signal[index]) == 0:
                bkg = 1
            elif side == "left":
                bkg = np.sum(
                    bkg_weight_array[
                        (bkg_array > x) & (bkg_mass_array > mass_window[0]) & (bkg_mass_array < mass_window[1])
                    ]
                )
            else:
                bkg = np.sum(
                    bkg_weight_array[
                        (bkg_array < x) & (bkg_mass_array > mass_window[0]) & (bkg_mass_array < mass_window[1])
                    ]
                )
            background[index] += bkg

    sig_bkg_ratio = (signal * ((signal + background) ** -1)) ** 0.5

    if full_sig == 0:
        full_sig = 1
    signal = signal / full_sig

    return sig_bkg_ratio, signal


def cut_analysis_plot(
    x_array: np.array,
    left: np.array,
    right: np.array,
    left_signal: np.array,
    right_signal: np.array,
    plot_window: Tuple[float, float] = None,
    x_label: str = "variable",
    filename: str = None,
    only_plot: str = None,
    logscale: bool = False,
) -> None:
    """
    plots the signal efficiency and signal purity for the cut analysis.
    red for right-sided cut
    blue for left-sided cut

    :param x_array: x_array
    :param left: left-sided signal purity
    :param right: right-sided signal purity
    :param left_signal: left-sided signal efficiency
    :param right_signal: right-sided signal efficiency
    :param plot_window: plot_window (None means full x-array)
    :param x_label: x_label
    :param filename: if a filename is given, then the plot will be stored there and not shown immediately
    :param only_plot: 'left' or 'right' to plot only one-sided cut
    :param logscale: True if x-axis should be log-scaled
    :return:
    """
    fig, ax1 = plt.subplots()
    if not only_plot == "right":
        ax1.plot(x_array, left, "b.", label="left-side s/s+b")
    if not only_plot == "left":
        ax1.plot(x_array, right, "r.", label="right-side s/s+b")

    if plot_window:
        ax1.set_xlim(plot_window)

    ax2 = ax1.twinx()
    if not only_plot == "right":
        ax2.plot(x_array, left_signal, "b-", label="left-side signal efficiency")
    if not only_plot == "left":
        ax2.plot(x_array, right_signal, "r-", label="right-side signal efficiency")

    fig.legend(frameon=False, bbox_to_anchor=(1.7, 1), bbox_transform=ax1.transAxes)

    ax1.grid(True)
    ax1.set_ylabel("signal purity")
    ax2.set_ylabel("signal eff")
    ax1.set_xlabel(x_label)

    if logscale:
        plt.xscale("log")

    if not filename:
        plt.show()
    else:
        fig.savefig(filename)
    plt.close()
