"""
Histogram class

Max Welsch 2018
"""

import numpy as np

from typing import Tuple, Optional


__all__ = [
    "Histogram",
]


class Histogram:
    """
    Simple histogram class for 1 dimensional data.
    """

    def __init__(
        self,
        n_bins: int,
        hist_range: Tuple[float, float],
        normed: bool = False,
        title: Optional[str] = None,
    ) -> None:
        """
        Constructor of histogram.

        :param n_bins: Number of bins.
        :param hist_range: Range that will be covered by the histogram. Values outside of the range won't be used.
        :param normed: Flag if the area of the histogram will be normalized to 1. Default: False.
        :param title: Histogram title.
        """
        self._n_bins = n_bins  # type: int
        self._bin_edges = np.linspace(hist_range[0], hist_range[1], n_bins + 1)  # type: np.ndarray
        self._range = hist_range  # type: Tuple[float,  float]
        self._normed = normed  # type: bool
        self._n_entries = 0  # type: int
        self.title = title  # type: Optional[str]

        self._bin_counts = np.zeros(self._n_bins)  # type: np.ndarray
        self._sum_w2 = np.zeros(self._n_bins)  # type: np.ndarray

    def _calc_sum_w2(
        self,
        data: np.ndarray,
        weights: np.ndarray,
    ) -> np.ndarray:
        """
        Calculates the sum of the weights squared in each bin.

        :param data: Data to be histogrammed.
        :param weights: Event weights for each entry in data. Should be of the same shape as data.
        :return: sum_w2
        """
        indices = np.digitize(data, self._bin_edges) - 1
        sum_w2 = np.zeros(self._n_bins)

        for b in range(self._n_bins):
            sum_w2[b] = np.sum(weights[indices == b] ** 2)

        return sum_w2

    def _calc_integral(self) -> np.ndarray:
        """
        Calculates the area of the histogram as sum(bin_count*bin_width).

        :return: Calculated value if the integral.
        """
        return np.sum(self._bin_counts * self.bin_width)

    def fill(
        self,
        data: np.ndarray,
        weights: Optional[np.ndarray] = None,
    ) -> None:
        """
        Fills the histogram with the given data. If no weights are provided,
        each events gets the weight 1. This method can be called several times
        and each time the entries in each bin are summed.

        :param data: Data to be histogramed.
        :type data: np.ndarray

        :param weights: Event weights for each entry in data. Should be of the
        same shape as param data.
        :type weights: np.ndarray
        """

        if weights is None:
            weights = np.ones(data.shape)

        bin_counts, _ = np.histogram(data, self._bin_edges, range=self._range, weights=weights)

        self._bin_counts += bin_counts
        self._sum_w2 += self._calc_sum_w2(data, weights)

        self._n_entries += len(data)

    @property
    def bin_width(self) -> np.ndarray:
        """
        Returns the bin width of each bin. If all bins have the same width, only
        a scalar will be returned. Otherwise a np.ndarray of shape (n_bins,)
        will be returned.

        :return: bin_width
        """
        bin_widths = self._bin_edges[1:] - self._bin_edges[:-1]
        if np.all(bin_widths == bin_widths[0]):
            return bin_widths[0]
        else:
            return bin_widths

    @property
    def bin_centers(self) -> np.ndarray:
        """
        Returns the centers of each bin as np.ndarray of shape (n_bins,).
        :return: bin_enters
        """
        return (self.bin_edges[1:] + self.bin_edges[:-1]) / 2

    @property
    def bin_edges(self) -> np.ndarray:
        """
        Returns the edges of each bin as np.ndarray of shape (n_bins + 1,).
        :return: bin_edges
        """
        return self._bin_edges

    @property
    def bin_counts(self) -> np.ndarray:
        """
        Returns the entries in each bin as np.ndarray of shape (n_bins,). If the
        histogram is normalized the normalized entries of the histogram are
        returned.
        :return: bin_counts
        """
        if self._normed:
            return self._bin_counts / self._calc_integral()
        else:
            return self._bin_counts

    @property
    def bin_errors(self) -> np.ndarray:
        """
        Returns the poisson errors in each bin as np.ndarray of shape (n_bins,).
        If the histogram is weighted, the sqrt(sum of the weights**2) is returned.
        If the histogram is normalized, the errors are scaled accordingly.
        :return: bin_errors
        """
        if self._normed:
            return np.sqrt(self._sum_w2) / self._calc_integral()
        else:
            return np.sqrt(self._sum_w2)
