"""
This module provides the methods to attach weights to data samples.

    The MC data set used for an analysis comprises several components, which
    might be taken from different sources, such as the collaboration wide
    MC production (Mixed, Charm, Charged, UDS, specialMC, rareMC) and privately
    produced MC.
    Depending on the source, a global weight has to be applied to the sample
    components. These are defined here.

Felix Metzner 2019
"""

import os
import copy
import numpy as np
import pandas as pd

from uncertainties import UFloat, nominal_value, ufloat
from typing import Optional, Union, Tuple, Dict, Callable

from analysistools.utilities import PathType
from analysistools.batch_tools.file_handler import get_n_events_from_file_dict


__all__ = [
    "belle_published_integrated_luminosity",
    "N_BB",
    "BR_GENERIC_MIXED",
    "BR_GENERIC_CHARGED",
    "FIXED_SPECIAL_WEIGHTS",
    "scale_weight_column_name",
    "expected_number_of_generic_MC",
    "off_to_on_resonance_scale_weight_column_name",
    "add_public_mc_weight",
    "add_private_mc_weight",
    "add_off_resonance_scaling_weight",
]


belle_published_integrated_luminosity = 711  # type: int  # See explanation below.
DATA_ON_RESONANCE = ufloat(702.623, 4.564)  # type: UFloat  # The usually cited 711 fb^-1 is not correct. See mattermost:
# M. Ziegler, 18th and 19th July 2016, belle2-analysis, posted answer from Nakao-san:
# Regarding the integrated luminosity, 703 fb-1 is the updated integrated
# luminosity after we have already used 711 fb-1 in publication. Since it
# is within error and most of the analysis we just use nBB, we decided not
# to correct it and use 711 fb-1 in publication. Those who use integrated
# luminosity in the analysis should be careful about it.

DATA_OFF_RESONANCE = ufloat(79.366, 0)  # type: UFloat
OFF_RESONANCE_UP_SCALING_FACTOR = DATA_ON_RESONANCE / DATA_OFF_RESONANCE  # type: UFloat
ON_RESONANCE_DOWN_SCALING_FACTOR = DATA_OFF_RESONANCE / DATA_ON_RESONANCE  # type: UFloat

N_BB = ufloat(771.581e6, 10.566e6)  # type: UFloat  # Source: https://belle.kek.jp/secured/nbb/nbb.html

BR_GENERIC_MIXED = ufloat(0.486, 0.006)  # type: UFloat
BR_GENERIC_CHARGED = ufloat(0.514, 0.006)  # type: UFloat

BR_CHARGEDRARE = 3.27945e-3  # type: float
BR_CHARGEDULNU = 3.670e-3  # type: float
BR_MIXEDRARE = 4.05703e-3  # type: float
BR_MIXEDULNU = 3.418e-3  # type: float

FIXED_SPECIAL_WEIGHTS = {
    "rareCharged": 50,
    "rareMixed": 50,
    "ulnuCharged": 20,
    "ulnuMixed": 20,
    "jpsiCharged": 100,
    "jpsiMixed": 100,
    "psi2sCharged": 100,
    "psi2sMixed": 100,
}  # type: Dict[str, int]

number_of_b_pairs_per_experiment = {  # in millions
    7: 6.4587,
    9: 4.7579,
    11: 8.8509,
    13: 11.6998,
    15: 13.5679,
    17: 12.4588,
    19: 27.1705,
    21: 4.3371,
    23: 6.4755,
    25: 28.0008,
    27: 28.1814,
    31: 19.6587,
    33: 19.3022,
    35: 18.5262,
    37: 67.1819,
    39: 47.0818,
    41: 64.0134,
    43: 61.5614,
    45: 14.3538,
    47: 41.2186,
    49: 29.7271,
    51: 41.8919,
    55: 80.2472,
    61: 37.4460,
    63: 35.6231,
    65: 41.7867,
}  # type: Dict[int, float]

scale_weight_column_name = "__scaling_weight__"  # type: str

expected_number_of_generic_MC = {
    "Charged": BR_GENERIC_CHARGED.nominal_value * N_BB.nominal_value,
    "Mixed": BR_GENERIC_MIXED.nominal_value * N_BB.nominal_value,
}  # type: Dict[str, float]

generic_charmless_types = (
    "rareCharged",
    "rareMixed",
    "ulnuCharged",
    "ulnuMixed",
)  # type: Tuple[str, ...]
assert all(dt in FIXED_SPECIAL_WEIGHTS.keys() for dt in generic_charmless_types), generic_charmless_types


def get_generic_mc_scaling_correction_factor(
    data_type: str,
    file_dict_dir_path: PathType,
) -> float:
    assert data_type in expected_number_of_generic_MC.keys(), (data_type, expected_number_of_generic_MC.keys())
    assert os.path.exists(file_dict_dir_path), file_dict_dir_path

    # Calculation is done with 10 streams, because we have 10 streams of generic MC in the dict file
    n_streams = 10.0  # type: float

    produced_number_of_events = get_n_events_from_file_dict(
        file_dict_path=os.path.join(file_dict_dir_path, f"evtgen_{data_type}/{data_type}.json")
    )  # type: float

    for special_data_type in generic_charmless_types:
        if special_data_type.endswith(data_type):
            spc_file_dict_path = os.path.join(file_dict_dir_path, f"special_{special_data_type}/{special_data_type}.json")
            spc_norm = FIXED_SPECIAL_WEIGHTS[special_data_type] / n_streams  # type: float
            produced_number_of_events += get_n_events_from_file_dict(file_dict_path=spc_file_dict_path) / spc_norm

    # 10 streams for Charged and Mixed
    expected_number_of_events = n_streams * expected_number_of_generic_MC[data_type]  # type: float
    mc_correction_factor = 1.0 * produced_number_of_events / expected_number_of_events  # type: float

    return mc_correction_factor


def add_public_mc_weight(
    df: pd.DataFrame,
    streams: float,
    data_type: str,
    file_dict_dir_path: Optional[PathType] = None,
) -> float:
    """
    Adds weight column for public MC samples, depending on how many streams of
    the MC type are used.

    The number of streams used can be specified via the parameter streams.
    This can also be a float below 1.0, if only a fraction of a stream is used.

    :param df: Pandas DataFrame to which the weight column is added.
    :param streams: Number or fraction of streams.
    :param data_type: The data type.
    :param file_dict_dir_path: The path to the directory containing the corresponding file dict
    :return: The calculated weight for this MC type.
    """
    assert streams > 0.0, streams
    if scale_weight_column_name in list(df.columns):
        raise RuntimeError("Provided data frame already has a component weight set!")

    mc_correction_factor = 1.0  # type: float
    if data_type in expected_number_of_generic_MC.keys():
        assert file_dict_dir_path is not None
        mc_correction_factor = get_generic_mc_scaling_correction_factor(
            data_type=data_type,
            file_dict_dir_path=file_dict_dir_path,
        )

    weight = 1.0 / (mc_correction_factor * streams)  # type: float
    df.loc[:, scale_weight_column_name] = weight

    return weight


def add_private_mc_weight(
    df: pd.DataFrame,
    produced_events: int,
    charged_B: bool,
    branching_fraction: Union[float, UFloat, None] = None,
    branching_fraction_formula: Optional[Callable] = None,
    *args,
    **kwargs,
) -> float:
    """
    Adds weight column for private MC samples, depending on the number of produced events
    and either a constant branch fraction value provided as float, or a callable function
    to calculate the branching fraction.

    Further arguments and keyword arguments are passed on to the callable.

    :param df: Pandas DataFrame to which the weight column is added.
    :param produced_events: Number of produced events for this MC component.
    :param charged_B: Whether the component is a charged B meson decay.
    :param branching_fraction: Branching fraction of this component.
    :param branching_fraction_formula: Callable to be used to calculate the branching fraction of this component.
    :param args: Arguments passed on to the branching_fraction_formula.
    :param kwargs: Keyword arguments passed on to the branching_fraction_formula.
    :return: The calculated weight for this MC type.
    """
    # Make sure that only either a branching fraction or a formula for one is provided, not both.
    if not ((branching_fraction is not None) != (branching_fraction_formula is not None)):
        raise ValueError(
            "Function add_private_mc_weight takes either a branching_fraction or "
            "a branching_fraction_formula, not both!"
        )

    if scale_weight_column_name in list(df.columns):
        raise RuntimeError("Provided data frame already has a component weight set!")

    if branching_fraction is not None:
        bf = (
            branching_fraction if isinstance(branching_fraction, float) else branching_fraction.nominal_value
        )  # type: float
    else:
        assert branching_fraction_formula is not None
        bf = branching_fraction_formula(*args, **kwargs)
    assert bf > 0.0

    if charged_B:
        y4s_to_b_branching_fraction = BR_GENERIC_CHARGED.nominal_value  # type: float
    else:
        y4s_to_b_branching_fraction = BR_GENERIC_MIXED.nominal_value

    weight = nominal_value(N_BB * 2 * y4s_to_b_branching_fraction * bf) * 1.0 / produced_events
    assert weight > 0.0

    df.loc[:, scale_weight_column_name] = weight

    return weight


off_to_on_resonance_scaling_plus_error_dict = {
    7: ufloat(9.830, 0.014),
    9: ufloat(2.516, 0.024),
    11: ufloat(4.6288, 0.0041),
    13: ufloat(4.3028, 0.0041),  # 13a (to Run 1199)
    14: ufloat(1.0458, 0.0013),  # 13b from Run 1200
    15: ufloat(4.2632, 0.0041),
    17: ufloat(3.6302, 0.0036),
    19: ufloat(9.4511, 0.0092),
    21: ufloat(1.8267, 0.0020),
    23: ufloat(2.6243, 0.0027),
    25: ufloat(3.9311, 0.0026),
    27: ufloat(3.8795, 0.0025),
    31: ufloat(7.32282, 0.00027),
    33: ufloat(6.34776, 0.00022),
    35: ufloat(8.46756, 0.00034),
    37: ufloat(9.95574, 0.00023),
    39: ufloat(6.79552, 0.00016),
    41: ufloat(10.37780, 0.00025),
    43: ufloat(8.59548, 0.00019),
    45: ufloat(5.54173, 0.00021),
    47: ufloat(10.74897, 0.00032),
    49: ufloat(10.31001, 0.00036),
    51: ufloat(7.93857, 0.00237),
    55: ufloat(9.24327, 0.00021),
    61: ufloat(13.6519, 0.0683),
    63: ufloat(6.4278, 0.0321),
    65: ufloat(8.6072, 0.0430),
}  # type: Dict[int, UFloat]

__off_to_on_resonance_scaling_plus_error_dict_v2_update = {
    31: ufloat(7.441, 0.052),
    33: ufloat(6.477, 0.036),
    35: ufloat(8.625, 0.048),
    37: ufloat(10.065, 0.042),
    39: ufloat(6.671, 0.028),
    41: ufloat(10.597, 0.037),
    43: ufloat(8.716, 0.024),
    45: ufloat(5.640, 0.015),
    47: ufloat(10.901, 0.03),
    49: ufloat(10.576, 0.044),
    51: ufloat(8.251, 0.069),
    55: ufloat(9.404, 0.026),
    61: ufloat(13.497, 0.047),
    63: ufloat(6.664, 0.037),
    65: ufloat(8.631, 0.024),
}  # type: Dict[int, UFloat]

__off_to_on_resonance_scaling_plus_error_dict_v2 = copy.copy(off_to_on_resonance_scaling_plus_error_dict)
__off_to_on_resonance_scaling_plus_error_dict_v2.update(__off_to_on_resonance_scaling_plus_error_dict_v2_update)
off_to_on_resonance_scaling_plus_error_dict_v2 = (
    __off_to_on_resonance_scaling_plus_error_dict_v2
)  # type: Dict[int, UFloat]

off_to_on_resonance_scaling_dict = {k: v.nominal_value for k, v in off_to_on_resonance_scaling_plus_error_dict.items()}
off_to_on_resonance_scaling_dict_v2 = {
    k: v.nominal_value for k, v in off_to_on_resonance_scaling_plus_error_dict_v2.items()
}  # type: Dict[int, float]


exp_13_split_info = {
    13: (0, 1199),  # 13a (to Run 1199)
    14: (1200, 9999),  # 13b from Run 1200 ufloat(1.0458, 0.0013)
}  # type: Dict[int, Tuple[int, int]]


off_to_on_resonance_scale_weight_column_name = "__off_to_on_resonance_scaling_weight__"  # type: str


def add_off_resonance_scaling_weight(
    df: pd.DataFrame,
    off_resonance_scaling_weight_column_name: str,
    exp_number_column_name: str,
    run_number_column_name: str,
    add_neutral_weight: bool = False,
    use_updated_scale_values: bool = False,
) -> None:
    assert exp_number_column_name in df.columns, exp_number_column_name
    assert run_number_column_name in df.columns, run_number_column_name

    assert off_resonance_scaling_weight_column_name not in df.columns, off_resonance_scaling_weight_column_name

    if add_neutral_weight:
        df.loc[:, off_resonance_scaling_weight_column_name] = 1.0
        return
    else:
        df.loc[:, off_resonance_scaling_weight_column_name] = np.nan

    if use_updated_scale_values:
        scaling_dict = off_to_on_resonance_scaling_dict_v2
    else:
        scaling_dict = off_to_on_resonance_scaling_dict

    for exp_num, scaling_weight in scaling_dict.items():
        if exp_num in exp_13_split_info:
            assert isinstance(exp_num, int), type(exp_num).__name__
            exp_cond = df[exp_number_column_name] == 13
            run_min, run_max = exp_13_split_info[exp_num]
            run_cond = (df[run_number_column_name] >= run_min) & (df[run_number_column_name] <= run_max)
            df.loc[exp_cond & run_cond, off_resonance_scaling_weight_column_name] = scaling_weight
        else:
            cond = df[exp_number_column_name] == exp_num
            df.loc[cond, off_resonance_scaling_weight_column_name] = scaling_weight

    assert not df[off_resonance_scaling_weight_column_name].isnull().any(), sum(
        df[off_resonance_scaling_weight_column_name].isnull()
    )
