"""
FourVector module. A FourVector represents a vector in a four dimensional vector
space with Minkowski norm.

Maximilian Welsch 2018.
"""

import numpy as np

from analysistools.physics_vectors.threevector import ThreeVector


def check_four_vector_dimensions(x):
    """
    Checks if the given array has the right shape of two dimensions and 4 components along axis 1.

    :param x: numpy.ndarray.
    :return: numpy.ndarray.
    """

    if x.ndim != 2:
        raise ValueError(f"Given array has not the shape for an array of four vectors! Dimension was {x.ndim}")
    if x.ndim == 2 and x.shape[1] != 4:
        raise ValueError(
            f"Given array has not the shape for an array of four vectors! "
            f"Dimension was {x.ndim}, Shape was {x.shape[1]}"
        )


class FourVector:
    def __init__(self, p):
        """
        Constructor.
        :param p: numpy.ndarray. Dimensions have to be (:,4). Represents an array of four vectors.
        """
        check_four_vector_dimensions(p)
        self._T = p[:, 0]
        self._Vec = ThreeVector(p[:, 1:])
        self.dimensions = p.shape

    def check_setter_input_dimension(self, value) -> None:
        if isinstance(value, np.ndarray):
            if not len(value.shape) == 1:
                raise ValueError(f"Input of setter must be 1-dimensional array, but shape was {value.shape}!")
            if not value.shape[0] == self.dimensions[0]:
                raise ValueError(
                    f"Input of setter must be 1-dimensional array of length {self.dimensions[0]}, "
                    f"but shape was {value.shape}!"
                )

    ##############
    # Properties #
    ##############

    @property
    def as_array(self):
        return np.concatenate(
            (self.T.reshape((-1, 1)), self.X.reshape((-1, 1)), self.Y.reshape((-1, 1)), self.Z.reshape((-1, 1))),
            axis=1,
        )

    @property
    def T(self):
        return self._T

    @T.setter
    def T(self, value):
        self.check_setter_input_dimension(value=value)
        self._T = value

    @property
    def Vec(self):
        """
        Get spatial components as ThreeVector
        :return: ThreeVector
        """
        return self._Vec

    @property
    def X(self):
        return self._Vec.X

    @X.setter
    def X(self, value):
        self.check_setter_input_dimension(value=value)
        self._Vec.X = value

    @property
    def Y(self):
        return self._Vec.Y

    @Y.setter
    def Y(self, value):
        self.check_setter_input_dimension(value=value)
        self._Vec.Y = value

    @property
    def Z(self):
        return self._Vec.Z

    @Z.setter
    def Z(self, value):
        self.check_setter_input_dimension(value=value)
        self._Vec.Z = value

    ####################
    # Member Functions #
    ####################

    def Angle(self, other):
        """
        Calculates the angle between the spatial component vectors.
        :param other: FourVector. Must have same dimensions.
        :return: numpy.array, shape: (-1, ).
        """
        return self.Vec.Angle(other)

    def M2(self):
        """
        Calculates the invariant mass of the four vector.
        :return: numpy.array, shape: (-1,).
        """
        return self.T ** 2 - self.Vec.Mag2()

    def M(self):
        """
        Calculates the square root of the invariant mass. If the invariant mass is < 0, -sqrt(- inv. mass) is returned.
        :return: numpy.array, shape: (-1, ).
        """
        m2 = self.M2()
        negative_m2 = m2 < 0
        m2[negative_m2] = -m2[negative_m2]
        m = np.sqrt(m2)
        m[negative_m2] = -m[negative_m2]
        return m

    def BoostVector(self):
        """
        Calculates the boost vector of the given four vector.
        :return: numpy.array, shape: (-1, ).
        """
        bx = self.X / self.T
        by = self.Y / self.T
        bz = self.Z / self.T
        return np.concatenate(
            (bx.reshape((-1, 1)), by.reshape((-1, 1)), bz.reshape((-1, 1))),
            axis=1,
        )

    def Boost(self, b):
        """
        Perform a Lorentz boost for all four vectors by the given boost vector.
        Analog to https://github.com/root-project/root/blob/master/math/physics/src/TLorentzVector.cxx
        :param b: numpy.array, shape: (-1, 3). Boost vector.
        """
        bx = b[:, 0]
        by = b[:, 1]
        bz = b[:, 2]

        b2 = bx ** 2 + by ** 2 + bz ** 2
        gamma = 1 / np.sqrt(1 - b2)
        bp = bx * self.X + by * self.Y + bz * self.Z
        gamma2 = np.zeros(b2.shape)
        gamma2[b2 > 0] = (gamma - 1) / b2
        gamma2[b2 <= 0] = 0

        self.X = self.X + gamma2 * bp * bx + gamma * bx * self.T
        self.X[np.abs(self.X) < 1e-15] = 0.0
        self.Y = self.Y + gamma2 * bp * by + gamma * by * self.T
        self.Y[np.abs(self.Y) < 1e-15] = 0.0
        self.Z = self.Z + gamma2 * bp * bz + gamma * bz * self.T
        self.Z[np.abs(self.Z) < 1e-15] = 0.0
        self.T = gamma * (self.T + bp)

    ########################
    # Overloaded Operators #
    ########################

    def __add__(self, other):
        T = self.T + other.T
        X = self.X + other.X
        Y = self.Y + other.Y
        Z = self.Z + other.Z
        return FourVector(
            np.concatenate(
                (T.reshape((-1, 1)), X.reshape((-1, 1)), Y.reshape((-1, 1)), Z.reshape((-1, 1))),
                axis=1,
            )
        )

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        T = self.T - other.T
        X = self.X - other.X
        Y = self.Y - other.Y
        Z = self.Z - other.Z
        return FourVector(
            np.concatenate(
                (T.reshape((-1, 1)), X.reshape((-1, 1)), Y.reshape((-1, 1)), Z.reshape((-1, 1))),
                axis=1,
            )
        )

    def __mul__(self, other):
        """
        Scalar product.
        :param other: FourVector
        :return: numpy.array, shape: (-1, ).
        """
        TT = self.T * other.T
        XX = self.X * other.X
        YY = self.Y * other.Y
        ZZ = self.Z * other.Z
        return TT - XX - YY - ZZ

    def __rmul__(self, other):
        """
        Scalar product.
        :param other: FourVector
        :return: numpy.array, shape: (-1, ).
        """
        return self.__mul__(other)
