"""
ThreeVector module. A ThreeVector represent a vector in a three dimensional vector
space with Euclidean norm.

Maximilian Welsch 2018.
"""

import numpy as np


def check_three_vector_dimensions(x):
    """
    Checks if the given array has the right shape of two dimensions and 4 components along axis 1.

    :param x: numpy.ndarray.
    :return: numpy.ndarray.
    """

    if x.ndim != 2:
        raise ValueError("Given array has not the shape for an array of three vectors!")
    if x.ndim == 2 and x.shape[1] != 3:
        raise ValueError("Given array has not the shape for an array of three vectors!")


class ThreeVector:
    def __init__(self, x):
        """
        Constructor
        :param x: numpy.array. Shape has to be (-1, 3). Represents an array of
         three component vectors
        """
        check_three_vector_dimensions(x)
        self._X = x[:, 0]
        self._Y = x[:, 1]
        self._Z = x[:, 2]
        self.dimensions = x.shape

    def check_setter_input_dimension(self, value) -> None:
        if isinstance(value, np.ndarray):
            if not len(value.shape) == 1:
                raise ValueError(f"Input of setter must be 1-dimensional array, but shape was {value.shape}!")
            if not value.shape[0] == self.dimensions[0]:
                raise ValueError(
                    f"Input of setter must be 1-dimensional array of length {self.dimensions[0]}, "
                    f"but shape was {value.shape}!"
                )

    ##############
    # Properties #
    ##############

    @property
    def as_array(self):
        return np.concatenate(
            (self.X.reshape((-1, 1)), self.Y.reshape((-1, 1)), self.Z.reshape((-1, 1))),
            axis=1,
        )

    @property
    def X(self):
        return self._X

    @X.setter
    def X(self, value):
        self.check_setter_input_dimension(value=value)
        self._X = value

    @property
    def Y(self):
        return self._Y

    @Y.setter
    def Y(self, value):
        self.check_setter_input_dimension(value=value)
        self._Y = value

    @property
    def Z(self):
        return self._Z

    @Z.setter
    def Z(self, value):
        self.check_setter_input_dimension(value=value)
        self._Z = value

    ####################
    # Member Functions #
    ####################

    def Scale(self, scalar: float):
        """
        Scales the vector
        :param scalar: Scaling parameter.
        :return: None.
        """
        self.X *= scalar
        self.Y *= scalar
        self.Z *= scalar
        return self

    def Angle(self, other):
        """
        Calculates angle between the vectors in radians.
        :param other: ThreeVector. Must have same dimensions
        """
        arg = self.Dot(other) / np.sqrt(self.Mag2() * other.Mag2())
        arg[arg > 1.0] = 1.0
        arg[arg < -1.0] = -1.0
        return np.arccos(arg)

    def Cross(self, other):
        X = self.Y * other.Z - self.Z * other.Y
        Y = self.Z * other.X - self.X * other.Z
        Z = self.X * other.Y - self.Y * other.X
        return ThreeVector(
            np.concatenate(
                (X.reshape(-1, 1), Y.reshape(-1, 1), Z.reshape(-1, 1)),
                axis=1,
            )
        )

    def Perp(self):
        """
        https://root.cern.ch/doc/master/TVector3_8cxx_source.html
        :return:
        """
        return np.sqrt(self.X * self.X + self.Y * self.Y)

    def Unit(self):
        """
        Calculates unit Vector
        :return: ThreeVector. Normalized three vector.
        """
        mag = self.Mag()
        X = self.X / mag
        Y = self.Y / mag
        Z = self.Z / mag

        return ThreeVector(
            np.concatenate(
                (X.reshape(-1, 1), Y.reshape(-1, 1), Z.reshape(-1, 1)),
                axis=1,
            )
        )

    def Dot(self, other):
        """
        Scalar product.
        :param other: ThreeVector
        :return: numpy.array, shape: (-1, ).
        """
        return self.__mul__(other)

    def Mag2(self):
        """
        Calculates the magnitude squared of the vector.
        :return: numpy.array, shape: (-1, )
        """
        return self.X ** 2 + self.Y ** 2 + self.Z ** 2

    def Mag(self):
        """
        Calculates the magnitude of the vector.
        :return: numpy.array, shape: (-1, )
        """
        return np.sqrt(self.Mag2())

    ########################
    # Overloaded Operators #
    ########################

    def __add__(self, other):
        X = self.X + other.X
        Y = self.Y + other.Y
        Z = self.Z + other.Z
        return ThreeVector(
            np.concatenate(
                (X.reshape((-1, 1)), Y.reshape((-1, 1)), Z.reshape((-1, 1))),
                axis=1,
            )
        )

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        X = self.X - other.X
        Y = self.Y - other.Y
        Z = self.Z - other.Z
        return ThreeVector(
            np.concatenate(
                (X.reshape((-1, 1)), Y.reshape((-1, 1)), Z.reshape((-1, 1))),
                axis=1,
            )
        )

    def __mul__(self, other):
        """
        Scalar product.
        :param other: ThreeVector
        :return: numpy.array, shape: (-1, ).
        """
        XX = self.X * other.X
        YY = self.Y * other.Y
        ZZ = self.Z * other.Z
        return XX + YY + ZZ

    def __rmul__(self, other):
        """
        Scalar product.
        :param other: ThreeVector
        :return: numpy.array, shape: (-1, ).
        """
        return self.__mul__(other)
