#include "Corrections.h"

// Constructor + destructor
//--------------------------------------------------------------------------

Corrections::Corrections(TString syst_source, int no_variedParam, UInt_t seed, int NsigmaFF){

  TEnv *config = OpenSettingsFile("../General/theory.config");

  _FFtheo = new TheoryClass(config);
  _SemiLepDss = new SemiLepDssWeights2D();
  _SemiLepD = new SemiLepWeights2D();
  _lepEffElec = new LepEff(0.5,0);
  _lepEffMu = new LepEff(0.9,1);
  _lepFake = new LepFake();
  _KPiEff = new KPiEff(0.6);

  std::ifstream inputData("../General/slowPionAllWeights.txt", std::ifstream::in);
  for(int line=0; line<24; line++){
    for(int column=0; column<18; column++){
      inputData >> _slowPiWeights[line][column];
    }
  }
  inputData.close();

  _source = syst_source;
  _paramNo = no_variedParam;
  _NsigmaFF = NsigmaFF;

  _seed = seed;
  _rand = new TRandom2(_seed);
  _systVar = _rand->Gaus(0.,1.);
}

Corrections::~Corrections(){
  delete _FFtheo;
  delete _SemiLepDss;
  delete _lepEffElec;
  delete _lepEffMu;
  delete _lepFake;
  delete _KPiEff;
  delete _rand;
}



// Public class functions
//--------------------------------------------------------------------------

// Correction of branching ratios to pdg values (2015)

float Corrections::BR(float lclass, float dclass, float true_Dstar_channel, float true_D_channel){

  float BRfactor = 1.;

  BRfactor *= BR_D(true_D_channel);
  BRfactor *= BR_Dstar(true_Dstar_channel);
  BRfactor *= BR_B(true_Dstar_channel);
  BRfactor *= BR_bkg(lclass, dclass);

  return BRfactor;

}


// Correction of from factor parameters to world average

float Corrections::FFDs_christian(int lclass, int dclass, float true_q2, float true_pl){
  if(abs(lclass)!=2 || dclass!=3) return 1.;

    float weight = 1.;

    if(_source=="DsFF"){
        if(sign(_paramNo)==1) _SemiLepD->setShape(1,_paramNo*2);
        else _SemiLepD->setShape(1,abs(_paramNo)*2-1);
    }
    
    weight = _SemiLepD->weight(1,0,0,true_q2, true_pl);
    
    return weight;  
}

float Corrections::FF(float lclass, float dclass, float true_w, float true_cv, float true_cl, float true_chi){

  float FFfactor = 1.;

  true_cl *= -1.; // Correction because of wrong sign in theoretical def. of cos_thl
  true_cv *= -1.;

  if(fabs(lclass)==2 && dclass==3){

    TString Btype;
    if(lclass==2) Btype = "Bch";
    else if(lclass==-2) Btype = "B0";

    if(Btype=="Bch" && true_w>1.5053) true_w = 1.505;
    else if(Btype=="B0" && true_w>1.5035) true_w = 1.503;
  
    std::vector<double> FF_MC;
    FF_MC.push_back(0.04);
    FF_MC.push_back(0.912);
    FF_MC.push_back(1.3);
    FF_MC.push_back(1.18);
    FF_MC.push_back(0.71);
    _FFtheo->SetValues(FF_MC, Btype);
    double Gamma_MC = _FFtheo->GammaTot();
    double dGamma_MC = _FFtheo->d4Gamma(true_w, true_cl, true_cv, true_chi);
  
    std::vector<double> FF_new;
    FF_new.push_back(0.04);
    FF_new.push_back(0.912);

    TVectorD par(3);
    par(0) = 1.207; par(1) = 1.401; par(2) = 0.854;
    if(_source=="DsFF") par = calcVariatedFF(_paramNo);
    FF_new.push_back(par(0));
    FF_new.push_back(par(1));
    FF_new.push_back(par(2));
    _FFtheo->SetValues(FF_new, Btype);
    double Gamma_new = _FFtheo->GammaTot();
    double dGamma_new = _FFtheo->d4Gamma(true_w, true_cl, true_cv, true_chi);

    FFfactor = (Gamma_MC/Gamma_new) * (dGamma_new/dGamma_MC);
  }

  if(FFfactor>10.) return 0.; // Do not consider big weights!
  return FFfactor;

}


// D form factor corrections (from Jan)

/*
float Corrections::DFF( int dclass, float q2_true, float plep_true, int iVarBin){
    
    if(dclass!=1) return 1.;

    float weight = 1.;

    if(_source=="DFF"){
        if(sign(_paramNo)==1) _SemiLepD->setShape(2,_paramNo*2);
        else _SemiLepD->setShape(2,abs(_paramNo)*2-1);
    }
    
    weight = _SemiLepD->weight(2,0,0,q2_true, plep_true);
    
    return weight;  
}
*/


// D** form factor corrections (from Jan)

float Corrections::DssFF(int dclass, float w_true, float costh_true, int iVarBin){

  float weight = 1.;

  if(_source=="DssFF"){
    if(sign(_paramNo)==1) _SemiLepDss->setShape(dclass-5,_paramNo*2);
    else _SemiLepDss->setShape(dclass-5,abs(_paramNo)*2-1);
  }

  weight = _SemiLepDss->weight(dclass-5, w_true, costh_true);

  return weight;

}



// Correction of slow pion efficiencies (from BN #1176)

float Corrections::slowPion(int exp, int truthMatching, int Dscharge, int Dcharge, float pPi){

  if(truthMatching==-999) return 1.; // Slow pions only in real D* decay

  double rho = 1.;

  for(int line=0; line<24; line++){
    if(abs(Dscharge-Dcharge)==_slowPiWeights[line][0] && exp>=_slowPiWeights[line][1] && exp<=_slowPiWeights[line][2] && pPi>=_slowPiWeights[line][3] && pPi<_slowPiWeights[line][4]){
      // Find right weight: either in 6th columns for no variation or after it (up/down alternate)
      if(_source!="slowPi" || line%6!=(abs(_paramNo)-1)%6) rho = _slowPiWeights[line][5];
      else if(sign(_paramNo)==1) rho = _slowPiWeights[line][4+((_paramNo-1)%6+1)*2];
      else if(sign(_paramNo)==-1) rho = _slowPiWeights[line][5+((abs(_paramNo)-1)%6+1)*2];
      break;
    }
  }

  if(rho>1.5) std::cout << rho << std::endl;

  return rho;
}


// Correction of Lepton ID efficiencies

float Corrections::lepEff(int fl_lep, int exp_no, float theta, float plep){

  if(fl_lep/10!=1+fl_lep%10) return 1.;

  float weight = 1.;

  theta  *= 180./TMath::Pi(); // angle in degrees
  int lep_gen = fl_lep%10;

  if(lep_gen==0){
    weight = _lepEffElec->weight(lep_gen, exp_no, theta, plep);
    if(_source=="lepEff_el"){
      _rand->SetSeed(weight*_seed*1000);
      weight += _rand->Gaus(0.,_lepEffElec->errStat(lep_gen, exp_no, theta, plep)) + _rand->Gaus(0.,_lepEffElec->errSyst1(lep_gen, exp_no, theta, plep)); // uncorrelated errors (individual per bin)
      weight += _systVar*_lepEffElec->errSyst2(lep_gen, exp_no, theta, plep); // correlated errors (same fraction of 1sigma interval for each bin)
    }
  }
  else if(lep_gen==1){
    weight = _lepEffMu->weight(lep_gen, exp_no, theta, plep);
    if(_source=="lepEff_mu"){
      _rand->SetSeed(weight*_seed*1000);
      weight += _rand->Gaus(0.,_lepEffMu->errStat(lep_gen, exp_no, theta, plep)) + _rand->Gaus(0.,_lepEffMu->errSyst1(lep_gen, exp_no, theta, plep)); // uncorrelated errors (individual per bin)
      weight += _systVar*_lepEffMu->errSyst2(lep_gen, exp_no, theta, plep); // correlated errors (same fraction of 1sigma interval for each bin)
    }
  }

  return weight;

}


float Corrections::lepFake(int fl_lep, float theta, float plep){

  float weight = 1.;

  theta  *= 180./TMath::Pi(); // angle in degrees
  int lep_gen = fl_lep%10;
  int lep_gen_true = fl_lep/10-1;

  weight = _lepFake->weight(lep_gen, lep_gen_true, theta, plep);
  if(_source=="lepFake"){
    _rand->SetSeed(weight*_seed*1000);
    weight += _rand->Gaus(0.,_lepFake->err(lep_gen, lep_gen_true, theta, plep));
  }

  return weight;

}


float Corrections::kPiEff(std::vector<int> *trkSep_list, std::vector<float> *trkTh_list, std::vector<float> *trkP_list, int exp){

  float weight = 1.;

  for(int i=0; i<trkSep_list->size(); i++){
    int isKaon=-1;

    if(trkSep_list->at(i)==33) isKaon = 0;
    else if(trkSep_list->at(i)==34) isKaon = 1;
    else if(trkSep_list->at(i)==44) isKaon = 2;
    else if(trkSep_list->at(i)==43) isKaon = 3;

    if(isKaon==-1) continue;
    else{
      weight *= _KPiEff->weight(isKaon, trkTh_list->at(i), trkP_list->at(i), exp);
      if(_source=="KPiEff"){
	_rand->SetSeed(weight*_seed*1000);
	weight += _rand->Gaus(0.,_KPiEff->errStat(isKaon, trkTh_list->at(i), trkP_list->at(i), exp)); // Statistical error (uncorrelated between bins)
	weight += _systVar*_KPiEff->errSyst(isKaon, trkTh_list->at(i), trkP_list->at(i), exp); // systematic error (correlated over all bins)
      }
    }

  }

  return weight;

}


float Corrections::tagging(float b_mode, float NB){

  float tagFactor = 1.;

  float tagCorrections[1120][6];

  std::ifstream inputData("../General/tagCorrectionWeights.txt", std::ifstream::in);

  for(int line=0; line<1120; line++){

    for(int column=0; column<6; column++){
      inputData >> tagCorrections[line][column];
    }

    if(abs(b_mode)==tagCorrections[line][0] && log(NB)>=tagCorrections[line][1] && log(NB)<=tagCorrections[line][2]){
      tagFactor = tagCorrections[line][3];
      if(_source=="tagEff"){
	_rand->SetSeed(tagCorrections[line][0]+_seed);
	tagFactor += _rand->Gaus(0.,1.)*tagCorrections[line][4]; // statistical error
	tagFactor += _systVar * tagCorrections[line][5]; // systematics error (100% correlated over all bins)
      }
      break;
    }

  }

  inputData.close();

  return tagFactor;

}



// Private class functions
//--------------------------------------------------------------------------

float Corrections::BR_D(float true_D_channel){

  float BR_factor = 1.;

  if(true_D_channel==42101100){
    BR_factor = 1.029;
    if(_source=="Ddecay" && abs(_paramNo==1)) BR_factor += sign(_paramNo)*0.010;
  }

  else if(true_D_channel==41101200){
    BR_factor = 0.995;
    if(_source=="Ddecay" && abs(_paramNo==2)) BR_factor += sign(_paramNo)*0.025;
  }

  else if(true_D_channel==42101300){
    BR_factor = 1.139;
    if(_source=="Ddecay" && abs(_paramNo==3)) BR_factor += sign(_paramNo)*0.028;
  }

  else if(true_D_channel==42101110){
    BR_factor = 1.093;
    if(_source=="Ddecay" && abs(_paramNo==4)) BR_factor += sign(_paramNo)*0.061;
  }

  return BR_factor;

}


float Corrections::BR_Dstar(float true_Dstar_channel){

  if(_source!="Dsdecay") return 1.; // nominal weight

  float weight = 1.;

  if(true_Dstar_channel==41310100 && _paramNo==1)      weight += sign(_paramNo)*0.007;
  else if(true_Dstar_channel==42310010 && _paramNo==2) weight += sign(_paramNo)*0.047;
  else if(true_Dstar_channel==41310010 && _paramNo==3) weight += sign(_paramNo)*0.016;
  else if(true_Dstar_channel==42310001 && _paramNo==4) weight += sign(_paramNo)*0.076;
  else if(true_Dstar_channel==41310001 && _paramNo==5) weight += sign(_paramNo)*0.250;

  return weight;

}


float Corrections::BR_B(float true_Dstar_channel){

  float weight = 1.;

  if(true_Dstar_channel==42310010 || true_Dstar_channel==42310001){
    weight = 0.9176;
    if(_source=="BtoDslnu" && abs(_paramNo)==1) weight +=  sign(_paramNo)*0.0206;
  }

  else if(true_Dstar_channel==41310100 || true_Dstar_channel==41310010 || true_Dstar_channel==41310001){
    weight = 0.9250;
    if(_source=="BtoDslnu" && abs(_paramNo)==2) weight +=  sign(_paramNo)*0.0206;
  }

  return weight;

}


float Corrections::BR_bkg(float lclass, float dclass){

  float weight = 1.;

  // B+/- -> D**lnu
  if(lclass==2 && dclass==5){
    weight = 0.3519 *3./2.*2.244; // factor 2.244 to fill up gap of D**
    if(_source=="BtoDsslnu" && abs(_paramNo)==1) weight += sign(_paramNo)*0.1728 *3./2.*2.244;
  }
  else if(lclass==2 && dclass==6){
    weight = 0.2026 *3./2.*2.54*2.244;
    if(_source=="BtoDsslnu" && abs(_paramNo)==2) weight += sign(_paramNo)*0.0205 *3./2.*2.54*2.244;
  }
  else if(lclass==2 && dclass==7){
    weight = 0.3187 *3./2.*2.244;
    if(_source=="BtoDsslnu" && abs(_paramNo)==3) weight += sign(_paramNo)*0.0549 *3./2.*2.244;
  }
  else if(lclass==2 && dclass==8){
    weight = 0.1605 *3./2.*2.244;
    if(_source=="BtoDsslnu" && abs(_paramNo)==4) weight += sign(_paramNo)*0.0494 *3./2.*2.244;
  }

  // B0 -> D**lnu
  else if(lclass==-2 && dclass==5){
    weight = 0.3581 *3./2.*2.601; // factor 2.601 to fill up gap of D**
    if(_source=="BtoDsslnu" && abs(_paramNo)==5) weight += sign(_paramNo)*0.1757 *3./2.*2.601;
  }
  else if(lclass==-2 && dclass==6){
    weight = 0.2000 *3./2.*2.54*2.601;
    if(_source=="BtoDsslnu" && abs(_paramNo)==6) weight += sign(_paramNo)*0.0194 *3./2.*2.54*2.601;
  }
  else if(lclass==-2 && dclass==7){
    weight = 0.3214 *3./2.*2.601;
    if(_source=="BtoDsslnu" && abs(_paramNo)==7) weight += sign(_paramNo)*0.0595 *3./2.*2.601;
  }
  else if(lclass==-2 && dclass==8){
    weight = 0.1622 *3./2.*2.601;
    if(_source=="BtoDsslnu" && abs(_paramNo)==8) weight += sign(_paramNo)*0.0494 *3./2.*2.601;
  }

  // B+/- -> D0lnu
  else if(lclass==2 && dclass==1){
    weight = 0.9827;
    if(_source=="BtoDlnu" && abs(_paramNo)==1) weight += sign(_paramNo)*0.0476;
  }

  // B0 -> Dlnu
  else if(lclass==-2 && dclass==1){
    weight = 1.0282;
    if(_source=="BtoDlnu" && abs(_paramNo)==2) weight += sign(_paramNo)*0.0423;
  }

  return weight;

}



TVectorD Corrections::calcVariatedFF(int variedPar){

  // Vector containing world averages of FF parameters
  TVectorD FFval(3);

  FFval(0) = 1.207;
  FFval(1) = 1.401;
  FFval(2) = 0.854;

  // Covariance matrix of FF using world avarages
  TMatrixD FFcov(3,3);

  FFcov(0,0) = 7.84e-4;    FFcov(0,1) = 5.23e-4;     FFcov(0,2) = -4.52e-4;
  FFcov(1,0) = 5.23e-4;    FFcov(1,1) = 1.089e-3;    FFcov(1,2) = -5.00e-4;
  FFcov(2,0) = -4.52e-4;   FFcov(2,1) = -5.00e-4;    FFcov(2,2) = 4.00e-4;

  // eigenvalues: elements of the diagonal form of covariance matrix
  TVectorD lambda;

  // matrix with eigenvector: transformation matrix between original covariance
  // and diagonal form: C' = P^-1 C P
  TMatrixD P = FFcov.EigenVectors(lambda);
  TMatrixD Pinvert = P; Pinvert.Invert();

  // Transform FF parameters in new basis
  FFval *= Pinvert;

  // Vary new FF parameters inside errors given by sqrt of eigenvalues
  FFval(abs(variedPar)-1) = FFval(abs(variedPar)-1) + _NsigmaFF*sign(1.*variedPar)*sqrt(lambda(abs(variedPar)-1));

  // Rotate back into original basis to apply event weights
  FFval *= P;

  return FFval;

}


int Corrections::sign(float n){
  if(n==0) return 0;
  else if(n>0) return 1;
  else if(n<0) return -1;
}

