#include "KPiEff.h"
#include "TMath.h"
#include <iostream>
#include <fstream>

void KPiEff::init(int varBins, const char* strVar, float fPid) {
  nVarBins = varBins;

  const double fMomBins[] = {0, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2.0, 2.1, 2.2, 2.3, 2.4, 2.5, 2.6, 2.7, 2.8, 2.9, 3.0, 3.2, 3.4, 3.6, 4.0, 4.5, 100.0};
  const double fThetaBins[] = {0,33,46,57,64,69,78,89,99,107,121,128,180};

  iPidCut = fPid*10;
  if(iPidCut<1 || iPidCut>9 )
    {std::cout<<"ERR: KPiEff::init: not supported PID cut: "<<fPid<<std::endl;exit(-1);}

  //need to read efficiency tables?
  if(fKPiEffSVD1 == nullptr && fKPiEffSVD2 == nullptr)
    {
      fKPiEffSVD1 = new float[152064];
      fKPiEffSVD2 = new float[152064];
      std::ifstream ifs;
      ifs.open("../General/LepCorrections/KPiEffSVD1.dat");
      if(!ifs.good())
        {std::cout<<"ERR: KPiEff::init: File not found: KPiEffSVD1.dat"<<std::endl;exit(-1);}
      float ftmp;
      for(int i = 0; ifs >> ftmp; i++)
        {
	  fKPiEffSVD1[i] = ftmp;
	  //        std::cout<<ftmp<<endl;
        }
      ifs.close();
      ifs.open("../General/LepCorrections/KPiEffSVD2.dat");
      if(!ifs.good())
        {std::cout<<"ERR: KPiEff::init: File not found: KPiEffSVD2.dat"<<std::endl;exit(-1);}
      for(int i = 0; ifs >> ftmp; i++)
        {
	  fKPiEffSVD2[i] = ftmp;
        }
      ifs.close();
    }

  // Pions
  axis_plep[0] = new TAxis(nMomBins, fMomBins);
  axis_theta[0] = new TAxis(nThetaBins, fThetaBins);

  // Kaons
  axis_plep[1] = new TAxis(nMomBins, fMomBins);
  axis_theta[1] = new TAxis(nThetaBins, fThetaBins);

  iPosWeight = 7;
  iPosErrStat = 8;
  iPosErrSyst = 9;
  iPosFlag = 10;

}

KPiEff::KPiEff(float fPid) {
  init(1, "default", fPid);
}

KPiEff::KPiEff(int varBins, const char* strVar,float fPid) {
  init(varBins, strVar, fPid);
}

KPiEff::~KPiEff() {
}

void KPiEff::Term() {
  delete[] fKPiEffSVD1;
  delete[] fKPiEffSVD2;
}

float KPiEff::lookup(int isKaon, int iBinTheta, int iBinMom,int exp, int iType) {
  if (iType < 0 || iType >= nColumns) return -1;
  if (iBinTheta < 0 || iBinTheta>=nThetaBins) return -1;
  if (iBinMom < 0 || iBinMom >= nMomBins) return -1;
  const float* fTable;
  if(exp<31)
    {
      fTable = fKPiEffSVD1;
      //	    if (isKaon == 0) fTable = fKaonEffSVD1;
      //	    else if (isKaon == 1) fTable = fKaonPionFakeSVD1;
      //	    else if (isKaon == 2) fTable = fPionEffSVD1;
      //	    else if (isKaon == 3) fTable = fPionKaonFakeSVD1;
      //	    else return -1;
    }
  else
    {
      fTable = fKPiEffSVD2;
      //	    if (isKaon == 0) fTable = fKaonEff31;
      //	    else if (isKaon == 1) fTable = fKaonPionFake31;
      //	    else if (isKaon == 2) fTable = fPionEff31;
      //	    else if (isKaon == 3) fTable = fPionKaonFake31;
      //	    else return -1;
    }
  int iPositionFlag = 3456*nColumns*isKaon + 384*nColumns*(iPidCut-1) + iBinMom*nColumns*nThetaBins + nColumns*iBinTheta + iPosFlag;
  if (int(fTable[iPositionFlag]) != 0) return -1;
  int iPosition = 3456*nColumns*isKaon + 384*nColumns*(iPidCut-1) + iBinMom*nColumns*nThetaBins + nColumns*iBinTheta + iType;
  return fTable[iPosition];
}

float KPiEff::lookup(int isKaon, float theta, float plep, int exp, int iType) {
  if ( isKaon < 0 ||  isKaon > 3) return -1;
  int iBinTheta = axis_theta[0]->FindBin(theta) - 1;
  int iBinMom = axis_plep[0]->FindBin(plep) - 1;
  return lookup( isKaon, iBinTheta, iBinMom, exp, iType);
}

float KPiEff::weight(int isKaon, float theta, float plep, int exp) {
  if(isKaon<0 || isKaon >3){ 
    std::cout<<"ERR: KPiEff::init: not supported particle type: "<<isKaon<<std::endl;
    exit(-1);
  }
  float fResult = lookup(isKaon, theta, plep, exp, iPosWeight);
  if (fResult < 0) return 1.;
  return fResult;
}

float KPiEff::errStat(int isKaon, float theta, float plep, int exp) {
  if(isKaon<0 || isKaon >3){
    std::cout<<"ERR: KPiEff::init: not supported particle type: "<<isKaon<<std::endl;
    exit(-1);
  }
  float fResult = lookup(isKaon, theta, plep, exp, iPosErrStat);
  if (fResult < 0) return 0.;
  return fResult;
}

float KPiEff::errSyst(int isKaon, float theta, float plep, int exp) {
  if(isKaon<0 || isKaon >3){
    std::cout<<"ERR: KPiEff::init: not supported particle type: "<<isKaon<<std::endl;
    exit(-1);
  }
  float fResult = lookup(isKaon, theta, plep, exp, iPosErrSyst);
  if (fResult < 0) return 0.;
  return fResult;
}


void KPiEff::reset() {
}

const int KPiEff::nMomBins = 32;
const int KPiEff::nThetaBins = 12;
const int KPiEff::nColumns = 11;

float *KPiEff::fKPiEffSVD1 = nullptr;

float *KPiEff::fKPiEffSVD2 = nullptr;

