"""
Particle ID correction for Belle

Calculates weights for data - MC corrections arising from the particle ID
See https://belle.kek.jp/secured/wiki/doku.php?id=software:systematics for more infos.

Felix Metzner 2019
"""

import os
import copy
import warnings
import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from collections import OrderedDict
from pandas.core.common import SettingWithCopyWarning
from typing import Optional, List, Tuple, Dict, Sequence

from analysistools.utilities import PathType
from analysistools.pid_corrections.table_manager import TableManager


__all__ = [
    "PIDCorrection",
    "ElectronIDCorrection",
    "ElectronFromHadronFakeIDCorrection",
    "MuonIDCorrection",
    "MuonFromHadronFakeIDCorrection",
    "KPiIDCorrection",
    "PionIDCorrection",
    "PionFromKaonFakeIDCorrection",
    "KaonIDCorrection",
    "KaonFromPionFakeIDCorrection",
]


_pid_table_manager = TableManager.instance()  # type: ignore
pid_table_manager = _pid_table_manager  # type: TableManager


class PIDCorrection(ABC):
    """
    Abstract base class for pID (fake) corrections
    """

    uncertainty_from_var_name = "uncertainty_from_var"  # type: str
    variation_name = "variation"  # type: str

    def __init__(self, variations: Optional[int] = None) -> None:
        if variations is not None:
            assert variations > 0, variations

        self.n_variations = variations  # type: Optional[int]
        self._columns = None  # type: Optional[List[str]]
        self._theta_bins = None  # type: Optional[List[float]]
        self._momentum_bins = None  # type: Optional[List[float]]
        self._pid_cut = None  # type: Optional[float]

        self._pid_base_seed = None  # type: Optional[int]
        self.base_seed_int = 10000  # type: int

        self._correction_columns = None  # type: Optional[List[str]]

        self._momentum_column = None  # type: Optional[str]
        self._theta_column = None  # type: Optional[str]
        self._cut_column = None  # type: Optional[str]
        self._flag_column = None  # type: Optional[str]

        self.base_path = os.path.dirname(os.path.abspath(__file__))  # type: PathType
        self._sub_dir = None  # type: Optional[str]

    @property
    def columns(self) -> List[str]:
        assert self._columns is not None
        return self._columns

    @property
    def theta_bins(self) -> List[float]:
        assert self._theta_bins is not None
        return self._theta_bins

    @property
    def momentum_bins(self) -> List[float]:
        assert self._momentum_bins is not None
        return self._momentum_bins

    @property
    def pid_cut(self) -> float:
        assert self._pid_cut is not None
        return self._pid_cut

    @property
    def pid_base_seed(self) -> int:
        assert self._pid_base_seed is not None
        return self._pid_base_seed

    @property
    def correction_columns(self) -> List[str]:
        assert self._correction_columns is not None
        return self._correction_columns

    @property
    def momentum_column(self) -> str:
        assert self._momentum_column is not None
        return self._momentum_column

    @property
    def theta_column(self) -> str:
        assert self._theta_column is not None
        return self._theta_column

    @property
    def cut_column(self) -> str:
        assert self._cut_column is not None
        return self._cut_column

    @property
    def flag_column(self) -> Optional[str]:
        return self._flag_column

    @property
    def sub_dir(self) -> str:
        assert self._sub_dir is not None
        return self._sub_dir

    @abstractmethod
    def get_pid_table_dict(self) -> Dict[str, Tuple[int, int]]:
        """
        Method to get pid table files and the experiment range for which they are
        valid. Must return a ordered dict, so that the experiment range is
        is considered correctly.

        :return: Ordered dictionary mapping file name to tuple of exp range
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_theta_bins() -> List[float]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_momentum_bins() -> List[float]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_exp_ranges() -> List[Tuple[int, int]]:
        raise NotImplementedError

    @abstractmethod
    def _get_action_for_pid_error_flag(self, flag_value) -> bool:
        raise NotImplementedError

    @abstractmethod
    def _get_pid_type_hint_on_error_flag(self) -> int:
        raise NotImplementedError

    def _get_correction_matrix(self, table_name: str, correction_col: str) -> np.ndarray:
        matrix_name = f"{table_name}_pid_cut{self.pid_cut}_{correction_col}"
        matrix = pid_table_manager.get_correction_matrix(name=matrix_name)

        if matrix is None:
            table = self._get_pid_table(table_name=table_name)
            matrix = self._build_correction_matrix(table=table, correction_col=correction_col)
            pid_table_manager.add_correction_matrix(name=matrix_name, new_matrix=matrix)

        return matrix

    def _get_varied_matrix(self, table_name: str, stat_seed: int, sys_seed: int) -> np.ndarray:
        matrix_name = f"{table_name}_pid_cut{self.pid_cut}_varied_{stat_seed}_{sys_seed}"  # type: str
        matrix = pid_table_manager.get_correction_matrix(name=matrix_name)

        if matrix is None:
            matrix = self._build_varied_matrix(table_name=table_name, stat_seed=stat_seed, sys_seed=sys_seed)
            pid_table_manager.add_correction_matrix(name=matrix_name, new_matrix=matrix)

        return matrix

    def _build_varied_matrix(self, table_name: str, stat_seed: int, sys_seed: int) -> np.ndarray:
        for correction_col in self.correction_columns:
            matrix_name = f"{table_name}_pid_cut{self.pid_cut}_{correction_col}"  # type: str
            assert pid_table_manager.contains_matrix(matrix_name), matrix_name

        return self._build_varied_pid_matrix(table_name=table_name, stat_seed=stat_seed, sys_seed=sys_seed)

    @abstractmethod
    def _build_varied_pid_matrix(self, table_name: str, stat_seed: int, sys_seed: int) -> np.ndarray:
        raise NotImplementedError

    def _get_pid_table(self, table_name: str) -> pd.DataFrame:
        table_path = os.path.join(self.base_path, self.sub_dir, f"{table_name}.csv")  # type: PathType

        return pid_table_manager.get_table(
            name=table_name,
            path=table_path,
            column_names=self.columns,
        )

    def _build_correction_matrix(self, table: pd.DataFrame, correction_col: str) -> np.ndarray:
        assert self.momentum_column in table.columns
        assert self.theta_column in table.columns
        assert correction_col in table.columns

        # Get subset of table that contains the information for the relevant cut value.
        sub_table = table.loc[table[self.cut_column] == self.pid_cut]

        # Ensure that theta-momentum combinations are unique in table:
        assert not (sub_table.groupby([self.momentum_column, self.theta_column]).size() != 1).any()

        matrix = np.full((len(self.momentum_bins[:-1]), len(self.theta_bins[:-1])), np.nan)  # type: np.ndarray

        for mom_index, mom_bin in enumerate(self.momentum_bins[:-1]):
            for theta_index, theta_bin in enumerate(self.theta_bins[:-1]):
                value_rows = sub_table.loc[
                    (sub_table[self.momentum_column] == mom_bin) & (sub_table[self.theta_column] == theta_bin)
                ]
                assert len(value_rows) == 1

                value = value_rows[correction_col].values[0]
                if self.flag_column is not None:
                    flag_value = value_rows[self.flag_column].values[0]
                    if self._get_action_for_pid_error_flag(flag_value=flag_value):
                        if correction_col == "stat":
                            value = self._get_pid_type_hint_on_error_flag()
                        else:
                            value = -1.0 * flag_value

                matrix[mom_index][theta_index] = value

        assert not np.isnan(matrix).any()
        return matrix

    @staticmethod
    def update_values_for_fail_flags(
        df: pd.DataFrame,
        column_name: str,
        new_column_name: str,
    ) -> None:
        if new_column_name in df.columns:
            return
        flag_choice = [-999, 1.0, 1.0, 1.0]  # type: List[float]
        assert len(df.loc[df[column_name] < 0, column_name].unique()) <= 3
        assert all([x in [-1.0, -2.0, -3.0] for x in df.loc[df[column_name] < 0, column_name].unique()])

        df.loc[df[column_name] < 0, new_column_name] = np.choose(
            -df.loc[df[column_name] < 0, column_name].astype("int64"),
            flag_choice,
        )
        df.loc[df[column_name] >= 0, new_column_name] = df.loc[df[column_name] >= 0, column_name]

    def get_pid_correction(
        self,
        df: pd.DataFrame,
        df_mom_col: str,
        df_theta_col: str,
        df_exp_col: str,
        convert_rad_to_deg: bool = False,
    ) -> pd.DataFrame:
        """
        Method returning the PID correction information for a certain correction
        type defined via the child classes of this abstract class.

        Given a data frame and the names of the columns containing the momentum and
        theta information for the particle to be corrected, as well as the column name
        of the column containing the experiment number, a data frame containing the
        correction values is returned.

        :param df: Data frame for which the corrections shall be returned.
        :param df_mom_col: Momentum column name for the particle of interest.
        :param df_theta_col: Theta column name for the particle of interest.
        :param df_exp_col: Column name of the column containing the experiment number.
        :param convert_rad_to_deg: Whether to convert the values in the theta column from radians to degrees.
        :return: Data frame containing the correction values.
        """
        if not len(df.index) > 0:
            raise RuntimeError("The dataframe you are passing to get_pid_correction is empty. Is this intended?")
        assert df_mom_col in df.columns
        assert df_theta_col in df.columns
        assert df_exp_col in df.columns
        assert df[df_mom_col].min() >= 0.0, df[df_mom_col].min()
        assert df[df_theta_col].min() >= 0.0, df[df_theta_col].min()

        table_names, exp_ranges = zip(*self.get_pid_table_dict().items())
        exp_bins = [exp for exp, _ in exp_ranges]
        exp_bins.append(exp_ranges[-1][1] + 1)

        exp_map = np.digitize(df[df_exp_col], exp_bins, right=False)
        assert (0 not in exp_map) and (len(exp_bins) not in exp_map)
        exp_map -= 1

        map_col_name = "__exp_map_for_pid_corr__"  # type: str
        assert map_col_name not in df.columns

        # Catching SettingWithCopyWarning from Pandas arising due to adding temporary column to df.
        with warnings.catch_warnings(record=True) as warning:
            warnings.simplefilter("always")
            # Actual command:
            df.loc[:, map_col_name] = exp_map
            for w in warning:
                assert issubclass(w.category, SettingWithCopyWarning)
                assert "A value is trying to be set on a copy" in str(w.message)

        theta_col_name = df_theta_col
        if convert_rad_to_deg:
            assert df[df_theta_col].max() <= np.pi, df[df_theta_col].max()
            new_theta_col_name = "__theta_in_deg_for_pid_corr__"
            assert new_theta_col_name not in df.columns
            # Catching SettingWithCopyWarning from Pandas arising due to adding temporary column to df.
            with warnings.catch_warnings(record=True) as warning:
                warnings.simplefilter("always")
                # Actual command:
                df.loc[:, new_theta_col_name] = df[df_theta_col] * 180.0 / np.pi
                for w in warning:
                    assert issubclass(w.category, SettingWithCopyWarning)
                    assert "A value is trying to be set on a copy" in str(w.message)
            theta_col_name = new_theta_col_name
        else:
            assert df[df_theta_col].max() <= 180.0, df[df_theta_col].max()
            if df[df_theta_col].max() <= np.pi:
                print(
                    f"WARNING: Theta column does not contain values above pi. "
                    f"The maximal value was {df[df_theta_col].max()}"
                    f"Are you sure you do not need to convert from radians to degrees?"
                )

        res_df = pd.DataFrame(df[map_col_name])
        res_column_names = []

        for correction_col in self.correction_columns:
            corr_res_column_name = self.get_weight_column_name(column_name=correction_col)
            res_column_names.append(corr_res_column_name)
            res_df.loc[:, corr_res_column_name] = np.nan

            for index, (table_name, exp_range) in enumerate(zip(table_names, exp_ranges)):
                if index not in exp_map:
                    continue
                assert np.isnan(res_df.loc[df[map_col_name] == index, corr_res_column_name]).all()

                corr_matrix = self._get_correction_matrix(table_name=table_name, correction_col=correction_col)
                df_for_exps = df[[df_mom_col, theta_col_name]].loc[df[map_col_name] == index]

                corr_for_exps = self._get_pid_correction(
                    df=df_for_exps,
                    df_mom_col=df_mom_col,
                    df_theta_col=theta_col_name,
                    correction_matrix=corr_matrix,
                )

                res_df.loc[res_df[map_col_name] == index, corr_res_column_name] = corr_for_exps

            assert not res_df.isnull().values.any()

        if self.n_variations is not None:
            variation_seed_pairs = self._get_variation_seed_pairs()

            uncert_column_name = PIDCorrection.get_uncert_from_var_col_name()
            res_column_names.append(uncert_column_name)
            res_df.loc[:, uncert_column_name] = np.nan

            for index, (table_name, exp_range) in enumerate(zip(table_names, exp_ranges)):
                if index not in exp_map:
                    continue
                df_for_exps = df[[df_mom_col, theta_col_name]].loc[df[map_col_name] == index]

                for variation_index, (sys_seed, stat_seed) in enumerate(variation_seed_pairs):
                    variation_column_name = self.get_var_col_name(variation_index=variation_index)
                    if variation_column_name not in res_df.columns:
                        res_column_names.append(variation_column_name)
                        res_df.loc[:, variation_column_name] = np.nan

                    assert np.isnan(res_df.loc[df[map_col_name] == index, variation_column_name]).all()
                    varied_matrix = self._get_varied_matrix(
                        table_name=table_name,
                        stat_seed=stat_seed + self.base_seed_int * 100 * exp_range[0],
                        sys_seed=sys_seed + self.base_seed_int * 1000,
                    )

                    variation_for_exps = self._get_pid_correction(
                        df=df_for_exps,
                        df_mom_col=df_mom_col,
                        df_theta_col=theta_col_name,
                        correction_matrix=varied_matrix,
                    )

                    res_df.loc[res_df[map_col_name] == index, variation_column_name] = variation_for_exps

                assert np.isnan(res_df.loc[df[map_col_name] == index, uncert_column_name]).all()

                uncertainty_matrix = self._get_uncertainty_matrix_from_variation(
                    table_name=table_name,
                    variation_seed_pairs=variation_seed_pairs,
                )

                uncertainty_from_variation_for_exps = self._get_pid_correction(
                    df=df_for_exps,
                    df_mom_col=df_mom_col,
                    df_theta_col=theta_col_name,
                    correction_matrix=uncertainty_matrix,
                )

                res_df.loc[res_df[map_col_name] == index, uncert_column_name] = uncertainty_from_variation_for_exps

            assert not res_df.isnull().values.any()

        del df[map_col_name]
        if convert_rad_to_deg:
            del df[theta_col_name]

        return res_df[res_column_names]

    def _get_pid_correction(
        self,
        df: pd.DataFrame,
        df_mom_col: str,
        df_theta_col: str,
        correction_matrix: np.ndarray,
    ) -> float:
        momentum_bins = self.momentum_bins.copy()
        momentum_bins[0] = 0.0
        momentum_bins[-1] = np.inf
        theta_bins = self.theta_bins.copy()
        theta_bins[0] = 0.0
        theta_bins[-1] = 180.0

        mom_indices = np.digitize(df[df_mom_col], momentum_bins, right=True)
        assert (0 not in mom_indices) and (len(momentum_bins) not in mom_indices)
        mom_indices -= 1
        theta_indices = np.digitize(df[df_theta_col], theta_bins, right=True)
        assert (0 not in theta_indices) and (len(theta_bins) not in theta_indices)
        theta_indices -= 1
        return correction_matrix[mom_indices, theta_indices]

    def _get_uncertainty_matrix_from_variation(
        self,
        table_name: str,
        variation_seed_pairs: List[Tuple[int, int]],
    ) -> np.ndarray:
        assert "ratio" in self.correction_columns

        nominal_matrix = self._get_correction_matrix(table_name=table_name, correction_col="ratio")
        var_matrix_name_beginning = f"{table_name}_pid_cut{self.pid_cut}_varied_"
        varied_matrices = pid_table_manager.get_matrices_starting_with(name_beginning=var_matrix_name_beginning)

        assert len(varied_matrices) == len(variation_seed_pairs)

        uncert_matrix = np.zeros_like(nominal_matrix)
        for varied_matrix in varied_matrices:
            uncert_matrix += np.square(varied_matrix - nominal_matrix)

        return np.sqrt(1.0 / len(variation_seed_pairs) * uncert_matrix)

    def _get_variation_seed_pairs(self) -> List[Tuple[int, int]]:
        nrs = np.random.RandomState(seed=self.pid_base_seed)
        assert self.n_variations is not None

        while True:
            sys_seeds = nrs.random_integers(
                low=self.base_seed_int * self.pid_base_seed,
                high=int(self.base_seed_int * (self.pid_base_seed + 0.5)) - 1,
                size=self.n_variations,
            )
            stat_seeds = nrs.random_integers(
                low=int(self.base_seed_int * (self.pid_base_seed + 0.5)),
                high=self.base_seed_int * (self.pid_base_seed + 1) - 1,
                size=self.n_variations,
            )
            if len(set(list(sys_seeds) + list(stat_seeds))) == 2 * self.n_variations:
                return list(zip(sys_seeds, stat_seeds))

    @staticmethod
    def get_weight_column_name(column_name: str) -> str:
        return f"__{PIDCorrection.get_weight_column_base_name()}_{column_name}__"

    @staticmethod
    def get_weight_column_base_name() -> str:
        return "weight_pid_corr"

    @staticmethod
    def get_weight_column_name_start() -> str:
        return f"__{PIDCorrection.get_weight_column_base_name()}_"

    @staticmethod
    def get_name_from_weight_column(weight_column: str) -> str:
        return weight_column.split(PIDCorrection.get_weight_column_name_start())[1].rsplit("__", 1)[0]

    @staticmethod
    def get_variation_columns(columns: Sequence[str]) -> List[Tuple[str, str]]:
        var_columns = [
            c
            for c in columns
            if c.startswith(f"{PIDCorrection.get_weight_column_name_start()}{PIDCorrection.variation_name}")
            or c.startswith(f"{PIDCorrection.get_weight_column_name_start()}{PIDCorrection.uncertainty_from_var_name}")
        ]

        var_names = [PIDCorrection.get_name_from_weight_column(c) for c in var_columns]

        return list(zip(var_columns, var_names))

    @staticmethod
    def get_var_col_name(variation_index) -> str:
        return PIDCorrection.get_weight_column_name(column_name=f"{PIDCorrection.variation_name}_{variation_index}")

    @staticmethod
    def get_uncert_from_var_col_name() -> str:
        return PIDCorrection.get_weight_column_name(column_name=PIDCorrection.uncertainty_from_var_name)


class LeptonIDCorrection(PIDCorrection):
    """
    Abstract class for lepton pID correction.
    """

    def __init__(self, variations: Optional[int] = None) -> None:
        super().__init__(variations=variations)
        self._columns = self.get_lepton_columns()  # type: List[str]
        self._momentum_bins = self.get_momentum_bins()  # type: List[float]

        self._correction_columns = self.get_lepton_correction_columns()  # type: List[str]

        self._momentum_column = "momentum_lower"  # type: str
        self._theta_column = "theta_lower"  # type: str
        self._cut_column = "cut"  # type: str

        self._sub_dir = "lepton_id_tables"  # type: str

    @abstractmethod
    def get_pid_table_dict(self) -> Dict[str, Tuple[int, int]]:
        raise NotImplementedError

    def _get_action_for_pid_error_flag(self, flag_value) -> bool:
        raise RuntimeError(
            f"This should not happen. pID error flags are not defined " f"for the {self.__class__.__name__} class!"
        )

    def _get_pid_type_hint_on_error_flag(self) -> int:
        raise RuntimeError(
            f"This should not happen. pID error flags are not defined " f"for the {self.__class__.__name__} class!"
        )

    def _build_varied_pid_matrix(self, table_name: str, stat_seed: int, sys_seed: int) -> np.ndarray:
        ratio_matrix_name = f"{table_name}_pid_cut{self.pid_cut}_ratio"  # type: str
        stat_matrix_name = f"{table_name}_pid_cut{self.pid_cut}_stat"  # type: str
        sys1_matrix_name = f"{table_name}_pid_cut{self.pid_cut}_sys1"  # type: str
        sys2_matrix_name = f"{table_name}_pid_cut{self.pid_cut}_sys2"  # type: str

        ratio_matrix = copy.deepcopy(pid_table_manager.get_correction_matrix(ratio_matrix_name))
        stat_matrix = copy.deepcopy(pid_table_manager.get_correction_matrix(stat_matrix_name))
        sys1_matrix = copy.deepcopy(pid_table_manager.get_correction_matrix(sys1_matrix_name))
        sys2_matrix = copy.deepcopy(pid_table_manager.get_correction_matrix(sys2_matrix_name))

        assert ratio_matrix is not None
        assert stat_matrix is not None
        assert sys1_matrix is not None
        assert sys2_matrix is not None

        sys_matrix = np.sqrt(sys1_matrix * sys1_matrix + sys2_matrix * sys2_matrix)  # type: np.ndarray

        sys_nrs = np.random.RandomState(seed=sys_seed)
        sys_smearing = sys_nrs.normal(loc=0.0, scale=1.0, size=1)
        smeared_sys = sys_matrix * sys_smearing  # type: np.ndarray

        stat_nrs = np.random.RandomState(seed=stat_seed)
        smeared_stat = stat_nrs.normal(loc=0.0, scale=stat_matrix)  # type: np.ndarray

        smeared_matrix = ratio_matrix + smeared_sys + smeared_stat  # type: np.ndarray
        return smeared_matrix

    @staticmethod
    def get_lepton_columns() -> List[str]:
        return [
            "cut",
            "theta_lower",
            "theta_upper",
            "momentum_lower",
            "momentum_upper",
            "ratio",
            "stat",
            "sys1",
            "sys2",
        ]

    @staticmethod
    def get_lepton_correction_columns() -> List[str]:
        return [
            "ratio",
            "stat",
            "sys1",
            "sys2",
        ]

    @staticmethod
    @abstractmethod
    def get_theta_bins() -> List[float]:
        raise NotImplementedError

    @staticmethod
    def get_momentum_bins() -> List[float]:
        return [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0]

    @staticmethod
    @abstractmethod
    def get_exp_ranges() -> List[Tuple[int, int]]:
        raise NotImplementedError


class ElectronIDCorrection(LeptonIDCorrection):
    """
    Class for electron pID correction.
    """

    def __init__(self, pid_cut: float, variations: Optional[int] = None) -> None:
        """
        Electron pID correction. Generates the pID correction information
        given a specific pID cut value.
        :param pid_cut: pID cut value
        :param variations: Number of variations to be performed via smearing of the weights.
        """
        if pid_cut not in self.get_e_pid_cut_values():
            raise ValueError(
                f"Electron ID Corrections are only available for "
                f"eID cuts in {self.get_e_pid_cut_values()}.\n"
                f"You try to apply them for the cut value {pid_cut}"
            )

        super().__init__(variations=variations)
        self._pid_cut = pid_cut  # type: float
        self._pid_base_seed = 1  # type: int
        self._theta_bins = self.get_theta_bins()  # type: List[float]

    def get_pid_table_dict(self) -> Dict[str, Tuple[int, int]]:
        return self._get_pid_table_dict()

    @staticmethod
    def _get_pid_table_dict() -> Dict[str, Tuple[int, int]]:
        return OrderedDict(
            [
                ("lepEffCorr_electron_svd1", (7, 27)),
                ("lepEffCorr_electron_svd2", (31, 65)),
            ]
        )

    @staticmethod
    def get_e_pid_cut_values() -> List[float]:
        return [0.1, 0.5, 0.6, 0.8, 0.9]

    @staticmethod
    def get_theta_bins() -> List[float]:
        return [18, 25, 35, 40, 60, 125, 132, 151]

    @staticmethod
    def get_exp_ranges() -> List[Tuple[int, int]]:
        return [v for v in ElectronIDCorrection._get_pid_table_dict().values()]


class MuonIDCorrection(LeptonIDCorrection):
    """
    Class for muon pID correction.
    """

    def __init__(self, pid_cut: float, variations: Optional[int] = None) -> None:
        """
        Muon pID correction. Generates the pID correction information
        given a specific pID cut value.
        :param pid_cut: pID cut value
        :param variations: Number of variations to be performed via smearing of the weights.
        """
        if pid_cut not in self.get_mu_pid_cut_values():
            raise ValueError(
                f"Muon ID Corrections are only available for "
                f"muID cuts in {self.get_mu_pid_cut_values()}.\n"
                f"You try to apply them for the cut value {pid_cut}"
            )

        super().__init__(variations=variations)
        self._pid_cut = pid_cut  # type: float
        self._pid_base_seed = 2  # type: int
        self._theta_bins = self.get_theta_bins()  # type: List[float]

    def get_pid_table_dict(self) -> Dict[str, Tuple[int, int]]:
        return self._get_pid_table_dict()

    @staticmethod
    def _get_pid_table_dict() -> Dict[str, Tuple[int, int]]:
        return OrderedDict(
            [
                ("lepEffCorr_muon_svd1", (7, 27)),
                ("lepEffCorr_muon_exp31", (31, 39)),
                ("lepEffCorr_muon_exp41", (41, 49)),
                ("lepEffCorr_muon_exp51", (51, 65)),
            ]
        )

    @staticmethod
    def get_mu_pid_cut_values() -> List[float]:
        return [0.1, 0.8, 0.9, 0.95, 0.97]

    @staticmethod
    def get_theta_bins() -> List[float]:
        return [17, 25, 37, 51, 117, 130, 145, 150]

    @staticmethod
    def get_exp_ranges() -> List[Tuple[int, int]]:
        return [v for v in MuonIDCorrection._get_pid_table_dict().values()]


class KPiIDCorrection(PIDCorrection):
    """
    Abstract class for kaon or pion pID correction
    """

    def __init__(self, pid_cut: float, variations: Optional[int] = None) -> None:
        """
        Kaon/pion pID correction. Generates the pID correction information
        given a specific pID cut value.
        :param pid_cut: pID cut value
        :param variations: Number of variations to be performed via smearing of the weights.
        """
        if pid_cut not in self.get_k_pi_pid_cut_values():
            raise ValueError(
                f"Kaon/Pion ID Corrections are only available for "
                f"kID/piID cuts in {self.get_k_pi_pid_cut_values()}.\n"
                f"You try to apply them for the cut value {pid_cut}"
            )

        super().__init__(variations=variations)
        self._columns = self.get_k_pi_columns()  # type: List[str]
        self._momentum_bins = self.get_momentum_bins()  # type: List[float]
        self._theta_bins = self.get_theta_bins()  # type: List[float]

        self._correction_columns = self.get_k_pi_correction_columns()  # type: List[str]

        self._momentum_column = "mom_lower"  # type: str
        self._theta_column = "theta_lower"  # type: str
        self._cut_column = "cut"  # type: str
        self._flag_column = "flag"  # type: str

        self._sub_dir = "kaon_pion_id_tables"  # type: str
        self._pid_cut = pid_cut  # type: float
        self._pid_base_seed = self.get_kaon_pion_base_seed()  # type: int

    @abstractmethod
    def get_pid_table_dict(self) -> Dict[str, Tuple[int, int]]:
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def get_pid_type_hint() -> int:
        raise NotImplementedError()

    def _get_action_for_pid_error_flag(self, flag_value) -> bool:
        return self.actions_for_k_pi_error_flag()[flag_value]

    def _get_pid_type_hint_on_error_flag(self) -> int:
        return self.get_pid_type_hint()

    def _build_varied_pid_matrix(self, table_name: str, stat_seed: int, sys_seed: int) -> np.ndarray:
        ratio_matrix_name = f"{table_name}_pid_cut{self.pid_cut}_ratio"  # type: str
        stat_matrix_name = f"{table_name}_pid_cut{self.pid_cut}_stat"  # type: str
        sys_matrix_name = f"{table_name}_pid_cut{self.pid_cut}_sys"  # type: str

        ratio_matrix = copy.deepcopy(pid_table_manager.get_correction_matrix(ratio_matrix_name))
        stat_matrix = copy.deepcopy(pid_table_manager.get_correction_matrix(stat_matrix_name))
        sys_matrix = copy.deepcopy(pid_table_manager.get_correction_matrix(sys_matrix_name))

        assert ratio_matrix is not None
        assert stat_matrix is not None
        assert sys_matrix is not None

        # Handling flagged table bins:
        ratio_matrix[ratio_matrix < 0] = 1.0
        # stat_matrix[stat_matrix < 0] = 1.1 * np.amax(stat_matrix)
        # sys_matrix[sys_matrix < 0] = 1.1 * np.amax(sys_matrix)
        stat_matrix[stat_matrix < 0] = 1.1 * np.quantile(np.asarray(stat_matrix[stat_matrix >= 0.0]).reshape(-1), 0.95)
        sys_matrix[sys_matrix < 0] = 1.1 * np.quantile(np.asarray(sys_matrix[sys_matrix >= 0.0]).reshape(-1), 0.95)

        sys_nrs = np.random.RandomState(seed=sys_seed)
        sys_smearing = sys_nrs.normal(loc=0.0, scale=1.0, size=1)  # type: np.ndarray
        smeared_sys = sys_matrix * sys_smearing  # type: np.ndarray

        stat_nrs = np.random.RandomState(seed=stat_seed)
        smeared_stat = stat_nrs.normal(loc=0.0, scale=stat_matrix)  # type: np.ndarray

        smeared_matrix = ratio_matrix + smeared_sys + smeared_stat  # type: np.ndarray
        return smeared_matrix

    @staticmethod
    def get_k_pi_original_columns() -> List[str]:
        return [
            "mode",
            "cut",
            "eff_data",
            "stat_data",
            "sys_data",
            "eff_mc",
            "stat_mc",
            "ratio",
            "stat",
            "sys",
            "flag",
        ]

    @staticmethod
    def get_k_pi_correction_columns() -> List[str]:
        return [
            "ratio",
            "stat",
            "sys",
        ]

    @staticmethod
    def get_k_pi_columns() -> List[str]:
        original_cols = KPiIDCorrection.get_k_pi_original_columns()
        extended_cols = original_cols + [
            "mom_lower",
            "mom_upper",
            "theta_lower",
            "theta_upper",
        ]
        return extended_cols

    @staticmethod
    def get_momentum_bins() -> List[float]:
        return [
            0,
            0.5,
            0.6,
            0.7,
            0.8,
            0.9,
            1.0,
            1.1,
            1.2,
            1.3,
            1.4,
            1.5,
            1.6,
            1.7,
            1.8,
            1.9,
            2.0,
            2.1,
            2.2,
            2.3,
            2.4,
            2.5,
            2.6,
            2.7,
            2.8,
            2.9,
            3.0,
            3.2,
            3.4,
            3.6,
            4.0,
            4.5,
            100.0,
        ]

    @staticmethod
    def get_theta_bins() -> List[float]:
        return [0, 33, 46, 57, 64, 69, 78, 89, 99, 107, 121, 128, 180]

    @staticmethod
    @abstractmethod
    def get_exp_ranges() -> List[Tuple[int, int]]:
        raise NotImplementedError

    @staticmethod
    def get_k_pi_mode() -> Dict[int, str]:
        return {
            0: "kaonEff",
            1: "kaon_pionFake",
            2: "pionEff",
            3: "pion_kaonFake",
        }

    @staticmethod
    def get_k_pi_flag() -> Dict[int, str]:
        return {
            0: "OK",
            1: "no_stats",
            2: "fit_failed",
            3: "fail_svd2",
        }

    @staticmethod
    def actions_for_k_pi_error_flag() -> Dict[int, bool]:
        return {
            0: False,
            1: True,
            2: True,
            3: True,
        }

    @staticmethod
    def get_k_pi_pid_cut_values() -> List[float]:
        return [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]

    @staticmethod
    @abstractmethod
    def get_kaon_pion_base_seed() -> int:
        raise NotImplementedError()


class KaonIDCorrection(KPiIDCorrection):
    """
    Class for kaon pID correction.
    """

    def get_pid_table_dict(self) -> Dict[str, Tuple[int, int]]:
        return self._get_pid_table_dict()

    @staticmethod
    def _get_pid_table_dict() -> Dict[str, Tuple[int, int]]:
        return OrderedDict(
            [
                ("kaonEff_SVD1", (7, 27)),
                ("kaonEff_SVD2", (31, 65)),
            ]
        )

    @staticmethod
    def get_pid_type_hint() -> int:
        return -1

    @staticmethod
    def get_kaon_pion_base_seed() -> int:
        return 3

    @staticmethod
    def get_exp_ranges() -> List[Tuple[int, int]]:
        return [v for v in KaonIDCorrection._get_pid_table_dict().values()]


class PionIDCorrection(KPiIDCorrection):
    """
    Class for pion pID correction.
    """

    def get_pid_table_dict(self) -> Dict[str, Tuple[int, int]]:
        return self._get_pid_table_dict()

    @staticmethod
    def _get_pid_table_dict() -> Dict[str, Tuple[int, int]]:
        return OrderedDict(
            [
                ("pionEff_SVD1", (7, 27)),
                ("pionEff_SVD2", (31, 65)),
            ]
        )

    @staticmethod
    def get_pid_type_hint() -> int:
        return -2

    @staticmethod
    def get_kaon_pion_base_seed() -> int:
        return 4

    @staticmethod
    def get_exp_ranges() -> List[Tuple[int, int]]:
        return [v for v in PionIDCorrection._get_pid_table_dict().values()]


class KaonFromPionFakeIDCorrection(KPiIDCorrection):
    """
    Class for kaon pID fakes from pions correction.
    """

    def get_pid_table_dict(self) -> Dict[str, Tuple[int, int]]:
        return self._get_pid_table_dict()

    @staticmethod
    def _get_pid_table_dict() -> Dict[str, Tuple[int, int]]:
        return OrderedDict(
            [
                ("kaon_pionFake_SVD1", (7, 27)),
                ("kaon_pionFake_SVD2", (31, 65)),
            ]
        )

    @staticmethod
    def get_pid_type_hint() -> int:
        return -3

    @staticmethod
    def get_kaon_pion_base_seed() -> int:
        return 5

    @staticmethod
    def get_exp_ranges() -> List[Tuple[int, int]]:
        return [v for v in KaonFromPionFakeIDCorrection._get_pid_table_dict().values()]


class PionFromKaonFakeIDCorrection(KPiIDCorrection):
    """
    Class for pion pID fakes from kaons correction.
    """

    def get_pid_table_dict(self) -> Dict[str, Tuple[int, int]]:
        return self._get_pid_table_dict()

    @staticmethod
    def _get_pid_table_dict() -> Dict[str, Tuple[int, int]]:
        return OrderedDict(
            [
                ("pion_kaonFake_SVD1", (7, 27)),
                ("pion_kaonFake_SVD2", (31, 65)),
            ]
        )

    @staticmethod
    def get_pid_type_hint() -> int:
        return -4

    @staticmethod
    def get_kaon_pion_base_seed() -> int:
        return 6

    @staticmethod
    def get_exp_ranges() -> List[Tuple[int, int]]:
        return [v for v in PionFromKaonFakeIDCorrection._get_pid_table_dict().values()]


class LeptonFromHadronFakeIDCorrection(PIDCorrection):
    """
    Abstract class for lepton pID fake correction for fakes from hadrons.
    """

    def __init__(self, variations: Optional[int] = None) -> None:
        super().__init__(variations=variations)

        self._columns = self.get_lepton_fake_columns()  # type: List[str]
        self._momentum_bins = self.get_momentum_bins()  # type: List[float]

        self._correction_columns = self.get_lepton_fake_correction_columns()  # type: List[str]

        self._momentum_column = "momentum_lower"  # type: str
        self._theta_column = "theta_lower"  # type: str
        self._cut_column = "cut"  # type: str

        self._sub_dir = "lepton_fake_tables"  # type: str

    @abstractmethod
    def get_pid_table_dict(self) -> Dict[str, Tuple[int, int]]:
        raise NotImplementedError

    def _get_action_for_pid_error_flag(self, flag_value) -> bool:
        raise RuntimeError(
            f"This should not happen. pID error flags are not defined " f"for the {self.__class__.__name__} class!"
        )

    def _get_pid_type_hint_on_error_flag(self) -> int:
        raise RuntimeError(
            f"This should not happen. pID error flags are not defined " f"for the {self.__class__.__name__} class!"
        )

    def _build_varied_pid_matrix(self, table_name: str, stat_seed: int, sys_seed: int) -> np.ndarray:
        ratio_matrix_name = f"{table_name}_pid_cut{self.pid_cut}_ratio"
        uncert_matrix_name = f"{table_name}_pid_cut{self.pid_cut}_uncertainty"

        ratio_matrix = copy.deepcopy(pid_table_manager.get_correction_matrix(ratio_matrix_name))
        uncert_matrix = copy.deepcopy(pid_table_manager.get_correction_matrix(uncert_matrix_name))

        assert ratio_matrix is not None
        assert uncert_matrix is not None

        uncert_nrs = np.random.RandomState(seed=stat_seed + sys_seed)
        smeared_uncert = uncert_nrs.normal(0, uncert_matrix)

        smeared_matrix = ratio_matrix + smeared_uncert
        return smeared_matrix

    @staticmethod
    def get_lepton_fake_columns() -> List[str]:
        return [
            "cut",
            "theta_lower",
            "theta_upper",
            "momentum_lower",
            "momentum_upper",
            "ratio",
            "uncertainty",
        ]

    @staticmethod
    def get_lepton_fake_correction_columns() -> List[str]:
        return [
            "ratio",
            "uncertainty",
        ]

    @staticmethod
    def get_momentum_bins() -> List[float]:
        return [0.0, 0.5, 0.75, 1.0, 1.25, 1.5, 1.75, 2.0, 2.25, 2.5, 3.0, 4.0]

    @staticmethod
    def get_exp_ranges() -> List[Tuple[int, int]]:
        return [(7, 65)]


class ElectronFromHadronFakeIDCorrection(LeptonFromHadronFakeIDCorrection):
    """
    Class for the electron pID fake correction for fakes from hadrons (pions or kaons).
    """

    def __init__(
        self,
        pid_cut: float,
        hadron_type: str,
        variations: Optional[int] = None,
    ) -> None:
        """
        Electron pID fake correction for fakes from hadrons (pions or kaons).
        pID cut value must be specified via pid_cut parameter.
        Hadron type must be set via the parameter hadron_type.
        :param pid_cut: pID cut value.
        :param hadron_type: Hadron type.
        :param variations: Number of variations to be performed via smearing of the weights.
        """
        if pid_cut not in self.get_e_fake_pid_cut_values():
            raise ValueError(
                f"Electron from Hadron Fake ID Corrections are only available for "
                f"eID cuts in {self.get_e_fake_pid_cut_values()}.\n"
                f"You try to apply them for the cut value {pid_cut}"
            )
        if hadron_type not in ["pion", "kaon"]:
            raise ValueError(
                f"Hadron type must be either 'kaon' or 'pion'\n." f"The provided hadron type was {hadron_type}"
            )

        super().__init__(variations=variations)

        self.hadron_type = hadron_type  # type: str

        self._theta_bins = self.get_theta_bins()  # type: List[float]
        self._pid_cut = pid_cut  # type: float

        self._pid_base_seed = 7  # type: int
        if self.hadron_type == "kaon":
            self._pid_base_seed = 8

    def get_pid_table_dict(self) -> Dict[str, Tuple[int, int]]:
        if self.hadron_type == "pion":
            return OrderedDict(
                [
                    ("LepFake_pi2e", (7, 65)),
                ]
            )
        elif self.hadron_type == "kaon":
            return OrderedDict(
                [
                    ("LepFake_K2e", (7, 65)),
                ]
            )
        else:
            raise ValueError(
                f"Hadron type must be either 'kaon' or 'pion'\n." f"The provided hadron type was {self.hadron_type}"
            )

    @staticmethod
    def get_e_fake_pid_cut_values() -> List[float]:
        return [0.6, 0.9]

    @staticmethod
    def get_theta_bins() -> List[float]:
        return [0, 18, 25, 35, 40, 60, 125, 132, 151]


class MuonFromHadronFakeIDCorrection(LeptonFromHadronFakeIDCorrection):
    """
    Class for the muon pID fake correction for fakes from hadrons (pions or kaons).
    """

    def __init__(
        self,
        pid_cut: float,
        hadron_type: str,
        variations: Optional[int] = None,
    ) -> None:
        """
        Moun pID fake correction for fakes from hadrons (pions or kaons).
        pID cut value must be specified via pid_cut parameter.
        Hadron type must be set via the parameter hadron_type.
        :param pid_cut: pID cut value.
        :param hadron_type: Hadron type.
        :param variations: Number of variations to be performed via smearing of the weights.
        """
        if pid_cut not in self.get_mu_fake_pid_cut_values():
            raise ValueError(
                f"Muon from Hadron Fake ID Corrections are only available for "
                f"muID cuts in {self.get_mu_fake_pid_cut_values()}.\n"
                f"You try to apply them for the cut value {pid_cut}"
            )
        if hadron_type not in ["pion", "kaon"]:
            raise ValueError(
                f"Hadron type must be either 'kaon' or 'pion'\n." f"The provided hadron type was {hadron_type}"
            )

        super().__init__(variations=variations)

        self.hadron_type = hadron_type  # type: str

        self._theta_bins = self.get_theta_bins()  # type: List[float]
        self._pid_cut = pid_cut  # type: float

        self._pid_base_seed = 9  # type: int
        if self.hadron_type == "kaon":
            self._pid_base_seed = 10

    def get_pid_table_dict(self) -> OrderedDict:
        if self.hadron_type == "pion":
            return OrderedDict(
                [
                    ("LepFake_pi2mu", (7, 65)),
                ]
            )
        elif self.hadron_type == "kaon":
            return OrderedDict(
                [
                    ("LepFake_K2mu", (7, 65)),
                ]
            )
        else:
            raise ValueError(
                f"Hadron type must be either 'kaon' or 'pion'\n." f"The provided hadron type was {self.hadron_type}"
            )

    @staticmethod
    def get_mu_fake_pid_cut_values() -> List[float]:
        return [0.9]

    @staticmethod
    def get_theta_bins() -> List[float]:
        return [0, 17, 25, 37, 51, 117, 130, 145, 150]
