"""
Management of evaluation of particle ID correction for Belle

Felix Metzner 2019
"""

import os
import json
import numpy as np
import pandas as pd

from collections.abc import MutableMapping
from typing import (
    Tuple,
    List,
    Dict,
    Optional,
    NamedTuple,
    Any,
    KeysView,
    ValuesView,
    ItemsView,
    Iterator,
    Mapping,
    Type,
)

from analysistools.utilities import PathType
from analysistools.pid_corrections import pid_corrector


__all__ = [
    "PIDWeightDefinitions",
    "PIDWeightInfo",
    "PIDWeightEvaluation",
    "PIDWeightInfoContainer",
]


class PIDWeightDefinitions:
    weight_type_column_name = "weight_type"  # type: str

    momentum_column_name = "momentum"  # type: str
    theta_column_name = "theta"  # type: str
    momentum_indices_column_name = "momentum_bin_indices"  # type: str
    theta_indices_column_name = "theta_bin_indices"  # type: str

    exp_column_name = "expNum"  # type: str

    hadron_particle_names = ("k", "pi")  # type: Tuple[str, ...]

    particle_map = {
        "e": "e",
        "mu": r"$\mu$",
        "k": "K",
        "pi": r"$\pi$",
    }  # type: Dict[str, str]

    pid_corrector_class_mapping = {
        ("e", None): pid_corrector.ElectronIDCorrection,
        ("e", "pi"): pid_corrector.ElectronFromHadronFakeIDCorrection,
        ("e", "k"): pid_corrector.ElectronFromHadronFakeIDCorrection,
        ("mu", None): pid_corrector.MuonIDCorrection,
        ("mu", "pi"): pid_corrector.MuonFromHadronFakeIDCorrection,
        ("mu", "k"): pid_corrector.MuonFromHadronFakeIDCorrection,
        ("k", None): pid_corrector.KaonIDCorrection,
        ("k", "pi"): pid_corrector.KaonFromPionFakeIDCorrection,
        ("pi", None): pid_corrector.PionIDCorrection,
        ("pi", "k"): pid_corrector.PionFromKaonFakeIDCorrection,
    }  # type: Dict[Tuple[str, Optional[str]], Type[pid_corrector.PIDCorrection]]


class PIDWeightInfoTuple(NamedTuple):
    name: str
    daughter_index: Optional[int]
    is_hadron: bool
    fake_from: Optional[str]


class PIDWeightInfo(NamedTuple):
    name: str
    condition: pd.Series
    binning: Tuple[List[float], List[float]]
    exp_ranges: List[Tuple[int, int]]
    is_hadron: bool
    fake_from: Optional[str]
    daughter_index: Optional[int]
    combined: bool
    hdf5_path: Optional[PathType] = None

    def add_to_info(
        self,
        name: str,
        condition: pd.Series,
        binning: Tuple[List[float], List[float]],
        exp_ranges: List[Tuple[int, int]],
        is_hadron: bool,
        fake_from: Optional[str],
        daughter_index: Optional[int],
        combined: bool,
        hdf5_path: Optional[PathType] = None,
    ) -> "PIDWeightInfo":
        assert self.exp_ranges == exp_ranges, (self.exp_ranges, exp_ranges)
        assert self.name == name, (self.name, name)
        assert self.is_hadron == is_hadron, (self.is_hadron, is_hadron)
        assert self.fake_from == fake_from, (self.fake_from, fake_from)
        assert self.hdf5_path == hdf5_path, (self.hdf5_path, hdf5_path)

        new_momentum_binning = PIDWeightInfo.update_binning(old=self.binning[0], new=binning[0])
        new_theta_binning = PIDWeightInfo.update_binning(old=self.binning[1], new=binning[1])
        new_binning = (new_momentum_binning, new_theta_binning)

        new_daughter_index = None  # type: Optional[int]
        if self.daughter_index is daughter_index:
            new_daughter_index = daughter_index

        return self._replace(
            condition=(condition | self.condition),
            binning=new_binning,
            combined=combined,
            daughter_index=new_daughter_index,
        )

    def add_data_frame_path(self, hdf5_path: PathType) -> "PIDWeightInfo":
        assert self.hdf5_path is None, hdf5_path
        return self._replace(hdf5_path=hdf5_path)

    def as_attribute_dict(self) -> Dict[str, Any]:
        return {fk: self.__getattribute__(fk) for fk in self._fields}

    def get_name_for_title(self) -> str:
        title = PIDWeightInfo.map_particle(self.name)  # type: str

        if self.fake_from is not None:
            title += fr" Fakes from {self.map_particle(self.fake_from)}"

        if self.daughter_index is not None:
            title += fr" (D Daughter {self.daughter_index + 1})"

        return fr"All {title}" if self.combined else title

    @staticmethod
    def map_particle(particle: str) -> str:
        assert particle in PIDWeightDefinitions.particle_map.keys(), (particle, PIDWeightDefinitions.particle_map.keys())
        return PIDWeightDefinitions.particle_map[particle]

    @staticmethod
    def map_pid_key_to_name(pid_key: str) -> str:
        name, _ = pid_key.split(sep="ID", maxsplit=1)
        assert name in PIDWeightDefinitions.particle_map.keys(), (name, list(PIDWeightDefinitions.particle_map.keys()))
        return name

    @staticmethod
    def get_binning_from_class(
        pid_cls: Type[pid_corrector.PIDCorrection],
    ) -> Tuple[List[float], List[float], List[Tuple[int, int]]]:
        momentum = pid_cls.get_momentum_bins()
        theta = pid_cls.get_theta_bins()
        exp_ranges = pid_cls.get_exp_ranges()
        return momentum, theta, exp_ranges

    @staticmethod
    def get_base_binning_for(
        particle: str,
        fake_from: Optional[str],
    ) -> Tuple[List[float], List[float], List[Tuple[int, int]]]:
        if (particle, fake_from) in PIDWeightDefinitions.pid_corrector_class_mapping.keys():
            pid_corrector_class = PIDWeightDefinitions.pid_corrector_class_mapping[(particle, fake_from)]
            return PIDWeightInfo.get_binning_from_class(pid_cls=pid_corrector_class)
        else:
            fake_string = ""
            if fake_from is not None:
                fake_string = f" and fake_from = {fake_from}"
            raise ValueError(f"No pID correction binning for particle = {particle}{fake_string} available!")

    @staticmethod
    def get_binning(
        df: pd.DataFrame,
        particle: str,
        fake_from: Optional[str] = None,
    ) -> Tuple[List[float], List[float], List[Tuple[int, int]]]:
        assert PIDWeightDefinitions.momentum_column_name in df.columns
        assert PIDWeightDefinitions.theta_column_name in df.columns

        momentum, theta, exp_ranges = PIDWeightInfo.get_base_binning_for(particle=particle, fake_from=fake_from)

        momentum[0] = 0.0
        momentum.append(np.ceil(df[PIDWeightDefinitions.momentum_column_name].max()))
        first_invalid_mom_bin = next(i for i, v in enumerate(momentum) if v >= momentum[-1]) + 1  # type: int
        if first_invalid_mom_bin < len(momentum):
            momentum = momentum[:first_invalid_mom_bin]

        theta[0] = 0.0
        theta[-1] = 180.0

        assert df[PIDWeightDefinitions.momentum_column_name].min() >= momentum[0]
        assert df[PIDWeightDefinitions.momentum_column_name].max() <= momentum[-1]
        assert all(momentum[k] < momentum[k + 1] for k in range(len(momentum) - 1)), momentum
        assert df[PIDWeightDefinitions.theta_column_name].min() >= theta[0]
        assert df[PIDWeightDefinitions.theta_column_name].max() <= theta[-1]
        assert all(theta[k] < theta[k + 1] for k in range(len(theta) - 1)), theta
        assert len(exp_ranges) > 0
        assert all(isinstance(exp_range, tuple) for exp_range in exp_ranges)

        return momentum, theta, exp_ranges

    @staticmethod
    def update_binning(old: List[float], new: List[float]) -> List[float]:
        assert old[0] == new[0], (old[0], new[0])
        max_index = min(len(old), len(new)) - 1  # type int
        assert all(old[i] == new[i] for i in range(max_index)), (old, new)

        if len(old) > len(new):
            selected = old  # type: List[float]
        elif len(new) > len(old):
            selected = new
        else:
            assert len(old) == len(new), (old, new)
            selected = old
            if old[max_index] < new[max_index]:
                selected[max_index] = new[max_index]

        assert min(selected) == min(old + new), (selected, old, new)
        assert max(selected) == max(old + new), (selected, old, new)
        assert len(selected) == max(len(old), len(new)), (selected, old, new)
        assert all(selected[k] < selected[k + 1] for k in range(len(selected) - 1)), (selected, old, new)
        return selected

    @staticmethod
    def convert_rad_to_degree(df: pd.DataFrame, hint: Optional[str]) -> None:
        assert (
            df[PIDWeightDefinitions.theta_column_name].max() < 5 or len(df.index) < 10
        ), f"{df[PIDWeightDefinitions.theta_column_name].max()}, {hint if hint is not None else 'None'}"
        df.loc[:, PIDWeightDefinitions.theta_column_name] = df[PIDWeightDefinitions.theta_column_name] / np.pi * 180.0

    @staticmethod
    def check_if_theta_is_in_degrees(df: pd.DataFrame, hint: Optional[str]) -> None:
        assert (
            df[PIDWeightDefinitions.theta_column_name].max() > 4 or len(df.index) < 10
        ), f"{df[PIDWeightDefinitions.theta_column_name].max()}, {hint if hint is not None else 'None'}"

    @staticmethod
    def map_pid_key_to_info(pid_key: str) -> PIDWeightInfoTuple:
        d_daughter_index = None  # type: Optional[int]
        fake_from = None  # type: Optional[str]

        _pid_key = pid_key.replace("from", "From")  # type: str
        particle, rest = _pid_key.split(sep="ID", maxsplit=1)  # type: str, str

        is_hadron = particle in PIDWeightDefinitions.hadron_particle_names  # type: bool

        if not rest:
            return PIDWeightInfoTuple(
                name=particle,
                daughter_index=d_daughter_index,
                is_hadron=is_hadron,
                fake_from=fake_from,
            )

        if rest[0].isdigit():
            d_daughter_index = int(rest[0])
            rest = rest[1:]

        if not rest:
            return PIDWeightInfoTuple(
                name=particle,
                daughter_index=d_daughter_index,
                is_hadron=is_hadron,
                fake_from=fake_from,
            )

        if rest.startswith("From"):
            fake_from = rest.split("From")[1].lower()
        else:
            raise RuntimeError(f"Encountered unexpected weight type: {_pid_key}")

        return PIDWeightInfoTuple(
            name=particle,
            daughter_index=d_daughter_index,
            is_hadron=is_hadron,
            fake_from=fake_from,
        )

    def as_dict(self, remove_condition_info: bool = True) -> Dict[str, Any]:
        return {
            "name": self.name,
            "binning": self.binning,
            "exp_ranges": self.exp_ranges,
            "is_hadron": self.is_hadron,
            "fake_from": self.fake_from,
            "daughter_index": self.daughter_index,
            "combined": self.combined,
            "condition": [sum(self.condition), len(self.condition)] if remove_condition_info else list(self.condition),
            "hdf5_path": self.hdf5_path,
        }

    @staticmethod
    def from_dict(weight_info_as_dict: Dict[str, Any]) -> "PIDWeightInfo":
        assert all(k in weight_info_as_dict.keys() for k in PIDWeightInfo._fields), [
            k for k in PIDWeightInfo._fields if k not in weight_info_as_dict.keys()
        ]
        assert all(k in PIDWeightInfo._fields for k in weight_info_as_dict.keys()), [
            k for k in weight_info_as_dict.keys() if k not in PIDWeightInfo._fields
        ]
        return PIDWeightInfo(**{k: pd.Series(v) if k == "condition" else v for k, v in weight_info_as_dict.items()})


class PIDWeightInfoContainer(MutableMapping):
    combined_key_appendix = "All"  # type: str

    def __init__(self) -> None:
        super().__init__()
        self.__pid_particle_infos = {}  # type: Dict[str, PIDWeightInfo]

    def __getitem__(self, item: str) -> PIDWeightInfo:
        assert isinstance(item, str), item.__class__.__name__
        assert item in self.__pid_particle_infos.keys(), item
        return self.__pid_particle_infos[item]

    def __setitem__(self, key: str, value: PIDWeightInfo) -> None:
        assert isinstance(key, str), key.__class__.__name__
        assert isinstance(value, PIDWeightInfo), value.__class__.__name__
        assert key not in self.__pid_particle_infos.keys(), key
        self.__pid_particle_infos[key] = value

    def __delitem__(self, item):
        assert isinstance(item, str), item.__class__.__name__
        assert item in self.__pid_particle_infos.keys(), item
        del self.__pid_particle_infos[item]

    def __iter__(self) -> Iterator[str]:
        return iter(self.__pid_particle_infos)

    def __len__(self) -> int:
        return len(self.__pid_particle_infos.keys())

    def keys(self) -> KeysView[str]:
        return self.__pid_particle_infos.keys()

    def values(self) -> ValuesView[PIDWeightInfo]:
        return self.__pid_particle_infos.values()

    def items(self) -> ItemsView[str, PIDWeightInfo]:
        return self.__pid_particle_infos.items()

    def update(self, __m: Mapping[str, PIDWeightInfo], **kwargs: PIDWeightInfo) -> None:  # type: ignore
        for k, v in __m.items():
            assert isinstance(k, str), k.__class__.__name__
            assert isinstance(v, PIDWeightInfo), v.__class__.__name__
            assert k not in self.__pid_particle_infos.keys(), k
            self.__pid_particle_infos[k] = v
        for k, v in kwargs.items():
            assert isinstance(k, str), k.__class__.__name__
            assert isinstance(v, PIDWeightInfo), v.__class__.__name__
            assert k not in self.__pid_particle_infos.keys(), k
            self.__pid_particle_infos[k] = v

    def update_info(self, key: str, particle_info: Optional[PIDWeightInfo] = None, **kwargs: Any) -> None:
        assert isinstance(key, str), key.__class__.__name__
        if key not in self.__pid_particle_infos.keys():
            if particle_info is not None:
                assert len(kwargs) == 0, kwargs
                self.__pid_particle_infos[key] = particle_info
            else:
                self.__pid_particle_infos[key] = PIDWeightInfo(**kwargs)
        else:
            if particle_info is not None:
                assert len(kwargs) == 0, kwargs
                new_info = self.__pid_particle_infos[key].add_to_info(**particle_info.as_attribute_dict())
            else:
                new_info = self.__pid_particle_infos[key].add_to_info(**kwargs)
            self.__pid_particle_infos[key] = new_info

    def explicit_info_update(
        self,
        key: str,
        name: str,
        condition: pd.Series,
        binning: Tuple[List[float], List[float]],
        exp_ranges: List[Tuple[int, int]],
        is_hadron: bool,
        fake_from: Optional[str],
        daughter_index: Optional[int],
        combined: bool,
    ) -> None:
        assert isinstance(key, str), key.__class__.__name__
        self.update_info(
            key=key,
            particle_info=PIDWeightInfo(
                name=name,
                condition=condition,
                binning=binning,
                exp_ranges=exp_ranges,
                is_hadron=is_hadron,
                fake_from=fake_from,
                daughter_index=daughter_index,
                combined=combined,
            ),
        )

    def update_from_weight_info_tuple(
        self,
        weight_key: str,
        info_tuple: PIDWeightInfoTuple,
        condition: pd.Series,
        binning: Tuple[List[float], List[float]],
        exp_ranges: List[Tuple[int, int]],
    ) -> None:
        assert weight_key not in self.keys(), (weight_key, self.keys())
        self.explicit_info_update(
            key=weight_key,
            name=info_tuple.name,
            condition=condition,
            binning=binning,
            exp_ranges=exp_ranges,
            is_hadron=info_tuple.is_hadron,
            fake_from=info_tuple.fake_from,
            daughter_index=info_tuple.daughter_index,
            combined=False,
        )

        key = self.get_key_for_combined(particle_name=info_tuple.name, fake_from=info_tuple.fake_from)  # type: str

        self.explicit_info_update(
            key=key,
            name=info_tuple.name,
            condition=condition,
            binning=binning,
            exp_ranges=exp_ranges,
            is_hadron=info_tuple.is_hadron,
            fake_from=info_tuple.fake_from,
            daughter_index=info_tuple.daughter_index,
            combined=key in self.keys(),
        )

    @staticmethod
    def get_key_for_combined(particle_name: str, fake_from: Optional[str]) -> str:
        fake_tag = f"From{fake_from.capitalize()}" if fake_from is not None else ""  # type: str
        return f"{particle_name}ID{fake_tag}{PIDWeightInfoContainer.combined_key_appendix}"

    def delete_duplicates(self) -> None:
        keys_of_combined = [k for k in self.keys() if k.endswith(self.combined_key_appendix)]
        duplicate_keys = [k for k in keys_of_combined if not self.__pid_particle_infos[k].combined]
        for duplicate_key in duplicate_keys:
            self.__delitem__(item=duplicate_key)

    def as_dict(self, remove_condition_info: bool = True) -> Dict[str, Dict[str, Any]]:
        return {k: wi.as_dict(remove_condition_info=remove_condition_info) for k, wi in self.__pid_particle_infos.items()}

    @staticmethod
    def from_dict(weight_infos_as_dict: Dict[str, Any]) -> "PIDWeightInfoContainer":
        weight_infos = PIDWeightInfoContainer()
        for k, wi in weight_infos_as_dict.items():
            weight_infos.update({k: PIDWeightInfo.from_dict(weight_info_as_dict=wi)})
        return weight_infos

    def get_pid_types(self) -> List[str]:
        return list(self.__pid_particle_infos.keys())

    def add_data_frame_path_to_info(self, key: str, hdf5_path: PathType) -> None:
        info_to_update = self.__getitem__(key)
        self.__pid_particle_infos[key] = info_to_update.add_data_frame_path(hdf5_path=hdf5_path)

    @staticmethod
    def get_pid_weight_collections(df: pd.DataFrame) -> "PIDWeightInfoContainer":
        assert len(df.index) > 0, "Empty data frame was provided!"
        assert PIDWeightDefinitions.weight_type_column_name in df.columns, PIDWeightDefinitions.weight_type_column_name

        weight_infos = PIDWeightInfoContainer()
        weight_types = list(df[PIDWeightDefinitions.weight_type_column_name].unique())  # type: List[str]

        for weight_type in weight_types:
            cond = df[PIDWeightDefinitions.weight_type_column_name] == weight_type
            assert isinstance(cond, pd.Series), type(cond).__name__

            weight_info_tuple = PIDWeightInfo.map_pid_key_to_info(pid_key=weight_type)

            momentum_bins, theta_bins, exp_ranges = PIDWeightInfo.get_binning(
                df=df.loc[cond],
                particle=weight_info_tuple.name,
                fake_from=weight_info_tuple.fake_from,
            )

            weight_infos.update_from_weight_info_tuple(
                weight_key=weight_type,
                info_tuple=weight_info_tuple,
                condition=cond,
                binning=(momentum_bins, theta_bins),
                exp_ranges=exp_ranges,
            )

        weight_infos.delete_duplicates()

        return weight_infos


class PIDWeightEvaluation:
    weight_type_column_name = PIDWeightDefinitions.weight_type_column_name  # type: str
    output_mapping_dict_output_key = "pid_info_dfs_mapping"  # type: str
    weight_info_dfs_sub_dir_name = "weight_info_data_frames"  # type: str

    def __init__(
        self,
        path_to_pid_info_data_frame: PathType,
        output_dir_path: PathType,
        with_selection: bool,
    ) -> None:
        self._df_path = path_to_pid_info_data_frame  # type: PathType
        self._output_dir = output_dir_path  # type: PathType
        self._with_selection = with_selection  # type: bool

        self._base_df = None  # type: Optional[pd.DataFrame]
        self._pid_weight_info_container = None  # type: Optional[PIDWeightInfoContainer]

    @property
    def base_df(self) -> pd.DataFrame:
        if self._base_df is None:
            self._base_df = pd.read_hdf(path_or_buf=self._df_path)
            assert isinstance(self._base_df, pd.DataFrame), type(self._base_df)
            PIDWeightInfo.convert_rad_to_degree(df=self._base_df, hint=f"Loading base data frame from {self._df_path}")
        return self._base_df

    @property
    def pid_weight_infos(self) -> PIDWeightInfoContainer:
        if self._pid_weight_info_container is None:
            self._pid_weight_info_container = PIDWeightInfoContainer.get_pid_weight_collections(df=self.base_df)
            assert isinstance(self._pid_weight_info_container, PIDWeightInfoContainer)

            for pid_w_type in self._pid_weight_info_container.keys():
                output_path = self._get_output_path(pid_type=pid_w_type)  # type: PathType
                self._pid_weight_info_container.add_data_frame_path_to_info(key=pid_w_type, hdf5_path=output_path)

            assert all(i.hdf5_path is not None for i in self._pid_weight_info_container.values())
            assert len(set(self._pid_weight_info_container.keys())) == len(
                set([v.hdf5_path for v in self._pid_weight_info_container.values()])
            ), "\n" + "\n\t - ".join([f"{k}, {i.hdf5_path}" for k, i in self._pid_weight_info_container.items()])

        return self._pid_weight_info_container

    def prepare_weight_evaluation_data_frames(self, dry_run: bool = False) -> Tuple[str, PathType]:
        output_key_path_pair = self._get_output_mapping_dict_output_key_path_pair()
        if dry_run:
            return output_key_path_pair

        os.makedirs(name=os.path.join(self._output_dir, self.weight_info_dfs_sub_dir_name), exist_ok=True)
        dfs_output_dict = self._split_pid_info_by_weight_type()

        assert len(dfs_output_dict.keys()) == len(self.pid_weight_infos.keys()), (
            len(dfs_output_dict.keys()),
            len(self.pid_weight_infos.keys()),
        )
        assert all(i.hdf5_path in dfs_output_dict.values() for i in self.pid_weight_infos.values())

        self._write_pid_weight_infos_dict()

        return output_key_path_pair

    def _split_pid_info_by_weight_type(self) -> Dict[str, PathType]:
        output_dict = {}  # type: Dict[str, PathType]

        for pid_w_type, pid_w_info in self.pid_weight_infos.items():

            df = self.base_df.loc[pid_w_info.condition]  # type: pd.DataFrame

            output_key = self.get_output_key(pid_type=pid_w_type, with_sel=self._with_selection)  # type: str
            output_path = self._get_output_path(pid_type=pid_w_type)  # type: PathType
            tmp_output_path = self._get_output_path(pid_type=pid_w_type, tmp=True)  # type: PathType
            assert pid_w_info.hdf5_path == output_path, (pid_w_info.hdf5_path, output_path)

            df.to_hdf(path_or_buf=tmp_output_path, key=output_key, format="table")
            os.rename(src=tmp_output_path, dst=output_path)

            assert output_key not in output_dict.keys(), output_key
            assert output_path not in output_dict.values(), output_path
            output_dict.update({output_key: output_path})

        return output_dict

    def _get_output_path(
        self,
        pid_type: str,
        tmp: bool = False,
    ) -> PathType:
        return self.get_output_path(
            pid_type=pid_type,
            dir_path=self._output_dir,
            with_sel=self._with_selection,
            tmp=tmp,
        )

    @staticmethod
    def get_output_path(
        pid_type: str,
        dir_path: PathType,
        with_sel: bool,
        tmp: bool = False,
    ) -> PathType:
        sel_tag = "_with_selection" if with_sel else ""  # type: str
        tmp_tag = "tmp_" if tmp else ""  # type: str
        return os.path.join(
            dir_path, PIDWeightEvaluation.weight_info_dfs_sub_dir_name, f"{tmp_tag}pid_weight_info_{pid_type}{sel_tag}.h5"
        )

    @staticmethod
    def get_output_key(pid_type: str, with_sel: bool = False) -> str:
        sel_tag = "_with_selection" if with_sel else ""  # type: str
        return f"pid_weight_info_{pid_type}{sel_tag}"

    def _write_pid_weight_infos_dict(self) -> None:
        tmp_file_path = self._get_output_mapping_dict_output_path(tmp=True)
        with open(tmp_file_path, "w") as result_fp:
            json.dump(self.pid_weight_infos.as_dict(remove_condition_info=True), result_fp, indent=2)

        os.rename(src=tmp_file_path, dst=self._get_output_mapping_dict_output_path())

    @staticmethod
    def get_output_mapping_dict_output_key(with_selection: bool) -> str:
        with_sel_tag = "_for_with_selection" if with_selection else ""  # type: str
        return f"{PIDWeightEvaluation.output_mapping_dict_output_key}{with_sel_tag}"

    def _get_output_mapping_dict_output_key(self) -> str:
        return self.get_output_mapping_dict_output_key(with_selection=self._with_selection)

    @staticmethod
    def get_output_mapping_dict_output_path(
        output_dir: PathType,
        with_selection: bool,
        tmp: bool = False,
    ) -> PathType:
        tmp_tag = "tmp_" if tmp else ""  # type: str
        file_name = f"{PIDWeightEvaluation.get_output_mapping_dict_output_key(with_selection=with_selection)}.json"
        return os.path.join(output_dir, f"{tmp_tag}{file_name}")

    def _get_output_mapping_dict_output_path(self, tmp: bool = False) -> PathType:
        return self.get_output_mapping_dict_output_path(
            output_dir=self._output_dir,
            with_selection=self._with_selection,
            tmp=tmp,
        )

    @staticmethod
    def get_output_mapping_dict_output_key_path_pair(
        output_dir: PathType,
        with_selection: bool,
    ) -> Tuple[str, PathType]:
        output_key = PIDWeightEvaluation.get_output_mapping_dict_output_key(with_selection=with_selection)  # type: str
        file_name = f"{output_key}.json"  # type: str
        output_path = os.path.join(output_dir, file_name)  # type: PathType
        return output_key, output_path

    def _get_output_mapping_dict_output_key_path_pair(self) -> Tuple[str, PathType]:
        return self.get_output_mapping_dict_output_key_path_pair(
            output_dir=self._output_dir,
            with_selection=self._with_selection,
        )

    @staticmethod
    def load_pid_weight_infos(weight_info_collection_path: PathType) -> PIDWeightInfoContainer:
        assert os.path.isfile(weight_info_collection_path), weight_info_collection_path
        with open(weight_info_collection_path, "r") as wi_file:
            weight_info_dict = json.load(wi_file)

        return PIDWeightInfoContainer.from_dict(weight_infos_as_dict=weight_info_dict)
