"""
Plot functions for evaluation of particle ID correction for Belle

Felix Metzner 2019
"""

import numpy as np
import pandas as pd

from typing import Tuple, List, Dict, Union, Optional

from matplotlib import pyplot as plt
from matplotlib import colors as colours
from matplotlib.offsetbox import AnchoredText

from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import SimpleHistogramPlot
from templatefitter.plotter.plot_utilities import FigureType, AxesType
from templatefitter.plotter.plot_style import KITColors, set_matplotlibrc_params

from analysistools.pid_corrections import pid_corrector
from analysistools.pid_corrections.pid_weight_evaluator import PIDWeightDefinitions


__all__ = [
    "plot_flag_heat_map",
    "plot_weight_heat_map",
    "plot_eqd_weight_heat_map",
    "plot_eqd_weight_val_heat_map",
    "plot_weight_distribution",
    "weight_box_plot",
    "weight_violin_plot",
]

# plt.switch_backend(newbackend="agg")  # This version might be better for the PNGs
plt.switch_backend(newbackend="pdf")

set_matplotlibrc_params()


def flag_meaning(flag: int) -> str:
    meaning = {
        0: "OK",
        1: "No Stats",
        2: "Fit Failed",
        3: "Fit Failed",
    }  # type: Dict[int, str]
    assert flag in meaning.keys()
    return meaning[flag]


def plot_flag_heat_map(
    df: pd.DataFrame,
    binning: Tuple[List[float], List[int]],
    particle_name_for_title: str,
) -> Tuple[FigureType, AxesType]:
    momentum_bins, theta_bins = binning

    pid_group = df.groupby(
        [PIDWeightDefinitions.momentum_indices_column_name, PIDWeightDefinitions.theta_indices_column_name]
    )["ratio"]
    assert all(pid_group.nunique() == 1)

    flag_map = np.full(shape=(len(momentum_bins) - 1, len(theta_bins) - 1), fill_value=np.nan)

    for i_mom in df[PIDWeightDefinitions.momentum_indices_column_name].unique():
        for i_theta in df[PIDWeightDefinitions.theta_indices_column_name].unique():
            assert np.isnan(flag_map[int(i_mom), int(i_theta)])
            try:
                val = pid_group.get_group((i_mom, i_theta)).unique()[0]
            except KeyError:
                val = np.nan
            flag_map[int(i_mom), int(i_theta)] = -val if val < 0 else 0

    c_map = colours.ListedColormap(
        [
            KITColors.kit_green,
            colours.to_rgba(KITColors.kit_black, 0.4),
            colours.to_rgba(KITColors.kit_black, 0.6),
            colours.to_rgba(KITColors.kit_black, 0.8),
        ]
    )

    bounds = [-0.5, 0.5, 1.5, 2.5, 3.5]
    fig = plt.figure(num=None, figsize=(6, 4.5), dpi=100, facecolor="w", edgecolor="k")  # type: FigureType
    img = plt.imshow(flag_map.T, cmap=c_map, aspect="auto", origin="lower", vmin=0, vmax=3)
    ax = plt.gca()  # type: AxesType

    ax.set_title(f"Particle ID Flags for {particle_name_for_title}")
    ax.set_xlabel(r"$p$ in GeV")
    ax.set_ylabel(r"$\theta$ in degrees")

    ax.set_xticks(np.arange(0, flag_map.shape[0] + 1, 1) - 0.5)
    ax.set_yticks(np.arange(0, flag_map.shape[1] + 1, 1) - 0.5)
    ax.set_xticklabels([f"{momentum_bins[int(x + 0.5)]}" for x in ax.get_xticks()], rotation=75)
    ax.set_yticklabels([f"{theta_bins[int(x + 0.5)]}" for x in ax.get_yticks()])

    plt.minorticks_off()

    c_bar = plt.colorbar(img, boundaries=bounds, ticks=[0, 1, 2, 3])
    c_bar.set_label("Flag Value", rotation=270, labelpad=20)

    all_events = len(df.index)
    info = f"{all_events} Events"
    for flag_value in [0, 1, 2, 3]:
        if flag_value == 0:
            n_events = len(df.loc[df.ratio >= flag_value])
        else:
            n_events = len(df.loc[df.ratio == -flag_value])
        if n_events > 0:
            info += f"\nFlag {flag_value} ({flag_meaning(flag_value)}): {n_events * 100. / all_events:3.2f}%"

    anchored_text = AnchoredText(info, loc="upper right")
    ax.add_artist(anchored_text)

    return fig, ax


def plot_weight_heat_map(
    df: pd.DataFrame,
    binning: Tuple[List[float], List[int]],
    particle_name_for_title: str,
) -> Tuple[FigureType, AxesType]:
    momentum_bins, theta_bins = binning

    fig = plt.figure(num=None, figsize=(6, 4.5), dpi=100, facecolor="w", edgecolor="k")  # type: FigureType
    ax = plt.gca()  # type: AxesType

    plt.hist2d(
        df[PIDWeightDefinitions.momentum_column_name],
        df[PIDWeightDefinitions.theta_column_name],
        bins=(momentum_bins, theta_bins),
        cmax=None,
        cmin=1,
    )

    ax.set_title(f"Particle ID Weights for {particle_name_for_title}")
    ax.set_xlabel(r"$p$ in GeV")
    ax.set_ylabel(r"$\theta$ in degrees")

    ax.set_xticks(momentum_bins)
    ax.set_yticks(theta_bins)
    ax.set_xticklabels(ax.get_xticks(), rotation=75)

    plt.minorticks_off()

    c_bar = plt.colorbar()
    c_bar.set_label("Number of Events", rotation=270, labelpad=20)

    info = f"{len(df.index)} Events"
    anchored_text = AnchoredText(info, loc="upper right", frameon=False)
    ax.add_artist(anchored_text)

    return fig, ax


def plot_eqd_weight_heat_map(
    df: pd.DataFrame,
    binning: Tuple[List[float], List[int]],
    particle_name_for_title: str,
) -> Tuple[FigureType, AxesType]:
    momentum_bins, theta_bins = binning

    fig = plt.figure(num=None, figsize=(6, 4.5), dpi=100, facecolor="w", edgecolor="k")  # type: FigureType
    ax = plt.gca()  # type: AxesType

    bins = (
        np.arange(0, len(df[PIDWeightDefinitions.momentum_indices_column_name].unique()) + 1, 1) - 0.5,
        np.arange(0, len(df[PIDWeightDefinitions.theta_indices_column_name].unique()) + 1, 1) - 0.5,
    )

    plt.hist2d(
        x=df[PIDWeightDefinitions.momentum_indices_column_name],
        y=df[PIDWeightDefinitions.theta_indices_column_name],
        bins=bins,
        cmax=None,
        cmin=1,
    )

    ax.set_title(f"Particle ID Weights for {particle_name_for_title}")
    ax.set_xlabel(r"$p$ in GeV")
    ax.set_ylabel(r"$\theta$ in degrees")

    ax.set_xticks(bins[0])
    ax.set_yticks(bins[1])
    ax.set_xticklabels([f"{momentum_bins[int(x + 0.5)]}" for x in ax.get_xticks()], rotation=75)
    ax.set_yticklabels([f"{theta_bins[int(x + 0.5)]}" for x in ax.get_yticks()])

    plt.minorticks_off()

    c_bar = plt.colorbar()
    c_bar.set_label("Number of Events", rotation=270, labelpad=20)

    info = f"{len(df.index)} Events"
    anchored_text = AnchoredText(info, loc="upper right", frameon=False)
    ax.add_artist(anchored_text)

    return fig, ax


def plot_eqd_weight_val_heat_map(
    df: pd.DataFrame,
    binning: Tuple[List[float], List[int]],
    particle_name_for_title: str,
    update_for_flag: bool = False,
) -> Tuple[FigureType, AxesType]:
    momentum_bins, theta_bins = binning

    column_name = "ratio"
    if update_for_flag:
        new_name = f"{column_name}_updated"
        pid_corrector.PIDCorrection.update_values_for_fail_flags(df, column_name=column_name, new_column_name=new_name)
        column_name = new_name

    pid_group = df.groupby(
        [PIDWeightDefinitions.momentum_indices_column_name, PIDWeightDefinitions.theta_indices_column_name]
    )[column_name]
    assert all(pid_group.nunique() == 1)

    value_map = np.full(shape=(len(momentum_bins) - 1, len(theta_bins) - 1), fill_value=np.nan)

    for i_mom in df[PIDWeightDefinitions.momentum_indices_column_name].unique():
        for i_theta in df[PIDWeightDefinitions.theta_indices_column_name].unique():
            assert np.isnan(value_map[int(i_mom), int(i_theta)])
            try:
                val = pid_group.get_group((i_mom, i_theta)).unique()[0]
            except KeyError:
                val = np.nan
            value_map[int(i_mom), int(i_theta)] = val
            # value_map[int(i_mom), int(i_theta)] = val if val > 0 or np.isnan(val) else 1

    fig = plt.figure(num=None, figsize=(6, 4.5), dpi=100, facecolor="w", edgecolor="k")  # type: FigureType
    img = plt.imshow(value_map.T, aspect="auto", origin="lower")

    ax = plt.gca()  # type: AxesType
    ax.set_title(f"Particle ID Weights for {particle_name_for_title}")
    ax.set_xlabel(r"$p$ in GeV")
    ax.set_ylabel(r"$\theta$ in degrees")

    ax.set_xticks(np.arange(0, value_map.shape[0] + 1, 1) - 0.5)
    ax.set_yticks(np.arange(0, value_map.shape[1] + 1, 1) - 0.5)
    ax.set_xticklabels([f"{momentum_bins[int(x + 0.5)]}" for x in ax.get_xticks()], rotation=75)
    ax.set_yticklabels([f"{theta_bins[int(x + 0.5)]}" for x in ax.get_yticks()])

    plt.minorticks_off()

    c_bar = plt.colorbar(img)
    c_bar.set_label("Particle ID Weight Value", rotation=270, labelpad=20)

    return fig, ax


def plot_weight_distribution(
    df: pd.DataFrame,
    particle_name_for_title: str,
    logarithmic: bool = False,
    update_for_flag: bool = False,
) -> Tuple[FigureType, AxesType]:
    column_name = "ratio"
    if update_for_flag:
        new_name = f"{column_name}_updated"
        pid_corrector.PIDCorrection.update_values_for_fail_flags(df, column_name=column_name, new_column_name=new_name)
        column_name = new_name

    info = (
        f"mean: {df[column_name].mean():2.3f}\n"
        f"median: {df[column_name].median():2.3f}\n"
        f"min: {df[column_name].min():2.3f}\n"
        f"max: {df[column_name].max():2.3f}"
    )

    anchored_text = AnchoredText(info, loc="upper left")
    scope = (df[column_name].min(), df[column_name].max())
    pid_weight = HistVariable(df_label=column_name, n_bins=40, scope=scope, var_name=r"Nominal pID Weight", unit="")

    fig = plt.figure(num=None, figsize=(6, 4.5), dpi=200, facecolor="w", edgecolor="k")  # type: FigureType
    ax = plt.gca()  # type: AxesType

    hw = SimpleHistogramPlot(pid_weight)
    hw.add_component("", df, color=KITColors.kit_blue)
    hw.plot_on(ax, normed=False, draw_legend=False)

    if logarithmic:
        ax.set_yscale("log")

    ax.set_title(f"Particle ID Weights for {particle_name_for_title}")
    ax.add_artist(anchored_text)

    return fig, ax


def translate_labels(column_name: str) -> str:
    label = ""  # type: str
    if "ratio" in column_name:
        label = "Nominal"
    elif "var_" in column_name:
        num = column_name.split("var_", 1)[1].split("_", 1)[0]  # type: str
        assert num.isdigit(), num
        label = f"Variation {int(num) + 1}"
    else:
        RuntimeError(f"Unable to parse {column_name}")
    return label


def weight_box_plot(
    df: pd.DataFrame,
    particle_name_for_title: str,
    whisker_quantile: float = 0.0,
    plot_fliers: bool = False,
    plot_mean: bool = True,
    cols: Optional[List[str]] = None,
    plot_zero_line: bool = True,
    update_for_flag: bool = False,
) -> Tuple[FigureType, AxesType]:
    assert 0.0 <= whisker_quantile < 0.5, whisker_quantile

    column_name = "ratio"
    if update_for_flag:
        new_name = f"{column_name}_updated"
        pid_corrector.PIDCorrection.update_values_for_fail_flags(df, column_name=column_name, new_column_name=new_name)
        column_name = new_name

    if cols is None:
        cols = [c for c in df.columns if c.startswith("var_") or c == column_name]

    assert len([c for c in cols if c.startswith("ratio")]) == 1

    if whisker_quantile == 0.0:
        whiskers = (0.0, 100.0)  # type: Tuple[float, float]
    else:
        whiskers = (0.0 + 100.0 * whisker_quantile, 100.0 - 100.0 * whisker_quantile)

    fig = plt.figure(num=None, figsize=(8.5, 14), dpi=100, facecolor="w", edgecolor="k")  # type: FigureType
    ax = plt.gca()  # type: AxesType

    box_plot_props = {
        "widths": 0.7,
        "notch": False,
        "vert": False,
        "boxprops": dict(color=KITColors.kit_black, linewidth=1.0, linestyle="-"),
        "whiskerprops": dict(color=KITColors.kit_black, linewidth=1.0, linestyle="-"),
        "capprops": dict(color=KITColors.kit_black, linewidth=1.0, linestyle="-"),
        "medianprops": dict(
            color=KITColors.kit_red,
            linewidth=1.4,
            linestyle="-",
            markerfacecolor=KITColors.kit_red,
            markeredgecolor=KITColors.kit_red,
            marker="|",
        ),
        "meanprops": dict(
            color=KITColors.kit_orange,
            linewidth=1.2,
            linestyle="--",
            markerfacecolor=KITColors.kit_orange,
            markeredgecolor=KITColors.kit_orange,
            marker="",
        ),
        "flierprops": dict(
            markerfacecolor=KITColors.kit_red,
            marker="+",
            markersize=3,
            markeredgecolor=KITColors.kit_red,
        ),
    }

    res = plt.boxplot(
        x=[df[col].values for col in cols],
        whis=whiskers,
        showfliers=plot_fliers,
        labels=cols,
        showmeans=plot_mean,
        meanline=True,
        **box_plot_props,
    )

    mean_x, median_x = None, None
    mean_objects = res["means"] if plot_mean else res["medians"]
    for data_index, (mean_obj, median_obj) in enumerate(zip(mean_objects, res["medians"])):
        if cols[data_index] == column_name:
            if plot_mean:
                mean_x = mean_obj._x[0]
            median_x = median_obj._x[0]
    if plot_mean:
        assert mean_x is not None
    assert median_x is not None

    if plot_mean:
        ax.vlines(
            x=mean_x,
            ymin=1.3,
            ymax=len(cols) + 0.5,
            color=KITColors.kit_orange,
            linewidth=0.8,
            linestyle="--",
            label="Mean of Nominal",
        )
    ax.vlines(
        x=median_x,
        ymin=1.3,
        ymax=len(cols) + 0.5,
        color=KITColors.kit_red,
        linewidth=0.8,
        linestyle="--",
        label="Median of Nominal",
    )

    legend_object1 = ax.legend(frameon=True, loc=3)

    legend_lines = [
        plt.Line2D([0], [0], color=KITColors.kit_red, lw=2, ls="-"),
        plt.Line2D([0], [0], color=KITColors.kit_orange, lw=2, ls="-"),
    ]

    if plot_mean:
        ax.legend(legend_lines, ["Medians", "Means"], frameon=True, loc=4)
    else:
        ax.legend(legend_lines[:1], ["Medians"], frameon=True, loc=4)

    ax.add_artist(legend_object1)

    x_lims = ax.get_xlim()
    scaling = (x_lims[1] - x_lims[0]) / 1000.0

    if plot_zero_line and x_lims[0] < 0.0 < x_lims[1]:
        ax.vlines(x=0.0, ymin=0.0, ymax=len(cols) + 0.5, color=KITColors.grey, linewidth=0.8, linestyle="-")

    if whisker_quantile > 0.0 and not plot_fliers:
        min_caps = [c for i, c in enumerate(res["caps"]) if i % 2 == 0]
        max_caps = [c for i, c in enumerate(res["caps"]) if i % 2 == 1]
        for data_index, (min_cap, max_cap) in enumerate(zip(min_caps, max_caps)):
            col = cols[data_index]
            min_x = min_cap._x[0]
            max_x = max_cap._x[0]
            mean_y = np.mean(min_cap._y)
            min_val = df[col].min()
            max_val = df[col].max()

            ax.text(x=min_x - 66 * scaling, y=mean_y, s=r"$\leftarrow$", fontsize=10, ha="right", va="center")
            ax.text(x=max_x + 60 * scaling, y=mean_y, s=r"$\rightarrow$", fontsize=10, ha="left", va="center")
            ax.text(x=min_x - 16 * scaling, y=mean_y, s=f"min\n{min_val:.2f}", fontsize=10, ha="right", va="center")
            ax.text(x=max_x + 16 * scaling, y=mean_y, s=f"max\n{max_val:.2f}", fontsize=10, ha="left", va="center")

    ax.set_yticklabels([translate_labels(c) for c in cols])

    ax.set_ylim(bottom=-0.6 if plot_mean else -0.1, top=len(cols) + 1.4)
    if whisker_quantile > 0.0:
        x_lims = ax.set_xlim(
            left=x_lims[0] - 0.10 * abs(x_lims[1] - x_lims[0]),
            right=x_lims[1] + 0.10 * abs(x_lims[1] - x_lims[0]),
        )
        info_text = (
            f"Whiskers contain weight values between the {int(whiskers[0])}% and " f"the {int(whiskers[1])}% quantiles."
        )
    else:
        x_lims = ax.set_xlim(
            left=x_lims[0] - 0.05 * abs(x_lims[1] - x_lims[0]),
            right=x_lims[1] + 0.05 * abs(x_lims[1] - x_lims[0]),
        )
        info_text = "Whiskers contain all weight values."

    ax.text(
        x=np.min(x_lims) + 0.5 * abs(x_lims[1] - x_lims[0]),
        y=len(cols) + 0.8,
        s=info_text,
        fontsize=12,
        ha="center",
        va="baseline",
    )

    ax.set_title(f"Box Plots of Particle ID Weights for {particle_name_for_title}")
    ax.set_xlabel("Weight Values")

    return fig, ax


def get_data_mask(
    df: pd.DataFrame,
    col: str,
    quantile: float = 0.05,
) -> Union[pd.Series, np.ndarray, bool, int]:
    assert 0.0 <= quantile < 0.5, quantile
    if quantile == 0.0:
        return np.full(len(df.index), True, dtype=bool)
    return (df[col] > df[col].quantile(0.0 + quantile)) & (df[col] < df[col].quantile(1.0 - quantile))


def weight_violin_plot(
    df: pd.DataFrame,
    particle_name_for_title: str,
    mask_quantile: float = 0.05,
    points: int = 60,
    cols: Optional[List[str]] = None,
    update_for_flag: bool = False,
    plot_zero_line: bool = True,
) -> Tuple[FigureType, AxesType]:
    column_name = "ratio"
    if update_for_flag:
        new_name = f"{column_name}_updated"
        pid_corrector.PIDCorrection.update_values_for_fail_flags(
            df,
            column_name=column_name,
            new_column_name=new_name,
        )
        column_name = new_name

    if cols is None:
        cols = [c for c in df.columns if c.startswith("var_") or c == column_name]

    fig = plt.figure(num=None, figsize=(8.5, 14), dpi=100, facecolor="w", edgecolor="k")  # type: FigureType
    ax = plt.gca()  # type: AxesType

    violin_plot_props = {
        "vert": False,
        "widths": 0.9,
        "points": points,
        "showmeans": True,
        "showextrema": True,
        "showmedians": True,
    }

    violin_plot_data = [df.loc[get_data_mask(df, col, quantile=mask_quantile), col].values for col in cols]
    if any(len(vdf) == 0 for vdf in violin_plot_data):
        ax.text(
            x=0.5,
            y=0.5,
            s="No Data",
            fontsize=18,
            ha="center",
            va="center",
            transform=ax.transAxes,
            bbox=dict(facecolor=KITColors.kit_red, alpha=0.5),
        )
        return fig, ax

    res = plt.violinplot(
        dataset=violin_plot_data,
        positions=np.arange(0, len(cols)),
        **violin_plot_props,
    )

    for vp_body in res["bodies"]:
        vp_body.set_color(KITColors.kit_blue)
        vp_body.set_alpha(0.9)

    res["cmedians"].set_color(KITColors.kit_red)
    res["cmeans"].set_color(KITColors.kit_orange)

    res["cbars"].set_color("none")

    res["cmins"].set_color(KITColors.kit_black)
    res["cmaxes"].set_color(KITColors.kit_black)

    x_lims = ax.get_xlim()
    scaling = (x_lims[1] - x_lims[0]) / 1000.0

    if plot_zero_line and x_lims[0] < 0.0 < x_lims[1]:
        ax.vlines(
            x=0.0,
            ymin=-0.4,
            ymax=len(cols) - 0.5,
            color=KITColors.grey,
            linewidth=0.8,
            linestyle="-",
        )

    mean_x, median_x = None, None
    for data_index, (mean_path, median_path) in enumerate(zip(res["cmeans"].get_paths(), res["cmedians"].get_paths())):
        if cols[data_index] == column_name:
            mean_x = mean_path.vertices[0][0]
            median_x = median_path.vertices[0][0]
    assert mean_x is not None and median_x is not None

    ax.vlines(
        x=mean_x,
        ymin=-0.4,
        ymax=len(cols) - 0.5,
        color=KITColors.kit_orange,
        linewidth=0.8,
        linestyle="--",
        label="Mean of Nominal",
    )
    ax.vlines(
        x=median_x,
        ymin=-0.4,
        ymax=len(cols) - 0.5,
        color=KITColors.kit_red,
        linewidth=0.8,
        linestyle="--",
        label="Median of Nominal",
    )

    if mask_quantile > 0.0:
        for data_index, (min_paths, max_paths) in enumerate(zip(res["cmins"].get_paths(), res["cmaxes"].get_paths())):
            col = cols[data_index]
            min_x = min_paths.vertices[0][0]
            max_x = max_paths.vertices[0][0]
            mean_y = np.mean(min_paths.vertices[:, 1])
            min_val = df[col].min()
            max_val = df[col].max()

            ax.text(x=min_x - 66 * scaling, y=mean_y, s=r"$\leftarrow$", fontsize=10, ha="right", va="center")
            ax.text(x=max_x + 60 * scaling, y=mean_y, s=r"$\rightarrow$", fontsize=10, ha="left", va="center")
            ax.text(x=min_x - 16 * scaling, y=mean_y, s=f"min\n{min_val:.2f}", fontsize=10, ha="right", va="center")
            ax.text(x=max_x + 16 * scaling, y=mean_y, s=f"max\n{max_val:.2f}", fontsize=10, ha="left", va="center")

    ax.set_yticks(np.arange(0, len(cols)))
    ax.set_yticklabels([translate_labels(c) for c in cols])

    ax.set_ylim(bottom=-1.6, top=len(cols) + 0.6)
    if mask_quantile > 0.0:
        x_lims = ax.set_xlim(
            left=x_lims[0] - 0.10 * abs(x_lims[1] - x_lims[0]),
            right=x_lims[1] + 0.10 * abs(x_lims[1] - x_lims[0]),
        )
        info_text = (
            f"Whiskers contain weight values between the {int(mask_quantile * 100)}% "
            f"and the {int((1. - mask_quantile) * 100)}% quantiles."
        )
    else:
        x_lims = ax.set_xlim(
            left=x_lims[0] - 0.05 * abs(x_lims[1] - x_lims[0]),
            right=x_lims[1] + 0.05 * abs(x_lims[1] - x_lims[0]),
        )
        info_text = "Whiskers contain all weight values."

    ax.text(
        x=np.min(x_lims) + 0.5 * abs(x_lims[1] - x_lims[0]),
        y=len(cols),
        s=info_text,
        fontsize=12,
        ha="center",
        va="baseline",
    )
    legend_object1 = ax.legend(frameon=True, loc=3)

    legend_lines = [
        plt.Line2D([0], [0], color=KITColors.kit_red, lw=2, ls="-"),
        plt.Line2D([0], [0], color=KITColors.kit_orange, lw=2, ls="-"),
    ]
    ax.legend(legend_lines, ["Medians", "Means"], frameon=True, loc=4)
    ax.add_artist(legend_object1)

    ax.set_title(f"Violin Plots of Particle ID Weights for {particle_name_for_title}")
    ax.set_xlabel("Weight Values")

    return fig, ax
