"""
Plotter for particle ID correction weights for Belle

Felix Metzner 2019
"""

import os
import copy
import inspect
import numpy as np
import pandas as pd

from matplotlib import pyplot as plt
from typing import Union, Optional, Tuple, List, Dict, Any, Callable, NamedTuple

from templatefitter.plotter.plot_utilities import export
from templatefitter.plotter.plot_style import set_matplotlibrc_params

from analysistools.utilities import PathType
from analysistools.pid_corrections.pid_weight_evaluator import PIDWeightDefinitions, PIDWeightInfo
from analysistools.pid_corrections.pid_weight_plots import (
    plot_weight_heat_map,
    plot_flag_heat_map,
    plot_eqd_weight_heat_map,
    plot_eqd_weight_val_heat_map,
    plot_weight_distribution,
    weight_box_plot,
    weight_violin_plot,
)


__all__ = [
    "PIDWeightsPlotter",
]

# plt.switch_backend(newbackend="agg")  # This version might be better for the PNGs
plt.switch_backend(newbackend="pdf")

set_matplotlibrc_params()


class PIDPlotInfoContainer(NamedTuple):
    name: str
    plot_function: Callable
    parameters: Dict[str, Any]
    only_for_hadrons: bool = False
    only_in_single_exp_range: bool = False


class PIDWeightsPlotter:

    __plots = tuple(
        [
            PIDPlotInfoContainer(
                name="pid_weight_plot",
                plot_function=plot_weight_heat_map,
                parameters={},
            ),
            PIDPlotInfoContainer(
                name="pid_weight_flag_plot",
                plot_function=plot_flag_heat_map,
                parameters={},
                only_for_hadrons=True,
                only_in_single_exp_range=True,  # Because flags are defined differently for different exp ranges!
            ),
            PIDPlotInfoContainer(
                name="pid_eqd_weight_plot",
                plot_function=plot_eqd_weight_heat_map,
                parameters={},
            ),
            PIDPlotInfoContainer(
                name="eqd_pid_weight_value_plot",
                plot_function=plot_eqd_weight_val_heat_map,
                parameters={"update_for_flag": True},
                only_in_single_exp_range=True,  # Because flags are defined differently for different exp ranges!
            ),
            PIDPlotInfoContainer(
                name="pid_weight_distribution_wF_plot",
                plot_function=plot_weight_distribution,
                parameters={"update_for_flag": False},
                only_for_hadrons=True,
            ),
            PIDPlotInfoContainer(
                name="pid_weight_distribution_plot",
                plot_function=plot_weight_distribution,
                parameters={"update_for_flag": True},
            ),
            PIDPlotInfoContainer(
                name="pid_weight_box_plot",
                plot_function=weight_box_plot,
                parameters={"whisker_quantile": 0.0, "plot_mean": False, "update_for_flag": True},
            ),
            PIDPlotInfoContainer(
                name="pid_weight_1to99_quantile_w_fliers_box_plot",
                plot_function=weight_box_plot,
                parameters={"whisker_quantile": 0.01, "plot_mean": False, "plot_fliers": True, "update_for_flag": True},
            ),
            PIDPlotInfoContainer(
                name="pid_weight_1to99_quantile_box_plot",
                plot_function=weight_box_plot,
                parameters={"whisker_quantile": 0.01, "update_for_flag": True},
            ),
            PIDPlotInfoContainer(
                name="pid_weight_5to95_quantile_box_plot",
                plot_function=weight_box_plot,
                parameters={"whisker_quantile": 0.05, "update_for_flag": True},
            ),
            PIDPlotInfoContainer(
                name="pid_weight_violin_plot",
                plot_function=weight_violin_plot,
                parameters={"mask_quantile": 0.0, "update_for_flag": True},
            ),
            PIDPlotInfoContainer(
                name="pid_weight_5to95_quantile_violin_plot",
                plot_function=weight_violin_plot,
                parameters={"mask_quantile": 0.05, "update_for_flag": True},
            ),
            PIDPlotInfoContainer(
                name="pid_weight_1to99_quantile_violin_plot",
                plot_function=weight_violin_plot,
                parameters={"mask_quantile": 0.01, "update_for_flag": True},
            ),
        ]
    )  # type: Tuple[PIDPlotInfoContainer, ...]

    def __init__(
        self,
        weight_info_key: str,
        weight_info: PIDWeightInfo,
        file_name_base: str,
        output_dir_path: PathType,
    ) -> None:
        self.weight_type = weight_info_key  # type: str
        self.file_name_base = file_name_base  # type: str
        self.output_dir_path = output_dir_path  # type: PathType

        self.weight_info = weight_info  # type: PIDWeightInfo
        assert isinstance(self.weight_info, PIDWeightInfo), type(self.weight_info).__name__

    def plot_pid_weights(
        self,
        df: pd.DataFrame,
        file_name_extension: Optional[str] = None,
        dry_run: bool = False,
    ) -> Dict[str, PathType]:
        if not dry_run:
            set_matplotlibrc_params()
            fne = file_name_extension if file_name_extension is not None else "None"  # type: str
            PIDWeightInfo.check_if_theta_is_in_degrees(df=df, hint=f"Theta unit check failed for {fne}")

        output_dict = {}  # type: Dict[str, PathType]

        if not dry_run:
            if len(self.weight_info.condition) == 2 and all(isinstance(c, int) for c in self.weight_info.condition):
                assert len(df.index) == self.weight_info.condition[0]

            momentum_bins, theta_bins = self.weight_info.binning  # type: List[float], List[Union[float, int]]

            mom_indices = np.digitize(
                x=df[PIDWeightDefinitions.momentum_column_name],
                bins=momentum_bins,
                right=True,
            )
            assert (0 not in mom_indices) and (len(momentum_bins) not in mom_indices)
            mom_indices -= 1
            df.loc[:, PIDWeightDefinitions.momentum_indices_column_name] = mom_indices

            theta_indices = np.digitize(
                x=df[PIDWeightDefinitions.theta_column_name],
                bins=theta_bins,
                right=True,
            )
            assert (0 not in theta_indices) and (len(theta_bins) not in theta_indices)
            theta_indices -= 1
            df.loc[:, PIDWeightDefinitions.theta_indices_column_name] = theta_indices

        _exp_ranges = copy.copy(self.weight_info.exp_ranges)  # type: List[Any]
        assert len(_exp_ranges) > 0, _exp_ranges
        if len(_exp_ranges) > 1:
            _exp_ranges.append(None)
        exp_ranges = _exp_ranges  # type: List[Optional[Tuple[int, int]]]

        for exp_range in exp_ranges:
            for plot_info in self.__plots:
                if plot_info.only_for_hadrons and not self.weight_info.is_hadron:
                    continue
                if plot_info.only_in_single_exp_range and exp_range is None:
                    continue

                if dry_run or exp_range is None:
                    exp_range_condition = None
                else:
                    exp_range_condition = (df[PIDWeightDefinitions.exp_column_name] >= exp_range[0]) & (
                        df[PIDWeightDefinitions.exp_column_name] <= exp_range[1]
                    )
                    assert isinstance(exp_range_condition, pd.Series), type(exp_range_condition).__name__

                output_dict_part = self.create_plot(
                    df=df if exp_range_condition is None else df.loc[exp_range_condition],
                    plot_info=plot_info,
                    exp_range=exp_range,
                    file_name_extension=file_name_extension,
                    dry_run=dry_run,
                )  # type: Dict[str, PathType]

                assert all(k not in output_dict.keys() for k in output_dict_part.keys()), "\n\t-" + "\n\t-".join(
                    list(output_dict_part.keys())
                )
                assert all(v not in output_dict.values() for v in output_dict_part.values()), "\n\t-" + "\n\t-".join(
                    [str(v) for v in output_dict_part.values()]
                )
                output_dict.update(output_dict_part)

        return output_dict

    def create_plot(
        self,
        df: pd.DataFrame,
        plot_info: PIDPlotInfoContainer,
        exp_range: Optional[Tuple[int, int]] = None,
        file_name_extension: Optional[str] = None,
        dry_run: bool = False,
    ) -> Dict[str, PathType]:
        file_name = self.build_plot_file_name(
            plot_name=plot_info.name,
            exp_range=exp_range,
            file_name_extension=file_name_extension,
        )  # type: str

        if not dry_run:
            fig, _ = plot_info.plot_function(
                df=df, **self.get_kwargs_for_plot_function(weight_info=self.weight_info, plot_info=plot_info)
            )
            export(fig=fig, filename=file_name, target_dir=self.output_dir_path, close_figure=True)
        return {f"{file_name}_{e}": os.path.join(self.output_dir_path, f"{file_name}.{e}") for e in ["pdf", "png"]}

    @staticmethod
    def get_kwargs_for_plot_function(weight_info: PIDWeightInfo, plot_info: PIDPlotInfoContainer) -> Dict[str, Any]:
        kwargs = {"particle_name_for_title": weight_info.get_name_for_title()}  # type: Dict[str, Any]
        if PIDWeightsPlotter.plot_function_requires_binning_arg(plot_function=plot_info.plot_function):
            kwargs.update({"binning": weight_info.binning})
        kwargs.update(plot_info.parameters)
        return kwargs

    @staticmethod
    def plot_function_requires_binning_arg(plot_function: Callable) -> bool:
        plot_func_signature = inspect.signature(plot_function)
        return "binning" in plot_func_signature.parameters.keys()

    def build_plot_file_name(
        self,
        plot_name: str,
        exp_range: Optional[Tuple[float, float]] = None,
        file_name_extension: Optional[str] = None,
    ) -> str:
        exp_range_info = ""
        if exp_range is not None:
            exp_range_info = f"_exp{exp_range[0]}_{exp_range[1]}"

        ext = ""
        if file_name_extension is not None:
            ext = f"_{file_name_extension}"

        return f"{self.file_name_base}{ext}_{self.weight_type}{exp_range_info}_{plot_name}"
