"""
Particle ID table formatter for Belle

Felix Metzner 2019
"""

import os
import pandas as pd

from typing import Optional, List

from analysistools.utilities import PathType
from analysistools.pid_corrections.pid_corrector import KPiIDCorrection


def load_correction_table(
    table_file: PathType,
    delimiter: str = ",",
):
    return pd.read_csv(table_file, header=None, delimiter=delimiter)


def k_pi_table_transform(
    original_table: pd.DataFrame,
    output_file: Optional[PathType] = None,
    verbose: bool = False,
):
    # Reformatting cut value x -> 0.x
    original_table["cut"] *= 0.1
    original_table["cut"].round(1)

    # Adding theta and momentum bin values
    k_pi_momentum_bins = KPiIDCorrection.get_momentum_bins()
    k_pi_theta_bins = KPiIDCorrection.get_theta_bins()
    k_pi_cuts = KPiIDCorrection.get_k_pi_pid_cut_values()

    original_table["mom_lower"] = 0.0
    original_table["mom_upper"] = 0.0
    original_table["theta_lower"] = 0.0
    original_table["theta_upper"] = 0.0

    theta_bins = len(k_pi_theta_bins) - 1
    momentum_bins = len(k_pi_momentum_bins) - 1

    for cut_bin, cut_value in enumerate(k_pi_cuts):
        cut_bin_start = cut_bin * theta_bins * momentum_bins
        cut_bin_end = (cut_bin + 1) * theta_bins * momentum_bins - 1

        if verbose:
            print(
                cut_bin,
                cut_value,
                cut_value * 10,
                cut_bin_start,
                cut_bin_end,
                int(original_table.iloc[[cut_bin_start]]["cut"]),
                int(original_table.iloc[[cut_bin_end]]["cut"]),
            )

        for mom_bin, mom_value in enumerate(k_pi_momentum_bins[:-1]):
            mom_bin_start = cut_bin_start + mom_bin * theta_bins
            mom_bin_end = cut_bin_start + (mom_bin + 1) * theta_bins - 1

            mom_value_next = k_pi_momentum_bins[mom_bin + 1]

            if verbose:
                print(
                    "\t",
                    mom_bin,
                    mom_value,
                    mom_value_next,
                    mom_bin_start,
                    mom_bin_end,
                    int(original_table.iloc[[mom_bin_start]]["cut"]),
                    int(original_table.iloc[[mom_bin_end]]["cut"]),
                )

            for theta_bin, theta_value in enumerate(k_pi_theta_bins[:-1]):
                theta_bin_start = mom_bin_start + theta_bin

                theta_value_next = k_pi_theta_bins[theta_bin + 1]

                if verbose:
                    print(
                        "\t\t",
                        theta_bin,
                        theta_value,
                        theta_value_next,
                        theta_bin_start,
                        int(original_table.iloc[[theta_bin_start]]["cut"]),
                    )

                original_table.at[theta_bin_start, "mom_lower"] = mom_value
                original_table.at[theta_bin_start, "mom_upper"] = mom_value_next
                original_table.at[theta_bin_start, "theta_lower"] = theta_value
                original_table.at[theta_bin_start, "theta_upper"] = theta_value_next

    assert not original_table.isnull().any().any()

    if output_file:
        original_table.to_csv(output_file, sep=",", header=False, index=False, float_format="%.4f")


def produce_extended_k_pi_tables(
    overwrite: bool = False,
):
    script_path = os.path.dirname(os.path.abspath(__file__))
    tables_path = os.path.join(script_path, "kaon_pion_id_tables")

    exp_ranges = ["SVD1", "SVD2"]  # type: List[str]

    for exp_range in exp_ranges:
        table_path = os.path.join(tables_path, f"KPiEff{exp_range}.csv")
        if not os.path.isfile(table_path):
            raise ValueError(f"Kaon Pion ID table file {table_path} does not exist!")

        k_pi_table = load_correction_table(table_file=table_path)
        k_pi_table.columns = KPiIDCorrection.get_k_pi_original_columns()

        for mode_id, table_name in KPiIDCorrection.get_k_pi_mode().items():
            new_table_path = os.path.join(tables_path, f"{table_name}_{exp_range}.csv")
            if os.path.isfile(new_table_path) and not overwrite:
                print(f"File {new_table_path} already exists. Skipping...")

            original_table = k_pi_table.loc[k_pi_table["mode"] == mode_id].copy().reset_index(drop=True)

            k_pi_table_transform(
                original_table=original_table,
                output_file=new_table_path,
            )


if __name__ == "__main__":
    produce_extended_k_pi_tables()
