"""
Particle ID table manager for Belle

Handles pID correction tables.

Felix Metzner 2019
"""
import numpy as np
import pandas as pd

from typing import Optional, List, Dict, Sequence

from analysistools.utilities import PathType
from analysistools.utilities.singleton_decorator import Singleton


__all__ = [
    "TableManager",
]


@Singleton
class TableManager:
    """
    Loads and keeps track of particle ID correction tables.
    """

    def __init__(self) -> None:
        self.tables = {}  # type: Dict[str, pd.DataFrame]
        self.matrices = {}  # type: Dict[str, np.ndarray]

    def load_table(
        self,
        name: str,
        path: PathType,
        column_names: Sequence[str],
    ) -> pd.DataFrame:
        table = pd.read_csv(path, header=None, delimiter=",")
        table.columns = column_names

        self.tables[name] = table
        return table

    def get_table(
        self,
        name: str,
        path: PathType,
        column_names: Sequence[str],
    ) -> pd.DataFrame:
        if name in self.tables:
            return self.tables[name]
        else:
            return self.load_table(name=name, path=path, column_names=column_names)

    def get_correction_matrix(self, name: str) -> Optional[np.ndarray]:
        if name in self.matrices:
            return self.matrices[name]
        else:
            return None

    def get_matrices_starting_with(self, name_beginning: str) -> List[np.ndarray]:
        return [v for k, v in self.matrices.items() if k.startswith(name_beginning)]

    def contains_matrix(self, name: str) -> bool:
        return name in self.matrices

    def add_correction_matrix(
        self,
        name: str,
        new_matrix: np.ndarray,
    ) -> None:
        if name in self.matrices:
            raise ValueError(f"The matrix name {name} is already registered!")
        self.matrices[name] = new_matrix
