"""
Provides methods to create latex tables from pandas data frames.

Felix Metzner 2019
"""

import os
import copy
import glob
import subprocess
import numpy as np
import pandas as pd

from typing import Optional, Union, Sequence, List, Tuple, Dict, Callable

from analysistools.utilities import PathType


__all__ = [
    "df_to_latex_table",
    "df_to_split_latex_table",
]


default_latex_table_header = r"""\documentclass{standalone}
\usepackage{siunitx}
\usepackage{amsfonts}
\usepackage{array}
\usepackage{booktabs}
\usepackage{siunitx}
\begin{document}
"""  # type: str
default_latex_table_tail = r"\end{document}"  # type: str


def df_to_latex_table(
    dfs: Sequence[pd.DataFrame],
    target_path: PathType,
    create_pdf: bool = False,
    row_formatters: Optional[Dict[str, Callable]] = None,
    col_formatters: Union[Dict[str, Callable], Sequence[Callable]] = None,
    custom_header: Optional[str] = None,
    custom_tail: Optional[str] = None,
    separator: bool = False,
    alignment: Union[str, List[str]] = "r",
    index_alignment: str = "r",
    placeholder_element: Optional[str] = None,
    **kwargs,
):
    assert all(isinstance(df, pd.DataFrame) for df in dfs)
    if len(dfs) > 1:
        assert all(len(df.columns) <= len(dfs[0].columns) for df in dfs)

    if not os.path.splitext(target_path)[1]:
        pure_table_latex_path = str(target_path) + "_pure_table.tex"  # type: PathType
        latex_path = str(target_path) + ".tex"  # type: PathType
        pdf_path = str(target_path) + ".pdf"  # type: PathType
    else:
        pure_table_latex_path = os.path.splitext(target_path)[0] + "_pure_table.tex"
        latex_path = os.path.splitext(target_path)[0] + ".tex"
        pdf_path = os.path.splitext(target_path)[0] + ".pdf"

    table_header = default_latex_table_header if custom_header is None else custom_header
    table_tail = default_latex_table_tail if custom_tail is None else custom_tail

    output = table_header + "\n"
    table_only_output = ""

    # Generating column format to be used.
    index_levels = 1
    if isinstance(dfs[0].index, pd.core.indexes.multi.MultiIndex):
        index_levels = len(dfs[0].index.levels)
    additional_cols = index_levels + 1 if separator else index_levels
    len_alignment = len(dfs[0].columns) + additional_cols
    len_alignment_wo_sep = len(dfs[0].columns) + index_levels

    if isinstance(alignment, list):
        alignment_wo_sep = [a for a in alignment if a != "|"]
        assert len(alignment_wo_sep) == len_alignment_wo_sep, (
            alignment,
            len(alignment),
            alignment_wo_sep,
            len(alignment_wo_sep),
            len_alignment,
            len_alignment_wo_sep,
        )
        col_format_list = alignment
    else:
        col_format_list = [alignment] * len_alignment
        col_format_list[:index_levels] = [
            index_alignment,
        ] * index_levels
    if separator:
        col_format_list[index_levels] = "|"
    col_format = "".join(col_format_list)

    for i, original_df in enumerate(dfs):
        df = copy.copy(original_df)

        # Applying row formatting.
        # multi index tuple has to be used as key for multi index data frames / tables with multi columns of multi rows.
        if row_formatters is not None:
            assert isinstance(row_formatters, dict), row_formatters
            temp_df = df.T
            for row_name in temp_df.columns:
                if row_name in row_formatters:
                    temp_df[row_name] = temp_df[row_name].apply(row_formatters[row_name])
            df = temp_df.T

        # Fixing column numbers if multiple data frames with different amount of columns are provided.
        added_columns = []  # type: List[Union[str, Tuple[str, ...]]]
        if len(df.columns) < len(dfs[0].columns):
            for c in range((len(dfs[0].columns) - len(df.columns))):
                if isinstance(df.columns, pd.MultiIndex):
                    added_column_tuple = tuple(
                        [r"${}$".format("\," * (c + 1))] * len(df.columns.levels)  # noqa: W605
                    )  # type: Tuple[str, ...]
                    df.loc[:, added_column_tuple] = placeholder_element if placeholder_element is not None else "---"
                    added_columns.append(added_column_tuple)
                else:
                    added_column = r"${}$".format("\," * c)  # type: str  # noqa: W605
                    df.loc[:, added_column] = placeholder_element if placeholder_element is not None else "---"
                    added_columns.append(added_column)

        # Applying own column formatting
        if col_formatters is not None:
            if isinstance(col_formatters, dict):
                for col_name in df.columns:
                    if col_name in col_formatters:
                        df[col_name] = df[col_name].apply(col_formatters[col_name])
            elif isinstance(col_formatters, (list, tuple)):
                assert len(col_formatters) == len(df.columns), (len(col_formatters), len(df.columns))
                for col_index, col_name in enumerate(df.columns):
                    df[col_name] = df[col_name].apply(col_formatters[col_index])
            else:
                raise RuntimeError("Unknown type for column formatters provided")

        for added_col in added_columns:
            df.loc[:, added_col] = r"$\,$"

        with pd.option_context("max_colwidth", 1000):
            latex_table = df.to_latex(escape=False, column_format=col_format, **kwargs)
        latex_table_list = latex_table.splitlines()

        # Removing first and/or last line if multiple tables are stacked.
        if 0 <= i < len(dfs) - 1:
            del latex_table_list[-1]
        if 0 < i <= len(dfs) - 1:
            del latex_table_list[0]

        table_only_output += "\n".join(latex_table_list)
        table_only_output += "\n"

    output += table_only_output
    output += table_tail

    with open(pure_table_latex_path, "w") as f:
        f.write(table_only_output)

    with open(latex_path, "w") as f:
        f.write(output)

    if create_pdf:
        cmd_list = [
            "pdflatex",
            "-halt-on-error",
            "-interaction=nonstopmode",
            f"-output-directory={os.path.dirname(pdf_path)}",
            str(latex_path),
        ]
        ret = subprocess.call(cmd_list)
        if ret != 0:
            raise RuntimeError(f"Compilation of latex table {os.path.basename(latex_path)} failed!\n{' '.join(cmd_list)}")

        exts = [".aux", ".log", ".nav", ".out", ".snm", ".toc", ".bbl", ".blg", ".fdb_latexmk", ".fls"]
        files = [fp for fp in glob.glob(f"{os.path.splitext(latex_path)[0]}.*") if os.path.splitext(fp)[1] in exts]
        for fp in files:
            os.remove(fp)


def df_to_split_latex_table(
    df: pd.DataFrame,
    table_fractions: Sequence[int],
    name_prefix: str,
    output_lookup_dict: Dict[str, PathType],
    placeholder_element: Optional[str] = None,
    remove_column_name_index: bool = False,
    split_col_formatters: Optional[Union[List[Callable], Tuple[Callable]]] = None,
    allow_smaller_sub_dicts: bool = False,
    **kwargs,
):
    if split_col_formatters is not None:
        assert "col_formatters" not in kwargs.keys(), (split_col_formatters, kwargs["col_formatters"])

    df_cols = copy.copy(df.columns)
    if isinstance(df.columns, pd.core.indexes.multi.MultiIndex):
        df_cols = copy.copy(df.columns.levels[0])

    for frac in table_fractions:
        adapted_kwargs = copy.copy(kwargs)
        assert len(df_cols) >= frac, (frac, len(df_cols))
        split = int(np.ceil(len(df_cols) / frac))
        if not allow_smaller_sub_dicts:
            assert len(df_cols) % frac == 0, (frac, len(df_cols))
            assert split == int(len(df_cols) / frac), (split, int(len(df_cols) / frac))
        formatter_factor = split
        dfs = [
            copy.copy(df.loc[:, df_cols[a:b]]) if b <= len(df_cols) else copy.copy(df.loc[:, df_cols[a:]])
            for a, b in zip(np.arange(frac) * split, (np.arange(frac) + 1) * split)
        ]

        if remove_column_name_index:
            for df_part in dfs:
                if isinstance(df_part.columns, pd.core.indexes.multi.MultiIndex):
                    df_part.rename(columns=lambda x: x.split("_", 1)[1], level=0, inplace=True)
                else:
                    df_part.rename(columns=lambda x: x.split("_", 1)[1], inplace=True)

        if placeholder_element is not None:
            for i, df_part in enumerate(dfs):
                t_df_part = df_part.T
                for c in t_df_part.columns:
                    if (t_df_part[c].astype(str) == placeholder_element).all():
                        del t_df_part[c]
                dfs[i] = t_df_part.T

        if split_col_formatters is not None:
            assert "col_formatters" not in adapted_kwargs.keys(), (split_col_formatters, adapted_kwargs["col_formatters"])
            adapted_kwargs.update({"col_formatters": split_col_formatters * formatter_factor})

        df_to_latex_table(
            dfs=dfs,
            target_path=output_lookup_dict[f"{name_prefix}_table_{frac}_part_latex"],
            **adapted_kwargs,
        )

        for i, df_part in enumerate(dfs):
            if split_col_formatters is not None:
                part_formatter_factor = len(df_part.columns)
                if isinstance(df_part.columns, pd.core.indexes.multi.MultiIndex):
                    part_formatter_factor = len(df_part.columns.levels[0])
                adapted_kwargs.update({"col_formatters": split_col_formatters * part_formatter_factor})
            df_to_latex_table(
                dfs=[df_part],
                target_path=output_lookup_dict[f"{name_prefix}_table_{i + 1}_of_{frac}_latex"],
                **adapted_kwargs,
            )
