"""
Definition of the best candidate selection (bcs) settings.

The settings are defined in a dictionary of the form shown by the example of a
best candidate selection of the Upsilon(4S) candidates based on the highest value
of the FEI signal probability of the tag side B meson which is the first daughter
of the Upsilon(4S). Only the best candidate is kept:

    bcs_dict_name = {
        "particle": "Upsilon(4S)",
        "rank_by": "highest",
        "variable": "daughter(0, extraInfo(SignalProbability))",
        "output_variable_name": "feiTagSideSignalProbability",
        "num_of_cands": 1,
    }

where the keys have the following meaning:

    particle: The particle this best candidate selections will be applied to.
    rank_by: Wherther to rank by "highest" or "lowest" value of the variable the bcs is based on.
    variable: The variable that is evaluated to determine the ranking of the candidates.
    output_variable_name: Variable name for the extra info that is added to the candidate which contains
                          the candidate"s rank and is accessible for VariablesToNtuple.
    num_of_cands: The number of best candidates according to the above explained ranking to keep.

The user can define own bcs dictionaries and apply those via the provided apply_bcs_from_dict function.

All used bcs dictionaries ca be collected in a dictionary, which can be used to print an overview of the bcs settings.


Max Welsch 2018
"""

import copy

from typing import List, Dict, Any
from collections import OrderedDict
from modularAnalysis import rankByHighest, rankByLowest


# bcs based on FEI signal probability for tag side B meson (which is the first daughter of the Upsilon(4s)).
upsilon4s_with_fei_probability_bcs = {
    "particle": "Upsilon(4S)",
    "rank_by": "highest",
    "variable": "daughter(0, extraInfo(SignalProbability))",
    "output_variable_name": "feiTagSideSignalProbabilityRank",
    "num_of_cands": 1,
    "multi_rank": False,
}


bcs_settings = {
    "upsilon4s_with_fei_probability_bcs": upsilon4s_with_fei_probability_bcs,
}


def apply_bcs_from_dict(path, bcs_dict: Dict[str, Any], particle_lists: List[str]) -> None:
    """
    Convenience function to apply a best candidate selection via the modularAnalysis functions rankByHighest or
    rankByLowest using predefined settings which are passed in form of a bcs dictionary as defined in
    best_cand_selection.py.

    :param path: basf2 path to append modules to.
    :param bcs_dict: The best candidate selection dictionary containing the settings for the best candidate selection
                     which shall be applied.
    :param particle_lists: List of particle lists on which the best candidate selection will be applied.
    :return: None
    """

    for particle_list in particle_lists:
        if not particle_list.startswith(bcs_dict["particle"]):
            raise ValueError(
                f"The provided bcs dictionary for the particle {bcs_dict['particle']}"
                f"is not meant to be applied to the particle {particle_list}!"
            )

        multi_ranking = bcs_dict.get("multi_rank", False)

        if bcs_dict["rank_by"] == "highest":
            rankByHighest(
                particleList=particle_list,
                variable=bcs_dict["variable"],
                numBest=bcs_dict["num_of_cands"],
                allowMultiRank=multi_ranking,
                outputVariable=bcs_dict["output_variable_name"],
                path=path,
            )
        elif bcs_dict["rank_by"] == "lowest":
            rankByLowest(
                particleList=particle_list,
                variable=bcs_dict["variable"],
                numBest=bcs_dict["num_of_cands"],
                allowMultiRank=multi_ranking,
                outputVariable=bcs_dict["output_variable_name"],
                path=path,
            )
        else:
            raise ValueError(
                f'The best candidate selection option "rank_by" = {bcs_dict["rank_by"]}'
                f"defined in the used best candidate selection dictionary is not valid!"
                f'The entry bcs_dict["rank_by"] must be either "highest" or "lowest"!'
            )


bcs_headers = OrderedDict(
    [
        ("bcs_setting", "bcs_setting"),
        ("particle", "particle"),
        ("rank_by", "rank_by"),
        ("variable", "variable"),
        ("output_variable_name", "output_variable_name"),
        ("num_of_cands", "num_of_cands"),
    ]
)


def print_bcs_settings(bcs_settings_dict, header_dict=None):
    """
    Print overview of available best candidate selection settings form bcs_settings_dict
    :param bcs_settings_dict: Dictionary containing the used bcs settings.
    :param header_dict: Ordered dictionary containing the header, so that
                        order of columns and their names are defined.
                        Default is best_cand_selection_utils.bcs_headers.
    :return: None
    """
    if header_dict is None:
        header_dict = copy.deepcopy(bcs_headers)

    column_headers = [key for key, _ in header_dict.items()]
    table_as_list = [column_headers]

    for key, bcs_dict in sorted(bcs_settings_dict.items()):
        bcs_row = []
        for _, column in header_dict.items():
            if column == "bcs_setting":
                bcs_row.append(str(key))
            if column == "particle":
                bcs_row.append(str(bcs_dict["particle"]))
            if column == "rank_by":
                bcs_row.append(str(bcs_dict["rank_by"]))
            if column == "variable":
                bcs_row.append(str(bcs_dict["variable"]))
            if column == "output_variable_name":
                bcs_row.append(str(bcs_dict["output_variable_name"]))
            if column == "num_of_cands":
                bcs_row.append(str(bcs_dict["num_of_cands"]))
        table_as_list.append(bcs_row)

    column_size = [max(map(len, col)) for col in zip(*table_as_list)]

    format_str = " | ".join("{{:<{i}}}".format(i=i) for i in column_size)
    table_as_list.insert(1, ["-" * i for i in column_size])

    for row in table_as_list:
        print(format_str.format(*row))
