"""
This module provides some tools to generate and evaluate the user defined
and analysis specific decay channels and their cuts.

The decay channels themselves and their cuts have to be defined specifically
for the analysis by the user under consideration of the following rules:

1.  The final goal is to provide a dictionary with the structure:

    decay_channel_dict = {
        "decay_channels": decay_channels,
        "generator_level_only_decay_channels": generator_level_only_decay_channels,
        "decay_channels_with_neutrinos": decay_channels_with_neutrinos,
        "decay_label_to_decay_id": decay_label_to_decay_id,
        "decay_id_decay_label": decay_id_decay_label,
        "decay_label_to_chain": decay_label_to_chain,
        "decay_label_to_cut_list": decay_label_to_cut_list,
        "decay_label_to_extra": decay_label_to_extra,
    }

    where the values themselves are dictionaries containing the respective information
    for all decay channels required in the analysis.

2.  These information are:
    - decay_channels: A dictionary with the decay channel tag as key and
        a simple string for illustration of the decay channel.
        E.g.: {"B0:D_e": "B0 -> D+ e-", "B0:D_mu": "B0 -> D+ mu-", ...}
    - generator_level_only_decay_channels: A dictionary containing the entries
        of the decay_channels dictionary of those decays that shall only be
        considered when generator level information is used to reconstruct
        the decay.
    - decay_channels_with_neutrinos: A dictionary containing the entries of
        the decay_channels dictionary of those decays that contain a neutrino.
        This is used to determine the correct Monte Carlo matching constrained
        when the generator level information is used to reconstruct the decay.
    - decay_label_to_decay_id: A dictionary containing the decay tag as key and
        an integer as unique decay ID for each decay.
        E.g.: {"B0:D_e": 1, "B0:D_mu": 2, ...}
    - decay_id_decay_label: inversion of decay_label_to_decay_id
    - decay_label_to_chain: A dictionary containing the decay tag as key and the
        basf2 decay string for the decay, which can be formatted to change the
        particle list tag (e.g. for the reconstruction using generator level info).
        E.g.: {"B0:D_e": "B0:D_e{suf} -> D+:all{suf} e-:fsp{suf}",
               "B0:D_mu": "B0:D_mu{suf} -> D+:all{suf} mu-:fsp{suf}", ...}
        where "suf" is used to format the decay string in reconstruction functions.
    - decay_label_to_cut_list: A dictionary containing the decay tag as key and a
        list of cut strings to define cuts on basf2 analysis variables that will
        be applied on the reconstructed particle.
        E.g.: {"B0:D_e": ["Mbc > 5.22", "-0.05 < deltaE < 0.1"],
               "B0:D_mu": ["Mbc > 5.22", "-0.05 < deltaE < 0.1"], ...}
    - decay_label_to_extra: Dictionary containing extra information for some reconstruction
                            steps, such as vertex fitting. E.g.:
                            {"B0:D_e": {"vertex_fit": {"function": "vertexTree", "kwargs": {}}}}
                            Does not need to be defined, as it will be set to be a empty
                            dictionary by default.

3.  It is important to note, that the final Y(4S) particle is always
    reconstructed in such a way, that the first daughter (index 0) is
    the tag B meson, and the second (index 1) is the signal B meson.

4.  How to define these dictionaries is up to the user.


Felix Metzner 2018
"""
import copy

from textwrap import TextWrapper
from typing import Optional, List, Dict, Any
from collections import OrderedDict, defaultdict


decays_channels_headers = OrderedDict(
    [
        ("decay_label", "decay_label"),
        ("decay_id", "decay_id"),
        ("decay", "decay"),
        ("decay_chain", "decay_chain"),
        ("cuts", "cuts"),
        ("extra", "extra"),
    ]
)


def dict_inversion(original_dict: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Inverting key and value of a dictionary.
    :param original_dict: original dictionary
    :return: inverted dictionary
    """
    inverted_dict = {y: x for x, y in original_dict.items()}

    if len(inverted_dict) != len(original_dict):
        raise ValueError(
            f"The inversion of the provided dictionary failed, "
            f"because the original dictionary is not bijective!"
            f"The original dictionary was: \n {original_dict}"
        )

    return inverted_dict


def generate_dict_collection_from_locals(
    local_vars: Dict[str, Any],
    key_word: str,
    default: bool = False,
) -> Dict[str, Dict[str, Any]]:
    """
    Generates a collection of decay dictionaries from all dictionaries defined in the
    provided local variables by taking those which fit the given key_word.
    The key word should be one of "_decay_channels", "_decay_label_to_decay_id",
    "_decay_label_to_chain", "_decay_label_to_cut_list" or "_decay_label_to_extra".
    During the collection, duplication of keys will be checked, so that no entry is
    unintentionally overwritten.
    The generated collection dictionary will be returned.
    If the boolean default is set to True a defaultdict will be used as base for the
    collection which has a dict as default value.
    :param local_vars: Local variables to be checked for the dictionaries to collect.
    :param key_word: Key word to identify the dictionaries that shall be collected.
    :param default: Whether to use a defaultdict as base for the dictionary collection.
    :return: dict or defaultdict: Dictionary collection
    """
    dict_collection = dict()  # type: Dict[str, Dict[str, Any]]
    if default:
        dict_collection = defaultdict(dict)
    for var in list(local_vars.keys()):
        if key_word in var:
            if not isinstance(local_vars[var], dict):
                raise TypeError(
                    f"Trying to add variable {var} to collection dictionary, which is not of type "
                    f"dict but has type {str(type(local_vars[var]))}."
                )
            for key in local_vars[var].keys():
                if key in dict_collection:
                    raise KeyError(
                        f"Trying to add key {key} to dictionary collection for keyword {key_word}, "
                        f"but key is already present and would be replaced!"
                    )
            dict_collection.update(local_vars[var])
    return dict_collection


def generate_dict_collection_from_list(
    dict_list,
    default=False,
    check_dict=None,
    check_message=None,
):
    """
    Generates a collection of decay dictionaries from a provided list of decay channel dictionaries.
    During the collection, duplication of keys will be checked, so that no entry is
    unintentionally overwritten.
    The generated collection dictionary will be returned.
    If the boolean default is set to True a defaultdict will be used as base for the
    collection which has a dict as default value.
    :param dict_list: List of decay channel dictionaries.
    :param default: Whether to use a defaultdict as base for the dictionary collection.
    :param check_dict: If check_dict is provided, the keys of the dicts in dict list must be also present in check_dict.
    :param check_message: Exception message printed if key is not also in check_dict.
    :return: dict or defaultdict: Dictionary collection
    """
    dict_collection = dict()
    if default:
        dict_collection = defaultdict(dict)
    for channel_dict in dict_list:
        if not isinstance(channel_dict, dict):
            raise TypeError(
                f"Trying to add value {channel_dict} to collection dictionary, which is not of type "
                f"dict but has type {str(type(channel_dict))}."
            )
        for key in channel_dict:
            if key in dict_collection:
                raise KeyError(
                    f"Trying to add key {key} to dictionary collection created from list, "
                    f"but key is already present and would be replaced!"
                )
            if check_dict and key not in check_dict:
                raise KeyError(
                    "{m}This is not the case for {k}.".format(m=check_message + "\n" if check_message else "", k=key)
                )
        dict_collection.update(channel_dict)

    return dict_collection


def prepare_decay_channel_dict(
    local_vars: Dict[str, Any],
    channels_with_neutrinos=None,
    gen_level_only_channels=None,
):
    """
    Convenience function to build the desired decay_channel_dict which is required for the analysistools
    package. The decay_channel_dict will be build from the provided local variables "local_vars=locals()"
    which should contain the dictionaries for the different decay channels.
    :param local_vars: Local variables = locals() of the analysis specific decay_channels module which contain
                       the analysis specific decay channel dictionaries.
    :param channels_with_neutrinos: List of decay_channels dicts which contain decay channels with neutrinos and
                                    thus mus be treated differently.
    :param gen_level_only_channels: List of decay_channels dicts which contain decay channels which shall be
                                    reconstructed only from generator level information and thus are treated
                                    differently.
    :return: dict: Dictionary collection with the structure of required decay_channel_dict.
    """
    if channels_with_neutrinos is None:
        channels_with_neutrinos = []
    if gen_level_only_channels is None:
        gen_level_only_channels = []

    decay_channels = generate_dict_collection_from_locals(local_vars, key_word="_decay_channels")
    decay_label_to_decay_id = generate_dict_collection_from_locals(local_vars, key_word="_decay_label_to_decay_id")
    decay_id_decay_label = dict_inversion(decay_label_to_decay_id)
    decay_label_to_chain = generate_dict_collection_from_locals(local_vars, key_word="_decay_label_to_chain")
    decay_label_to_cut_list = generate_dict_collection_from_locals(local_vars, key_word="_decay_label_to_cut_list")

    decay_channels_with_neutrinos = generate_dict_collection_from_list(
        dict_list=channels_with_neutrinos,
        check_dict=decay_channels,
        check_message="A decay channel with neutrinos must also be present in decay_channels!",
    )

    generator_level_only_decay_channels = generate_dict_collection_from_list(
        dict_list=gen_level_only_channels,
        check_dict=decay_channels,
        check_message="A generator level only decay channel must also be present in decay_channels!",
    )

    decay_label_to_extra = generate_dict_collection_from_locals(
        local_vars,
        key_word="_decay_label_to_extra",
        default=True,
    )

    decay_channel_dict = {
        "decay_channels": decay_channels,
        "generator_level_only_decay_channels": generator_level_only_decay_channels,
        "decay_channels_with_neutrinos": decay_channels_with_neutrinos,
        "decay_label_to_decay_id": decay_label_to_decay_id,
        "decay_id_decay_label": decay_id_decay_label,
        "decay_label_to_chain": decay_label_to_chain,
        "decay_label_to_cut_list": decay_label_to_cut_list,
        "decay_label_to_extra": decay_label_to_extra,
    }

    return decay_channel_dict


def print_decay_channels_table(
    decay_chl_dict: Dict[str, Any],
    header_dict: Optional[Dict[str, str]] = None,
) -> None:
    """
    Printing the information of the decay channels defined in this file
    (or any other channel defined in the same manner).
    :param decay_chl_dict: Decay channels dictionary.
    :param header_dict: Ordered dictionary containing the header, so that
                        order of columns and their names are defined.
                        Default is decay_channel_utils.decays_channels_headers
    :return: None
    """
    if header_dict is None:
        header_dict = copy.deepcopy(decays_channels_headers)

    wrapper = TextWrapper(50, break_long_words=True)

    column_headers = [key for key, _ in header_dict.items()]
    table_as_list = [column_headers]

    for key, _ in sorted(decay_chl_dict["decay_channels"].items()):
        cut_list = decay_chl_dict["decay_label_to_cut_list"][key]
        cut_string_list = wrapper.wrap(cut_list_to_string(cut_list))
        if not cut_string_list:
            # Else channels without any cuts will not be displayed
            cut_string_list = [""]

        for n, cut_string in enumerate(cut_string_list):
            decay_row = []
            for _, column in header_dict.items():
                if column == "decay_label":
                    decay_row.append(str(key) if n == 0 else "")
                if column == "decay_id":
                    decay_row.append(str(decay_chl_dict["decay_label_to_decay_id"][key]) if n == 0 else "")
                if column == "decay":
                    decay_row.append(str(decay_chl_dict["decay_channels"][key]) if n == 0 else "")
                if column == "decay_chain":
                    decay_row.append(str(decay_chl_dict["decay_label_to_chain"][key]) if n == 0 else "")
                if column == "cuts":
                    decay_row.append(cut_string)
                if column == "extra":
                    extras = [extra for extra in decay_chl_dict["decay_label_to_extra"][key].keys()]
                    decay_row.append(", ".join(extras) if n == 0 else "")
            table_as_list.append(decay_row)

    column_size = [max(map(len, col)) for col in zip(*table_as_list)]

    format_str = " | ".join("{{:<{i}}}".format(i=i) for i in column_size)
    table_as_list.insert(1, ["-" * i for i in column_size])

    for row in table_as_list:
        print(format_str.format(*row))


def cut_list_to_string(cut_list: List[str]) -> str:
    """
    Convenience function to turn list of cuts to cut-string.
    :param cut_list: List of cuts to be joined
    :return: cut-string
    """
    return " and ".join(cut_list)
