"""
This module provides the final state particles class as well as the
final state particle collection class.

These classes define the interface with which final state particles lists
can be created and used for an analysis.

A final state particle has the properties
    - label: Label of the final state particle type used for illustration.
    - particle: The name of the particle list (with a tag) which will be
                filled with the respective particles.
    - origin: Defines from where the particle list is filled. Can be either
              "reco" which means that the fsp will be taken from reconstructed
                     objects such as tracks or clusters
              "mc" which means that the fsp list will be filled from Monte Carlo
                   information.
              A list of other particle lists, e.g. ["pi0:mdst"] which means that
                     the fsp will be taken from an already available basf2
                     particle list such as the pi0:mdst list in the example, which
                     is created during the B2BII conversion process.
    - cuts: Cuts that are required for a particle to be considered for this
            fsp type.

Multiple final state particle objects can be combined in one collection to be
handed over to the anaysis tools and/or be used in the analysis.

Various final state particles with different cuts are already predefined in this
module and gatherd in the final state particle collection "fsp_collection".

Own particle collections can be created or the predefined one extended for the
needs of a specific analysis.

Using fsp_collection.get_final_state_particles(path, use_mc_info) the respective
final state particle lists will be created and filled in the given basf2 path.

Felix Metzner 2018
"""
import basf2
import tabulate

import modularAnalysis as mA

from collections import defaultdict

from typing import Union, Optional, List, Dict

from analysistools.processing.decay_channels_utils import cut_list_to_string
from analysistools.processing.observables_utils import remove_mc_blacklisted_variables


__all__ = [
    "FSRCorrectionSettings",
    "FinalStateParticle",
    "FinalStateParticleCollection",
    "fsp_collection",
    "electron",
    "muon",
    "tau",
    "charged_pion",
    "neutral_pion",
    "photon",
    "charged_kaon",
    "neutral_kaon",
]


class FSRCorrectionSettings:
    """
    The FSRCorrectionSettings class contains the settings required when a
    final state radiation correction is applied to electrons.
    """

    def __init__(
        self,
        fsr_gamma_list_name: str,
        angle_threshold: float = 1.5,
        energy_threshold: float = 1.0,
        theta_threshold: float = 1.0,
        momentum_factor: float = 1.0,
        method: str = "offsetCone",
        take_all_gammas: bool = False,
        corr_electron_list_name: str = "e-:fsp_fsr_corrected",
        write_out: bool = False,
    ):
        """
        Constructor of the FSRCorrectionSettings.
        Required is only the name of the photon particle list which contains
        the gammas to be considered as candidates for final state radiation.
        The remaining properties are set their default values if not provided.

        :param fsr_gamma_list_name: The particle list name for the fsr gamma candidates.
        :param angle_threshold: Maximal 1/2 of the opening angle of the cone in
        which to search for fsr gamma candidates. Default is 1.5 degrees.
        :param energy_threshold: Maximal energy of a fsr gamma candidate to be considered.
        Default is 1.0 GeV.
        :param theta_threshold: Maximal theta angle between lepton and fsr gamma candidate
        to be considered in arc method. Default is 1.0 degrees.
        :param momentum_factor: Factor for a check of the photon momentum to be lower than
        factor*lepton momentum.
        :param method: String defining the method to be used for the correction.
        :param take_all_gammas: Boolean flag to set whether all gammas fulfilling the
                                criteria will be added or only the best one.
        :param corr_electron_list_name: String defining the particleList name for the
                                        corrected electron list.
        :param write_out: Whether to write out the created list in the ROOT output.
        """
        self.fsr_gamma_list_name = fsr_gamma_list_name
        self.angle_threshold = angle_threshold
        self.energy_threshold = energy_threshold
        self.theta_threshold = theta_threshold
        self.momentum_factor = momentum_factor
        self.method = method
        self.take_all_gammas = take_all_gammas

        if not corr_electron_list_name.startswith("e-:"):
            raise RuntimeError(
                f"The particle list name for the corrected electron list must start with 'e-:', "
                f"but the list name you provided, which was '{corr_electron_list_name}', does not!"
            )
        self.corr_electron_list_name = corr_electron_list_name
        self.write_out = write_out

    def get_fsr_correction_settings(self) -> Dict[str, Union[str, float, bool]]:
        """
        :return: Dictionary containing the settings in the format expected by
        the final state radiation correction function correctFSR of the basf2
        modular analysis module.
        """
        return {
            "gammaListName": self.fsr_gamma_list_name,
            "angleThreshold": self.angle_threshold,
            "energyThreshold": self.energy_threshold,
            "thetaThreshold": self.theta_threshold,
            "momentumFactor": self.momentum_factor,
            "method": self.method,
            "takeAllGammas": self.take_all_gammas,
            "correctedElectronListName": self.corr_electron_list_name,
            # "writeOut": self.write_out
        }

    def description(self) -> str:
        """
        :return: String containing a description of the fsp correction settings.
        """
        return f"via {self.fsr_gamma_list_name}"


class FinalStateParticle:
    """
    This class describes properties of final state particles (FSP)
    used as base for the reconstruction of decay channels.
    """

    def __init__(
        self,
        label: str,
        particlelist: str,
        origin: Union[str, List[str]],
        cuts: List[str],
        electron_fsr_correction: Optional[FSRCorrectionSettings] = None,
    ) -> None:
        """

        Constructor.

        :param label: Label of the final state particle type
        used for illustration.
        :param particlelist: The name of the particle list
        (with a tag) which will be filled with the respective
        particles.
        :param origin:  Defines from where the particle list
        is filled. Can be either:
            "reco": which means that the fsp will be taken from
                reconstructed objects such as tracks or clusters
            "mc": which means that the fsp list will be filled
                from Monte Carlo information.
            A list of other particle lists: e.g. ["pi0:mdst"]
                 which means that the fsp will be taken from an
                 already available basf2 particle list such as
                 the pi0:mdst list in the example, which is
                 created during the B2BII conversion process.
        :param cuts: Cuts that are required for a particle to be
        considered for this fsp type.
        :param electron_fsr_correction: A FSRCorrectionSettings instance
        containing the setting for the final state radiation correction
        for electrons or False if no fsr correction shall be applied.
        Default is None.
        """
        self._label = label  # type: str
        self._particlelist = particlelist  # type: str
        self._origin = origin  # type: Union[str, List[str]]
        self._cuts = cuts  # type: List[str]

        self._electron_fsr_correction = None  # type: Optional[FSRCorrectionSettings]
        self.electron_fsr_correction = electron_fsr_correction  # type: Optional[FSRCorrectionSettings]

    @property
    def label(self) -> str:
        """
        Final state particle label. Used for illustration.
        :return: Label.
        """
        return self._label

    @property
    def particlelist(self) -> str:
        """
        Name of the particlelist which will be filled with this
        FSP.
        :return: Particle list name + tag
        """
        return self._particlelist

    @property
    def particle_type(self) -> str:
        """
        Returns particle typ of final state particle.
        The particle type is extracted from the particlelist name.
        E.g.: For the particlelist "e+:fsp" the function returns "e+".
        """
        return self.particlelist.split(":", 1)[0]

    @property
    def list_label(self) -> str:
        """
        Returns the label or tag given to the final state particle list.
        The label is extracted from the particlelist name.
        E.g.: For the particlelist "e+:fsp" the function returns "fsp".
        """
        return self.particlelist.split(":", 1)[1]

    @property
    def origin(self) -> Union[str, List[str]]:
        """
        Origin of the particle.

        :return: Particle origin.
        """
        return self._origin

    @property
    def cuts(self) -> List[str]:
        """
        Cuts that are required for the particle to pass when filled
        into a particle list.
        :return: List of cuts.
        """
        return self._cuts

    @property
    def electron_fsr_correction(self) -> Optional[FSRCorrectionSettings]:
        """
        Whether to apply final state radiation correction to electron fsps
        and if so,the dictionary with the settings for the final state particle
        correction.
        :return: False if no fsr correction or the settings for the fsr correction.
        """
        return self._electron_fsr_correction

    @electron_fsr_correction.setter
    def electron_fsr_correction(self, value: Optional[FSRCorrectionSettings]) -> None:
        if value is not None:
            assert self.particle_type in ["e-", "e+"]
            assert self.origin == "reco"
        self._electron_fsr_correction = value

    def electron_fsr_correction_description(self) -> str:
        if self.electron_fsr_correction is None:
            return "None"
        else:
            return self.electron_fsr_correction.description()

    def as_dict(self) -> Dict[str, Union[str, List[str]]]:
        """
        Dictionary representation of this class.

        :return: Dictionary with filled label, particlelist,
        origin and cuts.
        """
        return {
            "label": self.label,
            "particlelist": self.particlelist,
            "origin": self.origin,
            "cuts": self.cuts,
            "fsr correction": self.electron_fsr_correction_description(),
        }

    def __eq__(self, other: object) -> bool:
        """
        Overwrite equals operator to compare the values of the properties
        instead of instance identities, which still can be compared via ``is``.
        """
        if isinstance(other, self.__class__):
            return self.as_dict() == other.as_dict()
        return False

    def __str__(self) -> str:
        """
        String for pretty printing the FSP. Basically a more readable version of ``as_dict()``
        """
        fsp_string = "\nFinalStateParticle"
        fsp_string += "\n-----------------------------\n"
        fsp_string += tabulate.tabulate(self.as_dict(), headers="keys", tablefmt="orgtbl")
        return fsp_string


class FinalStateParticleCollection:
    """
    Custom collection for final state particles (FSP).
    """

    def __init__(self, **kwargs) -> None:
        """
        Constructor

        :param kwargs: Final state particles.

        :Example:

        >>> electron = FinalStateParticle("e", "e-:fsp", "reco", ["dr < 2"])
        >>> muon = FinalStateParticle("mu", "mu-:fsp", "reco", ["dr < 2"])
        >>> fsp_coll = FinalStateParticleCollection(e=electron, mu=muon)
        """

        self._reco_fsp = {}  # type: Dict[str, FinalStateParticle]
        self._mc_fsp = {}  # type: Dict[str, FinalStateParticle]
        self._list_fsp = {}  # type: Dict[str, FinalStateParticle]
        self._store_fsp(kwargs)

    def _store_fsp(self, kwargs_dict: Dict[str, FinalStateParticle]) -> None:
        """
        Helper function. Stores FSPs in the appropriate dictionary
        based on their origin.

        :param kwargs_dict: FSP dict.
        """

        for key, fsp in kwargs_dict.items():

            if not isinstance(fsp, FinalStateParticle):
                raise ValueError("Cannot add anything else than " "final state particles to this collection.")

            if fsp.origin == "reco":
                self._reco_fsp[key] = fsp
            elif fsp.origin == "mc":
                self._mc_fsp[key] = fsp
            else:
                self._list_fsp[key] = fsp

    def get_final_state_particles(self, path: basf2.Path, use_mc_info: bool) -> None:
        """
        Generates the particle lists for final state particles stored
        in the collection.
        Different approaches are used depending on whether the
        particle lists shall be filled from mc_information,
        reconstruction objects, or already available Belle particle
        lists provided by B2BII.

        :param path: basf2 path to append modules to.
        :param use_mc_info: Whether to use MC information for MC matching
                            and generator level particles.
        :return: None
        """
        for fsp in self.list_fsps:
            self.get_fsp_from_list(path, fsp, use_mc_info)

        for fsp in self.reco_fsps:
            self.get_fsp_from_reco(path, fsp, use_mc_info)

        if use_mc_info:
            for fsp in self.all_fsps:
                self.get_fsp_from_mc(path, fsp, use_mc_info)

    @staticmethod
    def get_fsp_from_list(
        path: basf2.Path,
        fsp: FinalStateParticle,
        use_mc_info: bool,
    ) -> None:
        """
        Generating final state particle lists from predefined Belle
        particle lists created during the B2BII conversion.

        :param path: basf2 path to append modules to.
        :param fsp: fs particle object containing the relevant information.
        :param use_mc_info: If true additional MC truth matching is performed
        :return: None
        """
        for index, mdst in enumerate(fsp.origin):
            mA.cutAndCopyList(
                outputListName=f"{fsp.particlelist}_{index}",
                inputListName=mdst,
                cut=cut_list_to_string(fsp.cuts),
                path=path,
            )

        mA.copyLists(
            outputListName=fsp.particlelist,
            inputListNames=[f"{fsp.particlelist}_{i}" for i in range(len(fsp.origin))],
            path=path,
        )

        if use_mc_info:
            mA.matchMCTruth(fsp.particlelist, path=path)

    def get_fsp_from_reco(
        self,
        path: basf2.Path,
        fsp: FinalStateParticle,
        use_mc_info: bool,
    ) -> None:
        """
        Generating final state particle lists from reconstruction
        objects.

        :param path: basf2 path to append modules to.
        :param fsp: FinalStateParticle containing the relevant information.
        :param use_mc_info: If true additional MC truth matching is performed.
        :return: None
        """
        if fsp.electron_fsr_correction is not None:
            assert fsp.electron_fsr_correction.fsr_gamma_list_name in self.list_fsp_particlelists()
            temp_list_name = f"{fsp.particlelist}_before_fsrc"
            particle_cut_tuple = (temp_list_name, cut_list_to_string(fsp.cuts))
            mA.fillParticleLists([particle_cut_tuple], path=path)
            if use_mc_info:
                mA.matchMCTruth(temp_list_name, path=path)

            fsr_corrector = mA.register_module("ConservativeFSRCorrection")
            fsr_corrector.set_name("ConservativeFSRCorrection_" + fsp.particlelist)
            fsr_corrector.param("leptonListName", temp_list_name)
            for param, value in fsp.electron_fsr_correction.get_fsr_correction_settings().items():
                fsr_corrector.param(param, value)
            path.add_module(fsr_corrector)

            if use_mc_info:
                mA.matchMCTruth(fsp.electron_fsr_correction.corr_electron_list_name, path=path)

            mA.copyLists(
                outputListName=fsp.particlelist,
                inputListNames=[temp_list_name, fsp.electron_fsr_correction.corr_electron_list_name],
                path=path,
            )
        else:
            particle_cut_tuple = (fsp.particlelist, cut_list_to_string(fsp.cuts))
            mA.fillParticleLists([particle_cut_tuple], path=path)

        if use_mc_info:
            mA.matchMCTruth(fsp.particlelist, path=path)

    @staticmethod
    def get_fsp_from_mc(
        path: basf2.Path,
        fsp: FinalStateParticle,
        use_mc_info: bool,
    ) -> None:
        """
        Generating final state particle lists from Monte Carlo
        objects.

        Cut strings containing variables which are listed in the
        mc_blacklisted_variables defined in observables_utils are
        removed for the application on the generator level particle
        lists, as these basf2 variables, despite working on
        reconstructed particles, do not yield the same or an at all
        sensible results for MC particles, which can lead to empty
        MC final state particle lists.

        :param path: basf2 path to append modules to.
        :param fsp: FinalStateParticle containing the relevant information.
        :param use_mc_info: If not true, function is not executed.
        :return: None
        """
        if not use_mc_info:
            return

        filtered_cuts = remove_mc_blacklisted_variables(fsp.cuts)
        particle_cut_tuple = (fsp.particlelist + "_gen", cut_list_to_string(filtered_cuts))

        mA.fillParticleListsFromMC(
            [particle_cut_tuple],
            addDaughters=True,
            path=path,
        )

        mA.matchMCTruth(fsp.particlelist + "_gen", path=path)

    def add_fsps(self, **kwargs) -> None:
        """
        Adds final state particles to the internal dictionaries.

        :param kwargs: key=FinalStateParticle
        :return:
        """
        self._store_fsp(kwargs)

    def remove_fsps(self, keys: List[str]) -> None:
        """
        Removes final state particles from the internal dictionaries.

        :param keys: List of keys of final state particles to be removed.
        :return:
        """
        for key in keys:
            v1 = self._reco_fsp.pop(key, None)
            v2 = self._mc_fsp.pop(key, None)
            v3 = self._list_fsp.pop(key, None)
            if v1 is None and v2 is None and v3 is None:
                raise KeyError(
                    f"The final state particle with the key {key} "
                    f"was not found in the collection ant thus not removed."
                )

    @property
    def reco_fsps(self) -> List[FinalStateParticle]:
        """
        :return: A list of all reco FSPs.
        """

        return list(self._reco_fsp.values())

    @property
    def mc_fsps(self) -> List[FinalStateParticle]:
        """
        :return: A list of all mc FSPs.
        """

        return list(self._mc_fsp.values())

    @property
    def list_fsps(self) -> List[FinalStateParticle]:
        """
        :return: A list of all list FSPs.
        """

        return list(self._list_fsp.values())

    @property
    def all_fsps(self) -> List[FinalStateParticle]:
        """
        :return: A combined list of all stored FSPs.
        """

        return self.reco_fsps + self.mc_fsps + self.list_fsps

    def reco_fsp_particlelists(self) -> List[str]:
        """
        :return: List of final state particle lists in reco_fsps
        """
        return [fsp.particlelist for fsp in self._reco_fsp.values()]

    def mc_fsp_particlelists(self) -> List[str]:
        """
        :return: List of final state particle lists in mc_fsps
        """
        return [fsp.particlelist for fsp in self._mc_fsp.values()]

    def list_fsp_particlelists(self) -> List[str]:
        """
        :return: List of final state particle lists in list_fsps
        """
        return [fsp.particlelist for fsp in self._list_fsp.values()]

    def as_dict(self) -> Dict[str, List[Union[str, List[str]]]]:
        """
        Creates a dictionary representation of all stored FSPs.

        :return: Dictionary with FSP properties.
        """
        dd = defaultdict(list)

        for fsp in self.all_fsps:
            for key, val in fsp.as_dict().items():
                dd[key].append(val)

        return dict(dd)

    def __str__(self) -> str:
        """
        String representation. Used for printing the collection.
        """
        return (
            "\nFinalStateParticle Collection"
            + "\n-----------------------------\n"
            + tabulate.tabulate(self.as_dict(), headers="keys", tablefmt="orgtbl")
        )


fsp_collection = FinalStateParticleCollection()

electron = FinalStateParticle(
    label="e",
    particlelist="e-:fsp",
    origin="reco",
    cuts=[
        "dr < 2",
        "abs(dz) < 4",
        "eIDBelle > 0.1",
        "useCMSFrame(pt) > 0.1",
    ],
    electron_fsr_correction=None,
)

fsp_collection.add_fsps(electron=electron)

muon = FinalStateParticle(
    label="mu",
    particlelist="mu-:fsp",
    origin="reco",
    cuts=[
        "dr < 2",
        "abs(dz) < 4",
        "muIDBelle > 0.1",
        "useCMSFrame(pt) > 0.1",
    ],
)

fsp_collection.add_fsps(muon=muon)

tau = FinalStateParticle(
    label="tau",
    particlelist="tau-:mc",
    origin="mc",
    cuts=[
        "dr < 2",
        "abs(dz) < 4",
    ],
)

fsp_collection.add_fsps(tau=tau)

charged_pion = FinalStateParticle(
    label="pi+",
    particlelist="pi+:fsp",
    origin="reco",
    cuts=[
        "dr < 2",
        "abs(dz) < 4",
        "atcPIDBelle(2,3) > 0.1",
        "useCMSFrame(pt) > 0.1",
    ],
)

fsp_collection.add_fsps(charged_pion=charged_pion)

neutral_pion = FinalStateParticle(
    label="pi0",
    particlelist="pi0:fsp",
    origin=["pi0:mdst"],
    cuts=[
        "0.104 < InvM < 0.165",
        "abs(SigMBF) < 3.",
        "E > 0.1",
        "daughter(0, goodBelleGamma) == 1",
        "daughter(1, goodBelleGamma) == 1",
    ],
)

fsp_collection.add_fsps(neutral_pion=neutral_pion)

photon = FinalStateParticle(
    label="gamma",
    particlelist="gamma:fsp",
    origin=["gamma:mdst", "gamma:v0mdst"],
    cuts=[
        "goodBelleGamma == 1",
    ],
)

fsp_collection.add_fsps(photon=photon)

charged_kaon = FinalStateParticle(
    label="K+",
    particlelist="K+:fsp",
    origin="reco",
    cuts=[
        "dr < 2",
        "abs(dz) < 4",
        "atcPIDBelle(3,2) > 0.1",
        "useCMSFrame(pt) > 0.1",
    ],
)

fsp_collection.add_fsps(charged_kaon=charged_kaon)

neutral_kaon = FinalStateParticle(
    label="K_S0",
    particlelist="K_S0:fsp",
    origin=["K_S0:mdst"],
    cuts=[
        "extraInfo(ksnbStandard) == 1",
        "abs(dM) < 0.1",
        "abs(SigMBF) < 3.",
    ],
)

fsp_collection.add_fsps(neutral_kaon=neutral_kaon)

if __name__ == "__main__":
    print(fsp_collection)
