"""
Command line interface for batch jobs.

Felix Metzner 2018
"""

import argparse

from typing import Union, Optional, List, Dict, Any, NamedTuple

from analysistools.utilities import PathType
from analysistools.batch_tools.file_handler import load_json_config

__all__ = [
    "BatchJobCMDLineArgs",
    "get_commandline_args",
]


class BatchJobCMDLineArgs(NamedTuple):
    input_files: List[PathType]
    input_sequences: List[str]
    output_dict: Dict[str, PathType]
    config: Dict[str, Any]
    sampling_rate: Optional[float]


def get_commandline_args() -> BatchJobCMDLineArgs:
    """
    Parser for commandline arguments to get file list and input sequence list.
    Additionally the targets for the different outputs are obtained from a sequence of
    strings such as --output_dict key1 value1 key2 value2...
    This list is then transformed to a dictionary which contains as keys the name of the
    output and as corresponding items the target path. E.g.:
    --output_dict Bsig_ntuple /path/to/job/dir/B_sig_ntuple.root Y4S_ntuple /path/to/job/dir/Y4S_ntuple.root
    is converted to the dictionary
    {
    "Bsig_ntuple": "/path/to/job/dir/B_sig_ntuple.root",
    "Y4S_ntuple": "/path/to/job/dir/Y4S_ntuple.root",
    }
    Furthermore the batch_config.json is read in. From this further parameters are made available for the
    different modules (e,g FEI options, use_mc_info, ..)
    :return: tuple of lists, a dict (file_list, input_sequence_list, output_dict) and the batch_config
    """

    parser = argparse.ArgumentParser(description="basf2 steering file with commandline input.")

    parser.add_argument("--config", type=str, required=True)
    parser.add_argument("--input_sequences", nargs="+", type=str, required=True)
    parser.add_argument("--output_dict", nargs="+", type=str, required=True)
    parser.add_argument("--input_files", nargs="+", type=str, required=True)

    args = vars(parser.parse_args())
    input_files = args["input_files"]  # type: List[PathType]
    input_sequences = args["input_sequences"]  # type: List[str]
    output_dict_list = args["output_dict"]  # type: List[Union[str, PathType]]
    batch_config_file = args["config"]  # type: PathType

    batch_config = load_json_config(batch_config_file, hint="Loading batch config in get_commandline_args")

    # Read global_config out of batch_config and merge the two dicts into one config dictionary
    global_config = load_json_config(batch_config["global_config"], hint="Loading batch config in get_commandline_args")

    config = batch_config.copy()
    config.update(global_config)
    config["batch_config"] = str(batch_config_file)

    if len(input_files) != len(input_sequences):
        raise ValueError(
            "Number of input files does not match number of provided input sequences!\n"
            "Input files are:\n {f}\n"
            "Input sequences are:\n {s}".format(f="\n".join([str(f) for f in input_files]), s="\n".join(input_sequences))
        )

    if len(output_dict_list) % 2 != 0:
        raise ValueError(
            f"There is something wrong with the signature of the provided output dictionary!\n"
            f"The provided list is {output_dict_list}\nThis list should contain pairs of "
            f"[key, value, key, value, ...] which are matched to create a dictionary."
            f"However, the number of elements is odd..."
        )

    sampling_rate = None if "sampling_rate" not in config.keys() else config["sampling_rate"]  # type: Optional[float]
    if sampling_rate is not None:
        if not 0.0 <= sampling_rate <= 1.0:
            raise ValueError(
                f"The optional argument '--sampling_rate' should be a float between 0.0 and 1.0 or None, "
                f"but is {sampling_rate}!"
            )

    output_dict = {}  # type: Dict[str, PathType]
    _output_dict = dict(zip(output_dict_list[0::2], output_dict_list[1::2]))
    if all(isinstance(k, str) for k in _output_dict.keys()):
        output_dict = _output_dict  # type: ignore
    else:
        raise RuntimeError(
            f"The keys of output_dict should be strings, "
            f"but are {list(set([type(k).__name__ for k in output_dict.keys()]))}"
        )

    return BatchJobCMDLineArgs(
        input_files=input_files,
        input_sequences=input_sequences,
        output_dict=output_dict,
        config=config,
        sampling_rate=sampling_rate,
    )
