"""
Convenience functions to retrieve some common sets of variables.

It is important to note, that the final Y(4S) particle is always
reconstructed in such a way, that the first daughter (index 0) is
the tag B meson, and the second (index 1) is the signal B meson.

Felix Metzner 2018
"""

import itertools
import tabulate

from abc import ABC, abstractmethod
from typing import List, Dict, Any, Set


# Set of variables which are ill-defined for use on basf2 MC Particles.
# The data structure is chosen to be a set, as this performs best for x in operations.
_mc_blacklisted_variables = {
    "goodBelleGamma",
    "SigMBF",
    "InvM",
    "eIDBelle",
    "muIDBelle",
    "muIDBelleQuality",
    "ksnbStandard",
    "goodKs",
    "atcPIDBelle",
}  # type: Set[str]


class VariableCollection:
    """
    Container to store observables used for specific particle lists
    in the analysis code.
    """

    def __init__(self, name=None, **kwargs) -> None:
        """
        Constructor. Only variables for this particle/particlelist
        should be given as arguments. If variables for daughter
        particles should be included, use the add_daughter_variables
        method.

        :param kwargs: Key/Value pair of variable lists and their keyword.
                       E.g. "event=EventVariables.get()", where
                       EventVariables.get() creates a list of variable names.
        """
        self._name = name  # type: str
        self._variables = {key: lobs for key, lobs in kwargs.items()}  # type: Dict[str, List[str]]

    def add_daughter_variables(
        self,
        var_list: List[str],
        *daughter_indices: int,
    ) -> None:
        """
        Adds variables to the collection used for the daughter specified
        in the daughter indices. See docstring of daughter function defined
        in the analysistools.observables_utils module.

        :param var_list: List of variables to extracted from the specified
                         child particle.
        :param daughter_indices: Daughter indices in reversed order.
        :return: None
        """
        _d_key = f"daughter_{str(daughter_indices[::-1])}"  # type: str
        if _d_key not in self._variables:
            self._variables[_d_key] = daughter(var_list, *daughter_indices)
        else:
            self._variables[_d_key].extend(daughter(var_list, *daughter_indices))

    def add_mc_daughter_variables(
        self,
        var_list: List[str],
        *daughter_indices: int,
    ) -> None:
        """
        Adds variables to the collection used for the mc daughter specified
        in the daughter indices. See docstring of mc_daughter function defined
        in the analysistools.observables_utils module.

        :param var_list: List of variables to extracted from the specified
                         MC child particle.
        :param daughter_indices: MC daughter indices in reversed order.
        :return: None
        """
        _d_key = f"mc_daughter_{str(daughter_indices[::-1])}"  # type: str
        if _d_key not in self._variables:
            self._variables[_d_key] = mc_daughter(var_list, *daughter_indices)
        else:
            self._variables[_d_key].extend(mc_daughter(var_list, *daughter_indices))

    def as_list(self) -> List[str]:
        """
        Creates a List from all variables specified as attributes.
        :return: Variable List.
        """
        return list(set(list(
            itertools.chain.from_iterable(self._variables.values())
        )))

    def as_dict(self) -> Dict[str, List[str]]:
        """
        Returns the added variables as dictionary.
        :return: Variable dictionary.
        """
        return self._variables

    def __getitem__(self, item: str) -> List[str]:
        """
        Returns a specific variable list belonging to the attribute
        specified as argument.
        :param item: Name of the variable list.
        :return: The wanted variable list.
        """
        return self._variables[item]

    def __str__(self) -> str:
        """
        String representation of the class as table. Column headings
        are the names provided as keyword arguments.

        :returns: Table of variables.
        """

        if self._name is not None:
            heading = f"\nVariableCollection {self._name} contains:\n"
        else:
            heading = "\n"

        return heading + str(tabulate.tabulate(self._variables, headers="keys", tablefmt="orgtbl"))


class VariableCollectionFactory(ABC):
    """
    Factory class which creates VariableCollections based on
    parameters defined in the reconstruction config.
    """

    def __init__(self, config: Dict[str, Any]) -> None:
        """
        Constructor. Sets parameters necessary for creating
        VariableCollections from VariableLists.

        :param config: The config dictionary used in the analysis
                       for setting flags like mc_info or use_b2bii.
        """
        self._use_mc_info = config["use_mc_info"]
        self._roe_mask = config["roe_mask"]
        self._use_b2bii = config["use_b2bii"]

    @abstractmethod
    def create(self, **kwargs) -> VariableCollection:
        """
        Creates the VariableCollection instance for later use.

        :return: An VariableCollection based on the config.
        """
        pass


class VariableListBase(ABC):
    """
    Base class for ObservableLists which can be defined in the
    analysis.
    """

    def __init__(
        self,
        use_mc_info: bool = False,
        use_b2bii: bool = False,
        roe_mask: str = None,
    ) -> None:
        """
        Constructor. The arguments are mainly based on flags used
        in the RDstar analysis. Maybe this will change in the future.

        :param use_mc_info: Whether to use MC info or not.
        :param use_b2bii: Whether to use b2bii conversion or not.
        :param roe_mask: Name of the desired ROE mask.
        """

        self._use_mc_info = use_mc_info
        self._use_b2bii = use_b2bii
        self._roe_mask = roe_mask

    @abstractmethod
    def get(self) -> List[str]:
        """
        Implements the variable list based on the flags defined as
        class attributes.

        :return: List of variables
        """
        pass


def daughter(variables: List[str], *daughter_indices: int) -> List[str]:
    """
    daughter(["M"], 1, 0, 2)
    Creates daughter(2, daughter(0, daughter(1, M)))
    In particular the order of indices is reversed!

    :param variables: List of variables to extracted from the specified child particle.
    :param daughter_indices: Daughter indices in reversed order.
    :return: List of modified variable strings.
    """
    if len(daughter_indices) == 0:
        return variables
    daughter_index, *daughter_indices = daughter_indices  # type: ignore
    assert isinstance(daughter_index, int), (daughter_index, type(daughter_index).__name__)
    return daughter([f"daughter({daughter_index}, {v})" for v in variables], *daughter_indices)


def mc_daughter(variables: List[str], *daughter_indices: int) -> List[str]:
    """
    mcDaughter(["M"], 1, 0, 2)
    Creates mcDaughter(2, mcDaughter(0, mcDaughter(1, M)))
    In particular the order of indices is reversed!

    :param variables: List of variables to extracted from the specified MC child particle.
    :param daughter_indices: MC daughter indices in reversed order.
    :return: List of modified variable strings.
    """
    if len(daughter_indices) == 0:
        return variables
    daughter_index, *daughter_indices = daughter_indices  # type: ignore
    assert isinstance(daughter_index, int), (daughter_index, type(daughter_index).__name__)
    return mc_daughter([f"mcDaughter({daughter_index}, {v})" for v in variables], *daughter_indices)


def makeROOTCompatible(variables: List[str]) -> List[str]:
    """
    Generates list of ROOT compatible variable names from
    list of basf2 variables by escaping specific characters.

    :param variables: List of basf2 variable names.
    :return: List of ROOT compatible variable names.
    """
    from ROOT import gSystem

    gSystem.Load("libanalysis.so")
    from ROOT import Belle2

    Belle2.Variable.Manager.Instance()
    return [Belle2.makeROOTCompatible(v) for v in variables]


def invert_make_ROOT_compatible(variables: List[str]) -> List[str]:
    """
    Invert escape for specific characters.

    :param variables: List of ROOT compatible variable names.
    :return: List of basf2 variable names.
    """
    from ROOT import gSystem

    gSystem.Load("libanalysis.so")
    from ROOT import Belle2

    Belle2.Variable.Manager.Instance()
    return [Belle2.invertMakeROOTCompatible(v) for v in variables]


def remove_mc_blacklisted_variables(variables: List[str]) -> List[str]:
    """
    Filters the mc blacklisted variables defined in this module out of a given
    list of basf2 variables to exclude variables which are ill-defined for use
    on MC particles.
    Variables which use blacklisted variables are also removed.

    :param variables: List of basf2 variables to be filtered.
    :return: Filtered list of variables.
    """

    def _contains_blacklisted_var(variable: str):
        separator_chars = [",", "=", "(", ")", "<", ">"]

        modified_var = variable
        for char in separator_chars:
            modified_var = modified_var.replace(char, " ")

        set_from_var = set(modified_var.split())
        return not set_from_var.isdisjoint(_mc_blacklisted_variables)

    return [var for var in variables if not _contains_blacklisted_var(var)]
