"""
Definition of handy functions to prepare the environment and the path for the analysis.

Felix Metzner 2018
"""

import os

from typing import Optional, Dict, Any
from basf2 import use_local_database, LogLevel, reset_database, use_database_chain

from analysistools.utilities import PathType


__all__ = [
    "setup_environment_vars",
    "setup_databases",
    "setup_database_for_fei",
    "fei_backward_compatibility",
]


def setup_environment_vars() -> None:
    """
    Setting up environment variables.
    :return: None
    """
    os.environ["BELLE_POSTGRES_SERVER"] = "bdb01"
    os.environ["USE_GRAND_REPROCESS_DATA"] = "1"
    os.environ["PGUSER"] = "g0db"
    os.environ["BELLE2_FILECATALOG"] = "NONE"


def setup_databases(config: Dict[str, Any]) -> None:
    """
    Setting up local databases for the FEI and - if b2bii is used - for the
    converted beam parameters to either be written to a personal local data base
    or to be not written anywhere.
    :param config: dictionary containing all configurations, like:
                    -use_b2bii: boolean indicating if B2BII is used.
                    -own_db_path: Path to own database directory.
                                  Defaults to None in which case no private local db is used.
                    -custom_fei_db_path: Provide alternative path to FEI data base location.
    :return: None
    """
    use_b2bii = config["use_b2bii"]
    own_db_path = config["own_local_mc_db_path"]
    custom_fei_db_path = config["custom_fei_db_path"]

    if not os.path.isdir(own_db_path):
        raise ValueError(
            f"The directory you provided for the private local data base does not exist!"
            f"You provided the path: {own_db_path}"
        )

    if use_b2bii:
        reset_database()
        use_database_chain()

    setup_database_for_fei(use_b2bii, custom_fei_db_path)

    if use_b2bii and own_db_path:
        use_local_database(
            os.path.join(own_db_path, "database.txt"),
            own_db_path,
            False,
            LogLevel.WARNING,
        )


def setup_database_for_fei(use_b2bii: bool, custom_fei_db_path: Optional[PathType] = None) -> None:
    """
    Setting up databases containing the FEI weight files.
    :param use_b2bii: boolean whether to use B2BII
    :param custom_fei_db_path: Provide alternative path to FEI data base location.
    :return: None
    """
    if custom_fei_db_path:
        use_local_database(
            os.path.join(custom_fei_db_path, "database.txt"),
            custom_fei_db_path,
            True,
            LogLevel.WARNING,
        )
        return

    # kekcc_2016_path_tkeck_path = "/home/belle2/tkeck/"  # type: str  # Old location before KEKcc upgrade.
    kekcc_2020_path_tkeck_path = "/home/belle2/metzner/tkeck_old_belle2_home/"  # type: str
    tkeck_path = kekcc_2020_path_tkeck_path  # type: str

    if use_b2bii:
        use_local_database(
            os.path.join(tkeck_path, "feiv4/Belle1_2017_convertedMC_Track14_2/localdb/database.txt"),
            os.path.join(tkeck_path, "feiv4/Belle1_2017_convertedMC_Track14_2/localdb/"),
            True,
            LogLevel.WARNING,
        )
    else:
        use_local_database(
            os.path.join(tkeck_path, "feiv4/Belle2_2017_MC7_Track14_2/localdb/database.txt"),
            os.path.join(tkeck_path, "feiv4/Belle2_2017_MC7_Track14_2/localdb/"),
            True,
            LogLevel.WARNING,
        )
        use_local_database(
            os.path.join(tkeck_path, "feiv4/Belle2_Generic_2017_MC7_WO_BB/localdb/database.txt"),
            os.path.join(tkeck_path, "feiv4/Belle2_Generic_2017_MC7_WO_BB/localdb/"),
            True,
            LogLevel.WARNING,
        )


def fei_backward_compatibility() -> None:
    from fei import backward_compatibility_layer

    backward_compatibility_layer.pid_renaming_oktober_2017()
