"""
Definition of convenience functions for the decay channel reconstruction.
Felix Metzner 2018
"""

import fei
import basf2
import vertex as basf2_vertex_fits

from modularAnalysis import (
    appendROEMasks,
    applyCuts,
    buildRestOfEvent,
    buildContinuumSuppression,
    create_path,
    copyLists,
    cutAndCopyList,
    discardFromROEMasks,
    fillParticleList,
    matchMCTruth,
    reconstructDecay,
    variablesToExtraInfo,
    variablesToDaughterExtraInfo,
)
from typing import Tuple, List, Dict, Set, Any, Callable


from analysistools.processing import roe_masks_utils
from analysistools.processing import decay_channels_utils
from analysistools.processing.observables_utils import remove_mc_blacklisted_variables


#################
#  B-Tag / FEI  #
#################


def reconstruct_tag_B(
    path: basf2.Path,
    config: Dict[str, Any],
    b_tag_chl_dict: Dict[str, Any],
    decay_chl_dict: Dict[str, Dict[str, Any]],
) -> None:
    """
    Adding FEI to provided path.
    :param path: the basf2 path to append to.
    :param config: dictionary containing all configurations, like:
                -use_b2bii: boolean indicating if B2BII is used.
                -all FEI related parameters
                -do_mc_matching: Whether to perform MC matching for B_tag candidates.
    :param b_tag_chl_dict: Decay channel dict containing the analysis specific B_Tag channels.
    :param decay_chl_dict: Nested dictionary containing the information about all analysis specific
                           decay channels as described in decay_channels_utils.py.
    :return: None
    """

    particles = fei.get_default_channels(
        convertedFromBelle=config["use_b2bii"],
        hadronic=config["FEI_options"]["hadronic"],
        semileptonic=config["FEI_options"]["semileptonic"],
        specific=config["FEI_options"]["specific"],
    )
    fei_config = fei.config.FeiConfiguration(
        prefix=config["FEI_options"]["FEI_tag"],
        b2bii=config["use_b2bii"],
        training=False,
        monitor=False,
    )

    fei_state = fei.get_path(particles, fei_config)
    path.add_path(fei_state.path)

    for particle in b_tag_chl_dict:
        cutAndCopyList(
            outputListName=particle,
            inputListName=decay_chl_dict["decay_label_to_chain"][particle],
            cut=decay_channels_utils.cut_list_to_string(decay_chl_dict["decay_label_to_cut_list"][particle]),
            writeOut=True,
            path=path,
        )

        if config["use_mc_info"]:
            path.add_module("MCMatcherParticles", listName=particle, looseMCMatching=True)

        apply_extra_reco_options(
            path=path,
            decay_label=particle,
            decay_chl_dict=decay_chl_dict,
            use_b2bii=config["use_b2bii"],
            gen_level_reco=False,
        )


########################
#    Rest of Event     #
########################


def build_roes(
    path: basf2.Path,
    particle_lists,
    mask_list: List[Callable],
    use_b2bii: bool,
    mask_name: str = None,
) -> None:
    """
    Building Rest Of Events for all particles in the provided particle list
    and either for all ROE masks defined in the provided analysis specific
    ROE masks list "mask_list" or for the one with the given name.
    The respective modules will be appended to the given basf2 path.
    :param path: basf2 path to which the modules will be appended.
    :param particle_lists: The particle lists to operate on.
    :param mask_list: List of all roe masks defined for this analysis.
    :param use_b2bii: If b2bii converted data is used.
    :param mask_name: The roe_mask definition to be used.
    :return: None
    """
    if mask_name:
        masks = roe_masks_utils.get_roe_mask(mask_list=mask_list, mask_name=mask_name, use_b2bii=use_b2bii)
    else:
        masks = roe_masks_utils.get_roe_masks(mask_list=mask_list, use_b2bii=use_b2bii)
    for particle in particle_lists:
        buildRestOfEvent(particle, path=path)
        appendROEMasks(particle, masks.values(), path=path)


#########################
# Continuum Suppression #
#########################


def build_cs(
    path: basf2.Path,
    particle_lists: List[str],
    roe_mask=None,
) -> None:
    """
    Builds Continuum Suppression Variables for all provided particle lists, appending the respective modules
    to the given basf2 path.
    :param path: basf2 path.
    :param particle_lists: list of particle lists to be considered.
    :param roe_mask: ROE mask to be used for the calculation of the continuum suppression variables. Default is None.
    :return: None
    """
    if roe_mask is None:
        roe_mask = ""

    for particle_list in particle_lists:
        buildContinuumSuppression(particle_list, roe_mask=roe_mask, path=path)


###############
# Vertex Fits #
###############


def do_vertex_fit(
    path: basf2.Path,
    particle_list: str,
    vertex_fit_function: str,
    **kwargs,
) -> None:
    """
    Convenience function to apply one of the vertex fit methods defined
    in the basf2 analysis package by providing the respective function name.
    :param path: basf2 path
    :param particle_list: basf2 particle list for which the fit shall be applied
    :param vertex_fit_function: function name of the vertex fit
    :return: None
    """
    conf_level = kwargs.pop("conf_level", -1)

    variables_changed_by_fit = ("E", "p", "px", "py", "pz")  # type: Tuple[str, ...]
    var_dict = {var: f"{var}_prefit" for var in variables_changed_by_fit}  # type: Dict[str, str]
    _var_dict_cms = {f"useCMSFrame({v})": f"{v}_cms_prefit" for v in variables_changed_by_fit}  # type: Dict[str, str]
    var_dict.update(_var_dict_cms)
    assert len(var_dict) == 2 * len(variables_changed_by_fit), var_dict
    assert len(set(var_dict.values())) == 2 * len(variables_changed_by_fit), var_dict

    variablesToExtraInfo(
        particleList=particle_list,
        variables=var_dict,
        option=0,
        path=path,
    )

    if vertex_fit_function != "my_dummy_vertex_fit":
        vertex_fitter = getattr(basf2_vertex_fits, vertex_fit_function)

        vertex_fitter(
            path=path,
            list_name=particle_list,
            conf_level=conf_level,
            **kwargs,
        )


########################
# Decay Reconstruction #
########################


def _contains_var_in_list(
    cut_string: str,
    var_list: List[str],
) -> bool:
    """
    Helper function used in reconstruct_decay_channel
    """
    separator_chars = [",", "=", "(", ")", "<", ">"]  # type: List[str]

    modified_var = cut_string
    for char in separator_chars:
        modified_var = modified_var.replace(char, " ")

    set_from_var = set(modified_var.split())
    return not set_from_var.isdisjoint(set(var_list))


def reconstruct_decay_channel(
    path: basf2.Path,
    decay_label: str,
    decay_chl_dict: Dict[str, Dict[str, Any]],
    use_mc_info: bool,
    use_b2bii: bool,
    do_gen_reco: bool,
    add_neutrinos: bool = False,
):
    """
    Reconstructs a decay channel and applies matchMCTruth on the generated particle list.
    In parallel the decays are also reconstructed from MC information. For this generator
    level reconstruction cut strings containing blacklisted basf2 variables as defined in
    mc_blacklisted_variables in the observables_utils module are removed from the given cuts
    and thus not applied. Furthermore only the correctly recombined particles are taken by
    adding a cut on the mc matching status.
    :param path: basf2 path to which the modules will be added
    :param decay_label: Decay label as described in decay_channels_utils.py and defined in
                        analysis specific decay_channels.py.
    :param decay_chl_dict: Nested dictionary containing the information about all analysis specific
                           decay channels as described in decay_channels_utils.py.
    :param use_mc_info: Switches on mc matching and reconstruction from generator level
                        information which yields lists with the suffix "_gen".
    :param use_b2bii: If b2bii converted data is used.
    :param do_gen_reco: Whether to run the reconstruction on generator level, too.
    :param add_neutrinos: For decays with one lepton and a neutrino format string, the corresponding
                          neutrino is added to the decay string if this option is activated.
    :return: None
    """
    second_step_vars = ["chiProb", "R2"]
    cuts_list = decay_chl_dict["decay_label_to_cut_list"][decay_label]
    reduced_cuts_list_1 = [v for v in cuts_list if not _contains_var_in_list(v, second_step_vars) and "ROE" not in v]
    reduced_cuts_list_2 = [v for v in cuts_list if v not in reduced_cuts_list_1]

    if decay_label not in decay_chl_dict["generator_level_only_decay_channels"]:
        reconstructDecay(
            decay_chl_dict["decay_label_to_chain"][decay_label].format(suf="", neutrino=""),
            cut=decay_channels_utils.cut_list_to_string(reduced_cuts_list_1),
            dmID=decay_chl_dict["decay_label_to_decay_id"][decay_label],
            path=path,
        )

        # Match MC truth
        if use_mc_info:
            matchMCTruth(decay_label, path=path)

        apply_extra_reco_options(
            path=path,
            decay_label=decay_label,
            decay_chl_dict=decay_chl_dict,
            use_b2bii=use_b2bii,
        )

        # Applying cuts again after cs, roe and vertex fit is done:
        if len(reduced_cuts_list_2) > 0:
            applyCuts(
                list_name=decay_label,
                cut=decay_channels_utils.cut_list_to_string(reduced_cuts_list_2),
                path=path,
            )

    if use_mc_info and do_gen_reco:
        # Reconstruct decay from generator level information:
        neutrino = ""
        if add_neutrinos:
            neutrino = get_suitable_neutrino(decay_chl_dict["decay_label_to_chain"][decay_label])
            if neutrino:
                neutrino = neutrino + "_gen"
        reconstructDecay(
            decay_chl_dict["decay_label_to_chain"][decay_label].format(suf="_gen", neutrino=neutrino),
            cut=decay_channels_utils.cut_list_to_string(remove_mc_blacklisted_variables(reduced_cuts_list_1)),
            dmID=decay_chl_dict["decay_label_to_decay_id"][decay_label],
            path=path,
        )

        matchMCTruth(f"{decay_label}_gen", path=path)

        # Ensure correctness of reconstruction from generator level particle lists:
        if decay_label in decay_chl_dict["decay_channels_with_neutrinos"] and not neutrino:
            truth_match_cut = "isSignalAcceptMissingNeutrino==1"
        else:
            truth_match_cut = "isSignal==1"

        # Apply MC truth cut before further reconstruction tasks are attempted.
        applyCuts(list_name=f"{decay_label}_gen", cut=truth_match_cut, path=path)

        apply_extra_reco_options(
            path=path,
            decay_label=decay_label,
            decay_chl_dict=decay_chl_dict,
            use_b2bii=use_b2bii,
            gen_level_reco=True,
        )

        # Applying cuts again after cs, roe and vertex fit is done:
        if len(reduced_cuts_list_2) > 0:
            applyCuts(
                list_name=f"{decay_label}_gen",
                cut=decay_channels_utils.cut_list_to_string(remove_mc_blacklisted_variables(reduced_cuts_list_2)),
                path=path,
            )


def apply_extra_reco_options(
    path: basf2.Path,
    decay_label: str,
    decay_chl_dict: Dict[str, Dict[str, Any]],
    use_b2bii: bool,
    gen_level_reco=False,
):
    """
    Applies based on the settings made in the decay_label_to_extra dictionary
    for the decay channel at hand the following tasks
        - ROE creation
        - Continuum Suppression variables calculation
        - Vertex Fit

    :param path: basf2 path to which the modules will be added
    :param decay_label: Decay label as described in decay_channels_utils.py and defined in
                        analysis specific decay_channels.py.
    :param decay_chl_dict: Nested dictionary containing the information about all analysis specific
                           decay channels as described in decay_channels_utils.py.
     :param use_b2bii: If b2bii converted data is used.
     :param gen_level_reco: Whether this is function is applied in the generator level reconstruction case.
    :return: None
    """
    particle_list = decay_label
    if gen_level_reco:
        particle_list = f"{decay_label}_gen"

    if "roe" in decay_chl_dict["decay_label_to_extra"][decay_label].keys():
        build_roes(
            path, [particle_list], use_b2bii=use_b2bii, **decay_chl_dict["decay_label_to_extra"][decay_label]["roe"]
        )

    if "cs" in decay_chl_dict["decay_label_to_extra"][decay_label].keys():
        build_cs(path, [particle_list], **decay_chl_dict["decay_label_to_extra"][decay_label]["cs"])

    if "vertex_fit" in decay_chl_dict["decay_label_to_extra"][decay_label].keys() and not gen_level_reco:
        do_vertex_fit(
            path=path,
            particle_list=particle_list,
            vertex_fit_function=decay_chl_dict["decay_label_to_extra"][decay_label]["vertex_fit"]["function"],
            **decay_chl_dict["decay_label_to_extra"][decay_label]["vertex_fit"]["kwargs"],
        )


def get_suitable_neutrino(decay_chain: str) -> str:
    """
    Convenience function to select correct neutrino type for lepton in decay string.
    Only works properly if only one lepton is in the decay string, thus only in this
    case a neutrino particle string is returned.
    :param decay_chain: basf2 decay string.
    :return: neutrino particle string for the decay or empty string
    """
    leptons = [" e+:", " e-:", " mu+:", " mu-:"]
    neutrinos = ["nu_e:mc", "anti-nu_e:mc", "nu_mu:mc", "anti-nu_mu:mc"]

    # Return empty string if none or multiple leptons in the decay string
    if not len(list(lepton_string for lepton_string in leptons if lepton_string in decay_chain)) == 1:
        return ""

    for lepton, neutrino in zip(leptons, neutrinos):
        if lepton in decay_chain:
            return neutrino

    return ""


def merge_particle_lists(
    path: basf2.Path,
    merge_channel_dict: Dict[str, str],
    decay_chl_dict: Dict[str, Dict[str, str]],
    use_mc_info: bool,
    do_gen_reco: bool,
    label: str,
    write_out: bool = True,
) -> Tuple[List[str], List[str]]:
    """
    Convenience function to merge related particle lists.
    Related particle lists are given via their common merge_channel_dict.
    If the merge_channel_dict comprises decays of different particles, separate particle
    lists will be generated.
    An additional label must be provided for the final particle list(s).
    Returns a list of the newly generated particle lists.
    :param path: The basf2 path object to which the modules will be appended.
    :param merge_channel_dict: The decay channel dictionary containing the decays to be merged.
    :param decay_chl_dict: Nested dictionary containing the information about all analysis specific
                           decay channels as described in decay_channels_utils.py.
    :param use_mc_info: if true additional generator level list are also merged
    :param do_gen_reco: Whether to run the reconstruction on generator level, too.
    :param label: A additional label for the new particle list as in particle:label.
    :param write_out: Whether to write out the new particle list.
    :return: tuple of list of new basf2 particle lists and list of new generator level particle lists.
    """
    particles = set([p.split(":")[0] for p in merge_channel_dict])  # type: Set[str]
    decays_reco_only = [p for p in merge_channel_dict if p not in decay_chl_dict["generator_level_only_decay_channels"]]
    suffixes = [""]  # type: List[str]
    decays = [decays_reco_only]  # type: List[List[str]]
    new_particle_lists = []  # type: List[str]
    new_gen_particle_lists = []  # type: List[str]
    new_lists = [new_particle_lists]  # type: List[List[str]]

    if use_mc_info and do_gen_reco:
        suffixes.append("_gen")
        decays.append(list(merge_channel_dict.keys()))
        new_lists.append(new_gen_particle_lists)

    for suffix, channels, new_list in zip(suffixes, decays, new_lists):
        for particle in particles:
            output_list_name = f"{particle}:{label}{suffix}"
            copyLists(
                outputListName=output_list_name,
                inputListNames=[f"{ch}{suffix}" for ch in channels if ch.startswith(particle)],
                writeOut=write_out,
                path=path,
            )
            new_list.append(output_list_name)

    return new_particle_lists, new_gen_particle_lists


def mark_curlers_belle(
    path: basf2.Path,
    label: str,
) -> None:
    """
    Marks potential curling tracks via a selection for a pion particle list
    which will carry the provided label as tag: pi+:<label>.

    The selection of the curling tracks is performed by recombining an arbitrary particle,
    here a D0 meson, by combining pi+ and pi+ or pi+ and pi-.
    The pions must be in the current ROE.
    The additional selection criteria on the transversal momentum pt, the difference in pt
    between the two tracks and the angle between the two tracks (which depends on their charge)
    are chosen to select two tracks which are clones of each other, originating from a curler.
    The one with the larger impact parameters is then marked as the clone that shall be removed.

    The track clone is marked by remaining in the created pion particle list pi+:<label>
    and can be removed from the ROE via the modular analysis function discardFromROEMasks.

    :param path: BASF2 path to which this functions shall be appended.
    :param label: Label to be used for particle lists generated by this function.
    :return: None
    """
    fillParticleList(f"pi+:{label}", "pt < 0.275 and isInRestOfEvent == 1", path=path)

    reconstructDecay(f"D0:{label}OS -> pi+:{label} pi-:{label}", "", 999901, path=path)
    reconstructDecay(f"D0:{label}SS -> pi+:{label} pi+:{label}", "", 999902, path=path)

    applyCuts(f"D0:{label}OS", "abs(formula(daughter(0,pt) - daughter(1,pt))) < 0.10", path=path)
    applyCuts(f"D0:{label}SS", "abs(formula(daughter(0,pt) - daughter(1,pt))) < 0.10", path=path)

    v_in = "belleTrackVar"
    variablesToDaughterExtraInfo(
        f"D0:{label}OS", "D0 -> ^pi+ pi-", {"daughter(0,formula(25*d0*d0+z0*z0))": v_in}, 1, path=path
    )
    variablesToDaughterExtraInfo(
        f"D0:{label}OS", "D0 -> pi+ ^pi-", {"daughter(1,formula(25*d0*d0+z0*z0))": v_in}, 1, path=path
    )
    variablesToDaughterExtraInfo(
        f"D0:{label}SS", "D0 -> ^pi+ pi+", {"daughter(0,formula(25*d0*d0+z0*z0))": v_in}, 1, path=path
    )
    variablesToDaughterExtraInfo(
        f"D0:{label}SS", "D0 -> pi+ ^pi+", {"daughter(1,formula(25*d0*d0+z0*z0))": v_in}, 1, path=path
    )

    applyCuts(f"D0:{label}OS", "daughterAngle(0,1) < -0.965926", path=path)
    applyCuts(f"D0:{label}SS", "daughterAngle(0,1) > 0.965926", path=path)

    v_i = f"extraInfo({v_in})"
    v_out = "belleTrackBool"
    overwrite_opt = 1
    variablesToDaughterExtraInfo(
        f"D0:{label}OS",
        "D0 -> ^pi+ pi-",
        {f"formula(daughter(0,{v_i}) - daughter(1,{v_i}))": v_out},
        overwrite_opt,
        path=path,
    )
    variablesToDaughterExtraInfo(
        f"D0:{label}OS",
        "D0 -> pi+ ^pi-",
        {f"formula(daughter(1,{v_i}) - daughter(0,{v_i}))": v_out},
        overwrite_opt,
        path=path,
    )
    variablesToDaughterExtraInfo(
        f"D0:{label}SS",
        "D0 -> ^pi+ pi+",
        {f"formula(daughter(0,{v_i}) - daughter(1,{v_i}))": v_out},
        overwrite_opt,
        path=path,
    )
    variablesToDaughterExtraInfo(
        f"D0:{label}SS",
        "D0 -> pi+ ^pi+",
        {f"formula(daughter(1,{v_i}) - daughter(0,{v_i}))": v_out},
        overwrite_opt,
        path=path,
    )

    applyCuts(f"pi+:{label}", "extraInfo(belleTrackBool) > 0", path=path)


def discard_curlers(
    path: basf2.Path,
    roe_mask: str,
    label: str = "curlers",
) -> None:
    """
    This function is a starting point for a simple version of a curler finder that operates
    on a ROE. This means that only tracks in the ROE are considered and compared.
    Thus, this method is only helpful for inclusively tagged analyses.

    The remaining tracks are not taken into account. Thus, this function is only sufficient,
    if e.g. a signal selection requires all tracks to be highly energetic. In such a case
    these tracks can not be curlers, and only the tracks in the ROE must be checked.

    For those cases, this function will mark potential curling tracks by calling
    mark_curlers_belle and remove the marked tracks from the ROE.
    As this function requires the signal side from which the ROE was created also to be
    free of curlers by design, one can remove all curlers with this function, if the
    requirement is fulfilled.

    :param path: BASF2 path to which this functions shall be appended.
    :param roe_mask: The roe_mask used for the rest of event that shall be checked.
    :param label: Label to be attached to particle lists generated with this function.
    :return: None
    """
    roe_curler_rejection_path = create_path()

    mark_curlers_belle(path=roe_curler_rejection_path, label=label)
    discardFromROEMasks(list_name=f"pi+:{label}", mask_names=[roe_mask], cut_string="", path=roe_curler_rejection_path)

    path.for_each("RestOfEvent", "RestOfEvents", roe_curler_rejection_path)
