"""
Utility functions to handle Rest Of Event (ROE) Masks which are used to clean the set of
tracks and ECL clusters which are considered for the ROE.

The ROE mask themselves must be defined in the analysis specific modules as list of dictionaries.
A ROE mask dictionary must have the following form:

    exampleROEMask = {
        "name": "exampleROEMask",
        "track_cuts": "dr < 0.4 and abs(dz) < 0.4 and nCDCHits > 0" if env.use_b2bii
        else "dr < 0.4 and abs(dz) < 0.4 and nCDCHits > 0",
        "cluster_cuts": : "goodBelleGamma == 1" if env.use_b2bii else "goodGamma == 1",
    }

The dictionary contains the keys "name", "track_cuts" and "cluster_cuts" with strings containing the respective
information as values.

All ROE mask must be collected in a list which is passed to the respective functions:

    roe_mask_list = [exampleROEMask]


Felix Metzner 2018
"""

from typing import Optional, Tuple, List, Dict, Callable


__all__ = [
    "ROEMaskInfoType",
    "get_roe_mask",
    "get_roe_masks",
    "print_roe_masks",
]


ROEMaskInfoType = Tuple[str, str, str]


def get_roe_masks(mask_list: List[Callable], use_b2bii: bool) -> Dict[str, ROEMaskInfoType]:
    """
    Get dictionary with all defined ROE Masks.
    :param mask_list: List containing all analysis specific ROE mask dictionaries.
    :param use_b2bii: If b2bii converted data is used.
    :return: dict with key "name", and tuple of "name", "track_cuts" and "cluster_cuts" as value.
    """
    roe_masks = {}
    for roe_mask in mask_list:
        mask = roe_mask(use_b2bii)
        roe_masks[mask["name"]] = (mask["name"], mask["track_cuts"], mask["cluster_cuts"])
    return roe_masks


def get_roe_mask(mask_list, mask_name: str, use_b2bii: bool) -> Dict[str, ROEMaskInfoType]:
    """
    Returns (mask_name, track_cuts, cluster_cuts) for given ROE mask name.
    :param mask_list: List containing all analysis specific ROE mask dictionaries.
    :param mask_name: Name of ROE mask as defined in roe_masks module.
    :param use_b2bii: If b2bii converted data is used.
    :return: dict with triplet of strings: {mask_name: (mask_name, track_cuts, cluster_cuts)}.
    """
    for roe_mask in mask_list:
        mask = roe_mask(use_b2bii)
        if mask["name"] == mask_name:
            return {mask["name"]: (mask["name"], mask["track_cuts"], mask["cluster_cuts"])}
    raise ValueError(
        f"The specified ROE mask name {mask_name} is not valid!\n"
        f"Must be one of analysis specific roe mask list: {mask_list}!"
    )


def print_roe_masks(
    mask_list: List[Callable],
    use_b2bii: bool,
    selected_mask: Optional[str] = None,
) -> None:
    """
    Function to print the names and cuts for all ROE Masks as defined in the
    roe_masks module.
    :param mask_list: List containing all analysis specific ROE mask dictionaries.
    :param use_b2bii: If b2bii converted data is used.
    :param selected_mask: Selected mask name to be used for the analysis.
                          Will be highlighted. Can be "all". Default is None.
    :return: None
    """
    column_headers = ["ROE Mask Name", "Track Cuts", "Cluster Cuts"]
    table_as_list = [column_headers]

    for roe_mask in mask_list:
        mask = roe_mask(use_b2bii)
        mask_name = mask["name"]
        if mask_name == selected_mask or selected_mask == "all":
            mask_name = "USED => {n}".format(n=mask_name)
        table_as_list.append([mask_name, mask["track_cuts"], mask["cluster_cuts"]])

    column_size = [max(map(len, col)) for col in zip(*table_as_list)]

    format_str = " | ".join("{{:<{i}}}".format(i=i) for i in column_size)
    table_as_list.insert(1, ["-" * i for i in column_size])

    for item in table_as_list:
        print(format_str.format(*item))
