"""
Utility functions for skim_cuts dictionaries.

The skim cut dictionaries themselves must be defined specifically for each analysis
following the guidelines stated below.

For each skim a dictionary is defined, which should have the following form

    pre_skim_cuts = {
        "description": "Event Level Skim Cuts.",
        "skim_name": "pre_event_skim",
        "target_lists": [""],
        "cuts": [
            ("nCleanedTracks(dr < 2 and abs(dz) < 4)", ">20"),
            ("countInList(gamma:fsp)", ">=20")
        ],
        "cands_in_lists": None,
    }

and contains:

    description: A short description of the skim.
    skim_name: String to be used for output file name.
    target_lists: List of names of the particle lists on which the
                  skim is performed. A empty string as target list
                  means that the skim is performed based on the whole
                  event (target_list = [""]).
    cuts: List of tuples of (variable, cut) which will be applied
          in the respective order!
    cands_in_lists: Cut criterion which will be used to reject every
                    event based on the number of events in the provided
                    target particle lists target_lists. E.g. with
                    target_lists = ["B+:sig"] and cands_in_lists = "<1")
                    every event which does not have at least one candidate
                    in the B+:sig particle list will be rejected.

The example given above can be used to skim by rejecting all events which have
more than 20 clean charged tracks and more than 20 candidates in the gamma:fsp
particle list.

It is important to note, that the final Y(4S) particle is always
reconstructed in such a way, that the first daughter (index 0) is
the tag B meson, and the second (index 1) is the signal B meson.

Felix Metzner 2018
"""

import os
import basf2

from typing import Optional, Dict, Sequence, Any
from modularAnalysis import variablesToNtuple, create_path, AfterConditionPath

from analysistools.utilities import PathType


__all__ = [
    "apply_skim_from_dict",
    "get_sampling_skim_cuts_dict",
    "apply_sampling_skim",
    "print_skim_cuts_dict",
    "print_skim_cuts",
]


def apply_skim_from_dict(
    path: basf2.Path,
    skim_dict: Dict[str, Any],
    write_out: bool = False,
    target_directory: Optional[PathType] = None,
) -> None:
    """
    Adds skim modules according to the settings provided in form of
    the skim_dict dictionary to the given basf2 path.
    If write_out parameter is set to True root files with the event counts
    and the number of candidates in the particle lists on which the skim
    cuts are performed, both before and after the cuts, are stored.
    :param path: basf2 path the modules are added to.
    :param skim_dict: Dictionary containing the settings to be used for the skim.
    :param write_out: Whether to write out counts into root file.
    :param target_directory: If write_out is true, a target directory for the output has to be provided.
    :return: None
    """
    if write_out and target_directory is None:
        raise ValueError("If skim information shall be written to file, a target directory must be provided!")

    vars_to_write = [var for var, _ in skim_dict["cuts"]]
    if skim_dict["cands_in_lists"]:
        vars_to_write.extend([f"countInList({p})" for p in skim_dict["target_lists"]])
    if write_out:
        assert target_directory is not None
        variablesToNtuple(
            decayString="",
            variables=vars_to_write,
            filename=os.path.join(target_directory, "events_before_{name}.root".format(name=skim_dict["skim_name"])),
            path=path,
        )

    empty_path = create_path()

    if skim_dict["cands_in_lists"]:
        counts_to_be_summed = [f"countInList({p_list})" for p_list in skim_dict["target_lists"]]
        sum_of_counts = "formula(" + " + ".join(counts_to_be_summed) + ")"
        skim_filter = basf2.register_module("VariableToReturnValue")
        skim_filter.param("variable", sum_of_counts)
        skim_filter.if_value(skim_dict["cands_in_lists"], empty_path, AfterConditionPath.END)
        path.add_module(skim_filter)

    for var, cut in skim_dict["cuts"]:
        skim_filter = basf2.register_module("VariableToReturnValue")
        skim_filter.param("variable", var)
        skim_filter.if_value(cut, empty_path, AfterConditionPath.END)
        path.add_module(skim_filter)

    if write_out:
        assert target_directory is not None
        variablesToNtuple(
            decayString="",
            variables=vars_to_write,
            filename=os.path.join(target_directory, "events_after_{name}.root".format(name=skim_dict["skim_name"])),
            path=path,
        )


def get_sampling_skim_cuts_dict(sampling_rate: Optional[float]) -> Optional[Dict[str, Any]]:
    if sampling_rate is None or sampling_rate == 1.0:
        return None
    if not 0.0 <= sampling_rate < 1.0:
        raise ValueError(f"Sampling rate must be float between 0.0 and 1.0 or None, but {sampling_rate} was provided!")

    print(f"Sampling Skim is active: Using sampling rate of {sampling_rate} for skim!")

    return {
        "description": "Sampling Skim --- Skims events based on given sampling rate and eventRandom variable.",
        "skim_name": f"sampling_event_skim_{sampling_rate:.5f}_rate",
        "target_lists": [""],
        "cuts": [
            (f"passesEventCut(eventRandom <= {sampling_rate})", "<1"),
        ],
        "cands_in_lists": None,
    }


def apply_sampling_skim(
    path: basf2.Path,
    sampling_rate: Optional[float],
) -> None:
    sampling_skim_dict = get_sampling_skim_cuts_dict(sampling_rate=sampling_rate)

    if sampling_skim_dict is not None:
        apply_skim_from_dict(
            path=path,
            skim_dict=sampling_skim_dict,
        )


def print_skim_cuts_dict(skim_dict: Dict[str, Any]) -> None:
    """
    Prints overview for the provided skim_cut dictionary.
    :param skim_dict:
    :return:
    """
    lists = str(skim_dict["target_lists"])
    table_1col_as_list = [
        skim_dict["description"] + " ({n})".format(n=skim_dict["skim_name"]),
        "target_lists: {lists}".format(lists=lists),
    ]
    column1_size = max([len(col) for col in table_1col_as_list])

    table_2col_as_list = [["Variable", "Reject event if"]]
    if skim_dict["cands_in_lists"]:
        table_2col_as_list.append(
            ["SkimFilter: Number of Candidates in lists  {lists}".format(lists=lists), str(skim_dict["cands_in_lists"])]
        )

    if len(skim_dict["cuts"]) > 0:
        for feature, cut in skim_dict["cuts"]:
            table_2col_as_list.append([str(feature), str(cut)])

    column2_size = [max(map(len, col)) for col in zip(*table_2col_as_list)]

    if sum(column2_size) + 3 < column1_size:
        diff = column1_size - 3 - sum(column2_size)
        column2_size = [val + diff + 1 for val in column2_size]
    else:
        column1_size = sum(column2_size) + 3

    table_1col_as_list.insert(1, "-" * column1_size)
    table_1col_as_list.append("-" * column1_size)
    for row in table_1col_as_list:
        print(row)

    table_2col_as_list.insert(1, ["-" * i for i in column2_size])
    if skim_dict["cands_in_lists"] and len(skim_dict["cuts"]) > 0:
        table_2col_as_list.insert(3, ["-" * i for i in column2_size])

    format_str = " | ".join("{{:<{i}}}".format(i=i) for i in column2_size)

    for row in table_2col_as_list:
        print(format_str.format(*row))


def print_skim_cuts(skim_dict_list: Sequence[Dict[str, Any]]) -> None:
    """
    Prints overview for all provided skim_cuts dictionaries.
    :param skim_dict_list: List of skim_dicts to print.
    :return: None
    """
    for skim_dict in skim_dict_list:
        print_skim_cuts_dict(skim_dict=skim_dict)
        print("\n")
