import json
import logging
import itertools
import numpy as np
import pandas as pd

from scipy.interpolate import Rbf
from typing import Optional, Tuple, List

from analysistools.utilities import PathType

logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = [
    "RbfReweighter",
]


class RbfReweighter:
    """
    RbfReweighter implements a method to calculate weights used
    to reweight monte carlo based on D dimensional histograms.

    Both, for the original and target monte carlo type, D dimensional
    histograms are calculated using the given variables, number of bins
    and limits for the bin edges.
    Bin-wise weights are determined by dividing the distribution of the
    target MC by the original MC. The weights for bins, where the
    denominator is zero are set to zero.
    Each bin weight is assigned to some coordinates corresponding to a
    D dimensional vector pointing to the bin mid. The resulting
    D dimensional weight function is given by an interpolating between
    these bins using the radial basis function method implemented in
    scipy. To ensure normalization, a scale factor is calculated on the
    given MC which ensure the normalization in the number of events
    after reweighting. It might happen that the weights take negative
    values due to the interpolation process. These weights are set to
    zero.

    :param variables: List of variable names. Must match the column names
    in the DataFrames later on.
    :param num_bins: List of integers denoting the number of bins in the
    histogram axis for each variable.
    :param limits: List of Tuple of floats denoting the minimum and maximum
    values the bin edges for each histogram axis.
    """

    def __init__(
        self,
        variables: Tuple[str, ...],
        num_bins: Tuple[int, ...],
        limits: Tuple[Tuple[float, float], ...],
    ) -> None:
        self._variables = variables
        self._num_bins = num_bins
        self._limits = limits

        self._bin_edges = [
            np.linspace(*limit, nbins + 1) for limit, nbins in zip(self._limits, self._num_bins)
        ]  # type: List[np.ndarray]
        self._bin_mids = list(map(lambda x: (x[:-1] + x[1:]) / 2, self._bin_edges))  # type: List[np.ndarray]
        self._weight_coords = np.array(list(itertools.product(*self._bin_mids)))  # type: np.ndarray

        self._bc_origin = None  # type: Optional[np.ndarray]
        self._bc_target = None  # type: Optional[np.ndarray]
        self._hist_weights = None  # type: Optional[np.ndarray]

        self._num_evts_origin = 0  # type: int
        self._num_evts_target = 0  # type: int

        self._scale_factor = None  # type: Optional[float]
        self._rbf = None  # type: Optional[Rbf]

    @property
    def variables(self) -> Tuple[str, ...]:
        return self._variables

    @property
    def bc_origin(self) -> np.ndarray:
        assert self._bc_origin is not None
        return self._bc_origin

    @property
    def bc_target(self) -> np.ndarray:
        assert self._bc_target is not None
        return self._bc_target

    @property
    def hist_weights(self) -> np.ndarray:
        assert self._hist_weights is not None
        return self._hist_weights

    @property
    def scale_factor(self) -> float:
        assert self._scale_factor is not None
        return self._scale_factor

    def create_interpolation(self, origin: pd.DataFrame, target: pd.DataFrame) -> None:
        """Creates the RBF instance and calculates the normalization
        factor. Has to called once after the class is created for the
        first time. If the class is instantiated from json, this step
        can be omiteed

        :param origin: DataFrame with the original MC.
        :param target: DataFrame with the target MC.
        """
        assert all(var in origin.columns for var in self._variables), [
            v for v in self._variables if v not in origin.columns
        ]
        assert all(var in target.columns for var in self._variables), [
            v for v in self._variables if v not in target.columns
        ]

        self._num_evts_origin = len(origin.index)
        self._num_evts_target = len(target.index)

        self._bc_origin, _ = np.histogramdd(origin[list(self._variables)].values, bins=self._bin_edges, density=True)
        self._bc_target, _ = np.histogramdd(target[list(self._variables)].values, bins=self._bin_edges, density=True)

        self._hist_weights = np.divide(
            self.bc_target, self.bc_origin, out=np.zeros_like(self.bc_origin), where=self.bc_origin != 0.0
        ).flatten()

        self._create_rbf()
        self._get_scale_factor(origin)

    def _get_scale_factor(self, origin_sample: pd.DataFrame) -> None:
        logging.debug("Calculating scale factor...")

        # Original approach (will have memory issues starting at a certain number of bins):
        assert self._rbf is not None
        weights = self._rbf(*[origin_sample[variable].values for variable in self._variables])

        # Dask approach to try to fix Memory issue for large number of bins:
        # Add `import dask.array as da` above, if you want to use this.
        # weights = None
        # dask_chunks = 10
        # if len(self._variables) == 1:
        #     i_v0 = da.from_array(origin_sample[self._variables[0]].values, chunks=dask_chunks)
        #     i_weights = da.map_blocks(self._rbf, i_v0)
        #     weights = i_weights.compute()
        # elif len(self._variables) == 2:
        #     i_v0 = da.from_array(origin_sample[self._variables[0]].values, chunks=dask_chunks)
        #     i_v1 = da.from_array(origin_sample[self._variables[1]].values, chunks=dask_chunks)
        #     i_weights = da.map_blocks(self._rbf, i_v0, i_v1)
        #     weights = i_weights.compute()
        # elif len(self._variables) == 3:
        #     i_v0 = da.from_array(origin_sample[self._variables[0]].values, chunks=dask_chunks)
        #     i_v1 = da.from_array(origin_sample[self._variables[1]].values, chunks=dask_chunks)
        #     i_v2 = da.from_array(origin_sample[self._variables[2]].values, chunks=dask_chunks)
        #     i_weights = da.map_blocks(self._rbf, i_v0, i_v1, i_v2)
        #     weights = i_weights.compute()
        # elif len(self._variables) == 4:
        #     i_v0 = da.from_array(origin_sample[self._variables[0]].values, chunks=dask_chunks)
        #     i_v1 = da.from_array(origin_sample[self._variables[1]].values, chunks=dask_chunks)
        #     i_v2 = da.from_array(origin_sample[self._variables[2]].values, chunks=dask_chunks)
        #     i_v3 = da.from_array(origin_sample[self._variables[3]].values, chunks=dask_chunks)
        #     i_weights = da.map_blocks(self._rbf, i_v0, i_v1, i_v2, i_v3)
        #     weights = i_weights.compute()
        # elif len(self._variables) == 5:
        #     i_v0 = da.from_array(origin_sample[self._variables[0]].values, chunks=dask_chunks)
        #     i_v1 = da.from_array(origin_sample[self._variables[1]].values, chunks=dask_chunks)
        #     i_v2 = da.from_array(origin_sample[self._variables[2]].values, chunks=dask_chunks)
        #     i_v3 = da.from_array(origin_sample[self._variables[3]].values, chunks=dask_chunks)
        #     i_v4 = da.from_array(origin_sample[self._variables[4]].values, chunks=dask_chunks)
        #     i_weights = da.map_blocks(self._rbf, i_v0, i_v1, i_v2, i_v3, i_v4)
        #     weights = i_weights.compute()
        # else:
        #     raise RuntimeError(f"Reweighting with {len(self._variables)} is not supported.")

        assert weights is not None
        assert isinstance(weights, np.ndarray), type(weights)
        assert weights.shape == origin_sample[self._variables[0]].values.shape, (
            weights.shape,
            origin_sample[self._variables[0]].values.shape,
        )

        # Numpy nditer Approach to try to fix Memory issue for large number of bins:
        #       For details look at
        #       https://stackoverflow.com/questions/51651956/memory-error-by-using-rbf-with-scipy
        #       and
        #       https://numpy.org/doc/stable/reference/generated/numpy.hsplit.html
        # weights = np.zeros(shape=len(origin_sample.index), dtype=np.float64)
        # for arrays in np.nditer(
        #         [weights] + [origin_sample[v].values for v in self._variables],
        #         flags=['external_loop', 'buffered'],
        #         op_flags=['readwrite']
        # ):
        #     assert len(arrays) == 1 + len(self._variables), (len(arrays), len(self._variables))
        #     i_weights = arrays[0]
        #
        #     assert all(isinstance(a, np.ndarray) for a in arrays), [type(a) for a in arrays]
        #     assert all(a.shape == i_weights.shape for a in arrays), [a.shape for a in arrays]
        #
        #     i_weights = self._rbf(*arrays[1:])
        #
        #     assert len(i_weights.shape) == 1, i_weights.shape

        weights[weights < 0] = 0.0
        num_origin_events = len(origin_sample.index)
        summed_weights = np.sum(weights)

        logging.debug(f"Number of original events: {num_origin_events}")
        logging.debug(f"Number of reweighed events: {summed_weights}")

        self._scale_factor = num_origin_events / np.sum(weights)

        logging.debug(f"Scale factor: {self.scale_factor}")

    @staticmethod
    def euclidean_norm_numpy(x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        return np.linalg.norm(x1 - x2, axis=0)

    def _create_rbf(self) -> None:
        self._rbf = Rbf(
            *[self._weight_coords[:, i] for i in range(len(self._variables))],
            self.hist_weights,
            function="cubic",
            # function="quintic",
            # norm=RbfReweighter.euclidean_norm_numpy,
        )

    def export_to_json(self, output_path: PathType) -> None:
        """
        Exports the necessary information of this class to a json file

        :param output_path: Absolute path to the json file.
        """
        data = {
            "variables": self._variables,
            "num_bins": self._num_bins,
            "limits": self._limits,
            "num_events": {
                "origin": self._num_evts_origin,
                "target": self._num_evts_target,
            },
            "bin_edges": [bin_edges.tolist() for bin_edges in self._bin_edges],
            "bin_mids:": [bin_mids.tolist() for bin_mids in self._bin_mids],
            "bc_origin": self.bc_origin.tolist(),
            "bc_target": self.bc_target.tolist(),
            "hist_weights": self.hist_weights.tolist(),
            "scale_factor": self.scale_factor,
        }

        with open(output_path, "w") as write_file:
            json.dump(obj=data, fp=write_file, indent=4)

    @classmethod
    def import_from_json(cls, file_path: PathType) -> "RbfReweighter":
        """Creates a class instance from a given json file.

        :param file_path: Absolute path to the json file.
        """
        with open(file_path, "r") as read_file:
            data = json.load(read_file)

        instance = cls(data["variables"], data["num_bins"], data["limits"])
        instance._bc_origin = data["bc_origin"]
        instance._bc_target = data["bc_target"]
        instance._hist_weights = data["hist_weights"]
        instance._create_rbf()
        instance._scale_factor = data["scale_factor"]
        return instance

    def __call__(self, *args) -> np.ndarray:
        """Calculates new weights using radial basis function
        interpolation of the histogram weights.

        :param: numpy.ndarrays of shape (`nevents`,).
        :return: Numpy ndarray with the new weights. Shape is
        (`num_events`,).
        """
        assert self._rbf is not None
        weights = self._rbf(*args)
        weights[weights < 0] = 0.0
        return self.scale_factor * weights
