import json
import numpy as np
import pandas as pd

from multiprocessing import Pool
from scipy.interpolate import Rbf
from typing import Union, Optional, List, Dict, Tuple, Iterable

from analysistools.utilities import PathType


__all__ = [
    "Weights",
    "get_weights_from_rbf",
    "rbf_helper",
    "get_weights_from_hist",
]


class Weights:
    """
    This class calculates bin-by-bin weights to calculate weight in order to
    reweight events from an original data sample to a target data sample. The
    weights are calculated as ratio of d dimensional bin counts which can be used
    later to apply interpolation methods.
    """

    def __init__(
        self,
        variables: List[str],
    ) -> None:
        self._vars = variables  # type: List[str]
        self._bin_count_old = None  # type: Optional[np.ndarray]
        self._bin_count_new = None  # type: Optional[np.ndarray]
        self._bin_edges = None  # type: Optional[List[np.ndarray]]
        self._bin_centers = None  # type: Optional[List[np.ndarray]]
        self._weights = None  # type: Optional[np.ndarray]

    @property
    def bin_count_old(self) -> np.ndarray:
        assert self._bin_count_old is not None
        return self._bin_count_old

    @property
    def bin_count_new(self) -> np.ndarray:
        assert self._bin_count_new is not None
        return self._bin_count_new

    @property
    def bin_edges(self) -> List[np.ndarray]:
        assert self._bin_edges is not None
        return self._bin_edges

    @property
    def bin_centers(self) -> List[np.ndarray]:
        assert self._bin_centers is not None
        return self._bin_centers

    @property
    def weights(self) -> np.ndarray:
        assert self._weights is not None
        return self._weights

    def calculate_weights(
        self,
        target: pd.DataFrame,
        original: pd.DataFrame,
        bins: Union[List[int], np.ndarray],
    ) -> None:
        """
        Calculates the weights for the reweighting form the original to the
        target data sample based on the ratio of the bin counts of d dimensional
        histograms.
        :param target: The target sample used as numerator in the ratio for the weight calculation.
        :param original: The original data sample used as denominator in the ratio for the weight calculation
        :param bins: List of ints or np.ndarrays used as bin edges for the d dimensional histogram.
                     Has to be either all int or all np.ndarrays.
        """

        if any([not isinstance(b, int) for b in bins]) and any([not isinstance(b, np.ndarray) for b in bins]):
            raise ValueError("Element in argument bins have to be either all integer or all np.ndarray!")

        if len(self._vars) != len(bins):
            raise ValueError("Please provide the same number of bins as variables used.")

        bin_count_old, bin_edges = np.histogramdd(
            original[self._vars].values,
            bins=bins,
            normed=True,
        )
        bin_count_new, _ = np.histogramdd(
            target[self._vars].values,
            bins=bin_edges,
            normed=True,
        )

        self._bin_count_old = bin_count_old
        self._bin_count_new = bin_count_new
        self._bin_edges = list(bin_edges)
        self._bin_centers = [(edges[1:] + edges[:-1]) / 2 for edges in bin_edges]

        with np.errstate(divide="ignore"):
            self._weights = np.divide(
                self._bin_count_new,
                self._bin_count_old,
            )
            # these bins aren't used at later stages when we interpolate between
            # the bin weights
            self._weights[self._bin_count_old == 0] = 1

    def export_as_json(self, path_to_file: PathType) -> None:
        """
        Exports the most important information as json file. Can be read in at
        later stages when the weights are used for new events.
        :param path_to_file: Name of the output json file.
        """
        data = dict()
        data["as_array"] = self.weights.tolist()
        data["bin_edges"] = [edges.tolist() for edges in self.bin_edges]
        data["bin_centers"] = [centers.tolist() for centers in self.bin_centers]
        data["variables"] = self._vars

        with open(path_to_file, "w") as out_file:
            json.dump(data, fp=out_file, indent=2)

    def import_from_json(self, path_to_file: PathType) -> None:
        """
        Imports the most important information, like the bin-by-bin weights,
        bin edges...
        :param path_to_file: Name of the input json file.
        """
        with open(path_to_file, "r") as in_file:
            data = json.load(fp=in_file)

        self._weights = np.array(data["as_array"])
        self._bin_edges = [np.array(edges) for edges in data["bin_edges"]]
        self._bin_centers = [np.array(centers) for centers in data["bin_centers"]]
        self._vars = data["variables"]

    @property
    def flat_coords_and_weights(self) -> Tuple[List[np.ndarray], np.ndarray]:
        """
        Returns the coordinates and the bin weights as flattened arrays. This
        is useful for most interpolation methods provided by scipy.
        """
        coord_matrices = np.meshgrid(*self.bin_centers, indexing="ij")  # type: Iterable[np.ndarray]
        coords_flat = [coord_matrix.flatten() for coord_matrix in coord_matrices]
        weights_flat = self.weights.flatten()

        # remove weights from empty denominator bins
        # not_empty = (weights_flat != -999)
        # coords_flat = list(map(lambda x: x[not_empty], coords_flat))
        # weights_flat = weights_flat[not_empty]

        return coords_flat, weights_flat

    @property
    def as_array(self) -> np.ndarray:
        """
        Returns the bin-bz-bin weights as d dimensional numpy array.
        """
        return self._weights

    @property
    def variables(self) -> List[str]:
        """
        Returns a list with the names of the variables used for reweighting.
        """
        return self._vars

    @property
    def bin_edge_dict(self) -> Dict[str, np.ndarray]:
        """
        Returns a dictionary with variable names as keys and np.ndarrays of the
        bin edges as values.
        """
        return {var: edges for var, edges in zip(self.variables, self.bin_edges)}


def get_weights_from_hist(
    df: pd.DataFrame,
    hist_weights: Weights,
    weight_kw: str = "weight",
) -> np.ndarray:
    """
    Calculates event-by-event weights based on the ratio of the bin weights in the given Weights class.
    """
    indices = []
    for var in hist_weights.variables:
        indices.append(np.digitize(df[var].values, hist_weights.bin_edge_dict[var]) - 1)

    # assume the old event weights are labeled as 'weight' in the df.
    # Used to fix the normalization
    try:
        new_weights = hist_weights.as_array[indices] * np.sum(df[weight_kw]) / np.sum(hist_weights.as_array[indices])
    except KeyError:
        new_weights = hist_weights.as_array[indices] * np.sum(len(df)) / np.sum(hist_weights.as_array[indices])
    return new_weights


def rbf_helper(rbf: Rbf, df: pd.DataFrame, variables: Iterable[str]) -> np.ndarray:
    """
    Helper function for calculating weights with the Rbf class from scipy
    :param rbf: A Rbf instance.
    :param df: pd.DataFrame use for reweighting.
    :param variables: List of variables used for reweighting.
    :return:
    """
    return rbf(*[df[var] for var in variables])


def get_weights_from_rbf(
    df: pd.DataFrame,
    hist_weights,
    n_chunks,
    n_processes=2,
    weight_kw: str = "weight",
) -> np.ndarray:

    coords, weights = hist_weights.flat_coords_and_weights
    rbf = Rbf(*coords, weights, function="cubic")

    df_chunks = np.array_split(ary=df, indices_or_sections=n_chunks, axis=0)

    args = [(rbf, chunk, hist_weights.variables) for chunk in df_chunks]

    with Pool(processes=n_processes) as pool:
        rbf_weights = pool.starmap(rbf_helper, args)  # type: np.ndarray

    rbf_weights = np.concatenate(rbf_weights)
    print(f"\tNumber of events with negative weights (will be set to zero): {np.sum(rbf_weights<0)}")
    rbf_weights[rbf_weights < 0] = 0

    try:
        new_weights = rbf_weights * np.sum(df[weight_kw]) / np.sum(rbf_weights)
    except KeyError:
        new_weights = rbf_weights * np.sum(len(df)) / np.sum(rbf_weights)
    return new_weights
