"""
This module contains functions and classes used for limit calculations.

Max Welsch 2018
"""
import scipy.stats
import scipy.special

from typing import Tuple, Union


__all__ = [
    "binomial_limit",
    "binomial_limit_approximation",
]


# has to be moved in a new module (e.g a statistics module)
def binomial_limit(
    n: Union[float, int],
    k: Union[float, int],
    sigma: float = 1,
) -> Tuple[float, float]:
    """
    Calculates the upper and lower limit for the probability p of a binomial
    distribution if an experiment yielded k successes for n trials.
    The confidence level for the limits is given in sigmas of the gaussian
    distribution.
    In contrast to other methods
    (see https://en.wikipedia.org/wiki/Binomial_proportion_confidence_interval),
    this method is exact in the sense that there is no approximation involved
    for the binomial distribution.

    NOTE:
    Not sure if this calculation is correct, especially the usage of the
    alpha as parameter when calculating the quantiles of the beta distribution.

    :param n: Number of trials.
    :param k: Number of successes
    :param sigma: Confidence level expressed in sigmas of a gaussian distribution
    (e.g. 2 sigma ~ 95% CL).
    :return confidence interval: Confidence interval for the binomial probability.
    """

    alpha = 2 * (1 - scipy.stats.norm.cdf(sigma))
    upper_limit = 1 - scipy.special.betaincinv(n - k + 1, k + 1, (1 - k / n) * alpha)
    lower_limit = 1 - scipy.special.betaincinv(n - k + 1, k + 1, 1 - k / n * alpha)

    return upper_limit, lower_limit


def binomial_limit_approximation(n, k, sigma=1):
    """
    Calculates the upper and lower limit for the probability p of a binomial
    distribution if an experiment yielded k successes for n trials.
    The confidence level for the limits is given in sigmas of the gaussian
    distribution.
    This method uses the approximation with the normal distribution.
    See "normal approximation interval".

    :param n: Number of trials.
    :param k: Number of successes
    :param sigma: Confidence level expressed in sigmas of a gaussian distribution
    (e.g. 2 sigma ~ 95% CL).
    :return confidence interval: Confidence interval for the binomial probability.
    """
    if not k >= 50 or not n - k >= 50:
        print(f"Warning: The binomial_limit_approximation does not provide a good approximation for {k} and {n}")

    alpha = 2 * (1 - scipy.stats.norm.cdf(sigma))
    c = scipy.stats.norm.ppf(1 - alpha / 2)
    p = k / n
    p_u = p + c * (p * (1 - p) / n) ** 0.5
    p_d = p - c * (p * (1 - p) / n) ** 0.5

    return p_u, p_d
