"""
Implementation of a weighted correlation calculation

Felix Metzner 2021
"""

import numpy as np

from typing import Tuple
from scipy.stats import rankdata


__all__ = [
    "weighted_corr",
    "valid_weight_corr_calc_methods",
]


def weighted_mean(a: np.ndarray, w: np.ndarray) -> float:
    return np.average(a, weights=w)


def weighted_cov(x: np.ndarray, y: np.ndarray, w: np.ndarray) -> float:
    mean_x = weighted_mean(a=x, w=w)  # type: float
    mean_y = weighted_mean(a=y, w=w)  # type: float
    return np.average((x - mean_x) * (y - mean_y), weights=w)


def weighted_pearson_corr(x: np.ndarray, y: np.ndarray, w: np.ndarray) -> float:
    return weighted_cov(x, y, w) / np.sqrt(weighted_cov(x, x, w) * weighted_cov(y, y, w))


def weighted_ranking(x: np.ndarray, w: np.ndarray) -> np.ndarray:
    (unique, arr_inv, counts) = np.unique(
        rankdata(x, method="average"),
        return_counts=True,
        return_inverse=True,
    )  # type: Tuple[np.ndarray, np.ndarray, np.ndarray]

    a = np.bincount(arr_inv, weights=w)  # type: np.ndarray
    return (np.cumsum(a) - a)[arr_inv] + ((counts + 1) / 2 * (a / counts))[arr_inv]


valid_weight_corr_calc_methods = (
    "pearson",
    "spearman",
)  # type: Tuple[str, ...]


def weighted_corr(
    x: np.ndarray,
    y: np.ndarray,
    w: np.ndarray,
    method: str = "pearson",
) -> float:
    assert all(isinstance(v, np.ndarray) for v in [x, y, w]), [type(v) for v in [x, y, w]]
    assert all(x.shape == v.shape for v in [y, w]), [v.shape for v in [x, y, w]]
    assert all(not np.isnan(v).any() for v in [x, y, w])

    if method == "pearson":
        return weighted_pearson_corr(x=x, y=y, w=w)
    elif method == "spearman":
        return weighted_pearson_corr(x=weighted_ranking(x, w=w), y=weighted_ranking(y, w=w), w=w)
    else:
        raise ValueError(
            f"Argument method should be one of {valid_weight_corr_calc_methods}, but '{method}' was provided!"
        )
