"""
Slow Pion Tracking correction for Belle

Calculates weights for data - MC corrections arising from the slow pion tracking efficiency

Sources:
General: https://belle.kek.jp/secured/wiki/doku.php?id=software:systematics
Data: https://belle.kek.jp/secured/belle_note/gn1176/
Belle Note: https://belle.kek.jp/secured/belle_note/gn1176/bn1176.pdf
Talk: https://kds.kek.jp/event/6029/contributions/138416/attachments/107716/127702/BAM-2010-11-15-WolfgangDungel.pdf

Felix Metzner 2019
"""

import os
import copy
import warnings
import numpy as np
import pandas as pd

from typing import Tuple, List, Dict
from pandas.core.common import SettingWithCopyWarning

from analysistools.utilities import PathType
from analysistools.pid_corrections.table_manager import TableManager


__all__ = [
    "SlowPionCorrection",
]


_slow_pi_table_manager = TableManager.instance()  # type: ignore
slow_pi_table_manager: TableManager = _slow_pi_table_manager


class SlowPionCorrection:
    number_of_variations: int = 12

    def __init__(self) -> None:
        self.weight_name: str = "slow_pi"

        self.table_name: str = "slow_pion_correction_table"
        self.base_path: PathType = os.path.dirname(os.path.abspath(__file__))

        self.columns = self.get_slow_pion_table_columns()

        self.momentum_column: str = "momentum_lower"
        self.charge_column: str = "charge"
        self.exp_column: str = "exp_lower"

        self.momentum_bins: List[float] = self.get_momentum_bins()

    def get_slow_pion_correction(
        self,
        df: pd.DataFrame,
        df_mom_col: str,
        df_exp_col: str,
        df_pdg_col: str,
    ) -> pd.DataFrame:
        if not len(df.index) > 0:
            raise RuntimeError("The dataframe you are passing to get_slow_pion_correction is empty. " "Is this intended?")
        assert df_mom_col in df.columns
        assert df_exp_col in df.columns
        assert df_pdg_col in df.columns
        assert df[df_mom_col].min() >= 0.0, df[df_mom_col].min()
        assert df[df_mom_col].max() <= self.momentum_bins[-1], df[df_mom_col].max()
        assert len(df[df_pdg_col].unique()) <= 3, f"{df[df_pdg_col].unique()}"
        assert all(
            abs(pdg) in self.get_pion_pdg_map().keys() for pdg in df[df_pdg_col].unique()
        ), f"{df[df_pdg_col].unique()}"

        exp_bins = self.get_exp_bins()
        exp_map = np.digitize(df[df_exp_col], exp_bins, right=False)
        assert (0 not in exp_map) and (len(exp_bins) not in exp_map)
        exp_map -= 1

        map_col_name = "__exp_map_for_slow_pion_corr__"
        assert map_col_name not in df.columns

        # Catching SettingWithCopyWarning from Pandas arising due to adding temporary column to df.
        with warnings.catch_warnings(record=True) as warning:
            warnings.simplefilter("always")
            # Actual command:
            df.loc[:, map_col_name] = exp_map
            for w in warning:
                assert issubclass(w.category, SettingWithCopyWarning)
                assert "A value is trying to be set on a copy" in str(w.message)

        pdg_map = copy.deepcopy(df[df_pdg_col].values)
        for pdg_code, pdg_bin in self.get_pion_pdg_map().items():
            pdg_map[abs(pdg_map) == pdg_code] = pdg_bin
        assert all(np.isin(pdg_map, test_elements=list(self.get_pion_pdg_map().values()))), np.unique(pdg_map)

        pdg_map_col_name = "__charge_map_for_slow_pion_corr__"
        assert pdg_map_col_name not in df.columns

        # Catching SettingWithCopyWarning from Pandas arising due to adding temporary column to df.
        with warnings.catch_warnings(record=True) as warning:
            warnings.simplefilter("always")
            # Actual command:
            df.loc[:, pdg_map_col_name] = pdg_map
            for w in warning:
                assert issubclass(w.category, SettingWithCopyWarning)
                assert "A value is trying to be set on a copy" in str(w.message)

        res_df = pd.DataFrame(df[[map_col_name, pdg_map_col_name]])
        res_column_names = []

        for correction_col in self.get_correction_cols():
            corr_res_column_name = self.get_weight_column_name(column_name=correction_col)
            res_column_names.append(corr_res_column_name)
            res_df.loc[:, corr_res_column_name] = np.nan

            for sub_table_name, sub_table_info in self.get_slow_pion_table_mapping().items():
                exp_bin = sub_table_info["exp_bin"]
                pdg_bin = sub_table_info["charge"]
                if exp_bin not in exp_map or pdg_bin not in pdg_map:
                    continue

                exp_cond = df[map_col_name] == exp_bin
                pdg_cond = df[pdg_map_col_name] == pdg_bin
                assert np.isnan(res_df.loc[exp_cond & pdg_cond, corr_res_column_name]).all()

                corr_matrix = self._get_correction_matrix(sub_table_name=sub_table_name, correction_col=correction_col)
                df_for_exps = df[[df_mom_col]].loc[exp_cond & pdg_cond]

                corr_for_exps = self._get_slow_pi_correction(
                    df=df_for_exps,
                    df_mom_col=df_mom_col,
                    correction_matrix=corr_matrix,
                )

                res_df.loc[exp_cond & pdg_cond, corr_res_column_name] = corr_for_exps

            assert not res_df.isnull().values.any()

        del df[map_col_name]
        del df[pdg_map_col_name]
        return res_df[res_column_names]

    def get_weight_column_name(self, column_name: str) -> str:
        return f"__weight_{self.weight_name}_{column_name}__"

    def _get_slow_pi_correction(
        self,
        df: pd.DataFrame,
        df_mom_col: str,
        correction_matrix: np.ndarray,
    ) -> np.ndarray:
        mom_indices = np.digitize(df[df_mom_col], self.momentum_bins, right=True)
        assert (0 not in mom_indices) and (len(self.momentum_bins) not in mom_indices)
        mom_indices -= 1

        return correction_matrix[mom_indices]

    def _get_correction_matrix(self, sub_table_name: str, correction_col: str) -> np.ndarray:
        matrix_name: str = f"{self.table_name}_{sub_table_name}_{correction_col}"
        matrix: np.ndarray = slow_pi_table_manager.get_correction_matrix(name=matrix_name)

        if matrix is None:
            table = self._get_slow_pi_table(sub_table_name=sub_table_name)
            matrix = self._build_correction_matrix(table=table, correction_col=correction_col)
            slow_pi_table_manager.add_correction_matrix(name=matrix_name, new_matrix=matrix)

        return matrix

    def _get_slow_pi_table(self, sub_table_name: str) -> pd.DataFrame:
        table_path: PathType = os.path.join(self.base_path, f"{self.table_name}.csv")

        table = slow_pi_table_manager.get_table(
            name=self.table_name,
            path=table_path,
            column_names=self.columns,
        )

        mapping = self.get_slow_pion_table_mapping()
        charge_condition = table[self.charge_column] == mapping[sub_table_name][self.charge_column]
        exp_condition = table[self.exp_column] == mapping[sub_table_name][self.exp_column]

        return table.loc[charge_condition & exp_condition]

    def _build_correction_matrix(
        self,
        table: pd.DataFrame,
        correction_col: str,
    ) -> np.ndarray:
        assert self.momentum_column in table.columns
        assert correction_col in table.columns

        # Ensure that momentum is unique in table:
        assert not (table.groupby(self.momentum_column).size() != 1).any()

        matrix: np.ndarray = np.full(len(self.momentum_bins[:-1]), np.nan)

        for mom_index, mom_bin in enumerate(self.momentum_bins[:-1]):
            value_rows = table.loc[(table[self.momentum_column] == mom_bin)]
            assert len(value_rows) == 1
            assert len(value_rows[correction_col]) == 1

            matrix[mom_index] = value_rows[correction_col].values[0]

        assert not np.isnan(matrix).any()
        return matrix

    @staticmethod
    def get_slow_pion_table_mapping() -> Dict[str, Dict[str, int]]:
        return {
            "charged_svd1": {"charge": 1, "exp_lower": 7, "exp_bin": 0},
            "charged_svd2": {"charge": 1, "exp_lower": 31, "exp_bin": 1},
            "neutral_svd1": {"charge": 0, "exp_lower": 7, "exp_bin": 0},
            "neutral_svd2": {"charge": 0, "exp_lower": 31, "exp_bin": 1},
        }

    @staticmethod
    def get_exp_ranges() -> List[Tuple[int, int]]:
        return [(7, 27), (31, 65)]

    @staticmethod
    def get_exp_bins() -> List[int]:
        exp_ranges = SlowPionCorrection.get_exp_ranges()
        exp_bins = [exp for exp, _ in exp_ranges]
        exp_bins.append(exp_ranges[-1][1] + 1)
        return exp_bins

    @staticmethod
    def get_slow_pion_table_columns() -> List[str]:
        cols = [
            "charge",
            "exp_lower",
            "exp_upper",
            "momentum_lower",
            "momentum_upper",
            "ratio",
        ]

        cols += [f"var_{i + 1}" for i in range(SlowPionCorrection.number_of_variations)]

        return cols

    @staticmethod
    def get_correction_cols() -> List[str]:
        cols = ["ratio"]
        cols += [f"var_{i + 1}" for i in range(SlowPionCorrection.number_of_variations)]
        return cols

    def get_correction_col_names(self) -> List[str]:
        return [self.get_weight_column_name(c) for c in self.get_correction_cols()]

    @staticmethod
    def get_momentum_bins() -> List[float]:
        return [
            0.05,
            0.075,
            0.1,
            0.125,
            0.15,
            0.175,
            0.2,
        ]

    @staticmethod
    def get_pion_pdg_map() -> Dict[int, int]:
        return {
            111: 0,
            211: 1,
        }
