"""
Tracking correction for Belle

Calculates
    - uncertainties for high momentum tracks
    - MC/Data correction factors for low momentum pion tracks

Felix Metzner 2019
"""

import numpy as np
import pandas as pd

from typing import Tuple, List, Dict

from analysistools.tracking_correction.slow_pion_correction import SlowPionCorrection


__all__ = [
    "TrackingCorrection",
]


class TrackingCorrection:
    def __init__(
        self,
        exp_column_name: str,
        consider_pi0_for_sys: bool = True,
    ) -> None:
        self.exp_column_name = exp_column_name

        self.name = "name"  # type: str
        self.pdg = "pdg_col"  # type: str
        self.mom = "mom_col"  # type: str
        self.theta = "theta_col"  # type: str

        # PDG codes of particles to be considered for high pt tracking correction.
        # Includes neutral pion, as for those the correction must also be applied, if slow pion correction is applied.
        self.track_pdg_codes = [11, 13, 211, 321]  # type: List[int]
        if consider_pi0_for_sys:
            self.track_pdg_codes.append(111)

        self.pion_pdg_s = list(SlowPionCorrection.get_pion_pdg_map().keys())  # type: List[int]

        self.low_pt_lower_bound = SlowPionCorrection.get_momentum_bins()[0]  # type: float
        self.low_pt_upper_bound = SlowPionCorrection.get_momentum_bins()[-1]  # type: float

        # Lower bound for high pt tracking efficiency systematic is given by upper bound of low pt systematic.
        self.high_pt_lower_bound = SlowPionCorrection.get_momentum_bins()[-1]  # type: float

        # Systematic uncertainty per track
        # Source: https://belle.kek.jp/secured/wiki/doku.php?id=software:systematics
        self.high_pt_sys = 0.0035  # type: float

        self.correction_info = {}  # type: Dict[int, Dict[str, str]]

    def add_track(
        self,
        name: str,
        pdg_column: str,
        momentum_column: str,
        theta_column: str,
        check_names: bool = True,
    ) -> None:
        is_new = self.check_track(
            name=name,
            pdg_column=pdg_column,
            momentum_column=momentum_column,
            theta_column=theta_column,
            check_names=check_names,
        )
        if not is_new:
            raise RuntimeError("Track info was already added. Was this intended?")
        index = len(self.correction_info) + 1
        self.correction_info.update(
            {
                index: {
                    self.name: name,
                    self.pdg: pdg_column,
                    self.mom: momentum_column,
                    self.theta: theta_column,
                }
            }
        )

    def get_corrections(self, df: pd.DataFrame) -> Tuple[np.array, np.array, pd.DataFrame, List[str]]:
        self.check_columns(columns=df.columns)
        sdf = df.loc[:, self.get_columns()]

        high_pt_tracks = self.get_n_high_pt_tracks(sdf)
        high_pt_corr = self.high_pt_sys * high_pt_tracks
        slow_pi_corr, slow_pi_col_names = self.get_slow_pion_correction(sdf)

        assert len(high_pt_tracks) == len(df.index)
        assert len(high_pt_corr) == len(df.index)
        assert len(slow_pi_corr) == len(df.index)

        return high_pt_corr, high_pt_tracks, slow_pi_corr, slow_pi_col_names

    def get_n_high_pt_tracks(self, df: pd.DataFrame):
        high_pt_tracks = np.zeros(len(df.index))

        for track_info in self.correction_info.values():
            is_track = abs(df[track_info[self.pdg]]).isin(self.track_pdg_codes)
            if self.high_pt_lower_bound > 0.0:
                is_high_pt = df[track_info[self.mom]] >= self.high_pt_lower_bound
            else:
                is_high_pt = True
            high_pt_tracks += is_track & is_high_pt

        return high_pt_tracks

    def get_high_pt_correction(self, df: pd.DataFrame) -> np.ndarray:
        high_pt_tracks = np.zeros(len(df.index))  # type: np.ndarray

        for track_info in self.correction_info.values():
            is_track = df[track_info[self.pdg]].isin(self.track_pdg_codes)
            if self.high_pt_lower_bound > 0.0:
                is_high_pt = df[track_info[self.mom]] >= self.high_pt_lower_bound
            else:
                is_high_pt = True
            high_pt_tracks += is_track & is_high_pt

        return high_pt_tracks * self.high_pt_sys

    def get_slow_pion_correction(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, List[str]]:
        slow_pion_corrector = SlowPionCorrection()

        corr_col_names = slow_pion_corrector.get_correction_col_names()
        slow_pi_corr = pd.DataFrame({c: np.ones(len(df.index)) for c in corr_col_names}, index=df.index)

        for track_info in self.correction_info.values():
            mom_cond = (df[track_info[self.mom]] >= self.low_pt_lower_bound) & (
                df[track_info[self.mom]] <= self.low_pt_upper_bound
            )
            pdg_cond = abs(df[track_info[self.pdg]]).isin(list(SlowPionCorrection.get_pion_pdg_map().keys()))

            df_with_pi = df.loc[mom_cond & pdg_cond]
            if not len(df_with_pi) > 0:
                continue
            corr_res = slow_pion_corrector.get_slow_pion_correction(
                df=df_with_pi,
                df_mom_col=track_info[self.mom],
                df_exp_col=self.exp_column_name,
                df_pdg_col=track_info[self.pdg],
            )
            assert len(df_with_pi.index) == len(corr_res.index)

            for col in corr_col_names:
                slow_pi_corr.loc[corr_res.index, col] *= corr_res[col]

        assert not pd.isnull(slow_pi_corr).values.any()

        return slow_pi_corr, corr_col_names

    def get_columns(self) -> List[str]:
        columns = [self.exp_column_name]
        for track_info in self.correction_info.values():
            columns.append(track_info[self.pdg])
            columns.append(track_info[self.mom])
            columns.append(track_info[self.theta])
        return columns

    def check_columns(self, columns: List[str]) -> None:
        for column in self.get_columns():
            assert column in columns, f"{column}"

    def check_track(
        self,
        name: str,
        pdg_column: str,
        momentum_column: str,
        theta_column: str,
        check_names: bool,
    ) -> bool:
        if check_names:
            self.check_column_names(pdg=pdg_column, mom=momentum_column, theta=theta_column)

        for track_info in self.correction_info.values():
            if track_info[self.name] == name:
                return False
            if track_info[self.pdg] == pdg_column:
                return False
            if track_info[self.mom] == momentum_column:
                return False
            if track_info[self.theta] == theta_column:
                return False
        return True

    @staticmethod
    def check_column_names(pdg: str, mom: str, theta: str) -> None:
        pdg_parts = pdg.split("daughter__bo")
        mom_parts = mom.split("daughter__bo")
        theta_parts = theta.split("daughter__bo")

        if not (len(pdg_parts) == len(mom_parts) and len(pdg_parts) == len(theta_parts)):
            raise RuntimeError(f"Column names do not match:\n {pdg}\n {mom}\n {theta}")

        pdg_daughters = [int(p[0]) for p in pdg_parts[1:] if p[0].isdigit()]
        mom_daughters = [int(p[0]) for p in mom_parts[1:] if p[0].isdigit()]
        theta_daughters = [int(p[0]) for p in theta_parts[1:] if p[0].isdigit()]

        if not (len(pdg_daughters) == len(mom_daughters) and len(pdg_daughters) == len(theta_daughters)):
            raise RuntimeError(f"Column names do not match:\n {pdg}\n {mom}\n {theta}")

        for a, b, c in zip(pdg_daughters, mom_daughters, theta_daughters):
            if not (a == b and a == c):
                raise RuntimeError(f"Column names do not match:\n {pdg}\n {mom}\n {theta}")
