import unittest

import numpy as np

from analysistools.physics_vectors.fourvector import FourVector
from analysistools.physics_vectors.fourvector import check_four_vector_dimensions


class TestFourVector(unittest.TestCase):
    def setUp(self):
        self.right_shape = np.array([[3, 1, 1, 1], [5, 1, 2, 3]])
        self.wrong_shape = np.array([3, 4, 5])

        self.p1 = FourVector(np.array([[5, 1, 2, 3], [7, 4, 3, 3]]))
        self.p2 = FourVector(np.array([[5, 6, 5, 1], [2, 2, 2, 2]]))
        self.p3 = FourVector(np.array([[1, 3, 3, 3]]))

    def test_add(self):
        add = self.p1 + self.p2
        radd = self.p2 + self.p1
        result = np.array([[10, 7, 7, 4], [9, 6, 5, 5]])
        np.testing.assert_array_equal(result, add.as_array)
        np.testing.assert_array_equal(result, radd.as_array)

    def test_sub(self):
        sub = self.p1 - self.p2
        result = np.array([[0, -5, -3, 2], [5, 2, 1, 1]])
        np.testing.assert_array_equal(result, sub.as_array)

    def test_mul(self):
        mul = self.p1 * self.p2
        exp_mul = np.array([6, -6])
        np.testing.assert_equal(mul, exp_mul)

    def test_T(self):
        np.testing.assert_array_equal(np.array([5, 7]), self.p1.T)

    def test_Vec(self):
        np.testing.assert_array_equal(np.array([[1, 2, 3], [4, 3, 3]]), self.p1.Vec.as_array)

    def test_as_array(self):
        np.testing.assert_array_equal(np.array([[5, 1, 2, 3], [7, 4, 3, 3]]), self.p1.as_array)

    def test_M2(self):
        exp_m2 = np.array([11, 15])
        np.testing.assert_array_equal(exp_m2, self.p1.M2())

    def test_M(self):
        exp_m = np.array([np.sqrt(11), np.sqrt(15)])
        exp_m2 = np.array([-np.sqrt(26)])
        np.testing.assert_equal(self.p1.M(), exp_m)
        np.testing.assert_equal(self.p3.M(), exp_m2)

    def test_BoostVector(self):
        exp_boostvector = np.array([[1 / 5, 2 / 5, 3 / 5], [4 / 7, 3 / 7, 3 / 7]])
        np.testing.assert_array_equal(exp_boostvector, self.p1.BoostVector())

    def test_Boost(self):
        exp_fourvector = np.array([[np.sqrt(11), 0, 0, 0], [np.sqrt(15), 0, 0, 0]])
        b = self.p1.BoostVector()
        self.p1.Boost(-b)
        np.testing.assert_almost_equal(exp_fourvector, self.p1.as_array)

    def test_Angle(self):
        exp_angle = np.array([0.8699, 0.1405])
        angles = self.p1.Angle(self.p2.Vec)
        np.testing.assert_almost_equal(exp_angle, angles, decimal=3)

    def test_check_dimensions(self):
        with self.assertRaises(ValueError):
            check_four_vector_dimensions(self.wrong_shape)
