import unittest
import numpy as np
import pandas as pd

from analysistools.reweighting.ndim_reweighter import NdimReweighter


class TestNdimReweighter(unittest.TestCase):
    def setUp(self):
        self.variables = ["x", "y", "z"]
        self.bins = [2, 2, 2]
        self.old_model = pd.DataFrame(
            np.array(
                [
                    [0.5, 0.5, 0.5],
                    [0.5, 0.5, 0.5],
                    [0.5, 0.5, 1.5],
                    [0.5, 0.5, 1.5],
                    [0.5, 1.5, 0.5],
                    [0.5, 1.5, 0.5],
                    [0.5, 1.5, 1.5],
                    [0.5, 1.5, 1.5],
                    [1.5, 0.5, 0.5],
                    [1.5, 0.5, 0.5],
                    [1.5, 0.5, 1.5],
                    [1.5, 0.5, 1.5],
                    [1.5, 1.5, 0.5],
                    [1.5, 1.5, 0.5],
                    [1.5, 1.5, 1.5],
                    [1.5, 1.5, 1.5],
                ]
            ),
            columns=self.variables,
        )
        self.new_model = pd.DataFrame(
            np.array(
                [
                    [0.5, 0.5, 0.5],
                    [0.5, 0.5, 0.5],
                    [0.5, 0.5, 1.5],
                    [0.5, 0.5, 1.5],
                    [0.5, 1.5, 0.5],
                    [0.5, 1.5, 0.5],
                    [0.5, 1.5, 1.5],
                    [0.5, 1.5, 1.5],
                    [1.5, 0.5, 0.5],
                    [1.5, 0.5, 0.5],
                    [1.5, 0.5, 1.5],
                    [1.5, 0.5, 1.5],
                    [1.5, 0.5, 1.5],
                    [1.5, 0.5, 1.5],
                    [1.5, 1.5, 0.5],
                    [1.5, 1.5, 0.5],
                    [1.5, 1.5, 1.5],
                    [1.5, 1.5, 1.5],
                ]
            ),
            columns=self.variables,
        )

        self.h_old_exp = np.array([[[2, 2], [2, 2]], [[2, 2], [2, 2]]]) / 16 / 0.5 ** 3
        self.h_new_exp = np.array([[[2, 2], [2, 2]], [[2, 4], [2, 2]]]) / 18 / 0.5 ** 3

        self.models = {"old": self.old_model, "new": self.new_model}

    def test_histograms(self):
        rw = NdimReweighter(self.models, "new", "old", self.variables, self.bins)
        edges = rw.bin_edges
        np.testing.assert_equal(edges, [np.array([0.5, 1.0, 1.5]), np.array([0.5, 1.0, 1.5]), np.array([0.5, 1.0, 1.5])])

        h_old = rw.h_old
        np.testing.assert_equal(h_old, self.h_old_exp)

        h_new = rw.h_new
        np.testing.assert_equal(h_new, np.array([[[2, 2], [2, 2]], [[2, 4], [2, 2]]]) / 18 / 0.5 ** 3)

    def test_bin_edges_as_array(self):
        rw = NdimReweighter(
            self.models, "new", "old", self.variables, [np.array([0, 1, 2]), np.array([0, 1, 2]), np.array([0, 1, 2])]
        )

        np.testing.assert_equal(rw.bin_edges, [np.array([0, 1, 2]), np.array([0, 1, 2]), np.array([0, 1, 2])])

    def test_weights(self):
        rw = NdimReweighter(self.models, "new", "old", self.variables, self.bins)
        np.testing.assert_almost_equal(
            rw.weights, np.array([[[0.888, 0.888], [0.888, 0.888]], [[0.888, 1.777], [0.888, 0.888]]]), decimal=3
        )

    def test_add_variations(self):
        rw = NdimReweighter(self.models, "new", "old", self.variables, self.bins)

        variation = pd.DataFrame(
            np.array(
                [
                    [0.5, 0.5, 0.5],
                    [0.5, 0.5, 0.5],
                    [0.5, 0.5, 1.5],
                    [0.5, 0.5, 1.5],
                    [0.5, 1.5, 0.5],
                    [0.5, 1.5, 0.5],
                    [0.5, 1.5, 1.5],
                    [0.5, 1.5, 1.5],
                    [1.5, 0.5, 0.5],
                    [1.5, 0.5, 0.5],
                    [1.5, 0.5, 1.5],
                    [1.5, 0.5, 1.5],
                    [1.5, 0.5, 1.5],
                    [1.5, 0.5, 1.5],
                    [1.5, 1.5, 0.5],
                    [1.5, 1.5, 0.5],
                    [1.5, 1.5, 1.5],
                    [1.5, 1.5, 1.5],
                    [1.5, 1.5, 1.5],
                    [1.5, 1.5, 1.5],
                ]
            ),
            columns=self.variables,
        )

        rw.add_variations({"new_var_1": variation}, "new")
        weights = rw(np.array([0.5, 1.5]), np.array([0.5, 1.5]), np.array([0.5, 1.5]), return_variations=True)
        self.assertEqual(2, len(weights.columns))
