""""
Unit tests for final_state_particles_utils module.

Max Welsch 2018
"""

import unittest
import unittest.mock as mock

from analysistools.processing.final_state_particles_utils import (
    FinalStateParticle,
    FinalStateParticleCollection,
)


class TestFinalStateParticleCollection(unittest.TestCase):
    def setUp(self):
        self.fsp_e = FinalStateParticle("e", "e-:fsp", "reco", ["dr < 2"])
        self.fsp_tau = FinalStateParticle("tau", "tau-:mc", "mc", ["dr < 2", "abs(dz) < 4"])
        self.fsp_pi0 = FinalStateParticle(
            "pi0", "pi0:fsp", ["pi0:mdst"], ["abs(dM) < 0.01", "abs(SigMBF) < 3.", "E > 0.1"]
        )

        self.fsp_coll = FinalStateParticleCollection(e=self.fsp_e, tau=self.fsp_tau, pi0=self.fsp_pi0)

    def test_as_dict(self):
        coll_as_dict = {
            "label": ["e", "tau", "pi0"],
            "particlelist": ["e-:fsp", "tau-:mc", "pi0:fsp"],
            "origin": ["reco", "mc", ["pi0:mdst"]],
            "cuts": [["dr < 2"], ["dr < 2", "abs(dz) < 4"], ["abs(dM) < 0.01", "abs(SigMBF) < 3.", "E > 0.1"]],
        }

        self.assertDictEqual(self.fsp_coll.as_dict(), coll_as_dict)

    def test_as_list(self):
        self.assertListEqual(self.fsp_coll.reco_fsps, [self.fsp_e])
        self.assertListEqual(self.fsp_coll.mc_fsps, [self.fsp_tau])
        self.assertListEqual(self.fsp_coll.list_fsps, [self.fsp_pi0])

    def test_add(self):
        piplus = FinalStateParticle("pi+", "pi+:fsp", "reco", ["dr < 2"])
        photon = FinalStateParticle("gamma", "gamma:fsp", ["gamma:mdst", "gamma:v0mdst"], ["goodBelleGamma == 1"])

        self.fsp_coll.add_fsps(piplus=piplus, photon=photon)

        self.assertListEqual(self.fsp_coll.reco_fsps, [self.fsp_e, piplus])
        self.assertListEqual(self.fsp_coll.mc_fsps, [self.fsp_tau])
        self.assertListEqual(self.fsp_coll.list_fsps, [self.fsp_pi0, photon])

    def test_no_fsp(self):
        with self.assertRaises(ValueError):
            _ = FinalStateParticleCollection(bla=1.0)

    @mock.patch("analysistools.processing.final_state_particles_utils.mA.fillParticleLists")
    @mock.patch("analysistools.processing.final_state_particles_utils.basf2.Path")
    def test_get_fsp_from_reco(self, mock_path, mock_fillpl):
        fsp_coll = FinalStateParticleCollection()

        fsp_coll.get_fsp_from_reco(mock_path, self.fsp_e, False)

        mock_fillpl.assert_called_with([("e-:fsp", "dr < 2")], path=mock_path)
