"""
Unit tests for oberservables_utils module.

Max Welsch 2018
"""
import unittest

from analysistools.processing.observables_utils import VariableCollection, VariableListBase, VariableCollectionFactory


class TestVariableCollection(unittest.TestCase):
    def setUp(self):
        self.kinematic_vars = ["px", "py", "pz", "E"]
        self.event_vars = ["evt_num"]

    def test_as_list(self):

        oc = VariableCollection(kinematic=self.kinematic_vars, event=self.event_vars)

        expected = ["px", "py", "pz", "E", "evt_num"]
        print(expected)
        self.assertListEqual(oc.as_list(), expected)

    def test_as_dictionary(self):
        oc = VariableCollection(kinematic=self.kinematic_vars, event=self.event_vars)

        expected = {"kinematic": self.kinematic_vars, "event": self.event_vars}

        self.assertDictEqual(oc.as_dict(), expected)

    def test_get_item_op(self):
        oc = VariableCollection(kinematic=self.kinematic_vars, event=self.event_vars)
        self.assertListEqual(oc["kinematic"], self.kinematic_vars)

    def test_print(self):
        oc = VariableCollection(name="Test", kinematic=self.kinematic_vars, event=self.event_vars)

        print(oc)

    def test_daughter_variables(self):
        vc = VariableCollection("test", kinematic=self.kinematic_vars)
        vc.add_daughter_variables(self.kinematic_vars, 1)
        vc.add_mc_daughter_variables(self.event_vars, 2, 1)

        expected = [
            "px",
            "py",
            "pz",
            "E",
            "daughter(1, px)",
            "daughter(1, py)",
            "daughter(1, pz)",
            "daughter(1, E)",
            "mcDaughter(1, mcDaughter(2, evt_num))",
        ]
        print(vc)
        self.assertListEqual(expected, vc.as_list())


class TestVariableList(unittest.TestCase):
    def setUp(self):
        class TestVariableList(VariableListBase):
            def __init__(self, use_mc_info: bool = False, use_b2bii: bool = False, roe_mask: str = None):
                super().__init__(use_mc_info=use_mc_info, use_b2bii=use_b2bii, roe_mask=roe_mask)

        self.TestObservableList = TestVariableList

        class MCVariableList(VariableListBase):
            def __init__(self, use_mc_info: bool = True, use_b2bii: bool = False, roe_mask: str = None):
                super().__init__(use_mc_info=use_mc_info, use_b2bii=use_b2bii, roe_mask=roe_mask)

            def get(self):

                if not self._use_mc_info:
                    return []

                return ["MC_var_1", "MC_var_2"]

        self.MCObservableList = MCVariableList

    def test_not_implemented(self):

        with self.assertRaises(TypeError):
            _ = self.TestObservableList()

    def test_mc_observables(self):

        mc_observables = self.MCObservableList()
        mc_observables_if_false = self.MCObservableList(use_mc_info=False)

        self.assertListEqual(mc_observables.get(), ["MC_var_1", "MC_var_2"])
        self.assertListEqual(mc_observables_if_false.get(), [])


class TestVariableCollectionFactory(unittest.TestCase):
    def setUp(self):
        class EventVariables(VariableListBase):
            def get(self):
                event_variables = [
                    "belleECLEnergy" if self._use_b2bii else "",
                    "nKLMClusters",
                    "KLMEnergy",
                    "Ecms",
                    "isContinuumEvent" if self._use_mc_info else "",
                    "isNotContinuumEvent" if self._use_mc_info else "",
                ]

                return list(filter(lambda x: x != "", event_variables))

        self.event_variables = EventVariables

    def test_mc_factory(self):

        mock_config = {
            "use_mc_info": True,
            "use_b2bii": True,
            "roe_mask": "CleanROE",
        }

        eo = self.event_variables

        class DummyFactory(VariableCollectionFactory):
            def create(self):
                evt_obs = eo(use_mc_info=mock_config["use_mc_info"], use_b2bii=mock_config["use_b2bii"])
                return VariableCollection("Dummy", event=evt_obs.get())

        df = DummyFactory(mock_config)

        dummy_coll = df.create()
        expected = ["belleECLEnergy", "nKLMClusters", "KLMEnergy", "Ecms", "isContinuumEvent", "isNotContinuumEvent"]
        self.assertListEqual(dummy_coll.as_list(), expected)

    def test_data_factory(self):
        mock_config = {
            "use_mc_info": False,
            "use_b2bii": False,
            "roe_mask": "CleanROE",
        }

        eo = self.event_variables

        class DummyFactory(VariableCollectionFactory):
            def create(self):
                evt_vars = eo(use_mc_info=mock_config["use_mc_info"], use_b2bii=mock_config["use_b2bii"])

                return VariableCollection("Dummy", event=evt_vars.get())

        df = DummyFactory(mock_config)

        dummy_coll = df.create()
        expected = ["nKLMClusters", "KLMEnergy", "Ecms"]
        self.assertListEqual(dummy_coll.as_list(), expected)
