#include "LepEff.h"
#include "TMath.h"
#include<iostream>

void LepEff::init(int varBins, const char* strVar, float fPid, int lep_gen) {
  nVarBins = varBins;

  fPidCut = fPid;
  if(lep_gen == 0){
    const float fPidCuts[] = { .1, .5, .6, .8, .9};
    iPidCut = -1;
    for(int i = 0; i<5;++i){
      if(fPidCut == fPidCuts[i]){
	iPidCut = i;
	break;
      }
    }
  }
  else if(lep_gen ==1){
    const float fPidCuts[] = {.1, .8, .9, .95, .97};
    iPidCut = -1;
    for(int i = 0; i<5;++i){
      if(fPidCut == fPidCuts[i]){
	iPidCut = i;
	break;
      }
    }
  } 
  else{
    std::cout<<"ERR: LepEff::init: not supported particle type: "<<lep_gen<<std::endl;exit(-1);
  }

  if(iPidCut<0){
    std::cout<<"ERR: LepEff::init: not supported PID cut: "<<fPidCut<<std::endl;exit(-1);
  }


  iPidCutOffset = 70*9;

  const double fMomBins[] = {0.0,  0.5,  1.0,  1.5,  2.0,  2.5,  3.0,  3.5,  4.0,  4.5,  5.0};

  iError = 0;

  // Electrons
  axis_plep[0] = new TAxis(nEffMomBins, fMomBins);
  const double fThetaBinsElec[] = {18, 25, 35,  40,  60,  125,  132,  151};
  axis_theta[0] = new TAxis(nEffThetaBins, fThetaBinsElec);

  // Muons
  axis_plep[1] = new TAxis(nEffMomBins, fMomBins);
  const double fThetaBinsMuon[] = {17, 25, 37,  51,  117,  130,  145,  150};
  axis_theta[1] = new TAxis(nEffThetaBins, fThetaBinsMuon);

  iPosWeight = 5;
  iPosErrStat = 6;
  iPosErrSyst1 = 7;
  iPosErrSyst2 = 8;

}

LepEff::LepEff(float fPid, int lep_gen) {
  init(1, "default", fPid, lep_gen);
}

LepEff::LepEff(int varBins, const char* strVar, float fPid, int lep_gen) {
  init(varBins, strVar, fPid, lep_gen);
}

LepEff::~LepEff() {
}

float LepEff::lookup(int lep_gen, int exp_no, int iBinTheta, int iBinLepMom, int iType) {
  if (iType < 0 || iType >= nColumns) return -1;
  if (iBinTheta < 0 || iBinTheta >= nEffThetaBins) return -1;
  if (iBinLepMom < 0 || iBinLepMom >= nEffMomBins) return -1;
  int iPosition = iPidCut*iPidCutOffset + iBinTheta*nColumns*nEffMomBins+iBinLepMom*nColumns+iType;
  if (lep_gen == 0 && exp_no>27) return fElecEff[iPosition];
  else if (lep_gen == 0 && exp_no <= 27) return fElecEffSVD1[iPosition];
  else if (exp_no <= 27) return fMuonEffSVD1[iPosition];
  else if (exp_no <= 39) return fMuonEff31[iPosition];
  else if (exp_no <= 49) return fMuonEff41[iPosition];
  else return fMuonEff51[iPosition];
  return -1;
}

float LepEff::lookup(int lep_gen, int exp_no, float theta, float plep, int iType) {
  if (lep_gen < 0 || lep_gen > 1) return -1;
  int iBinTheta = axis_theta[lep_gen]->FindBin(theta) - 1;
  int iBinLepMom = axis_plep[lep_gen]->FindBin(plep) - 1;
  return lookup(lep_gen, exp_no, iBinTheta, iBinLepMom, iType);
}



float LepEff::weight(int lep_gen, int exp_no, float theta, float plep){

  int i = -1;
  if (lep_gen == 0 && exp_no > 27) i = 0;
  else if (lep_gen == 0 && exp_no <= 27) i = 1;
  else if (exp_no <= 27) i = 2;
  else if (exp_no <= 39) i = 3;
  else if (exp_no <= 49) i = 4;
  else i = 5;

  float fResult = lookup(lep_gen, exp_no, theta, plep, iPosWeight);
  if (fResult < 0) return 1.;
  else return fResult;

}



float LepEff::errStat(int lep_gen, int exp_no, float theta, float plep) {
  float fResult = lookup(lep_gen, exp_no, theta, plep, iPosErrStat);
  if (fResult < 0) return 0.;
  return fResult;
}

float LepEff::errSyst1(int lep_gen, int exp_no, float theta, float plep) {
  float fResult = lookup(lep_gen, exp_no, theta, plep, iPosErrSyst1);
  if (fResult < 0) return 0.;
  return fResult;
}

float LepEff::errSyst2(int lep_gen, int exp_no, float theta, float plep) {
  float fResult = lookup(lep_gen, exp_no, theta, plep, iPosErrSyst2);
  if (fResult < 0) return 0.;
  return fResult;
}


void LepEff::reset() {
  iError = 0;
}

const int LepEff::nEffMomBins = 10;
const int LepEff::nEffThetaBins = 7;
const int LepEff::nColumns = 9;


const int LepEff::exp_range[] = {11,35,11,35,45,55};
const float LepEff::fElecEff[] = {
  0.10, 18.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0024, 0.0415,
  0.10, 18.0, 25.0, 0.50, 1.00, 0.9818, 0.0003, 0.0024, 0.0415,
  0.10, 18.0, 25.0, 1.00, 1.50, 0.9874, 0.0002, 0.0024, 0.0415,
  0.10, 18.0, 25.0, 1.50, 2.00, 0.9884, 0.0002, 0.0024, 0.0135,
  0.10, 18.0, 25.0, 2.00, 2.50, 0.9867, 0.0003, 0.0024, 0.0135,
  0.10, 18.0, 25.0, 2.50, 3.00, 0.9855, 0.0006, 0.0024, 0.0135,
  0.10, 18.0, 25.0, 3.00, 3.50, 0.9852, 0.0008, 0.0024, 0.0135,
  0.10, 18.0, 25.0, 3.50, 4.00, 0.9863, 0.0013, 0.0024, 0.0135,
  0.10, 18.0, 25.0, 4.00, 4.50, 0.9866, 0.0064, 0.0024, 0.0135,
  0.10, 18.0, 25.0, 4.50, 5.00, 0.9389, 0.0091, 0.0024, 0.0135,
  0.10, 25.0, 35.0, 0.00, 0.50, 1.0000, 0.0000, 0.0061, 0.0415,
  0.10, 25.0, 35.0, 0.50, 1.00, 0.9475, 0.0005, 0.0061, 0.0415,
  0.10, 25.0, 35.0, 1.00, 1.50, 0.9619, 0.0006, 0.0061, 0.0415,
  0.10, 25.0, 35.0, 1.50, 2.00, 0.9793, 0.0009, 0.0061, 0.0135,
  0.10, 25.0, 35.0, 2.00, 2.50, 0.9825, 0.0016, 0.0061, 0.0135,
  0.10, 25.0, 35.0, 2.50, 3.00, 0.9759, 0.0025, 0.0061, 0.0135,
  0.10, 25.0, 35.0, 3.00, 3.50, 0.9720, 0.0042, 0.0061, 0.0135,
  0.10, 25.0, 35.0, 3.50, 4.00, 0.9587, 0.0100, 0.0061, 0.0135,
  0.10, 25.0, 35.0, 4.00, 4.50, 0.9611, 0.0056, 0.0061, 0.0135,
  0.10, 25.0, 35.0, 4.50, 5.00, 1.0000, 0.0000, 0.0061, 0.0135,
  0.10, 35.0, 40.0, 0.00, 0.50, 0.8287, 0.0107, 0.0064, 0.0415,
  0.10, 35.0, 40.0, 0.50, 1.00, 0.9004, 0.0010, 0.0064, 0.0415,
  0.10, 35.0, 40.0, 1.00, 1.50, 0.9685, 0.0011, 0.0064, 0.0415,
  0.10, 35.0, 40.0, 1.50, 2.00, 0.9827, 0.0014, 0.0064, 0.0135,
  0.10, 35.0, 40.0, 2.00, 2.50, 0.9890, 0.0019, 0.0064, 0.0135,
  0.10, 35.0, 40.0, 2.50, 3.00, 0.9837, 0.0023, 0.0064, 0.0135,
  0.10, 35.0, 40.0, 3.00, 3.50, 0.9744, 0.0033, 0.0064, 0.0135,
  0.10, 35.0, 40.0, 3.50, 4.00, 0.9346, 0.0088, 0.0064, 0.0135,
  0.10, 35.0, 40.0, 4.00, 4.50, 1.0000, 0.0000, 0.0064, 0.0135,
  0.10, 35.0, 40.0, 4.50, 5.00, 1.0000, 0.0000, 0.0064, 0.0135,
  0.10, 40.0, 60.0, 0.00, 0.50, 0.9563, 0.0019, 0.0035, 0.0415,
  0.10, 40.0, 60.0, 0.50, 1.00, 0.9495, 0.0005, 0.0035, 0.0415,
  0.10, 40.0, 60.0, 1.00, 1.50, 0.9763, 0.0006, 0.0035, 0.0415,
  0.10, 40.0, 60.0, 1.50, 2.00, 0.9844, 0.0008, 0.0035, 0.0135,
  0.10, 40.0, 60.0, 2.00, 2.50, 0.9831, 0.0011, 0.0035, 0.0135,
  0.10, 40.0, 60.0, 2.50, 3.00, 0.9777, 0.0021, 0.0035, 0.0135,
  0.10, 40.0, 60.0, 3.00, 3.50, 0.9747, 0.0093, 0.0035, 0.0135,
  0.10, 40.0, 60.0, 3.50, 4.00, 1.0000, 0.0000, 0.0035, 0.0135,
  0.10, 40.0, 60.0, 4.00, 4.50, 1.0000, 0.0000, 0.0035, 0.0135,
  0.10, 40.0, 60.0, 4.50, 5.00, 1.0000, 0.0000, 0.0035, 0.0135,
  0.10, 60.0, 125.0, 0.00, 0.50, 0.9659, 0.0005, 0.0033, 0.0098,
  0.10, 60.0, 125.0, 0.50, 1.00, 0.9762, 0.0002, 0.0033, 0.0098,
  0.10, 60.0, 125.0, 1.00, 1.50, 0.9840, 0.0004, 0.0033, 0.0098,
  0.10, 60.0, 125.0, 1.50, 2.00, 0.9830, 0.0005, 0.0033, 0.0138,
  0.10, 60.0, 125.0, 2.00, 2.50, 0.9791, 0.0012, 0.0033, 0.0138,
  0.10, 60.0, 125.0, 2.50, 3.00, 0.9478, 0.0031, 0.0033, 0.0138,
  0.10, 60.0, 125.0, 3.00, 3.50, 1.0000, 0.0000, 0.0033, 0.0138,
  0.10, 60.0, 125.0, 3.50, 4.00, 1.0000, 0.0000, 0.0033, 0.0138,
  0.10, 60.0, 125.0, 4.00, 4.50, 1.0000, 0.0000, 0.0033, 0.0138,
  0.10, 60.0, 125.0, 4.50, 5.00, 1.0000, 0.0000, 0.0033, 0.0138,
  0.10, 125.0, 132.0, 0.00, 0.50, 0.7862, 0.0034, 0.0216, 0.0262,
  0.10, 125.0, 132.0, 0.50, 1.00, 1.0144, 0.0022, 0.0216, 0.0262,
  0.10, 125.0, 132.0, 1.00, 1.50, 0.9960, 0.0055, 0.0216, 0.0262,
  0.10, 125.0, 132.0, 1.50, 2.00, 0.9891, 0.0138, 0.0216, 0.0391,
  0.10, 125.0, 132.0, 2.00, 2.50, 0.9857, 0.0463, 0.0216, 0.0391,
  0.10, 125.0, 132.0, 2.50, 3.00, 1.0000, 0.0000, 0.0216, 0.0391,
  0.10, 125.0, 132.0, 3.00, 3.50, 1.0000, 0.0000, 0.0216, 0.0391,
  0.10, 125.0, 132.0, 3.50, 4.00, 1.0000, 0.0000, 0.0216, 0.0391,
  0.10, 125.0, 132.0, 4.00, 4.50, 1.0000, 0.0000, 0.0216, 0.0391,
  0.10, 125.0, 132.0, 4.50, 5.00, 1.0000, 0.0000, 0.0216, 0.0391,
  0.10, 132.0, 151.0, 0.00, 0.50, 1.0146, 0.0036, 0.0042, 0.0262,
  0.10, 132.0, 151.0, 0.50, 1.00, 1.0138, 0.0005, 0.0042, 0.0262,
  0.10, 132.0, 151.0, 1.00, 1.50, 0.9878, 0.0006, 0.0042, 0.0262,
  0.10, 132.0, 151.0, 1.50, 2.00, 0.9777, 0.0015, 0.0042, 0.0391,
  0.10, 132.0, 151.0, 2.00, 2.50, 0.9630, 0.0070, 0.0042, 0.0391,
  0.10, 132.0, 151.0, 2.50, 3.00, 1.0000, 0.0000, 0.0042, 0.0391,
  0.10, 132.0, 151.0, 3.00, 3.50, 1.0000, 0.0000, 0.0042, 0.0391,
  0.10, 132.0, 151.0, 3.50, 4.00, 1.0000, 0.0000, 0.0042, 0.0391,
  0.10, 132.0, 151.0, 4.00, 4.50, 1.0000, 0.0000, 0.0042, 0.0391,
  0.10, 132.0, 151.0, 4.50, 5.00, 1.0000, 0.0000, 0.0042, 0.0391,
  0.50, 18.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0029, 0.0449,
  0.50, 18.0, 25.0, 0.50, 1.00, 0.9747, 0.0004, 0.0029, 0.0449,
  0.50, 18.0, 25.0, 1.00, 1.50, 0.9814, 0.0002, 0.0029, 0.0449,
  0.50, 18.0, 25.0, 1.50, 2.00, 0.9845, 0.0002, 0.0029, 0.015,
  0.50, 18.0, 25.0, 2.00, 2.50, 0.9821, 0.0004, 0.0029, 0.015,
  0.50, 18.0, 25.0, 2.50, 3.00, 0.9791, 0.0008, 0.0029, 0.015,
  0.50, 18.0, 25.0, 3.00, 3.50, 0.9783, 0.0011, 0.0029, 0.015,
  0.50, 18.0, 25.0, 3.50, 4.00, 0.9789, 0.0017, 0.0029, 0.015,
  0.50, 18.0, 25.0, 4.00, 4.50, 0.9804, 0.0075, 0.0029, 0.015,
  0.50, 18.0, 25.0, 4.50, 5.00, 0.9214, 0.0103, 0.0029, 0.015,
  0.50, 25.0, 35.0, 0.00, 0.50, 1.0000, 0.0000, 0.0068, 0.0449,
  0.50, 25.0, 35.0, 0.50, 1.00, 0.9346, 0.0005, 0.0068, 0.0449,
  0.50, 25.0, 35.0, 1.00, 1.50, 0.9548, 0.0006, 0.0068, 0.0449,
  0.50, 25.0, 35.0, 1.50, 2.00, 0.9792, 0.0010, 0.0068, 0.015,
  0.50, 25.0, 35.0, 2.00, 2.50, 0.9829, 0.0016, 0.0068, 0.015,
  0.50, 25.0, 35.0, 2.50, 3.00, 0.9755, 0.0026, 0.0068, 0.015,
  0.50, 25.0, 35.0, 3.00, 3.50, 0.9723, 0.0044, 0.0068, 0.015,
  0.50, 25.0, 35.0, 3.50, 4.00, 0.9628, 0.0109, 0.0068, 0.015,
  0.50, 25.0, 35.0, 4.00, 4.50, 1.0163, 0.0659, 0.0068, 0.015,
  0.50, 25.0, 35.0, 4.50, 5.00, 1.0000, 0.0000, 0.0068, 0.015,
  0.50, 35.0, 40.0, 0.00, 0.50, 0.8236, 0.0107, 0.0097, 0.0449,
  0.50, 35.0, 40.0, 0.50, 1.00, 0.8934, 0.0010, 0.0097, 0.0449,
  0.50, 35.0, 40.0, 1.00, 1.50, 0.9658, 0.0012, 0.0097, 0.0449,
  0.50, 35.0, 40.0, 1.50, 2.00, 0.9867, 0.0015, 0.0097, 0.015,
  0.50, 35.0, 40.0, 2.00, 2.50, 0.9958, 0.0023, 0.0097, 0.015,
  0.50, 35.0, 40.0, 2.50, 3.00, 0.9867, 0.0030, 0.0097, 0.015,
  0.50, 35.0, 40.0, 3.00, 3.50, 0.9802, 0.0058, 0.0097, 0.015,
  0.50, 35.0, 40.0, 3.50, 4.00, 0.9145, 0.0099, 0.0097, 0.015,
  0.50, 35.0, 40.0, 4.00, 4.50, 1.0000, 0.0000, 0.0097, 0.015,
  0.50, 35.0, 40.0, 4.50, 5.00, 1.0000, 0.0000, 0.0097, 0.015,
  0.50, 40.0, 60.0, 0.00, 0.50, 0.9548, 0.0019, 0.0045, 0.0449,
  0.50, 40.0, 60.0, 0.50, 1.00, 0.9448, 0.0005, 0.0045, 0.0449,
  0.50, 40.0, 60.0, 1.00, 1.50, 0.9721, 0.0007, 0.0045, 0.0449,
  0.50, 40.0, 60.0, 1.50, 2.00, 0.9881, 0.0009, 0.0045, 0.015,
  0.50, 40.0, 60.0, 2.00, 2.50, 0.9874, 0.0014, 0.0045, 0.015,
  0.50, 40.0, 60.0, 2.50, 3.00, 0.9810, 0.0027, 0.0045, 0.015,
  0.50, 40.0, 60.0, 3.00, 3.50, 0.9734, 0.0107, 0.0045, 0.015,
  0.50, 40.0, 60.0, 3.50, 4.00, 1.0000, 0.0000, 0.0045, 0.015,
  0.50, 40.0, 60.0, 4.00, 4.50, 1.0000, 0.0000, 0.0045, 0.015,
  0.50, 40.0, 60.0, 4.50, 5.00, 1.0000, 0.0000, 0.0045, 0.015,
  0.50, 60.0, 125.0, 0.00, 0.50, 0.9614, 0.0006, 0.0046, 0.0091,
  0.50, 60.0, 125.0, 0.50, 1.00, 0.9715, 0.0003, 0.0046, 0.0091,
  0.50, 60.0, 125.0, 1.00, 1.50, 0.9807, 0.0005, 0.0046, 0.0091,
  0.50, 60.0, 125.0, 1.50, 2.00, 0.9845, 0.0007, 0.0046, 0.0123,
  0.50, 60.0, 125.0, 2.00, 2.50, 0.9816, 0.0016, 0.0046, 0.0123,
  0.50, 60.0, 125.0, 2.50, 3.00, 0.9423, 0.0032, 0.0046, 0.0123,
  0.50, 60.0, 125.0, 3.00, 3.50, 1.0000, 0.0000, 0.0046, 0.0123,
  0.50, 60.0, 125.0, 3.50, 4.00, 1.0000, 0.0000, 0.0046, 0.0123,
  0.50, 60.0, 125.0, 4.00, 4.50, 1.0000, 0.0000, 0.0046, 0.0123,
  0.50, 60.0, 125.0, 4.50, 5.00, 1.0000, 0.0000, 0.0046, 0.0123,
  0.50, 125.0, 132.0, 0.00, 0.50, 0.7858, 0.0035, 0.0261, 0.0185,
  0.50, 125.0, 132.0, 0.50, 1.00, 1.0575, 0.0026, 0.0261, 0.0185,
  0.50, 125.0, 132.0, 1.00, 1.50, 0.9872, 0.0060, 0.0261, 0.0185,
  0.50, 125.0, 132.0, 1.50, 2.00, 1.0002, 0.0146, 0.0261, 0.0448,
  0.50, 125.0, 132.0, 2.00, 2.50, 1.0073, 0.0503, 0.0261, 0.0448,
  0.50, 125.0, 132.0, 2.50, 3.00, 1.0000, 0.0000, 0.0261, 0.0448,
  0.50, 125.0, 132.0, 3.00, 3.50, 1.0000, 0.0000, 0.0261, 0.0448,
  0.50, 125.0, 132.0, 3.50, 4.00, 1.0000, 0.0000, 0.0261, 0.0448,
  0.50, 125.0, 132.0, 4.00, 4.50, 1.0000, 0.0000, 0.0261, 0.0448,
  0.50, 125.0, 132.0, 4.50, 5.00, 1.0000, 0.0000, 0.0261, 0.0448,
  0.50, 132.0, 151.0, 0.00, 0.50, 1.0231, 0.0040, 0.0071, 0.0185,
  0.50, 132.0, 151.0, 0.50, 1.00, 1.0374, 0.0007, 0.0071, 0.0185,
  0.50, 132.0, 151.0, 1.00, 1.50, 0.9864, 0.0009, 0.0071, 0.0185,
  0.50, 132.0, 151.0, 1.50, 2.00, 0.9691, 0.0018, 0.0071, 0.0448,
  0.50, 132.0, 151.0, 2.00, 2.50, 0.9609, 0.0096, 0.0071, 0.0448,
  0.50, 132.0, 151.0, 2.50, 3.00, 1.0000, 0.0000, 0.0071, 0.0448,
  0.50, 132.0, 151.0, 3.00, 3.50, 1.0000, 0.0000, 0.0071, 0.0448,
  0.50, 132.0, 151.0, 3.50, 4.00, 1.0000, 0.0000, 0.0071, 0.0448,
  0.50, 132.0, 151.0, 4.00, 4.50, 1.0000, 0.0000, 0.0071, 0.0448,
  0.50, 132.0, 151.0, 4.50, 5.00, 1.0000, 0.0000, 0.0071, 0.0448,
  0.60, 18.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0031, 0.0453,
  0.60, 18.0, 25.0, 0.50, 1.00, 0.9731, 0.0004, 0.0031, 0.0453,
  0.60, 18.0, 25.0, 1.00, 1.50, 0.9799, 0.0002, 0.0031, 0.0453,
  0.60, 18.0, 25.0, 1.50, 2.00, 0.9839, 0.0003, 0.0031, 0.0154,
  0.60, 18.0, 25.0, 2.00, 2.50, 0.9813, 0.0005, 0.0031, 0.0154,
  0.60, 18.0, 25.0, 2.50, 3.00, 0.9775, 0.0008, 0.0031, 0.0154,
  0.60, 18.0, 25.0, 3.00, 3.50, 0.9771, 0.0012, 0.0031, 0.0154,
  0.60, 18.0, 25.0, 3.50, 4.00, 0.9758, 0.0017, 0.0031, 0.0154,
  0.60, 18.0, 25.0, 4.00, 4.50, 0.9766, 0.0075, 0.0031, 0.0154,
  0.60, 18.0, 25.0, 4.50, 5.00, 0.9185, 0.0104, 0.0031, 0.0154,
  0.60, 25.0, 35.0, 0.00, 0.50, 1.0000, 0.0000, 0.0072, 0.0453,
  0.60, 25.0, 35.0, 0.50, 1.00, 0.9312, 0.0006, 0.0072, 0.0453,
  0.60, 25.0, 35.0, 1.00, 1.50, 0.9526, 0.0006, 0.0072, 0.0453,
  0.60, 25.0, 35.0, 1.50, 2.00, 0.9791, 0.0010, 0.0072, 0.0154,
  0.60, 25.0, 35.0, 2.00, 2.50, 0.9826, 0.0016, 0.0072, 0.0154,
  0.60, 25.0, 35.0, 2.50, 3.00, 0.9758, 0.0026, 0.0072, 0.0154,
  0.60, 25.0, 35.0, 3.00, 3.50, 0.9713, 0.0044, 0.0072, 0.0154,
  0.60, 25.0, 35.0, 3.50, 4.00, 0.9659, 0.0114, 0.0072, 0.0154,
  0.60, 25.0, 35.0, 4.00, 4.50, 1.0092, 0.0655, 0.0072, 0.0154,
  0.60, 25.0, 35.0, 4.50, 5.00, 1.0000, 0.0000, 0.0072, 0.0154,
  0.60, 35.0, 40.0, 0.00, 0.50, 0.8230, 0.0108, 0.0103, 0.0453,
  0.60, 35.0, 40.0, 0.50, 1.00, 0.8919, 0.0011, 0.0103, 0.0453,
  0.60, 35.0, 40.0, 1.00, 1.50, 0.9648, 0.0012, 0.0103, 0.0453,
  0.60, 35.0, 40.0, 1.50, 2.00, 0.9881, 0.0016, 0.0103, 0.0154,
  0.60, 35.0, 40.0, 2.00, 2.50, 0.9961, 0.0024, 0.0103, 0.0154,
  0.60, 35.0, 40.0, 2.50, 3.00, 0.9882, 0.0032, 0.0103, 0.0154,
  0.60, 35.0, 40.0, 3.00, 3.50, 0.9798, 0.0060, 0.0103, 0.0154,
  0.60, 35.0, 40.0, 3.50, 4.00, 0.9082, 0.0102, 0.0103, 0.0154,
  0.60, 35.0, 40.0, 4.00, 4.50, 1.0000, 0.0000, 0.0103, 0.0154,
  0.60, 35.0, 40.0, 4.50, 5.00, 1.0000, 0.0000, 0.0103, 0.0154,
  0.60, 40.0, 60.0, 0.00, 0.50, 0.9546, 0.0019, 0.0046, 0.0453,
  0.60, 40.0, 60.0, 0.50, 1.00, 0.9435, 0.0005, 0.0046, 0.0453,
  0.60, 40.0, 60.0, 1.00, 1.50, 0.9710, 0.0007, 0.0046, 0.0453,
  0.60, 40.0, 60.0, 1.50, 2.00, 0.9888, 0.0009, 0.0046, 0.0154,
  0.60, 40.0, 60.0, 2.00, 2.50, 0.9879, 0.0015, 0.0046, 0.0154,
  0.60, 40.0, 60.0, 2.50, 3.00, 0.9803, 0.0028, 0.0046, 0.0154,
  0.60, 40.0, 60.0, 3.00, 3.50, 0.9716, 0.0107, 0.0046, 0.0154,
  0.60, 40.0, 60.0, 3.50, 4.00, 1.0000, 0.0000, 0.0046, 0.0154,
  0.60, 40.0, 60.0, 4.00, 4.50, 1.0000, 0.0000, 0.0046, 0.0154,
  0.60, 40.0, 60.0, 4.50, 5.00, 1.0000, 0.0000, 0.0046, 0.0154,
  0.60, 60.0, 125.0, 0.00, 0.50, 0.9606, 0.0006, 0.0049, 0.0088,
  0.60, 60.0, 125.0, 0.50, 1.00, 0.9701, 0.0003, 0.0049, 0.0088,
  0.60, 60.0, 125.0, 1.00, 1.50, 0.9797, 0.0005, 0.0049, 0.0088,
  0.60, 60.0, 125.0, 1.50, 2.00, 0.9846, 0.0007, 0.0049, 0.0122,
  0.60, 60.0, 125.0, 2.00, 2.50, 0.9812, 0.0017, 0.0049, 0.0122,
  0.60, 60.0, 125.0, 2.50, 3.00, 0.9404, 0.0033, 0.0049, 0.0122,
  0.60, 60.0, 125.0, 3.00, 3.50, 1.0000, 0.0000, 0.0049, 0.0122,
  0.60, 60.0, 125.0, 3.50, 4.00, 1.0000, 0.0000, 0.0049, 0.0122,
  0.60, 60.0, 125.0, 4.00, 4.50, 1.0000, 0.0000, 0.0049, 0.0122,
  0.60, 60.0, 125.0, 4.50, 5.00, 1.0000, 0.0000, 0.0049, 0.0122,
  0.60, 125.0, 132.0, 0.00, 0.50, 0.7861, 0.0035, 0.0275, 0.0159,
  0.60, 125.0, 132.0, 0.50, 1.00, 1.0723, 0.0028, 0.0275, 0.0159,
  0.60, 125.0, 132.0, 1.00, 1.50, 0.9835, 0.0061, 0.0275, 0.0159,
  0.60, 125.0, 132.0, 1.50, 2.00, 1.0073, 0.0149, 0.0275, 0.0445,
  0.60, 125.0, 132.0, 2.00, 2.50, 1.0110, 0.0511, 0.0275, 0.0445,
  0.60, 125.0, 132.0, 2.50, 3.00, 1.0000, 0.0000, 0.0275, 0.0445,
  0.60, 125.0, 132.0, 3.00, 3.50, 1.0000, 0.0000, 0.0275, 0.0445,
  0.60, 125.0, 132.0, 3.50, 4.00, 1.0000, 0.0000, 0.0275, 0.0445,
  0.60, 125.0, 132.0, 4.00, 4.50, 1.0000, 0.0000, 0.0275, 0.0445,
  0.60, 125.0, 132.0, 4.50, 5.00, 1.0000, 0.0000, 0.0275, 0.0445,
  0.60, 132.0, 151.0, 0.00, 0.50, 1.0268, 0.0041, 0.0078, 0.0159,
  0.60, 132.0, 151.0, 0.50, 1.00, 1.0457, 0.0007, 0.0078, 0.0159,
  0.60, 132.0, 151.0, 1.00, 1.50, 0.9860, 0.0010, 0.0078, 0.0159,
  0.60, 132.0, 151.0, 1.50, 2.00, 0.9687, 0.0020, 0.0078, 0.0445,
  0.60, 132.0, 151.0, 2.00, 2.50, 0.9575, 0.0100, 0.0078, 0.0445,
  0.60, 132.0, 151.0, 2.50, 3.00, 1.0000, 0.0000, 0.0078, 0.0445,
  0.60, 132.0, 151.0, 3.00, 3.50, 1.0000, 0.0000, 0.0078, 0.0445,
  0.60, 132.0, 151.0, 3.50, 4.00, 1.0000, 0.0000, 0.0078, 0.0445,
  0.60, 132.0, 151.0, 4.00, 4.50, 1.0000, 0.0000, 0.0078, 0.0445,
  0.60, 132.0, 151.0, 4.50, 5.00, 1.0000, 0.0000, 0.0078, 0.0445,
  0.80, 18.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0041, 0.0483,
  0.80, 18.0, 25.0, 0.50, 1.00, 0.9688, 0.0005, 0.0041, 0.0483,
  0.80, 18.0, 25.0, 1.00, 1.50, 0.9755, 0.0003, 0.0041, 0.0483,
  0.80, 18.0, 25.0, 1.50, 2.00, 0.9830, 0.0003, 0.0041, 0.0158,
  0.80, 18.0, 25.0, 2.00, 2.50, 0.9794, 0.0006, 0.0041, 0.0158,
  0.80, 18.0, 25.0, 2.50, 3.00, 0.9747, 0.0009, 0.0041, 0.0158,
  0.80, 18.0, 25.0, 3.00, 3.50, 0.9732, 0.0015, 0.0041, 0.0158,
  0.80, 18.0, 25.0, 3.50, 4.00, 0.9692, 0.0020, 0.0041, 0.0158,
  0.80, 18.0, 25.0, 4.00, 4.50, 0.9708, 0.0084, 0.0041, 0.0158,
  0.80, 18.0, 25.0, 4.50, 5.00, 0.9068, 0.0111, 0.0041, 0.0158,
  0.80, 25.0, 35.0, 0.00, 0.50, 1.0000, 0.0000, 0.0081, 0.0483,
  0.80, 25.0, 35.0, 0.50, 1.00, 0.9228, 0.0006, 0.0081, 0.0483,
  0.80, 25.0, 35.0, 1.00, 1.50, 0.9457, 0.0006, 0.0081, 0.0483,
  0.80, 25.0, 35.0, 1.50, 2.00, 0.9796, 0.0010, 0.0081, 0.0158,
  0.80, 25.0, 35.0, 2.00, 2.50, 0.9821, 0.0017, 0.0081, 0.0158,
  0.80, 25.0, 35.0, 2.50, 3.00, 0.9757, 0.0027, 0.0081, 0.0158,
  0.80, 25.0, 35.0, 3.00, 3.50, 0.9687, 0.0045, 0.0081, 0.0158,
  0.80, 25.0, 35.0, 3.50, 4.00, 0.9630, 0.0116, 0.0081, 0.0158,
  0.80, 25.0, 35.0, 4.00, 4.50, 1.0013, 0.0651, 0.0081, 0.0158,
  0.80, 25.0, 35.0, 4.50, 5.00, 1.0000, 0.0000, 0.0081, 0.0158,
  0.80, 35.0, 40.0, 0.00, 0.50, 0.8209, 0.0109, 0.0120, 0.0483,
  0.80, 35.0, 40.0, 0.50, 1.00, 0.8884, 0.0011, 0.0120, 0.0483,
  0.80, 35.0, 40.0, 1.00, 1.50, 0.9621, 0.0013, 0.0120, 0.0483,
  0.80, 35.0, 40.0, 1.50, 2.00, 0.9911, 0.0017, 0.0120, 0.0158,
  0.80, 35.0, 40.0, 2.00, 2.50, 0.9979, 0.0026, 0.0120, 0.0158,
  0.80, 35.0, 40.0, 2.50, 3.00, 0.9893, 0.0036, 0.0120, 0.0158,
  0.80, 35.0, 40.0, 3.00, 3.50, 0.9830, 0.0073, 0.0120, 0.0158,
  0.80, 35.0, 40.0, 3.50, 4.00, 0.8994, 0.0107, 0.0120, 0.0158,
  0.80, 35.0, 40.0, 4.00, 4.50, 1.0000, 0.0000, 0.0120, 0.0158,
  0.80, 35.0, 40.0, 4.50, 5.00, 1.0000, 0.0000, 0.0120, 0.0158,
  0.80, 40.0, 60.0, 0.00, 0.50, 0.9538, 0.0019, 0.0055, 0.0483,
  0.80, 40.0, 60.0, 0.50, 1.00, 0.9399, 0.0005, 0.0055, 0.0483,
  0.80, 40.0, 60.0, 1.00, 1.50, 0.9677, 0.0007, 0.0055, 0.0483,
  0.80, 40.0, 60.0, 1.50, 2.00, 0.9910, 0.0010, 0.0055, 0.0158,
  0.80, 40.0, 60.0, 2.00, 2.50, 0.9921, 0.0017, 0.0055, 0.0158,
  0.80, 40.0, 60.0, 2.50, 3.00, 0.9845, 0.0033, 0.0055, 0.0158,
  0.80, 40.0, 60.0, 3.00, 3.50, 0.9711, 0.0120, 0.0055, 0.0158,
  0.80, 40.0, 60.0, 3.50, 4.00, 1.0000, 0.0000, 0.0055, 0.0158,
  0.80, 40.0, 60.0, 4.00, 4.50, 1.0000, 0.0000, 0.0055, 0.0158,
  0.80, 40.0, 60.0, 4.50, 5.00, 1.0000, 0.0000, 0.0055, 0.0158,
  0.80, 60.0, 125.0, 0.00, 0.50, 0.9587, 0.0006, 0.0059, 0.0076,
  0.80, 60.0, 125.0, 0.50, 1.00, 0.9666, 0.0003, 0.0059, 0.0076,
  0.80, 60.0, 125.0, 1.00, 1.50, 0.9769, 0.0005, 0.0059, 0.0076,
  0.80, 60.0, 125.0, 1.50, 2.00, 0.9858, 0.0008, 0.0059, 0.0106,
  0.80, 60.0, 125.0, 2.00, 2.50, 0.9825, 0.0019, 0.0059, 0.0106,
  0.80, 60.0, 125.0, 2.50, 3.00, 0.9358, 0.0034, 0.0059, 0.0106,
  0.80, 60.0, 125.0, 3.00, 3.50, 1.0000, 0.0000, 0.0059, 0.0106,
  0.80, 60.0, 125.0, 3.50, 4.00, 1.0000, 0.0000, 0.0059, 0.0106,
  0.80, 60.0, 125.0, 4.00, 4.50, 1.0000, 0.0000, 0.0059, 0.0106,
  0.80, 60.0, 125.0, 4.50, 5.00, 1.0000, 0.0000, 0.0059, 0.0106,
  0.80, 125.0, 132.0, 0.00, 0.50, 0.7911, 0.0037, 0.0309, 0.01,
  0.80, 125.0, 132.0, 0.50, 1.00, 1.1144, 0.0032, 0.0309, 0.01,
  0.80, 125.0, 132.0, 1.00, 1.50, 0.9725, 0.0063, 0.0309, 0.01,
  0.80, 125.0, 132.0, 1.50, 2.00, 1.0125, 0.0154, 0.0309, 0.049,
  0.80, 125.0, 132.0, 2.00, 2.50, 1.0128, 0.0524, 0.0309, 0.049,
  0.80, 125.0, 132.0, 2.50, 3.00, 1.0000, 0.0000, 0.0309, 0.049,
  0.80, 125.0, 132.0, 3.00, 3.50, 1.0000, 0.0000, 0.0309, 0.049,
  0.80, 125.0, 132.0, 3.50, 4.00, 1.0000, 0.0000, 0.0309, 0.049,
  0.80, 125.0, 132.0, 4.00, 4.50, 1.0000, 0.0000, 0.0309, 0.049,
  0.80, 125.0, 132.0, 4.50, 5.00, 1.0000, 0.0000, 0.0309, 0.049,
  0.80, 132.0, 151.0, 0.00, 0.50, 1.0393, 0.0045, 0.0108, 0.01,
  0.80, 132.0, 151.0, 0.50, 1.00, 1.0772, 0.0009, 0.0108, 0.01,
  0.80, 132.0, 151.0, 1.00, 1.50, 0.9840, 0.0011, 0.0108, 0.01,
  0.80, 132.0, 151.0, 1.50, 2.00, 0.9649, 0.0023, 0.0108, 0.049,
  0.80, 132.0, 151.0, 2.00, 2.50, 0.9566, 0.0117, 0.0108, 0.049,
  0.80, 132.0, 151.0, 2.50, 3.00, 1.0000, 0.0000, 0.0108, 0.049,
  0.80, 132.0, 151.0, 3.00, 3.50, 1.0000, 0.0000, 0.0108, 0.049,
  0.80, 132.0, 151.0, 3.50, 4.00, 1.0000, 0.0000, 0.0108, 0.049,
  0.80, 132.0, 151.0, 4.00, 4.50, 1.0000, 0.0000, 0.0108, 0.049,
  0.80, 132.0, 151.0, 4.50, 5.00, 1.0000, 0.0000, 0.0108, 0.049,
  0.90, 18.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0054, 0.0496,
  0.90, 18.0, 25.0, 0.50, 1.00, 0.9641, 0.0006, 0.0054, 0.0496,
  0.90, 18.0, 25.0, 1.00, 1.50, 0.9709, 0.0003, 0.0054, 0.0496,
  0.90, 18.0, 25.0, 1.50, 2.00, 0.9830, 0.0004, 0.0054, 0.0156,
  0.90, 18.0, 25.0, 2.00, 2.50, 0.9785, 0.0007, 0.0054, 0.0156,
  0.90, 18.0, 25.0, 2.50, 3.00, 0.9739, 0.0011, 0.0054, 0.0156,
  0.90, 18.0, 25.0, 3.00, 3.50, 0.9732, 0.0019, 0.0054, 0.0156,
  0.90, 18.0, 25.0, 3.50, 4.00, 0.9700, 0.0030, 0.0054, 0.0156,
  0.90, 18.0, 25.0, 4.00, 4.50, 0.9661, 0.0097, 0.0054, 0.0156,
  0.90, 18.0, 25.0, 4.50, 5.00, 0.8865, 0.0121, 0.0054, 0.0156,
  0.90, 25.0, 35.0, 0.00, 0.50, 1.0000, 0.0000, 0.0091, 0.0496,
  0.90, 25.0, 35.0, 0.50, 1.00, 0.9144, 0.0006, 0.0091, 0.0496,
  0.90, 25.0, 35.0, 1.00, 1.50, 0.9381, 0.0007, 0.0091, 0.0496,
  0.90, 25.0, 35.0, 1.50, 2.00, 0.9805, 0.0010, 0.0091, 0.0156,
  0.90, 25.0, 35.0, 2.00, 2.50, 0.9827, 0.0018, 0.0091, 0.0156,
  0.90, 25.0, 35.0, 2.50, 3.00, 0.9755, 0.0028, 0.0091, 0.0156,
  0.90, 25.0, 35.0, 3.00, 3.50, 0.9676, 0.0047, 0.0091, 0.0156,
  0.90, 25.0, 35.0, 3.50, 4.00, 0.9565, 0.0117, 0.0091, 0.0156,
  0.90, 25.0, 35.0, 4.00, 4.50, 0.9924, 0.0645, 0.0091, 0.0156,
  0.90, 25.0, 35.0, 4.50, 5.00, 1.0000, 0.0000, 0.0091, 0.0156,
  0.90, 35.0, 40.0, 0.00, 0.50, 0.8205, 0.0111, 0.0128, 0.0496,
  0.90, 35.0, 40.0, 0.50, 1.00, 0.8865, 0.0012, 0.0128, 0.0496,
  0.90, 35.0, 40.0, 1.00, 1.50, 0.9607, 0.0014, 0.0128, 0.0496,
  0.90, 35.0, 40.0, 1.50, 2.00, 0.9946, 0.0018, 0.0128, 0.0156,
  0.90, 35.0, 40.0, 2.00, 2.50, 1.0004, 0.0028, 0.0128, 0.0156,
  0.90, 35.0, 40.0, 2.50, 3.00, 0.9902, 0.0039, 0.0128, 0.0156,
  0.90, 35.0, 40.0, 3.00, 3.50, 0.9821, 0.0081, 0.0128, 0.0156,
  0.90, 35.0, 40.0, 3.50, 4.00, 0.8868, 0.0112, 0.0128, 0.0156,
  0.90, 35.0, 40.0, 4.00, 4.50, 1.0000, 0.0000, 0.0128, 0.0156,
  0.90, 35.0, 40.0, 4.50, 5.00, 1.0000, 0.0000, 0.0128, 0.0156,
  0.90, 40.0, 60.0, 0.00, 0.50, 0.9537, 0.0019, 0.0066, 0.0496,
  0.90, 40.0, 60.0, 0.50, 1.00, 0.9369, 0.0005, 0.0066, 0.0496,
  0.90, 40.0, 60.0, 1.00, 1.50, 0.9639, 0.0008, 0.0066, 0.0496,
  0.90, 40.0, 60.0, 1.50, 2.00, 0.9951, 0.0011, 0.0066, 0.0156,
  0.90, 40.0, 60.0, 2.00, 2.50, 0.9957, 0.0019, 0.0066, 0.0156,
  0.90, 40.0, 60.0, 2.50, 3.00, 0.9844, 0.0035, 0.0066, 0.0156,
  0.90, 40.0, 60.0, 3.00, 3.50, 0.9754, 0.0143, 0.0066, 0.0156,
  0.90, 40.0, 60.0, 3.50, 4.00, 1.0000, 0.0000, 0.0066, 0.0156,
  0.90, 40.0, 60.0, 4.00, 4.50, 1.0000, 0.0000, 0.0066, 0.0156,
  0.90, 40.0, 60.0, 4.50, 5.00, 1.0000, 0.0000, 0.0066, 0.0156,
  0.90, 60.0, 125.0, 0.00, 0.50, 0.9573, 0.0006, 0.0070, 0.0067,
  0.90, 60.0, 125.0, 0.50, 1.00, 0.9630, 0.0003, 0.0070, 0.0067,
  0.90, 60.0, 125.0, 1.00, 1.50, 0.9741, 0.0006, 0.0070, 0.0067,
  0.90, 60.0, 125.0, 1.50, 2.00, 0.9872, 0.0009, 0.0070, 0.0089,
  0.90, 60.0, 125.0, 2.00, 2.50, 0.9838, 0.0021, 0.0070, 0.0089,
  0.90, 60.0, 125.0, 2.50, 3.00, 0.9315, 0.0035, 0.0070, 0.0089,
  0.90, 60.0, 125.0, 3.00, 3.50, 1.0000, 0.0000, 0.0070, 0.0089,
  0.90, 60.0, 125.0, 3.50, 4.00, 1.0000, 0.0000, 0.0070, 0.0089,
  0.90, 60.0, 125.0, 4.00, 4.50, 1.0000, 0.0000, 0.0070, 0.0089,
  0.90, 60.0, 125.0, 4.50, 5.00, 1.0000, 0.0000, 0.0070, 0.0089,
  0.90, 125.0, 132.0, 0.00, 0.50, 0.7952, 0.0038, 0.0337, 0.0103,
  0.90, 125.0, 132.0, 0.50, 1.00, 1.1593, 0.0037, 0.0337, 0.0103,
  0.90, 125.0, 132.0, 1.00, 1.50, 0.9670, 0.0067, 0.0337, 0.0103,
  0.90, 125.0, 132.0, 1.50, 2.00, 1.0222, 0.0161, 0.0337, 0.0534,
  0.90, 125.0, 132.0, 2.00, 2.50, 1.0066, 0.0527, 0.0337, 0.0534,
  0.90, 125.0, 132.0, 2.50, 3.00, 1.0000, 0.0000, 0.0337, 0.0534,
  0.90, 125.0, 132.0, 3.00, 3.50, 1.0000, 0.0000, 0.0337, 0.0534,
  0.90, 125.0, 132.0, 3.50, 4.00, 1.0000, 0.0000, 0.0337, 0.0534,
  0.90, 125.0, 132.0, 4.00, 4.50, 1.0000, 0.0000, 0.0337, 0.0534,
  0.90, 125.0, 132.0, 4.50, 5.00, 1.0000, 0.0000, 0.0337, 0.0534,
  0.90, 132.0, 151.0, 0.00, 0.50, 1.0482, 0.0049, 0.0147, 0.0103,
  0.90, 132.0, 151.0, 0.50, 1.00, 1.1289, 0.0012, 0.0147, 0.0103,
  0.90, 132.0, 151.0, 1.00, 1.50, 0.9839, 0.0013, 0.0147, 0.0103,
  0.90, 132.0, 151.0, 1.50, 2.00, 0.9630, 0.0026, 0.0147, 0.0534,
  0.90, 132.0, 151.0, 2.00, 2.50, 0.9645, 0.0139, 0.0147, 0.0534,
  0.90, 132.0, 151.0, 2.50, 3.00, 1.0000, 0.0000, 0.0147, 0.0534,
  0.90, 132.0, 151.0, 3.00, 3.50, 1.0000, 0.0000, 0.0147, 0.0534,
  0.90, 132.0, 151.0, 3.50, 4.00, 1.0000, 0.0000, 0.0147, 0.0534,
  0.90, 132.0, 151.0, 4.00, 4.50, 1.0000, 0.0000, 0.0147, 0.0534,
  0.90, 132.0, 151.0, 4.50, 5.00, 1.0000, 0.0000, 0.0147, 0.0534
};
const float LepEff::fElecEffSVD1[] = {
  0.10, 18.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0039, 0.0536,
  0.10, 18.0, 25.0, 0.50, 1.00, 0.9833, 0.0012, 0.0039, 0.0536,
  0.10, 18.0, 25.0, 1.00, 1.50, 0.9892, 0.0008, 0.0039, 0.0536,
  0.10, 18.0, 25.0, 1.50, 2.00, 0.9878, 0.0009, 0.0039, 0.0168,
  0.10, 18.0, 25.0, 2.00, 2.50, 0.9879, 0.0028, 0.0039, 0.0168,
  0.10, 18.0, 25.0, 2.50, 3.00, 0.9797, 0.0062, 0.0039, 0.0168,
  0.10, 18.0, 25.0, 3.00, 3.50, 0.9565, 0.0075, 0.0039, 0.0168,
  0.10, 18.0, 25.0, 3.50, 4.00, 0.9695, 0.0198, 0.0039, 0.0168,
  0.10, 18.0, 25.0, 4.00, 4.50, 0.9003, 0.0123, 0.0039, 0.0168,
  0.10, 18.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0039, 0.0168,
  0.10, 25.0, 35.0, 0.00, 0.50, 1.0000, 0.0000, 0.0062, 0.0536,
  0.10, 25.0, 35.0, 0.50, 1.00, 0.9570, 0.0028, 0.0062, 0.0536,
  0.10, 25.0, 35.0, 1.00, 1.50, 0.9781, 0.0023, 0.0062, 0.0536,
  0.10, 25.0, 35.0, 1.50, 2.00, 0.9811, 0.0027, 0.0062, 0.0168,
  0.10, 25.0, 35.0, 2.00, 2.50, 0.9761, 0.0038, 0.0062, 0.0168,
  0.10, 25.0, 35.0, 2.50, 3.00, 0.9787, 0.0064, 0.0062, 0.0168,
  0.10, 25.0, 35.0, 3.00, 3.50, 0.9922, 0.0132, 0.0062, 0.0168,
  0.10, 25.0, 35.0, 3.50, 4.00, 1.0213, 0.0375, 0.0062, 0.0168,
  0.10, 25.0, 35.0, 4.00, 4.50, 0.9368, 0.0184, 0.0062, 0.0168,
  0.10, 25.0, 35.0, 4.50, 5.00, 1.0000, 0.0000, 0.0062, 0.0168,
  0.10, 35.0, 40.0, 0.00, 0.50, 0.7986, 0.0691, 0.0206, 0.0536,
  0.10, 35.0, 40.0, 0.50, 1.00, 0.8711, 0.0053, 0.0206, 0.0536,
  0.10, 35.0, 40.0, 1.00, 1.50, 0.9668, 0.0041, 0.0206, 0.0536,
  0.10, 35.0, 40.0, 1.50, 2.00, 0.9847, 0.0040, 0.0206, 0.0168,
  0.10, 35.0, 40.0, 2.00, 2.50, 0.9884, 0.0045, 0.0206, 0.0168,
  0.10, 35.0, 40.0, 2.50, 3.00, 0.9805, 0.0054, 0.0206, 0.0168,
  0.10, 35.0, 40.0, 3.00, 3.50, 1.0110, 0.0208, 0.0206, 0.0168,
  0.10, 35.0, 40.0, 3.50, 4.00, 0.9483, 0.0168, 0.0206, 0.0168,
  0.10, 35.0, 40.0, 4.00, 4.50, 1.0000, 0.0000, 0.0206, 0.0168,
  0.10, 35.0, 40.0, 4.50, 5.00, 1.0000, 0.0000, 0.0206, 0.0168,
  0.10, 40.0, 60.0, 0.00, 0.50, 0.9445, 0.0136, 0.0032, 0.0536,
  0.10, 40.0, 60.0, 0.50, 1.00, 0.9157, 0.0023, 0.0032, 0.0536,
  0.10, 40.0, 60.0, 1.00, 1.50, 0.9722, 0.0021, 0.0032, 0.0536,
  0.10, 40.0, 60.0, 1.50, 2.00, 0.9846, 0.0022, 0.0032, 0.0168,
  0.10, 40.0, 60.0, 2.00, 2.50, 0.9805, 0.0027, 0.0032, 0.0168,
  0.10, 40.0, 60.0, 2.50, 3.00, 0.9724, 0.0037, 0.0032, 0.0168,
  0.10, 40.0, 60.0, 3.00, 3.50, 0.9685, 0.0033, 0.0032, 0.0168,
  0.10, 40.0, 60.0, 3.50, 4.00, 1.0000, 0.0000, 0.0032, 0.0168,
  0.10, 40.0, 60.0, 4.00, 4.50, 1.0000, 0.0000, 0.0032, 0.0168,
  0.10, 40.0, 60.0, 4.50, 5.00, 1.0000, 0.0000, 0.0032, 0.0168,
  0.10, 60.0, 125.0, 0.00, 0.50, 0.9444, 0.0045, 0.0039, 0.0134,
  0.10, 60.0, 125.0, 0.50, 1.00, 0.9704, 0.0014, 0.0039, 0.0134,
  0.10, 60.0, 125.0, 1.00, 1.50, 0.9865, 0.0016, 0.0039, 0.0134,
  0.10, 60.0, 125.0, 1.50, 2.00, 0.9842, 0.0019, 0.0039, 0.0134,
  0.10, 60.0, 125.0, 2.00, 2.50, 0.9767, 0.0029, 0.0039, 0.0134,
  0.10, 60.0, 125.0, 2.50, 3.00, 0.9486, 0.0083, 0.0039, 0.0134,
  0.10, 60.0, 125.0, 3.00, 3.50, 1.0000, 0.0000, 0.0039, 0.0134,
  0.10, 60.0, 125.0, 3.50, 4.00, 1.0000, 0.0000, 0.0039, 0.0134,
  0.10, 60.0, 125.0, 4.00, 4.50, 1.0000, 0.0000, 0.0039, 0.0134,
  0.10, 60.0, 125.0, 4.50, 5.00, 1.0000, 0.0000, 0.0039, 0.0134,
  0.10, 125.0, 132.0, 0.00, 0.50, 0.7380, 0.0293, 0.0809, 0.0401,
  0.10, 125.0, 132.0, 0.50, 1.00, 1.0415, 0.0182, 0.0809, 0.0401,
  0.10, 125.0, 132.0, 1.00, 1.50, 1.0082, 0.0274, 0.0809, 0.0401,
  0.10, 125.0, 132.0, 1.50, 2.00, 1.0518, 0.0659, 0.0809, 0.1085,
  0.10, 125.0, 132.0, 2.00, 2.50, 0.8152, 0.1263, 0.0809, 0.1085,
  0.10, 125.0, 132.0, 2.50, 3.00, 1.0000, 0.0000, 0.0809, 0.1085,
  0.10, 125.0, 132.0, 3.00, 3.50, 1.0000, 0.0000, 0.0809, 0.1085,
  0.10, 125.0, 132.0, 3.50, 4.00, 1.0000, 0.0000, 0.0809, 0.1085,
  0.10, 125.0, 132.0, 4.00, 4.50, 1.0000, 0.0000, 0.0809, 0.1085,
  0.10, 125.0, 132.0, 4.50, 5.00, 1.0000, 0.0000, 0.0809, 0.1085,
  0.10, 132.0, 151.0, 0.00, 0.50, 1.0228, 0.0349, 0.0275, 0.0401,
  0.10, 132.0, 151.0, 0.50, 1.00, 1.0305, 0.0062, 0.0275, 0.0401,
  0.10, 132.0, 151.0, 1.00, 1.50, 1.0037, 0.0069, 0.0275, 0.0401,
  0.10, 132.0, 151.0, 1.50, 2.00, 0.9520, 0.0091, 0.0275, 0.1085,
  0.10, 132.0, 151.0, 2.00, 2.50, 0.9093, 0.0447, 0.0275, 0.1085,
  0.10, 132.0, 151.0, 2.50, 3.00, 1.0000, 0.0000, 0.0275, 0.1085,
  0.10, 132.0, 151.0, 3.00, 3.50, 1.0000, 0.0000, 0.0275, 0.1085,
  0.10, 132.0, 151.0, 3.50, 4.00, 1.0000, 0.0000, 0.0275, 0.1085,
  0.10, 132.0, 151.0, 4.00, 4.50, 1.0000, 0.0000, 0.0275, 0.1085,
  0.10, 132.0, 151.0, 4.50, 5.00, 1.0000, 0.0000, 0.0275, 0.1085,
  0.50, 18.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0048, 0.0576,
  0.50, 18.0, 25.0, 0.50, 1.00, 0.9777, 0.0018, 0.0048, 0.0576,
  0.50, 18.0, 25.0, 1.00, 1.50, 0.9848, 0.0011, 0.0048, 0.0576,
  0.50, 18.0, 25.0, 1.50, 2.00, 0.9881, 0.0013, 0.0048, 0.0168,
  0.50, 18.0, 25.0, 2.00, 2.50, 0.9869, 0.0033, 0.0048, 0.0168,
  0.50, 18.0, 25.0, 2.50, 3.00, 0.9787, 0.0072, 0.0048, 0.0168,
  0.50, 18.0, 25.0, 3.00, 3.50, 0.9581, 0.0119, 0.0048, 0.0168,
  0.50, 18.0, 25.0, 3.50, 4.00, 0.9591, 0.0196, 0.0048, 0.0168,
  0.50, 18.0, 25.0, 4.00, 4.50, 0.8936, 0.0127, 0.0048, 0.0168,
  0.50, 18.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0048, 0.0168,
  0.50, 25.0, 35.0, 0.00, 0.50, 1.0000, 0.0000, 0.0077, 0.0576,
  0.50, 25.0, 35.0, 0.50, 1.00, 0.9476, 0.0030, 0.0077, 0.0576,
  0.50, 25.0, 35.0, 1.00, 1.50, 0.9730, 0.0025, 0.0077, 0.0576,
  0.50, 25.0, 35.0, 1.50, 2.00, 0.9811, 0.0028, 0.0077, 0.0168,
  0.50, 25.0, 35.0, 2.00, 2.50, 0.9770, 0.0039, 0.0077, 0.0168,
  0.50, 25.0, 35.0, 2.50, 3.00, 0.9813, 0.0068, 0.0077, 0.0168,
  0.50, 25.0, 35.0, 3.00, 3.50, 0.9883, 0.0133, 0.0077, 0.0168,
  0.50, 25.0, 35.0, 3.50, 4.00, 1.0161, 0.0373, 0.0077, 0.0168,
  0.50, 25.0, 35.0, 4.00, 4.50, 0.9368, 0.0184, 0.0077, 0.0168,
  0.50, 25.0, 35.0, 4.50, 5.00, 1.0000, 0.0000, 0.0077, 0.0168,
  0.50, 35.0, 40.0, 0.00, 0.50, 0.7889, 0.0683, 0.0271, 0.0576,
  0.50, 35.0, 40.0, 0.50, 1.00, 0.8676, 0.0055, 0.0271, 0.0576,
  0.50, 35.0, 40.0, 1.00, 1.50, 0.9677, 0.0044, 0.0271, 0.0576,
  0.50, 35.0, 40.0, 1.50, 2.00, 0.9871, 0.0043, 0.0271, 0.0168,
  0.50, 35.0, 40.0, 2.00, 2.50, 0.9969, 0.0057, 0.0271, 0.0168,
  0.50, 35.0, 40.0, 2.50, 3.00, 0.9868, 0.0076, 0.0271, 0.0168,
  0.50, 35.0, 40.0, 3.00, 3.50, 1.0153, 0.0243, 0.0271, 0.0168,
  0.50, 35.0, 40.0, 3.50, 4.00, 0.9368, 0.0184, 0.0271, 0.0168,
  0.50, 35.0, 40.0, 4.00, 4.50, 1.0000, 0.0000, 0.0271, 0.0168,
  0.50, 35.0, 40.0, 4.50, 5.00, 1.0000, 0.0000, 0.0271, 0.0168,
  0.50, 40.0, 60.0, 0.00, 0.50, 0.9454, 0.0137, 0.0050, 0.0576,
  0.50, 40.0, 60.0, 0.50, 1.00, 0.9128, 0.0024, 0.0050, 0.0576,
  0.50, 40.0, 60.0, 1.00, 1.50, 0.9714, 0.0024, 0.0050, 0.0576,
  0.50, 40.0, 60.0, 1.50, 2.00, 0.9910, 0.0027, 0.0050, 0.0168,
  0.50, 40.0, 60.0, 2.00, 2.50, 0.9866, 0.0036, 0.0050, 0.0168,
  0.50, 40.0, 60.0, 2.50, 3.00, 0.9745, 0.0055, 0.0050, 0.0168,
  0.50, 40.0, 60.0, 3.00, 3.50, 0.9624, 0.0036, 0.0050, 0.0168,
  0.50, 40.0, 60.0, 3.50, 4.00, 1.0000, 0.0000, 0.0050, 0.0168,
  0.50, 40.0, 60.0, 4.00, 4.50, 1.0000, 0.0000, 0.0050, 0.0168,
  0.50, 40.0, 60.0, 4.50, 5.00, 1.0000, 0.0000, 0.0050, 0.0168,
  0.50, 60.0, 125.0, 0.00, 0.50, 0.9448, 0.0046, 0.0025, 0.0134,
  0.50, 60.0, 125.0, 0.50, 1.00, 0.9720, 0.0016, 0.0025, 0.0134,
  0.50, 60.0, 125.0, 1.00, 1.50, 0.9861, 0.0019, 0.0025, 0.0134,
  0.50, 60.0, 125.0, 1.50, 2.00, 0.9892, 0.0025, 0.0025, 0.0116,
  0.50, 60.0, 125.0, 2.00, 2.50, 0.9773, 0.0038, 0.0025, 0.0116,
  0.50, 60.0, 125.0, 2.50, 3.00, 0.9486, 0.0083, 0.0025, 0.0116,
  0.50, 60.0, 125.0, 3.00, 3.50, 1.0000, 0.0000, 0.0025, 0.0116,
  0.50, 60.0, 125.0, 3.50, 4.00, 1.0000, 0.0000, 0.0025, 0.0116,
  0.50, 60.0, 125.0, 4.00, 4.50, 1.0000, 0.0000, 0.0025, 0.0116,
  0.50, 60.0, 125.0, 4.50, 5.00, 1.0000, 0.0000, 0.0025, 0.0116,
  0.50, 125.0, 132.0, 0.00, 0.50, 0.7454, 0.0306, 0.0866, 0.04,
  0.50, 125.0, 132.0, 0.50, 1.00, 1.1060, 0.0225, 0.0866, 0.04,
  0.50, 125.0, 132.0, 1.00, 1.50, 0.9927, 0.0288, 0.0866, 0.04,
  0.50, 125.0, 132.0, 1.50, 2.00, 1.0790, 0.0710, 0.0866, 0.1151,
  0.50, 125.0, 132.0, 2.00, 2.50, 0.8127, 0.1259, 0.0866, 0.1151,
  0.50, 125.0, 132.0, 2.50, 3.00, 1.0000, 0.0000, 0.0866, 0.1151,
  0.50, 125.0, 132.0, 3.00, 3.50, 1.0000, 0.0000, 0.0866, 0.1151,
  0.50, 125.0, 132.0, 3.50, 4.00, 1.0000, 0.0000, 0.0866, 0.1151,
  0.50, 125.0, 132.0, 4.00, 4.50, 1.0000, 0.0000, 0.0866, 0.1151,
  0.50, 125.0, 132.0, 4.50, 5.00, 1.0000, 0.0000, 0.0866, 0.1151,
  0.50, 132.0, 151.0, 0.00, 0.50, 1.0610, 0.0423, 0.0347, 0.04,
  0.50, 132.0, 151.0, 0.50, 1.00, 1.0708, 0.0088, 0.0347, 0.04,
  0.50, 132.0, 151.0, 1.00, 1.50, 1.0110, 0.0092, 0.0347, 0.04,
  0.50, 132.0, 151.0, 1.50, 2.00, 0.9692, 0.0141, 0.0347, 0.1151,
  0.50, 132.0, 151.0, 2.00, 2.50, 0.8866, 0.0437, 0.0347, 0.1151,
  0.50, 132.0, 151.0, 2.50, 3.00, 1.0000, 0.0000, 0.0347, 0.1151,
  0.50, 132.0, 151.0, 3.00, 3.50, 1.0000, 0.0000, 0.0347, 0.1151,
  0.50, 132.0, 151.0, 3.50, 4.00, 1.0000, 0.0000, 0.0347, 0.1151,
  0.50, 132.0, 151.0, 4.00, 4.50, 1.0000, 0.0000, 0.0347, 0.1151,
  0.50, 132.0, 151.0, 4.50, 5.00, 1.0000, 0.0000, 0.0347, 0.1151,
  0.60, 18.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0052, 0.0554,
  0.60, 18.0, 25.0, 0.50, 1.00, 0.9763, 0.0019, 0.0052, 0.0554,
  0.60, 18.0, 25.0, 1.00, 1.50, 0.9839, 0.0012, 0.0052, 0.0554,
  0.60, 18.0, 25.0, 1.50, 2.00, 0.9886, 0.0015, 0.0052, 0.0172,
  0.60, 18.0, 25.0, 2.00, 2.50, 0.9862, 0.0033, 0.0052, 0.0172,
  0.60, 18.0, 25.0, 2.50, 3.00, 0.9764, 0.0072, 0.0052, 0.0172,
  0.60, 18.0, 25.0, 3.00, 3.50, 0.9597, 0.0130, 0.0052, 0.0172,
  0.60, 18.0, 25.0, 3.50, 4.00, 0.9558, 0.0196, 0.0052, 0.0172,
  0.60, 18.0, 25.0, 4.00, 4.50, 0.8919, 0.0128, 0.0052, 0.0172,
  0.60, 18.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0052, 0.0172,
  0.60, 25.0, 35.0, 0.00, 0.50, 1.0000, 0.0000, 0.0081, 0.0554,
  0.60, 25.0, 35.0, 0.50, 1.00, 0.9458, 0.0030, 0.0081, 0.0554,
  0.60, 25.0, 35.0, 1.00, 1.50, 0.9715, 0.0025, 0.0081, 0.0554,
  0.60, 25.0, 35.0, 1.50, 2.00, 0.9808, 0.0028, 0.0081, 0.0172,
  0.60, 25.0, 35.0, 2.00, 2.50, 0.9764, 0.0039, 0.0081, 0.0172,
  0.60, 25.0, 35.0, 2.50, 3.00, 0.9821, 0.0070, 0.0081, 0.0172,
  0.60, 25.0, 35.0, 3.00, 3.50, 0.9910, 0.0138, 0.0081, 0.0172,
  0.60, 25.0, 35.0, 3.50, 4.00, 1.0149, 0.0373, 0.0081, 0.0172,
  0.60, 25.0, 35.0, 4.00, 4.50, 0.9368, 0.0184, 0.0081, 0.0172,
  0.60, 25.0, 35.0, 4.50, 5.00, 1.0000, 0.0000, 0.0081, 0.0172,
  0.60, 35.0, 40.0, 0.00, 0.50, 0.7861, 0.0680, 0.0335, 0.0554,
  0.60, 35.0, 40.0, 0.50, 1.00, 0.8652, 0.0055, 0.0335, 0.0554,
  0.60, 35.0, 40.0, 1.00, 1.50, 0.9667, 0.0044, 0.0335, 0.0554,
  0.60, 35.0, 40.0, 1.50, 2.00, 0.9879, 0.0044, 0.0335, 0.0172,
  0.60, 35.0, 40.0, 2.00, 2.50, 0.9985, 0.0059, 0.0335, 0.0172,
  0.60, 35.0, 40.0, 2.50, 3.00, 0.9873, 0.0080, 0.0335, 0.0172,
  0.60, 35.0, 40.0, 3.00, 3.50, 1.0259, 0.0276, 0.0335, 0.0172,
  0.60, 35.0, 40.0, 3.50, 4.00, 0.9368, 0.0184, 0.0335, 0.0172,
  0.60, 35.0, 40.0, 4.00, 4.50, 1.0000, 0.0000, 0.0335, 0.0172,
  0.60, 35.0, 40.0, 4.50, 5.00, 1.0000, 0.0000, 0.0335, 0.0172,
  0.60, 40.0, 60.0, 0.00, 0.50, 0.9462, 0.0138, 0.0054, 0.0554,
  0.60, 40.0, 60.0, 0.50, 1.00, 0.9120, 0.0025, 0.0054, 0.0554,
  0.60, 40.0, 60.0, 1.00, 1.50, 0.9729, 0.0025, 0.0054, 0.0554,
  0.60, 40.0, 60.0, 1.50, 2.00, 0.9930, 0.0028, 0.0054, 0.0172,
  0.60, 40.0, 60.0, 2.00, 2.50, 0.9865, 0.0037, 0.0054, 0.0172,
  0.60, 40.0, 60.0, 2.50, 3.00, 0.9753, 0.0059, 0.0054, 0.0172,
  0.60, 40.0, 60.0, 3.00, 3.50, 0.9599, 0.0037, 0.0054, 0.0172,
  0.60, 40.0, 60.0, 3.50, 4.00, 1.0000, 0.0000, 0.0054, 0.0172,
  0.60, 40.0, 60.0, 4.00, 4.50, 1.0000, 0.0000, 0.0054, 0.0172,
  0.60, 40.0, 60.0, 4.50, 5.00, 1.0000, 0.0000, 0.0054, 0.0172,
  0.60, 60.0, 125.0, 0.00, 0.50, 0.9466, 0.0046, 0.0026, 0.0134,
  0.60, 60.0, 125.0, 0.50, 1.00, 0.9709, 0.0016, 0.0026, 0.0134,
  0.60, 60.0, 125.0, 1.00, 1.50, 0.9865, 0.0020, 0.0026, 0.0134,
  0.60, 60.0, 125.0, 1.50, 2.00, 0.9900, 0.0026, 0.0026, 0.0116,
  0.60, 60.0, 125.0, 2.00, 2.50, 0.9761, 0.0038, 0.0026, 0.0116,
  0.60, 60.0, 125.0, 2.50, 3.00, 0.9458, 0.0086, 0.0026, 0.0116,
  0.60, 60.0, 125.0, 3.00, 3.50, 1.0000, 0.0000, 0.0026, 0.0116,
  0.60, 60.0, 125.0, 3.50, 4.00, 1.0000, 0.0000, 0.0026, 0.0116,
  0.60, 60.0, 125.0, 4.00, 4.50, 1.0000, 0.0000, 0.0026, 0.0116,
  0.60, 60.0, 125.0, 4.50, 5.00, 1.0000, 0.0000, 0.0026, 0.0116,
  0.60, 125.0, 132.0, 0.00, 0.50, 0.7530, 0.0315, 0.0855, 0.0402,
  0.60, 125.0, 132.0, 0.50, 1.00, 1.1155, 0.0234, 0.0855, 0.0402,
  0.60, 125.0, 132.0, 1.00, 1.50, 0.9872, 0.0291, 0.0855, 0.0402,
  0.60, 125.0, 132.0, 1.50, 2.00, 1.0745, 0.0707, 0.0855, 0.1202,
  0.60, 125.0, 132.0, 2.00, 2.50, 0.8117, 0.1257, 0.0855, 0.1202,
  0.60, 125.0, 132.0, 2.50, 3.00, 1.0000, 0.0000, 0.0855, 0.1202,
  0.60, 125.0, 132.0, 3.00, 3.50, 1.0000, 0.0000, 0.0855, 0.1202,
  0.60, 125.0, 132.0, 3.50, 4.00, 1.0000, 0.0000, 0.0855, 0.1202,
  0.60, 125.0, 132.0, 4.00, 4.50, 1.0000, 0.0000, 0.0855, 0.1202,
  0.60, 125.0, 132.0, 4.50, 5.00, 1.0000, 0.0000, 0.0855, 0.1202,
  0.60, 132.0, 151.0, 0.00, 0.50, 1.0555, 0.0421, 0.0361, 0.0402,
  0.60, 132.0, 151.0, 0.50, 1.00, 1.0802, 0.0094, 0.0361, 0.0402,
  0.60, 132.0, 151.0, 1.00, 1.50, 1.0120, 0.0097, 0.0361, 0.0402,
  0.60, 132.0, 151.0, 1.50, 2.00, 0.9789, 0.0160, 0.0361, 0.1202,
  0.60, 132.0, 151.0, 2.00, 2.50, 0.8768, 0.0432, 0.0361, 0.1202,
  0.60, 132.0, 151.0, 2.50, 3.00, 1.0000, 0.0000, 0.0361, 0.1202,
  0.60, 132.0, 151.0, 3.00, 3.50, 1.0000, 0.0000, 0.0361, 0.1202,
  0.60, 132.0, 151.0, 3.50, 4.00, 1.0000, 0.0000, 0.0361, 0.1202,
  0.60, 132.0, 151.0, 4.00, 4.50, 1.0000, 0.0000, 0.0361, 0.1202,
  0.60, 132.0, 151.0, 4.50, 5.00, 1.0000, 0.0000, 0.0361, 0.1202,
  0.80, 18.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0087, 0.0551,
  0.80, 18.0, 25.0, 0.50, 1.00, 0.9759, 0.0025, 0.0087, 0.0551,
  0.80, 18.0, 25.0, 1.00, 1.50, 0.9807, 0.0015, 0.0087, 0.0551,
  0.80, 18.0, 25.0, 1.50, 2.00, 0.9889, 0.0017, 0.0087, 0.0185,
  0.80, 18.0, 25.0, 2.00, 2.50, 0.9911, 0.0040, 0.0087, 0.0185,
  0.80, 18.0, 25.0, 2.50, 3.00, 0.9757, 0.0079, 0.0087, 0.0185,
  0.80, 18.0, 25.0, 3.00, 3.50, 0.9636, 0.0161, 0.0087, 0.0185,
  0.80, 18.0, 25.0, 3.50, 4.00, 0.9654, 0.0281, 0.0087, 0.0185,
  0.80, 18.0, 25.0, 4.00, 4.50, 0.8818, 0.0133, 0.0087, 0.0185,
  0.80, 18.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0087, 0.0185,
  0.80, 25.0, 35.0, 0.00, 0.50, 1.0000, 0.0000, 0.0086, 0.0551,
  0.80, 25.0, 35.0, 0.50, 1.00, 0.9374, 0.0032, 0.0086, 0.0551,
  0.80, 25.0, 35.0, 1.00, 1.50, 0.9674, 0.0026, 0.0086, 0.0551,
  0.80, 25.0, 35.0, 1.50, 2.00, 0.9814, 0.0028, 0.0086, 0.0185,
  0.80, 25.0, 35.0, 2.00, 2.50, 0.9760, 0.0040, 0.0086, 0.0185,
  0.80, 25.0, 35.0, 2.50, 3.00, 0.9807, 0.0071, 0.0086, 0.0185,
  0.80, 25.0, 35.0, 3.00, 3.50, 0.9933, 0.0145, 0.0086, 0.0185,
  0.80, 25.0, 35.0, 3.50, 4.00, 1.0076, 0.0370, 0.0086, 0.0185,
  0.80, 25.0, 35.0, 4.00, 4.50, 0.9310, 0.0192, 0.0086, 0.0185,
  0.80, 25.0, 35.0, 4.50, 5.00, 1.0000, 0.0000, 0.0086, 0.0185,
  0.80, 35.0, 40.0, 0.00, 0.50, 0.7952, 0.0706, 0.0341, 0.0551,
  0.80, 35.0, 40.0, 0.50, 1.00, 0.8621, 0.0058, 0.0341, 0.0551,
  0.80, 35.0, 40.0, 1.00, 1.50, 0.9721, 0.0049, 0.0341, 0.0551,
  0.80, 35.0, 40.0, 1.50, 2.00, 0.9945, 0.0048, 0.0341, 0.0185,
  0.80, 35.0, 40.0, 2.00, 2.50, 1.0033, 0.0065, 0.0341, 0.0185,
  0.80, 35.0, 40.0, 2.50, 3.00, 0.9905, 0.0092, 0.0341, 0.0185,
  0.80, 35.0, 40.0, 3.00, 3.50, 1.0211, 0.0274, 0.0341, 0.0185,
  0.80, 35.0, 40.0, 3.50, 4.00, 0.9310, 0.0192, 0.0341, 0.0185,
  0.80, 35.0, 40.0, 4.00, 4.50, 1.0000, 0.0000, 0.0341, 0.0185,
  0.80, 35.0, 40.0, 4.50, 5.00, 1.0000, 0.0000, 0.0341, 0.0185,
  0.80, 40.0, 60.0, 0.00, 0.50, 0.9476, 0.0139, 0.0042, 0.0551,
  0.80, 40.0, 60.0, 0.50, 1.00, 0.9109, 0.0026, 0.0042, 0.0551,
  0.80, 40.0, 60.0, 1.00, 1.50, 0.9742, 0.0028, 0.0042, 0.0551,
  0.80, 40.0, 60.0, 1.50, 2.00, 0.9949, 0.0030, 0.0042, 0.0185,
  0.80, 40.0, 60.0, 2.00, 2.50, 0.9884, 0.0041, 0.0042, 0.0185,
  0.80, 40.0, 60.0, 2.50, 3.00, 0.9828, 0.0079, 0.0042, 0.0185,
  0.80, 40.0, 60.0, 3.00, 3.50, 0.9556, 0.0039, 0.0042, 0.0185,
  0.80, 40.0, 60.0, 3.50, 4.00, 1.0000, 0.0000, 0.0042, 0.0185,
  0.80, 40.0, 60.0, 4.00, 4.50, 1.0000, 0.0000, 0.0042, 0.0185,
  0.80, 40.0, 60.0, 4.50, 5.00, 1.0000, 0.0000, 0.0042, 0.0185,
  0.80, 60.0, 125.0, 0.00, 0.50, 0.9478, 0.0047, 0.0033, 0.0135,
  0.80, 60.0, 125.0, 0.50, 1.00, 0.9701, 0.0018, 0.0033, 0.0135,
  0.80, 60.0, 125.0, 1.00, 1.50, 0.9860, 0.0023, 0.0033, 0.0135,
  0.80, 60.0, 125.0, 1.50, 2.00, 0.9941, 0.0030, 0.0033, 0.0117,
  0.80, 60.0, 125.0, 2.00, 2.50, 0.9761, 0.0043, 0.0033, 0.0117,
  0.80, 60.0, 125.0, 2.50, 3.00, 0.9415, 0.0089, 0.0033, 0.0117,
  0.80, 60.0, 125.0, 3.00, 3.50, 1.0000, 0.0000, 0.0033, 0.0117,
  0.80, 60.0, 125.0, 3.50, 4.00, 1.0000, 0.0000, 0.0033, 0.0117,
  0.80, 60.0, 125.0, 4.00, 4.50, 1.0000, 0.0000, 0.0033, 0.0117,
  0.80, 60.0, 125.0, 4.50, 5.00, 1.0000, 0.0000, 0.0033, 0.0117,
  0.80, 125.0, 132.0, 0.00, 0.50, 0.7653, 0.0334, 0.0864, 0.0398,
  0.80, 125.0, 132.0, 0.50, 1.00, 1.1422, 0.0261, 0.0864, 0.0398,
  0.80, 125.0, 132.0, 1.00, 1.50, 0.9852, 0.0308, 0.0864, 0.0398,
  0.80, 125.0, 132.0, 1.50, 2.00, 1.0767, 0.0720, 0.0864, 0.1266,
  0.80, 125.0, 132.0, 2.00, 2.50, 0.8052, 0.1247, 0.0864, 0.1266,
  0.80, 125.0, 132.0, 2.50, 3.00, 1.0000, 0.0000, 0.0864, 0.1266,
  0.80, 125.0, 132.0, 3.00, 3.50, 1.0000, 0.0000, 0.0864, 0.1266,
  0.80, 125.0, 132.0, 3.50, 4.00, 1.0000, 0.0000, 0.0864, 0.1266,
  0.80, 125.0, 132.0, 4.00, 4.50, 1.0000, 0.0000, 0.0864, 0.1266,
  0.80, 125.0, 132.0, 4.50, 5.00, 1.0000, 0.0000, 0.0864, 0.1266,
  0.80, 132.0, 151.0, 0.00, 0.50, 1.0804, 0.0480, 0.0512, 0.0398,
  0.80, 132.0, 151.0, 0.50, 1.00, 1.0992, 0.0110, 0.0512, 0.0398,
  0.80, 132.0, 151.0, 1.00, 1.50, 1.0189, 0.0113, 0.0512, 0.0398,
  0.80, 132.0, 151.0, 1.50, 2.00, 0.9802, 0.0177, 0.0512, 0.1266,
  0.80, 132.0, 151.0, 2.00, 2.50, 0.8884, 0.0544, 0.0512, 0.1266,
  0.80, 132.0, 151.0, 2.50, 3.00, 1.0000, 0.0000, 0.0512, 0.1266,
  0.80, 132.0, 151.0, 3.00, 3.50, 1.0000, 0.0000, 0.0512, 0.1266,
  0.80, 132.0, 151.0, 3.50, 4.00, 1.0000, 0.0000, 0.0512, 0.1266,
  0.80, 132.0, 151.0, 4.00, 4.50, 1.0000, 0.0000, 0.0512, 0.1266,
  0.80, 132.0, 151.0, 4.50, 5.00, 1.0000, 0.0000, 0.0512, 0.1266,
  0.90, 18.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0103, 0.0567,
  0.90, 18.0, 25.0, 0.50, 1.00, 0.9765, 0.0031, 0.0103, 0.0567,
  0.90, 18.0, 25.0, 1.00, 1.50, 0.9754, 0.0018, 0.0103, 0.0567,
  0.90, 18.0, 25.0, 1.50, 2.00, 0.9936, 0.0022, 0.0103, 0.0184,
  0.90, 18.0, 25.0, 2.00, 2.50, 0.9923, 0.0044, 0.0103, 0.0184,
  0.90, 18.0, 25.0, 2.50, 3.00, 0.9727, 0.0084, 0.0103, 0.0184,
  0.90, 18.0, 25.0, 3.00, 3.50, 0.9684, 0.0188, 0.0103, 0.0184,
  0.90, 18.0, 25.0, 3.50, 4.00, 0.9548, 0.0278, 0.0103, 0.0184,
  0.90, 18.0, 25.0, 4.00, 4.50, 0.8581, 0.0143, 0.0103, 0.0184,
  0.90, 18.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0103, 0.0184,
  0.90, 25.0, 35.0, 0.00, 0.50, 1.0000, 0.0000, 0.0100, 0.0567,
  0.90, 25.0, 35.0, 0.50, 1.00, 0.9306, 0.0034, 0.0100, 0.0567,
  0.90, 25.0, 35.0, 1.00, 1.50, 0.9628, 0.0027, 0.0100, 0.0567,
  0.90, 25.0, 35.0, 1.50, 2.00, 0.9818, 0.0029, 0.0100, 0.0184,
  0.90, 25.0, 35.0, 2.00, 2.50, 0.9772, 0.0042, 0.0100, 0.0184,
  0.90, 25.0, 35.0, 2.50, 3.00, 0.9817, 0.0075, 0.0100, 0.0184,
  0.90, 25.0, 35.0, 3.00, 3.50, 0.9996, 0.0156, 0.0100, 0.0184,
  0.90, 25.0, 35.0, 3.50, 4.00, 1.0167, 0.0406, 0.0100, 0.0184,
  0.90, 25.0, 35.0, 4.00, 4.50, 0.9253, 0.0199, 0.0100, 0.0184,
  0.90, 25.0, 35.0, 4.50, 5.00, 1.0000, 0.0000, 0.0100, 0.0184,
  0.90, 35.0, 40.0, 0.00, 0.50, 0.8188, 0.0765, 0.0379, 0.0567,
  0.90, 35.0, 40.0, 0.50, 1.00, 0.8577, 0.0060, 0.0379, 0.0567,
  0.90, 35.0, 40.0, 1.00, 1.50, 0.9768, 0.0053, 0.0379, 0.0567,
  0.90, 35.0, 40.0, 1.50, 2.00, 0.9985, 0.0051, 0.0379, 0.0184,
  0.90, 35.0, 40.0, 2.00, 2.50, 1.0071, 0.0071, 0.0379, 0.0184,
  0.90, 35.0, 40.0, 2.50, 3.00, 0.9988, 0.0108, 0.0379, 0.0184,
  0.90, 35.0, 40.0, 3.00, 3.50, 1.0247, 0.0304, 0.0379, 0.0184,
  0.90, 35.0, 40.0, 3.50, 4.00, 0.9195, 0.0206, 0.0379, 0.0184,
  0.90, 35.0, 40.0, 4.00, 4.50, 1.0000, 0.0000, 0.0379, 0.0184,
  0.90, 35.0, 40.0, 4.50, 5.00, 1.0000, 0.0000, 0.0379, 0.0184,
  0.90, 40.0, 60.0, 0.00, 0.50, 0.9499, 0.0141, 0.0104, 0.0567,
  0.90, 40.0, 60.0, 0.50, 1.00, 0.9099, 0.0028, 0.0104, 0.0567,
  0.90, 40.0, 60.0, 1.00, 1.50, 0.9754, 0.0030, 0.0104, 0.0567,
  0.90, 40.0, 60.0, 1.50, 2.00, 1.0006, 0.0033, 0.0104, 0.0184,
  0.90, 40.0, 60.0, 2.00, 2.50, 0.9878, 0.0044, 0.0104, 0.0184,
  0.90, 40.0, 60.0, 2.50, 3.00, 0.9841, 0.0088, 0.0104, 0.0184,
  0.90, 40.0, 60.0, 3.00, 3.50, 0.9923, 0.0424, 0.0104, 0.0184,
  0.90, 40.0, 60.0, 3.50, 4.00, 1.0000, 0.0000, 0.0104, 0.0184,
  0.90, 40.0, 60.0, 4.00, 4.50, 1.0000, 0.0000, 0.0104, 0.0184,
  0.90, 40.0, 60.0, 4.50, 5.00, 1.0000, 0.0000, 0.0104, 0.0184,
  0.90, 60.0, 125.0, 0.00, 0.50, 0.9507, 0.0048, 0.0036, 0.0134,
  0.90, 60.0, 125.0, 0.50, 1.00, 0.9698, 0.0019, 0.0036, 0.0134,
  0.90, 60.0, 125.0, 1.00, 1.50, 0.9865, 0.0025, 0.0036, 0.0134,
  0.90, 60.0, 125.0, 1.50, 2.00, 0.9957, 0.0033, 0.0036, 0.0118,
  0.90, 60.0, 125.0, 2.00, 2.50, 0.9771, 0.0049, 0.0036, 0.0118,
  0.90, 60.0, 125.0, 2.50, 3.00, 0.9387, 0.0091, 0.0036, 0.0118,
  0.90, 60.0, 125.0, 3.00, 3.50, 1.0000, 0.0000, 0.0036, 0.0118,
  0.90, 60.0, 125.0, 3.50, 4.00, 1.0000, 0.0000, 0.0036, 0.0118,
  0.90, 60.0, 125.0, 4.00, 4.50, 1.0000, 0.0000, 0.0036, 0.0118,
  0.90, 60.0, 125.0, 4.50, 5.00, 1.0000, 0.0000, 0.0036, 0.0118,
  0.90, 125.0, 132.0, 0.00, 0.50, 0.7732, 0.0348, 0.0856, 0.0402,
  0.90, 125.0, 132.0, 0.50, 1.00, 1.1903, 0.0299, 0.0856, 0.0402,
  0.90, 125.0, 132.0, 1.00, 1.50, 0.9754, 0.0322, 0.0856, 0.0402,
  0.90, 125.0, 132.0, 1.50, 2.00, 1.0789, 0.0733, 0.0856, 0.1331,
  0.90, 125.0, 132.0, 2.00, 2.50, 0.8006, 0.1240, 0.0856, 0.1331,
  0.90, 125.0, 132.0, 2.50, 3.00, 1.0000, 0.0000, 0.0856, 0.1331,
  0.90, 125.0, 132.0, 3.00, 3.50, 1.0000, 0.0000, 0.0856, 0.1331,
  0.90, 125.0, 132.0, 3.50, 4.00, 1.0000, 0.0000, 0.0856, 0.1331,
  0.90, 125.0, 132.0, 4.00, 4.50, 1.0000, 0.0000, 0.0856, 0.1331,
  0.90, 125.0, 132.0, 4.50, 5.00, 1.0000, 0.0000, 0.0856, 0.1331,
  0.90, 132.0, 151.0, 0.00, 0.50, 1.1238, 0.0560, 0.0746, 0.0402,
  0.90, 132.0, 151.0, 0.50, 1.00, 1.1207, 0.0129, 0.0746, 0.0402,
  0.90, 132.0, 151.0, 1.00, 1.50, 1.0218, 0.0128, 0.0746, 0.0402,
  0.90, 132.0, 151.0, 1.50, 2.00, 0.9995, 0.0215, 0.0746, 0.1331,
  0.90, 132.0, 151.0, 2.00, 2.50, 0.8969, 0.0646, 0.0746, 0.1331,
  0.90, 132.0, 151.0, 2.50, 3.00, 1.0000, 0.0000, 0.0746, 0.1331,
  0.90, 132.0, 151.0, 3.00, 3.50, 1.0000, 0.0000, 0.0746, 0.1331,
  0.90, 132.0, 151.0, 3.50, 4.00, 1.0000, 0.0000, 0.0746, 0.1331,
  0.90, 132.0, 151.0, 4.00, 4.50, 1.0000, 0.0000, 0.0746, 0.1331,
  0.90, 132.0, 151.0, 4.50, 5.00, 1.0000, 0.0000, 0.0746, 0.1331
};


// eff(data) / eff(MC) SVD1
const float LepEff::fMuonEffSVD1[] = {
  0.10, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0173, 0.0335,
  0.10, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0173, 0.0335,
  0.10, 17.0, 25.0, 1.00, 1.50, 0.5450, 0.0334, 0.0173, 0.0335,
  0.10, 17.0, 25.0, 1.50, 2.00, 0.7505, 0.0091, 0.0173, 0.0335,
  0.10, 17.0, 25.0, 2.00, 2.50, 0.7327, 0.0059, 0.0173, 0.0335,
  0.10, 17.0, 25.0, 2.50, 3.00, 0.7357, 0.0061, 0.0173, 0.0335,
  0.10, 17.0, 25.0, 3.00, 3.50, 0.6888, 0.0071, 0.0173, 0.0335,
  0.10, 17.0, 25.0, 3.50, 4.00, 0.6880, 0.0107, 0.0173, 0.0335,
  0.10, 17.0, 25.0, 4.00, 4.50, 0.6341, 0.0212, 0.0173, 0.0335,
  0.10, 17.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0173, 0.0335,
  0.10, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0035, 0.0335,
  0.10, 25.0, 37.0, 0.50, 1.00, 0.9260, 0.0124, 0.0035, 0.0335,
  0.10, 25.0, 37.0, 1.00, 1.50, 0.9850, 0.0025, 0.0035, 0.0335,
  0.10, 25.0, 37.0, 1.50, 2.00, 0.9957, 0.0019, 0.0035, 0.0335,
  0.10, 25.0, 37.0, 2.00, 2.50, 0.9846, 0.0021, 0.0035, 0.0335,
  0.10, 25.0, 37.0, 2.50, 3.00, 0.9766, 0.0027, 0.0035, 0.0335,
  0.10, 25.0, 37.0, 3.00, 3.50, 0.9662, 0.0039, 0.0035, 0.0335,
  0.10, 25.0, 37.0, 3.50, 4.00, 0.9699, 0.0071, 0.0035, 0.0335,
  0.10, 25.0, 37.0, 4.00, 4.50, 0.8734, 0.0316, 0.0035, 0.0335,
  0.10, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0035, 0.0335,
  0.10, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0033, 0.0335,
  0.10, 37.0, 51.0, 0.50, 1.00, 0.8784, 0.0067, 0.0033, 0.0335,
  0.10, 37.0, 51.0, 1.00, 1.50, 0.9737, 0.0018, 0.0033, 0.0335,
  0.10, 37.0, 51.0, 1.50, 2.00, 0.9964, 0.0019, 0.0033, 0.0335,
  0.10, 37.0, 51.0, 2.00, 2.50, 0.9982, 0.0027, 0.0033, 0.0335,
  0.10, 37.0, 51.0, 2.50, 3.00, 0.9891, 0.0040, 0.0033, 0.0335,
  0.10, 37.0, 51.0, 3.00, 3.50, 0.9790, 0.0068, 0.0033, 0.0335,
  0.10, 37.0, 51.0, 3.50, 4.00, 1.0238, 0.0408, 0.0033, 0.0335,
  0.10, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0033, 0.0335,
  0.10, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0033, 0.0335,
  0.10, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0012, 0.017,
  0.10, 51.0, 117.0, 0.50, 1.00, 0.9886, 0.0015, 0.0012, 0.017,
  0.10, 51.0, 117.0, 1.00, 1.50, 0.9810, 0.0007, 0.0012, 0.017,
  0.10, 51.0, 117.0, 1.50, 2.00, 0.9759, 0.0010, 0.0012, 0.017,
  0.10, 51.0, 117.0, 2.00, 2.50, 0.9639, 0.0017, 0.0012, 0.017,
  0.10, 51.0, 117.0, 2.50, 3.00, 0.9437, 0.0049, 0.0012, 0.017,
  0.10, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0012, 0.017,
  0.10, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0012, 0.017,
  0.10, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0012, 0.017,
  0.10, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0012, 0.017,
  0.10, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0054, 0.0327,
  0.10, 117.0, 130.0, 0.50, 1.00, 0.9340, 0.0063, 0.0054, 0.0327,
  0.10, 117.0, 130.0, 1.00, 1.50, 0.9828, 0.0025, 0.0054, 0.0327,
  0.10, 117.0, 130.0, 1.50, 2.00, 0.9967, 0.0030, 0.0054, 0.0327,
  0.10, 117.0, 130.0, 2.00, 2.50, 1.0038, 0.0052, 0.0054, 0.0327,
  0.10, 117.0, 130.0, 2.50, 3.00, 1.0387, 0.0122, 0.0054, 0.0327,
  0.10, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0054, 0.0327,
  0.10, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0054, 0.0327,
  0.10, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0054, 0.0327,
  0.10, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0054, 0.0327,
  0.10, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0055, 0.0327,
  0.10, 130.0, 145.0, 0.50, 1.00, 0.9273, 0.0079, 0.0055, 0.0327,
  0.10, 130.0, 145.0, 1.00, 1.50, 1.0214, 0.0025, 0.0055, 0.0327,
  0.10, 130.0, 145.0, 1.50, 2.00, 1.0070, 0.0023, 0.0055, 0.0327,
  0.10, 130.0, 145.0, 2.00, 2.50, 0.9975, 0.0032, 0.0055, 0.0327,
  0.10, 130.0, 145.0, 2.50, 3.00, 0.9973, 0.0056, 0.0055, 0.0327,
  0.10, 130.0, 145.0, 3.00, 3.50, 1.0440, 0.0255, 0.0055, 0.0327,
  0.10, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0055, 0.0327,
  0.10, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0055, 0.0327,
  0.10, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0055, 0.0327,
  0.10, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0179, 0.0327,
  0.10, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0179, 0.0327,
  0.10, 145.0, 150.0, 1.00, 1.50, 0.4505, 0.0318, 0.0179, 0.0327,
  0.10, 145.0, 150.0, 1.50, 2.00, 0.7707, 0.0083, 0.0179, 0.0327,
  0.10, 145.0, 150.0, 2.00, 2.50, 0.8065, 0.0058, 0.0179, 0.0327,
  0.10, 145.0, 150.0, 2.50, 3.00, 0.8325, 0.0080, 0.0179, 0.0327,
  0.10, 145.0, 150.0, 3.00, 3.50, 0.7881, 0.0227, 0.0179, 0.0327,
  0.10, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0179, 0.0327,
  0.10, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0179, 0.0327,
  0.10, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0179, 0.0327,
  0.80, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0195, 0.0355,
  0.80, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0195, 0.0355,
  0.80, 17.0, 25.0, 1.00, 1.50, 0.5406, 0.0334, 0.0195, 0.0355,
  0.80, 17.0, 25.0, 1.50, 2.00, 0.7002, 0.0091, 0.0195, 0.0355,
  0.80, 17.0, 25.0, 2.00, 2.50, 0.6512, 0.0058, 0.0195, 0.0355,
  0.80, 17.0, 25.0, 2.50, 3.00, 0.6955, 0.0061, 0.0195, 0.0355,
  0.80, 17.0, 25.0, 3.00, 3.50, 0.6643, 0.0071, 0.0195, 0.0355,
  0.80, 17.0, 25.0, 3.50, 4.00, 0.6756, 0.0108, 0.0195, 0.0355,
  0.80, 17.0, 25.0, 4.00, 4.50, 0.6166, 0.0212, 0.0195, 0.0355,
  0.80, 17.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0195, 0.0355,
  0.80, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0044, 0.0355,
  0.80, 25.0, 37.0, 0.50, 1.00, 0.8786, 0.0137, 0.0044, 0.0355,
  0.80, 25.0, 37.0, 1.00, 1.50, 0.9593, 0.0029, 0.0044, 0.0355,
  0.80, 25.0, 37.0, 1.50, 2.00, 0.9797, 0.0022, 0.0044, 0.0355,
  0.80, 25.0, 37.0, 2.00, 2.50, 0.9745, 0.0023, 0.0044, 0.0355,
  0.80, 25.0, 37.0, 2.50, 3.00, 0.9671, 0.0030, 0.0044, 0.0355,
  0.80, 25.0, 37.0, 3.00, 3.50, 0.9601, 0.0042, 0.0044, 0.0355,
  0.80, 25.0, 37.0, 3.50, 4.00, 0.9631, 0.0076, 0.0044, 0.0355,
  0.80, 25.0, 37.0, 4.00, 4.50, 0.8603, 0.0318, 0.0044, 0.0355,
  0.80, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0044, 0.0355,
  0.80, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0046, 0.0355,
  0.80, 37.0, 51.0, 0.50, 1.00, 0.8615, 0.0071, 0.0046, 0.0355,
  0.80, 37.0, 51.0, 1.00, 1.50, 0.9649, 0.0022, 0.0046, 0.0355,
  0.80, 37.0, 51.0, 1.50, 2.00, 0.9961, 0.0021, 0.0046, 0.0355,
  0.80, 37.0, 51.0, 2.00, 2.50, 0.9995, 0.0035, 0.0046, 0.0355,
  0.80, 37.0, 51.0, 2.50, 3.00, 0.9960, 0.0057, 0.0046, 0.0355,
  0.80, 37.0, 51.0, 3.00, 3.50, 0.9959, 0.0092, 0.0046, 0.0355,
  0.80, 37.0, 51.0, 3.50, 4.00, 1.0238, 0.0408, 0.0046, 0.0355,
  0.80, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0046, 0.0355,
  0.80, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0046, 0.0355,
  0.80, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0018, 0.0186,
  0.80, 51.0, 117.0, 0.50, 1.00, 0.9685, 0.0018, 0.0018, 0.0186,
  0.80, 51.0, 117.0, 1.00, 1.50, 0.9817, 0.0009, 0.0018, 0.0186,
  0.80, 51.0, 117.0, 1.50, 2.00, 0.9791, 0.0012, 0.0018, 0.0186,
  0.80, 51.0, 117.0, 2.00, 2.50, 0.9664, 0.0021, 0.0018, 0.0186,
  0.80, 51.0, 117.0, 2.50, 3.00, 0.9522, 0.0056, 0.0018, 0.0186,
  0.80, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0018, 0.0186,
  0.80, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0018, 0.0186,
  0.80, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0018, 0.0186,
  0.80, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0018, 0.0186,
  0.80, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0068, 0.0325,
  0.80, 117.0, 130.0, 0.50, 1.00, 0.9105, 0.0068, 0.0068, 0.0325,
  0.80, 117.0, 130.0, 1.00, 1.50, 0.9778, 0.0029, 0.0068, 0.0325,
  0.80, 117.0, 130.0, 1.50, 2.00, 0.9954, 0.0035, 0.0068, 0.0325,
  0.80, 117.0, 130.0, 2.00, 2.50, 0.9975, 0.0079, 0.0068, 0.0325,
  0.80, 117.0, 130.0, 2.50, 3.00, 1.0540, 0.0175, 0.0068, 0.0325,
  0.80, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0068, 0.0325,
  0.80, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0068, 0.0325,
  0.80, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0068, 0.0325,
  0.80, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0068, 0.0325,
  0.80, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0072, 0.0325,
  0.80, 130.0, 145.0, 0.50, 1.00, 0.8622, 0.0089, 0.0072, 0.0325,
  0.80, 130.0, 145.0, 1.00, 1.50, 0.9966, 0.0030, 0.0072, 0.0325,
  0.80, 130.0, 145.0, 1.50, 2.00, 0.9939, 0.0025, 0.0072, 0.0325,
  0.80, 130.0, 145.0, 2.00, 2.50, 0.9838, 0.0035, 0.0072, 0.0325,
  0.80, 130.0, 145.0, 2.50, 3.00, 0.9809, 0.0059, 0.0072, 0.0325,
  0.80, 130.0, 145.0, 3.00, 3.50, 0.9688, 0.0345, 0.0072, 0.0325,
  0.80, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0072, 0.0325,
  0.80, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0072, 0.0325,
  0.80, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0072, 0.0325,
  0.80, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0254, 0.0325,
  0.80, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0254, 0.0325,
  0.80, 145.0, 150.0, 1.00, 1.50, 0.4408, 0.0321, 0.0254, 0.0325,
  0.80, 145.0, 150.0, 1.50, 2.00, 0.7281, 0.0086, 0.0254, 0.0325,
  0.80, 145.0, 150.0, 2.00, 2.50, 0.6949, 0.0062, 0.0254, 0.0325,
  0.80, 145.0, 150.0, 2.50, 3.00, 0.7663, 0.0086, 0.0254, 0.0325,
  0.80, 145.0, 150.0, 3.00, 3.50, 0.7233, 0.0240, 0.0254, 0.0325,
  0.80, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0254, 0.0325,
  0.80, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0254, 0.0325,
  0.80, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0254, 0.0325,
  0.90, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0214, 0.0294,
  0.90, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0214, 0.0294,
  0.90, 17.0, 25.0, 1.00, 1.50, 0.5460, 0.0346, 0.0214, 0.0294,
  0.90, 17.0, 25.0, 1.50, 2.00, 0.6682, 0.0091, 0.0214, 0.0294,
  0.90, 17.0, 25.0, 2.00, 2.50, 0.6055, 0.0058, 0.0214, 0.0294,
  0.90, 17.0, 25.0, 2.50, 3.00, 0.6639, 0.0062, 0.0214, 0.0294,
  0.90, 17.0, 25.0, 3.00, 3.50, 0.6437, 0.0072, 0.0214, 0.0294,
  0.90, 17.0, 25.0, 3.50, 4.00, 0.6614, 0.0108, 0.0214, 0.0294,
  0.90, 17.0, 25.0, 4.00, 4.50, 0.6074, 0.0213, 0.0214, 0.0294,
  0.90, 17.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0214, 0.0294,
  0.90, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0049, 0.0294,
  0.90, 25.0, 37.0, 0.50, 1.00, 0.8347, 0.0141, 0.0049, 0.0294,
  0.90, 25.0, 37.0, 1.00, 1.50, 0.9397, 0.0031, 0.0049, 0.0294,
  0.90, 25.0, 37.0, 1.50, 2.00, 0.9680, 0.0023, 0.0049, 0.0294,
  0.90, 25.0, 37.0, 2.00, 2.50, 0.9661, 0.0025, 0.0049, 0.0294,
  0.90, 25.0, 37.0, 2.50, 3.00, 0.9560, 0.0031, 0.0049, 0.0294,
  0.90, 25.0, 37.0, 3.00, 3.50, 0.9514, 0.0045, 0.0049, 0.0294,
  0.90, 25.0, 37.0, 3.50, 4.00, 0.9590, 0.0083, 0.0049, 0.0294,
  0.90, 25.0, 37.0, 4.00, 4.50, 0.8683, 0.0369, 0.0049, 0.0294,
  0.90, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0049, 0.0294,
  0.90, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0052, 0.0294,
  0.90, 37.0, 51.0, 0.50, 1.00, 0.8481, 0.0075, 0.0052, 0.0294,
  0.90, 37.0, 51.0, 1.00, 1.50, 0.9556, 0.0024, 0.0052, 0.0294,
  0.90, 37.0, 51.0, 1.50, 2.00, 0.9947, 0.0022, 0.0052, 0.0294,
  0.90, 37.0, 51.0, 2.00, 2.50, 0.9878, 0.0037, 0.0052, 0.0294,
  0.90, 37.0, 51.0, 2.50, 3.00, 0.9834, 0.0061, 0.0052, 0.0294,
  0.90, 37.0, 51.0, 3.00, 3.50, 0.9895, 0.0100, 0.0052, 0.0294,
  0.90, 37.0, 51.0, 3.50, 4.00, 1.0445, 0.0442, 0.0052, 0.0294,
  0.90, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0052, 0.0294,
  0.90, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0052, 0.0294,
  0.90, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0022, 0.0217,
  0.90, 51.0, 117.0, 0.50, 1.00, 0.9573, 0.0019, 0.0022, 0.0217,
  0.90, 51.0, 117.0, 1.00, 1.50, 0.9808, 0.0010, 0.0022, 0.0217,
  0.90, 51.0, 117.0, 1.50, 2.00, 0.9823, 0.0013, 0.0022, 0.0217,
  0.90, 51.0, 117.0, 2.00, 2.50, 0.9704, 0.0024, 0.0022, 0.0217,
  0.90, 51.0, 117.0, 2.50, 3.00, 0.9532, 0.0061, 0.0022, 0.0217,
  0.90, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0022, 0.0217,
  0.90, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0022, 0.0217,
  0.90, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0022, 0.0217,
  0.90, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0022, 0.0217,
  0.90, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0068, 0.0327,
  0.90, 117.0, 130.0, 0.50, 1.00, 0.8968, 0.0072, 0.0068, 0.0327,
  0.90, 117.0, 130.0, 1.00, 1.50, 0.9711, 0.0032, 0.0068, 0.0327,
  0.90, 117.0, 130.0, 1.50, 2.00, 0.9880, 0.0037, 0.0068, 0.0327,
  0.90, 117.0, 130.0, 2.00, 2.50, 0.9743, 0.0086, 0.0068, 0.0327,
  0.90, 117.0, 130.0, 2.50, 3.00, 1.0287, 0.0195, 0.0068, 0.0327,
  0.90, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0068, 0.0327,
  0.90, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0068, 0.0327,
  0.90, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0068, 0.0327,
  0.90, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0068, 0.0327,
  0.90, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0093, 0.0327,
  0.90, 130.0, 145.0, 0.50, 1.00, 0.8140, 0.0093, 0.0093, 0.0327,
  0.90, 130.0, 145.0, 1.00, 1.50, 0.9803, 0.0033, 0.0093, 0.0327,
  0.90, 130.0, 145.0, 1.50, 2.00, 0.9873, 0.0026, 0.0093, 0.0327,
  0.90, 130.0, 145.0, 2.00, 2.50, 0.9702, 0.0038, 0.0093, 0.0327,
  0.90, 130.0, 145.0, 2.50, 3.00, 0.9710, 0.0063, 0.0093, 0.0327,
  0.90, 130.0, 145.0, 3.00, 3.50, 0.9203, 0.0404, 0.0093, 0.0327,
  0.90, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0093, 0.0327,
  0.90, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0093, 0.0327,
  0.90, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0093, 0.0327,
  0.90, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0256, 0.0327,
  0.90, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0256, 0.0327,
  0.90, 145.0, 150.0, 1.00, 1.50, 0.4337, 0.0328, 0.0256, 0.0327,
  0.90, 145.0, 150.0, 1.50, 2.00, 0.7070, 0.0089, 0.0256, 0.0327,
  0.90, 145.0, 150.0, 2.00, 2.50, 0.6382, 0.0064, 0.0256, 0.0327,
  0.90, 145.0, 150.0, 2.50, 3.00, 0.7267, 0.0093, 0.0256, 0.0327,
  0.90, 145.0, 150.0, 3.00, 3.50, 0.6940, 0.0255, 0.0256, 0.0327,
  0.90, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0256, 0.0327,
  0.90, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0256, 0.0327,
  0.90, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0256, 0.0327,
  0.95, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0225, 0.0288,
  0.95, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0225, 0.0288,
  0.95, 17.0, 25.0, 1.00, 1.50, 0.5523, 0.0368, 0.0225, 0.0288,
  0.95, 17.0, 25.0, 1.50, 2.00, 0.6419, 0.0094, 0.0225, 0.0288,
  0.95, 17.0, 25.0, 2.00, 2.50, 0.5603, 0.0059, 0.0225, 0.0288,
  0.95, 17.0, 25.0, 2.50, 3.00, 0.6248, 0.0062, 0.0225, 0.0288,
  0.95, 17.0, 25.0, 3.00, 3.50, 0.6170, 0.0072, 0.0225, 0.0288,
  0.95, 17.0, 25.0, 3.50, 4.00, 0.6462, 0.0111, 0.0225, 0.0288,
  0.95, 17.0, 25.0, 4.00, 4.50, 0.5840, 0.0212, 0.0225, 0.0288,
  0.95, 17.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0225, 0.0288,
  0.95, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0065, 0.0288,
  0.95, 25.0, 37.0, 0.50, 1.00, 0.8068, 0.0156, 0.0065, 0.0288,
  0.95, 25.0, 37.0, 1.00, 1.50, 0.9139, 0.0035, 0.0065, 0.0288,
  0.95, 25.0, 37.0, 1.50, 2.00, 0.9495, 0.0026, 0.0065, 0.0288,
  0.95, 25.0, 37.0, 2.00, 2.50, 0.9491, 0.0027, 0.0065, 0.0288,
  0.95, 25.0, 37.0, 2.50, 3.00, 0.9415, 0.0035, 0.0065, 0.0288,
  0.95, 25.0, 37.0, 3.00, 3.50, 0.9314, 0.0049, 0.0065, 0.0288,
  0.95, 25.0, 37.0, 3.50, 4.00, 0.9504, 0.0093, 0.0065, 0.0288,
  0.95, 25.0, 37.0, 4.00, 4.50, 0.8593, 0.0417, 0.0065, 0.0288,
  0.95, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0065, 0.0288,
  0.95, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0059, 0.0288,
  0.95, 37.0, 51.0, 0.50, 1.00, 0.8312, 0.0080, 0.0059, 0.0288,
  0.95, 37.0, 51.0, 1.00, 1.50, 0.9426, 0.0028, 0.0059, 0.0288,
  0.95, 37.0, 51.0, 1.50, 2.00, 0.9901, 0.0023, 0.0059, 0.0288,
  0.95, 37.0, 51.0, 2.00, 2.50, 0.9729, 0.0040, 0.0059, 0.0288,
  0.95, 37.0, 51.0, 2.50, 3.00, 0.9627, 0.0066, 0.0059, 0.0288,
  0.95, 37.0, 51.0, 3.00, 3.50, 0.9699, 0.0110, 0.0059, 0.0288,
  0.95, 37.0, 51.0, 3.50, 4.00, 1.0445, 0.0442, 0.0059, 0.0288,
  0.95, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0059, 0.0288,
  0.95, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0059, 0.0288,
  0.95, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0029, 0.0253,
  0.95, 51.0, 117.0, 0.50, 1.00, 0.9496, 0.0022, 0.0029, 0.0253,
  0.95, 51.0, 117.0, 1.00, 1.50, 0.9773, 0.0011, 0.0029, 0.0253,
  0.95, 51.0, 117.0, 1.50, 2.00, 0.9855, 0.0015, 0.0029, 0.0253,
  0.95, 51.0, 117.0, 2.00, 2.50, 0.9751, 0.0028, 0.0029, 0.0253,
  0.95, 51.0, 117.0, 2.50, 3.00, 0.9528, 0.0069, 0.0029, 0.0253,
  0.95, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0029, 0.0253,
  0.95, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0029, 0.0253,
  0.95, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0029, 0.0253,
  0.95, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0029, 0.0253,
  0.95, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0064, 0.0321,
  0.95, 117.0, 130.0, 0.50, 1.00, 0.8880, 0.0080, 0.0064, 0.0321,
  0.95, 117.0, 130.0, 1.00, 1.50, 0.9588, 0.0036, 0.0064, 0.0321,
  0.95, 117.0, 130.0, 1.50, 2.00, 0.9787, 0.0041, 0.0064, 0.0321,
  0.95, 117.0, 130.0, 2.00, 2.50, 0.9424, 0.0094, 0.0064, 0.0321,
  0.95, 117.0, 130.0, 2.50, 3.00, 0.9940, 0.0217, 0.0064, 0.0321,
  0.95, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0064, 0.0321,
  0.95, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0064, 0.0321,
  0.95, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0064, 0.0321,
  0.95, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0064, 0.0321,
  0.95, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0102, 0.0321,
  0.95, 130.0, 145.0, 0.50, 1.00, 0.7725, 0.0101, 0.0102, 0.0321,
  0.95, 130.0, 145.0, 1.00, 1.50, 0.9470, 0.0036, 0.0102, 0.0321,
  0.95, 130.0, 145.0, 1.50, 2.00, 0.9753, 0.0029, 0.0102, 0.0321,
  0.95, 130.0, 145.0, 2.00, 2.50, 0.9536, 0.0041, 0.0102, 0.0321,
  0.95, 130.0, 145.0, 2.50, 3.00, 0.9575, 0.0070, 0.0102, 0.0321,
  0.95, 130.0, 145.0, 3.00, 3.50, 0.9319, 0.0425, 0.0102, 0.0321,
  0.95, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0102, 0.0321,
  0.95, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0102, 0.0321,
  0.95, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0102, 0.0321,
  0.95, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0269, 0.0321,
  0.95, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0269, 0.0321,
  0.95, 145.0, 150.0, 1.00, 1.50, 0.4092, 0.0330, 0.0269, 0.0321,
  0.95, 145.0, 150.0, 1.50, 2.00, 0.6787, 0.0094, 0.0269, 0.0321,
  0.95, 145.0, 150.0, 2.00, 2.50, 0.5880, 0.0067, 0.0269, 0.0321,
  0.95, 145.0, 150.0, 2.50, 3.00, 0.6743, 0.0096, 0.0269, 0.0321,
  0.95, 145.0, 150.0, 3.00, 3.50, 0.6452, 0.0261, 0.0269, 0.0321,
  0.95, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0269, 0.0321,
  0.95, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0269, 0.0321,
  0.95, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0269, 0.0321,
  0.97, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0284, 0.0255,
  0.97, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0284, 0.0255,
  0.97, 17.0, 25.0, 1.00, 1.50, 0.5477, 0.0379, 0.0284, 0.0255,
  0.97, 17.0, 25.0, 1.50, 2.00, 0.6075, 0.0095, 0.0284, 0.0255,
  0.97, 17.0, 25.0, 2.00, 2.50, 0.5139, 0.0061, 0.0284, 0.0255,
  0.97, 17.0, 25.0, 2.50, 3.00, 0.5762, 0.0063, 0.0284, 0.0255,
  0.97, 17.0, 25.0, 3.00, 3.50, 0.5667, 0.0072, 0.0284, 0.0255,
  0.97, 17.0, 25.0, 3.50, 4.00, 0.6032, 0.0111, 0.0284, 0.0255,
  0.97, 17.0, 25.0, 4.00, 4.50, 0.5526, 0.0213, 0.0284, 0.0255,
  0.97, 17.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0284, 0.0255,
  0.97, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0095, 0.0255,
  0.97, 25.0, 37.0, 0.50, 1.00, 0.7787, 0.0166, 0.0095, 0.0255,
  0.97, 25.0, 37.0, 1.00, 1.50, 0.8904, 0.0038, 0.0095, 0.0255,
  0.97, 25.0, 37.0, 1.50, 2.00, 0.9324, 0.0029, 0.0095, 0.0255,
  0.97, 25.0, 37.0, 2.00, 2.50, 0.9326, 0.0030, 0.0095, 0.0255,
  0.97, 25.0, 37.0, 2.50, 3.00, 0.9259, 0.0039, 0.0095, 0.0255,
  0.97, 25.0, 37.0, 3.00, 3.50, 0.9130, 0.0056, 0.0095, 0.0255,
  0.97, 25.0, 37.0, 3.50, 4.00, 0.9359, 0.0111, 0.0095, 0.0255,
  0.97, 25.0, 37.0, 4.00, 4.50, 0.8642, 0.0538, 0.0095, 0.0255,
  0.97, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0095, 0.0255,
  0.97, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0056, 0.0255,
  0.97, 37.0, 51.0, 0.50, 1.00, 0.7856, 0.0085, 0.0056, 0.0255,
  0.97, 37.0, 51.0, 1.00, 1.50, 0.9230, 0.0031, 0.0056, 0.0255,
  0.97, 37.0, 51.0, 1.50, 2.00, 0.9834, 0.0024, 0.0056, 0.0255,
  0.97, 37.0, 51.0, 2.00, 2.50, 0.9615, 0.0044, 0.0056, 0.0255,
  0.97, 37.0, 51.0, 2.50, 3.00, 0.9426, 0.0073, 0.0056, 0.0255,
  0.97, 37.0, 51.0, 3.00, 3.50, 0.9551, 0.0124, 0.0056, 0.0255,
  0.97, 37.0, 51.0, 3.50, 4.00, 1.0551, 0.0459, 0.0056, 0.0255,
  0.97, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0056, 0.0255,
  0.97, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0056, 0.0255,
  0.97, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0039, 0.0265,
  0.97, 51.0, 117.0, 0.50, 1.00, 0.9268, 0.0025, 0.0039, 0.0265,
  0.97, 51.0, 117.0, 1.00, 1.50, 0.9757, 0.0012, 0.0039, 0.0265,
  0.97, 51.0, 117.0, 1.50, 2.00, 0.9895, 0.0016, 0.0039, 0.0265,
  0.97, 51.0, 117.0, 2.00, 2.50, 0.9780, 0.0032, 0.0039, 0.0265,
  0.97, 51.0, 117.0, 2.50, 3.00, 0.9501, 0.0078, 0.0039, 0.0265,
  0.97, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0039, 0.0265,
  0.97, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0039, 0.0265,
  0.97, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0039, 0.0265,
  0.97, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0039, 0.0265,
  0.97, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0081, 0.0317,
  0.97, 117.0, 130.0, 0.50, 1.00, 0.8505, 0.0088, 0.0081, 0.0317,
  0.97, 117.0, 130.0, 1.00, 1.50, 0.9426, 0.0039, 0.0081, 0.0317,
  0.97, 117.0, 130.0, 1.50, 2.00, 0.9688, 0.0045, 0.0081, 0.0317,
  0.97, 117.0, 130.0, 2.00, 2.50, 0.9137, 0.0103, 0.0081, 0.0317,
  0.97, 117.0, 130.0, 2.50, 3.00, 0.9452, 0.0234, 0.0081, 0.0317,
  0.97, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0081, 0.0317,
  0.97, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0081, 0.0317,
  0.97, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0081, 0.0317,
  0.97, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0081, 0.0317,
  0.97, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0113, 0.0317,
  0.97, 130.0, 145.0, 0.50, 1.00, 0.7366, 0.0107, 0.0113, 0.0317,
  0.97, 130.0, 145.0, 1.00, 1.50, 0.9161, 0.0040, 0.0113, 0.0317,
  0.97, 130.0, 145.0, 1.50, 2.00, 0.9630, 0.0031, 0.0113, 0.0317,
0.97, 130.0, 145.0, 2.00, 2.50, 0.9364, 0.0045, 0.0113, 0.0317,
0.97, 130.0, 145.0, 2.50, 3.00, 0.9412, 0.0077, 0.0113, 0.0317,
0.97, 130.0, 145.0, 3.00, 3.50, 0.9266, 0.0443, 0.0113, 0.0317,
0.97, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0113, 0.0317,
0.97, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0113, 0.0317,
0.97, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0113, 0.0317,
0.97, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0265, 0.0317,
0.97, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0265, 0.0317,
0.97, 145.0, 150.0, 1.00, 1.50, 0.4142, 0.0346, 0.0265, 0.0317,
0.97, 145.0, 150.0, 1.50, 2.00, 0.6507, 0.0098, 0.0265, 0.0317,
0.97, 145.0, 150.0, 2.00, 2.50, 0.5414, 0.0069, 0.0265, 0.0317,
0.97, 145.0, 150.0, 2.50, 3.00, 0.6247, 0.0098, 0.0265, 0.0317,
0.97, 145.0, 150.0, 3.00, 3.50, 0.5887, 0.0271, 0.0265, 0.0317,
0.97, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0265, 0.0317,
0.97, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0265, 0.0317,
0.97, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0265, 0.0317
};


// eff(data) / eff(MC) exp 31..39 and exp 45a run 1-220
const float LepEff::fMuonEff31[] = {
0.10, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0279, 0.0279,
0.10, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0279, 0.0279,
0.10, 17.0, 25.0, 1.00, 1.50, 0.5566, 0.0314, 0.0279, 0.0279,
0.10, 17.0, 25.0, 1.50, 2.00, 0.6606, 0.0063, 0.0279, 0.0279,
0.10, 17.0, 25.0, 2.00, 2.50, 0.5188, 0.0032, 0.0279, 0.0279,
0.10, 17.0, 25.0, 2.50, 3.00, 0.4920, 0.0030, 0.0279, 0.0279,
0.10, 17.0, 25.0, 3.00, 3.50, 0.4649, 0.0035, 0.0279, 0.0279,
0.10, 17.0, 25.0, 3.50, 4.00, 0.4290, 0.0046, 0.0279, 0.0279,
0.10, 17.0, 25.0, 4.00, 4.50, 0.3578, 0.0083, 0.0279, 0.0279,
0.10, 17.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0279, 0.0279,
0.10, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0056, 0.0279,
0.10, 25.0, 37.0, 0.50, 1.00, 0.9108, 0.0081, 0.0056, 0.0279,
0.10, 25.0, 37.0, 1.00, 1.50, 0.9732, 0.0016, 0.0056, 0.0279,
0.10, 25.0, 37.0, 1.50, 2.00, 0.9481, 0.0012, 0.0056, 0.0279,
0.10, 25.0, 37.0, 2.00, 2.50, 0.9205, 0.0014, 0.0056, 0.0279,
0.10, 25.0, 37.0, 2.50, 3.00, 0.8938, 0.0018, 0.0056, 0.0279,
0.10, 25.0, 37.0, 3.00, 3.50, 0.8821, 0.0025, 0.0056, 0.0279,
0.10, 25.0, 37.0, 3.50, 4.00, 0.8475, 0.0042, 0.0056, 0.0279,
0.10, 25.0, 37.0, 4.00, 4.50, 0.7860, 0.0199, 0.0056, 0.0279,
0.10, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0056, 0.0279,
0.10, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0037, 0.0279,
0.10, 37.0, 51.0, 0.50, 1.00, 0.9524, 0.0050, 0.0037, 0.0279,
0.10, 37.0, 51.0, 1.00, 1.50, 0.9785, 0.0012, 0.0037, 0.0279,
0.10, 37.0, 51.0, 1.50, 2.00, 0.9932, 0.0012, 0.0037, 0.0279,
0.10, 37.0, 51.0, 2.00, 2.50, 0.9924, 0.0018, 0.0037, 0.0279,
0.10, 37.0, 51.0, 2.50, 3.00, 0.9864, 0.0027, 0.0037, 0.0279,
0.10, 37.0, 51.0, 3.00, 3.50, 0.9710, 0.0042, 0.0037, 0.0279,
0.10, 37.0, 51.0, 3.50, 4.00, 0.9253, 0.0402, 0.0037, 0.0279,
0.10, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0037, 0.0279,
0.10, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0037, 0.0279,
0.10, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0026, 0.0087,
0.10, 51.0, 117.0, 0.50, 1.00, 1.0212, 0.0010, 0.0026, 0.0087,
0.10, 51.0, 117.0, 1.00, 1.50, 0.9839, 0.0004, 0.0026, 0.0087,
0.10, 51.0, 117.0, 1.50, 2.00, 0.9758, 0.0007, 0.0026, 0.0087,
0.10, 51.0, 117.0, 2.00, 2.50, 0.9641, 0.0011, 0.0026, 0.0087,
0.10, 51.0, 117.0, 2.50, 3.00, 0.9494, 0.0032, 0.0026, 0.0087,
0.10, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0026, 0.0087,
0.10, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0026, 0.0087,
0.10, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0026, 0.0087,
0.10, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0026, 0.0087,
0.10, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0032, 0.0355,
0.10, 117.0, 130.0, 0.50, 1.00, 0.9911, 0.0045, 0.0032, 0.0355,
0.10, 117.0, 130.0, 1.00, 1.50, 0.9813, 0.0016, 0.0032, 0.0355,
0.10, 117.0, 130.0, 1.50, 2.00, 0.9944, 0.0020, 0.0032, 0.0355,
0.10, 117.0, 130.0, 2.00, 2.50, 1.0040, 0.0034, 0.0032, 0.0355,
0.10, 117.0, 130.0, 2.50, 3.00, 1.0229, 0.0079, 0.0032, 0.0355,
0.10, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0032, 0.0355,
0.10, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0032, 0.0355,
0.10, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0032, 0.0355,
0.10, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0032, 0.0355,
0.10, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0056, 0.0355,
0.10, 130.0, 145.0, 0.50, 1.00, 0.8922, 0.0049, 0.0056, 0.0355,
0.10, 130.0, 145.0, 1.00, 1.50, 1.0137, 0.0017, 0.0056, 0.0355,
0.10, 130.0, 145.0, 1.50, 2.00, 0.9974, 0.0015, 0.0056, 0.0355,
0.10, 130.0, 145.0, 2.00, 2.50, 0.9839, 0.0022, 0.0056, 0.0355,
0.10, 130.0, 145.0, 2.50, 3.00, 0.9913, 0.0044, 0.0056, 0.0355,
0.10, 130.0, 145.0, 3.00, 3.50, 0.9856, 0.0288, 0.0056, 0.0355,
0.10, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0056, 0.0355,
0.10, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0056, 0.0355,
0.10, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0056, 0.0355,
0.10, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0270, 0.0355,
0.10, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0270, 0.0355,
0.10, 145.0, 150.0, 1.00, 1.50, 0.4072, 0.0177, 0.0270, 0.0355,
0.10, 145.0, 150.0, 1.50, 2.00, 0.6457, 0.0048, 0.0270, 0.0355,
0.10, 145.0, 150.0, 2.00, 2.50, 0.6616, 0.0035, 0.0270, 0.0355,
0.10, 145.0, 150.0, 2.50, 3.00, 0.6855, 0.0056, 0.0270, 0.0355,
0.10, 145.0, 150.0, 3.00, 3.50, 0.7310, 0.0206, 0.0270, 0.0355,
0.10, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0270, 0.0355,
0.10, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0270, 0.0355,
0.10, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0270, 0.0355,
0.80, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0303, 0.0351,
0.80, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0303, 0.0351,
0.80, 17.0, 25.0, 1.00, 1.50, 0.5675, 0.0329, 0.0303, 0.0351,
0.80, 17.0, 25.0, 1.50, 2.00, 0.6372, 0.0063, 0.0303, 0.0351,
0.80, 17.0, 25.0, 2.00, 2.50, 0.4961, 0.0032, 0.0303, 0.0351,
0.80, 17.0, 25.0, 2.50, 3.00, 0.4728, 0.0030, 0.0303, 0.0351,
0.80, 17.0, 25.0, 3.00, 3.50, 0.4440, 0.0034, 0.0303, 0.0351,
0.80, 17.0, 25.0, 3.50, 4.00, 0.4103, 0.0045, 0.0303, 0.0351,
0.80, 17.0, 25.0, 4.00, 4.50, 0.3410, 0.0081, 0.0303, 0.0351,
0.80, 17.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0303, 0.0351,
0.80, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0064, 0.0351,
0.80, 25.0, 37.0, 0.50, 1.00, 0.8524, 0.0091, 0.0064, 0.0351,
0.80, 25.0, 37.0, 1.00, 1.50, 0.9413, 0.0018, 0.0064, 0.0351,
0.80, 25.0, 37.0, 1.50, 2.00, 0.9268, 0.0013, 0.0064, 0.0351,
0.80, 25.0, 37.0, 2.00, 2.50, 0.9021, 0.0015, 0.0064, 0.0351,
0.80, 25.0, 37.0, 2.50, 3.00, 0.8750, 0.0018, 0.0064, 0.0351,
0.80, 25.0, 37.0, 3.00, 3.50, 0.8639, 0.0026, 0.0064, 0.0351,
0.80, 25.0, 37.0, 3.50, 4.00, 0.8283, 0.0043, 0.0064, 0.0351,
0.80, 25.0, 37.0, 4.00, 4.50, 0.7605, 0.0199, 0.0064, 0.0351,
0.80, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0064, 0.0351,
0.80, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0052, 0.0351,
0.80, 37.0, 51.0, 0.50, 1.00, 0.9412, 0.0053, 0.0052, 0.0351,
0.80, 37.0, 51.0, 1.00, 1.50, 0.9718, 0.0014, 0.0052, 0.0351,
0.80, 37.0, 51.0, 1.50, 2.00, 0.9891, 0.0013, 0.0052, 0.0351,
0.80, 37.0, 51.0, 2.00, 2.50, 0.9852, 0.0022, 0.0052, 0.0351,
0.80, 37.0, 51.0, 2.50, 3.00, 0.9810, 0.0036, 0.0052, 0.0351,
0.80, 37.0, 51.0, 3.00, 3.50, 0.9836, 0.0059, 0.0052, 0.0351,
0.80, 37.0, 51.0, 3.50, 4.00, 0.8309, 0.0511, 0.0052, 0.0351,
0.80, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0052, 0.0351,
0.80, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0052, 0.0351,
0.80, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0029, 0.0058,
0.80, 51.0, 117.0, 0.50, 1.00, 1.0258, 0.0012, 0.0029, 0.0058,
0.80, 51.0, 117.0, 1.00, 1.50, 0.9886, 0.0006, 0.0029, 0.0058,
0.80, 51.0, 117.0, 1.50, 2.00, 0.9779, 0.0008, 0.0029, 0.0058,
0.80, 51.0, 117.0, 2.00, 2.50, 0.9661, 0.0013, 0.0029, 0.0058,
0.80, 51.0, 117.0, 2.50, 3.00, 0.9531, 0.0036, 0.0029, 0.0058,
0.80, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0029, 0.0058,
0.80, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0029, 0.0058,
0.80, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0029, 0.0058,
0.80, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0029, 0.0058,
0.80, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0043, 0.039,
0.80, 117.0, 130.0, 0.50, 1.00, 0.9747, 0.0049, 0.0043, 0.039,
0.80, 117.0, 130.0, 1.00, 1.50, 0.9741, 0.0018, 0.0043, 0.039,
0.80, 117.0, 130.0, 1.50, 2.00, 0.9913, 0.0023, 0.0043, 0.039,
0.80, 117.0, 130.0, 2.00, 2.50, 0.9900, 0.0050, 0.0043, 0.039,
0.80, 117.0, 130.0, 2.50, 3.00, 1.0114, 0.0110, 0.0043, 0.039,
0.80, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0043, 0.039,
0.80, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0043, 0.039,
0.80, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0043, 0.039,
0.80, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0043, 0.039,
0.80, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0068, 0.039,
0.80, 130.0, 145.0, 0.50, 1.00, 0.7890, 0.0054, 0.0068, 0.039,
0.80, 130.0, 145.0, 1.00, 1.50, 0.9797, 0.0020, 0.0068, 0.039,
0.80, 130.0, 145.0, 1.50, 2.00, 0.9807, 0.0016, 0.0068, 0.039,
0.80, 130.0, 145.0, 2.00, 2.50, 0.9663, 0.0023, 0.0068, 0.039,
0.80, 130.0, 145.0, 2.50, 3.00, 0.9759, 0.0046, 0.0068, 0.039,
0.80, 130.0, 145.0, 3.00, 3.50, 0.9680, 0.0304, 0.0068, 0.039,
0.80, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0068, 0.039,
0.80, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0068, 0.039,
0.80, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0068, 0.039,
0.80, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0294, 0.039,
0.80, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0294, 0.039,
0.80, 145.0, 150.0, 1.00, 1.50, 0.3901, 0.0176, 0.0294, 0.039,
0.80, 145.0, 150.0, 1.50, 2.00, 0.6230, 0.0049, 0.0294, 0.039,
0.80, 145.0, 150.0, 2.00, 2.50, 0.6276, 0.0037, 0.0294, 0.039,
0.80, 145.0, 150.0, 2.50, 3.00, 0.6570, 0.0057, 0.0294, 0.039,
0.80, 145.0, 150.0, 3.00, 3.50, 0.6884, 0.0204, 0.0294, 0.039,
0.80, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0294, 0.039,
0.80, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0294, 0.039,
0.80, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0294, 0.039,
0.90, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0326, 0.0403,
0.90, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0326, 0.0403,
0.90, 17.0, 25.0, 1.00, 1.50, 0.5579, 0.0327, 0.0326, 0.0403,
0.90, 17.0, 25.0, 1.50, 2.00, 0.6132, 0.0063, 0.0326, 0.0403,
0.90, 17.0, 25.0, 2.00, 2.50, 0.4828, 0.0033, 0.0326, 0.0403,
0.90, 17.0, 25.0, 2.50, 3.00, 0.4623, 0.0030, 0.0326, 0.0403,
0.90, 17.0, 25.0, 3.00, 3.50, 0.4341, 0.0034, 0.0326, 0.0403,
0.90, 17.0, 25.0, 3.50, 4.00, 0.4013, 0.0045, 0.0326, 0.0403,
0.90, 17.0, 25.0, 4.00, 4.50, 0.3328, 0.0081, 0.0326, 0.0403,
0.90, 17.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0326, 0.0403,
0.90, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0070, 0.0403,
0.90, 25.0, 37.0, 0.50, 1.00, 0.8130, 0.0096, 0.0070, 0.0403,
0.90, 25.0, 37.0, 1.00, 1.50, 0.9142, 0.0020, 0.0070, 0.0403,
0.90, 25.0, 37.0, 1.50, 2.00, 0.9072, 0.0014, 0.0070, 0.0403,
0.90, 25.0, 37.0, 2.00, 2.50, 0.8860, 0.0015, 0.0070, 0.0403,
0.90, 25.0, 37.0, 2.50, 3.00, 0.8615, 0.0019, 0.0070, 0.0403,
0.90, 25.0, 37.0, 3.00, 3.50, 0.8507, 0.0026, 0.0070, 0.0403,
0.90, 25.0, 37.0, 3.50, 4.00, 0.8162, 0.0045, 0.0070, 0.0403,
0.90, 25.0, 37.0, 4.00, 4.50, 0.7386, 0.0203, 0.0070, 0.0403,
0.90, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0070, 0.0403,
0.90, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0061, 0.0403,
0.90, 37.0, 51.0, 0.50, 1.00, 0.9349, 0.0056, 0.0061, 0.0403,
0.90, 37.0, 51.0, 1.00, 1.50, 0.9645, 0.0016, 0.0061, 0.0403,
0.90, 37.0, 51.0, 1.50, 2.00, 0.9849, 0.0014, 0.0061, 0.0403,
0.90, 37.0, 51.0, 2.00, 2.50, 0.9729, 0.0024, 0.0061, 0.0403,
0.90, 37.0, 51.0, 2.50, 3.00, 0.9630, 0.0039, 0.0061, 0.0403,
0.90, 37.0, 51.0, 3.00, 3.50, 0.9771, 0.0066, 0.0061, 0.0403,
0.90, 37.0, 51.0, 3.50, 4.00, 0.8059, 0.0540, 0.0061, 0.0403,
0.90, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0061, 0.0403,
0.90, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0061, 0.0403,
0.90, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0030, 0.0044,
0.90, 51.0, 117.0, 0.50, 1.00, 1.0260, 0.0014, 0.0030, 0.0044,
0.90, 51.0, 117.0, 1.00, 1.50, 0.9901, 0.0006, 0.0030, 0.0044,
0.90, 51.0, 117.0, 1.50, 2.00, 0.9802, 0.0008, 0.0030, 0.0044,
0.90, 51.0, 117.0, 2.00, 2.50, 0.9694, 0.0015, 0.0030, 0.0044,
0.90, 51.0, 117.0, 2.50, 3.00, 0.9575, 0.0039, 0.0030, 0.0044,
0.90, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0030, 0.0044,
0.90, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0030, 0.0044,
0.90, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0030, 0.0044,
0.90, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0030, 0.0044,
0.90, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0046, 0.0429,
0.90, 117.0, 130.0, 0.50, 1.00, 0.9633, 0.0052, 0.0046, 0.0429,
0.90, 117.0, 130.0, 1.00, 1.50, 0.9655, 0.0020, 0.0046, 0.0429,
0.90, 117.0, 130.0, 1.50, 2.00, 0.9833, 0.0024, 0.0046, 0.0429,
0.90, 117.0, 130.0, 2.00, 2.50, 0.9642, 0.0055, 0.0046, 0.0429,
0.90, 117.0, 130.0, 2.50, 3.00, 0.9658, 0.0118, 0.0046, 0.0429,
0.90, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0046, 0.0429,
0.90, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0046, 0.0429,
0.90, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0046, 0.0429,
0.90, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0046, 0.0429,
0.90, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0063, 0.0429,
0.90, 130.0, 145.0, 0.50, 1.00, 0.7397, 0.0057, 0.0063, 0.0429,
0.90, 130.0, 145.0, 1.00, 1.50, 0.9526, 0.0021, 0.0063, 0.0429,
0.90, 130.0, 145.0, 1.50, 2.00, 0.9698, 0.0017, 0.0063, 0.0429,
0.90, 130.0, 145.0, 2.00, 2.50, 0.9514, 0.0024, 0.0063, 0.0429,
0.90, 130.0, 145.0, 2.50, 3.00, 0.9612, 0.0047, 0.0063, 0.0429,
0.90, 130.0, 145.0, 3.00, 3.50, 0.9503, 0.0315, 0.0063, 0.0429,
0.90, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0063, 0.0429,
0.90, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0063, 0.0429,
0.90, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0063, 0.0429,
0.90, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0319, 0.0429,
0.90, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0319, 0.0429,
0.90, 145.0, 150.0, 1.00, 1.50, 0.3813, 0.0177, 0.0319, 0.0429,
0.90, 145.0, 150.0, 1.50, 2.00, 0.6031, 0.0050, 0.0319, 0.0429,
0.90, 145.0, 150.0, 2.00, 2.50, 0.6049, 0.0038, 0.0319, 0.0429,
0.90, 145.0, 150.0, 2.50, 3.00, 0.6383, 0.0058, 0.0319, 0.0429,
0.90, 145.0, 150.0, 3.00, 3.50, 0.6647, 0.0203, 0.0319, 0.0429,
0.90, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0319, 0.0429,
0.90, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0319, 0.0429,
0.90, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0319, 0.0429,
0.95, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0344, 0.0496,
0.95, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0344, 0.0496,
0.95, 17.0, 25.0, 1.00, 1.50, 0.5467, 0.0331, 0.0344, 0.0496,
0.95, 17.0, 25.0, 1.50, 2.00, 0.5934, 0.0064, 0.0344, 0.0496,
0.95, 17.0, 25.0, 2.00, 2.50, 0.4590, 0.0033, 0.0344, 0.0496,
0.95, 17.0, 25.0, 2.50, 3.00, 0.4388, 0.0030, 0.0344, 0.0496,
0.95, 17.0, 25.0, 3.00, 3.50, 0.4119, 0.0034, 0.0344, 0.0496,
0.95, 17.0, 25.0, 3.50, 4.00, 0.3816, 0.0044, 0.0344, 0.0496,
0.95, 17.0, 25.0, 4.00, 4.50, 0.3159, 0.0079, 0.0344, 0.0496,
0.95, 17.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0344, 0.0496,
0.95, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0078, 0.0496,
0.95, 25.0, 37.0, 0.50, 1.00, 0.7655, 0.0102, 0.0078, 0.0496,
0.95, 25.0, 37.0, 1.00, 1.50, 0.8753, 0.0021, 0.0078, 0.0496,
0.95, 25.0, 37.0, 1.50, 2.00, 0.8666, 0.0015, 0.0078, 0.0496,
0.95, 25.0, 37.0, 2.00, 2.50, 0.8454, 0.0016, 0.0078, 0.0496,
0.95, 25.0, 37.0, 2.50, 3.00, 0.8219, 0.0020, 0.0078, 0.0496,
0.95, 25.0, 37.0, 3.00, 3.50, 0.8137, 0.0028, 0.0078, 0.0496,
0.95, 25.0, 37.0, 3.50, 4.00, 0.7750, 0.0047, 0.0078, 0.0496,
0.95, 25.0, 37.0, 4.00, 4.50, 0.7021, 0.0205, 0.0078, 0.0496,
0.95, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0078, 0.0496,
0.95, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0069, 0.0496,
0.95, 37.0, 51.0, 0.50, 1.00, 0.9191, 0.0060, 0.0069, 0.0496,
0.95, 37.0, 51.0, 1.00, 1.50, 0.9501, 0.0018, 0.0069, 0.0496,
0.95, 37.0, 51.0, 1.50, 2.00, 0.9782, 0.0015, 0.0069, 0.0496,
0.95, 37.0, 51.0, 2.00, 2.50, 0.9511, 0.0025, 0.0069, 0.0496,
0.95, 37.0, 51.0, 2.50, 3.00, 0.9327, 0.0041, 0.0069, 0.0496,
0.95, 37.0, 51.0, 3.00, 3.50, 0.9554, 0.0073, 0.0069, 0.0496,
0.95, 37.0, 51.0, 3.50, 4.00, 0.8046, 0.0556, 0.0069, 0.0496,
0.95, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0069, 0.0496,
0.95, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0069, 0.0496,
0.95, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0031, 0.002,
0.95, 51.0, 117.0, 0.50, 1.00, 1.0321, 0.0015, 0.0031, 0.002,
0.95, 51.0, 117.0, 1.00, 1.50, 0.9906, 0.0007, 0.0031, 0.002,
0.95, 51.0, 117.0, 1.50, 2.00, 0.9841, 0.0009, 0.0031, 0.002,
0.95, 51.0, 117.0, 2.00, 2.50, 0.9749, 0.0018, 0.0031, 0.002,
0.95, 51.0, 117.0, 2.50, 3.00, 0.9635, 0.0044, 0.0031, 0.002,
0.95, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0031, 0.002,
0.95, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0031, 0.002,
0.95, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0031, 0.002,
0.95, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0031, 0.002,
0.95, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0059, 0.0508,
0.95, 117.0, 130.0, 0.50, 1.00, 0.9603, 0.0058, 0.0059, 0.0508,
0.95, 117.0, 130.0, 1.00, 1.50, 0.9504, 0.0023, 0.0059, 0.0508,
0.95, 117.0, 130.0, 1.50, 2.00, 0.9688, 0.0026, 0.0059, 0.0508,
0.95, 117.0, 130.0, 2.00, 2.50, 0.9203, 0.0060, 0.0059, 0.0508,
0.95, 117.0, 130.0, 2.50, 3.00, 0.9099, 0.0126, 0.0059, 0.0508,
0.95, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0059, 0.0508,
0.95, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0059, 0.0508,
0.95, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0059, 0.0508,
0.95, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0059, 0.0508,
0.95, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0078, 0.0508,
0.95, 130.0, 145.0, 0.50, 1.00, 0.6879, 0.0061, 0.0078, 0.0508,
0.95, 130.0, 145.0, 1.00, 1.50, 0.9117, 0.0023, 0.0078, 0.0508,
0.95, 130.0, 145.0, 1.50, 2.00, 0.9514, 0.0018, 0.0078, 0.0508,
0.95, 130.0, 145.0, 2.00, 2.50, 0.9224, 0.0025, 0.0078, 0.0508,
0.95, 130.0, 145.0, 2.50, 3.00, 0.9356, 0.0048, 0.0078, 0.0508,
0.95, 130.0, 145.0, 3.00, 3.50, 0.9344, 0.0329, 0.0078, 0.0508,
0.95, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0078, 0.0508,
0.95, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0078, 0.0508,
0.95, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0078, 0.0508,
0.95, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0329, 0.0508,
0.95, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0329, 0.0508,
0.95, 145.0, 150.0, 1.00, 1.50, 0.3713, 0.0183, 0.0329, 0.0508,
0.95, 145.0, 150.0, 1.50, 2.00, 0.5812, 0.0051, 0.0329, 0.0508,
0.95, 145.0, 150.0, 2.00, 2.50, 0.5745, 0.0039, 0.0329, 0.0508,
0.95, 145.0, 150.0, 2.50, 3.00, 0.6116, 0.0059, 0.0329, 0.0508,
0.95, 145.0, 150.0, 3.00, 3.50, 0.6274, 0.0201, 0.0329, 0.0508,
0.95, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0329, 0.0508,
0.95, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0329, 0.0508,
0.95, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0329, 0.0508,
0.97, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0399, 0.066,
0.97, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0399, 0.066,
0.97, 17.0, 25.0, 1.00, 1.50, 0.5390, 0.0341, 0.0399, 0.066,
0.97, 17.0, 25.0, 1.50, 2.00, 0.5516, 0.0063, 0.0399, 0.066,
0.97, 17.0, 25.0, 2.00, 2.50, 0.3884, 0.0031, 0.0399, 0.066,
0.97, 17.0, 25.0, 2.50, 3.00, 0.3734, 0.0028, 0.0399, 0.066,
0.97, 17.0, 25.0, 3.00, 3.50, 0.3490, 0.0031, 0.0399, 0.066,
0.97, 17.0, 25.0, 3.50, 4.00, 0.3246, 0.0041, 0.0399, 0.066,
0.97, 17.0, 25.0, 4.00, 4.50, 0.2644, 0.0072, 0.0399, 0.066,
0.97, 17.0, 25.0, 4.50, 5.00, 1.0000, 0.0000, 0.0399, 0.066,
0.97, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0092, 0.066,
0.97, 25.0, 37.0, 0.50, 1.00, 0.7218, 0.0106, 0.0092, 0.066,
0.97, 25.0, 37.0, 1.00, 1.50, 0.8284, 0.0023, 0.0092, 0.066,
0.97, 25.0, 37.0, 1.50, 2.00, 0.7956, 0.0015, 0.0092, 0.066,
0.97, 25.0, 37.0, 2.00, 2.50, 0.7633, 0.0016, 0.0092, 0.066,
0.97, 25.0, 37.0, 2.50, 3.00, 0.7323, 0.0021, 0.0092, 0.066,
0.97, 25.0, 37.0, 3.00, 3.50, 0.7145, 0.0029, 0.0092, 0.066,
0.97, 25.0, 37.0, 3.50, 4.00, 0.6579, 0.0050, 0.0092, 0.066,
0.97, 25.0, 37.0, 4.00, 4.50, 0.5689, 0.0226, 0.0092, 0.066,
0.97, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0092, 0.066,
0.97, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0084, 0.066,
0.97, 37.0, 51.0, 0.50, 1.00, 0.8886, 0.0065, 0.0084, 0.066,
0.97, 37.0, 51.0, 1.00, 1.50, 0.9327, 0.0020, 0.0084, 0.066,
0.97, 37.0, 51.0, 1.50, 2.00, 0.9679, 0.0015, 0.0084, 0.066,
0.97, 37.0, 51.0, 2.00, 2.50, 0.9301, 0.0027, 0.0084, 0.066,
0.97, 37.0, 51.0, 2.50, 3.00, 0.9048, 0.0045, 0.0084, 0.066,
0.97, 37.0, 51.0, 3.00, 3.50, 0.9339, 0.0080, 0.0084, 0.066,
0.97, 37.0, 51.0, 3.50, 4.00, 0.7822, 0.0584, 0.0084, 0.066,
0.97, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0084, 0.066,
0.97, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0084, 0.066,
0.97, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0035, 0.0018,
0.97, 51.0, 117.0, 0.50, 1.00, 1.0217, 0.0018, 0.0035, 0.0018,
0.97, 51.0, 117.0, 1.00, 1.50, 0.9928, 0.0008, 0.0035, 0.0018,
0.97, 51.0, 117.0, 1.50, 2.00, 0.9878, 0.0010, 0.0035, 0.0018,
0.97, 51.0, 117.0, 2.00, 2.50, 0.9796, 0.0020, 0.0035, 0.0018,
0.97, 51.0, 117.0, 2.50, 3.00, 0.9648, 0.0050, 0.0035, 0.0018,
0.97, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0035, 0.0018,
0.97, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0035, 0.0018,
0.97, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0035, 0.0018,
0.97, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0035, 0.0018,
0.97, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0071, 0.0611,
0.97, 117.0, 130.0, 0.50, 1.00, 0.9310, 0.0065, 0.0071, 0.0611,
0.97, 117.0, 130.0, 1.00, 1.50, 0.9353, 0.0025, 0.0071, 0.0611,
0.97, 117.0, 130.0, 1.50, 2.00, 0.9512, 0.0029, 0.0071, 0.0611,
0.97, 117.0, 130.0, 2.00, 2.50, 0.8954, 0.0066, 0.0071, 0.0611,
0.97, 117.0, 130.0, 2.50, 3.00, 0.8570, 0.0134, 0.0071, 0.0611,
0.97, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0071, 0.0611,
0.97, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0071, 0.0611,
0.97, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0071, 0.0611,
0.97, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0071, 0.0611,
0.97, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0086, 0.0611,
0.97, 130.0, 145.0, 0.50, 1.00, 0.6478, 0.0064, 0.0086, 0.0611,
0.97, 130.0, 145.0, 1.00, 1.50, 0.8691, 0.0025, 0.0086, 0.0611,
0.97, 130.0, 145.0, 1.50, 2.00, 0.9262, 0.0020, 0.0086, 0.0611,
0.97, 130.0, 145.0, 2.00, 2.50, 0.8847, 0.0027, 0.0086, 0.0611,
0.97, 130.0, 145.0, 2.50, 3.00, 0.9022, 0.0051, 0.0086, 0.0611,
0.97, 130.0, 145.0, 3.00, 3.50, 0.8749, 0.0356, 0.0086, 0.0611,
0.97, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0086, 0.0611,
0.97, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0086, 0.0611,
0.97, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0086, 0.0611,
0.97, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0374, 0.0611,
0.97, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0374, 0.0611,
0.97, 145.0, 150.0, 1.00, 1.50, 0.3617, 0.0186, 0.0374, 0.0611,
0.97, 145.0, 150.0, 1.50, 2.00, 0.5590, 0.0053, 0.0374, 0.0611,
0.97, 145.0, 150.0, 2.00, 2.50, 0.5225, 0.0039, 0.0374, 0.0611,
0.97, 145.0, 150.0, 2.50, 3.00, 0.5553, 0.0059, 0.0374, 0.0611,
0.97, 145.0, 150.0, 3.00, 3.50, 0.5646, 0.0199, 0.0374, 0.0611,
0.97, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0374, 0.0611,
0.97, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0374, 0.0611,
0.97, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0374, 0.0611
};



// eff(data) / eff(MC) exp 41..49 except exp 45a run 1-220
const float LepEff::fMuonEff41[] = {
  0.10, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0753, 0.0279,
  0.10, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0753, 0.0279,
0.10, 17.0, 25.0, 1.00, 1.50, 0.4359, 0.0276, 0.0753, 0.0279,
0.10, 17.0, 25.0, 1.50, 2.00, 0.5270, 0.0057, 0.0753, 0.0279,
0.10, 17.0, 25.0, 2.00, 2.50, 0.3954, 0.0028, 0.0753, 0.0279,
0.10, 17.0, 25.0, 2.50, 3.00, 0.3735, 0.0026, 0.0753, 0.0279,
0.10, 17.0, 25.0, 3.00, 3.50, 0.3446, 0.0029, 0.0753, 0.0279,
0.10, 17.0, 25.0, 3.50, 4.00, 0.3273, 0.0040, 0.0753, 0.0279,
0.10, 17.0, 25.0, 4.00, 4.50, 0.2510, 0.0066, 0.0753, 0.0279,
0.10, 17.0, 25.0, 4.50, 5.00, 0.1388, 0.1672, 0.0753, 0.0279,
0.10, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0158, 0.0279,
0.10, 25.0, 37.0, 0.50, 1.00, 0.8840, 0.0090, 0.0158, 0.0279,
0.10, 25.0, 37.0, 1.00, 1.50, 0.9624, 0.0019, 0.0158, 0.0279,
0.10, 25.0, 37.0, 1.50, 2.00, 0.9137, 0.0014, 0.0158, 0.0279,
0.10, 25.0, 37.0, 2.00, 2.50, 0.8795, 0.0016, 0.0158, 0.0279,
0.10, 25.0, 37.0, 2.50, 3.00, 0.8532, 0.0020, 0.0158, 0.0279,
0.10, 25.0, 37.0, 3.00, 3.50, 0.8346, 0.0027, 0.0158, 0.0279,
0.10, 25.0, 37.0, 3.50, 4.00, 0.8054, 0.0046, 0.0158, 0.0279,
0.10, 25.0, 37.0, 4.00, 4.50, 0.6792, 0.0182, 0.0158, 0.0279,
0.10, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0158, 0.0279,
0.10, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0036, 0.0279,
0.10, 37.0, 51.0, 0.50, 1.00, 0.9718, 0.0061, 0.0036, 0.0279,
0.10, 37.0, 51.0, 1.00, 1.50, 0.9752, 0.0014, 0.0036, 0.0279,
0.10, 37.0, 51.0, 1.50, 2.00, 0.9945, 0.0015, 0.0036, 0.0279,
0.10, 37.0, 51.0, 2.00, 2.50, 0.9933, 0.0021, 0.0036, 0.0279,
0.10, 37.0, 51.0, 2.50, 3.00, 0.9761, 0.0029, 0.0036, 0.0279,
0.10, 37.0, 51.0, 3.00, 3.50, 0.9811, 0.0050, 0.0036, 0.0279,
0.10, 37.0, 51.0, 3.50, 4.00, 1.0662, 0.0322, 0.0036, 0.0279,
0.10, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0036, 0.0279,
0.10, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0036, 0.0279,
0.10, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0028, 0.0087,
0.10, 51.0, 117.0, 0.50, 1.00, 1.0257, 0.0012, 0.0028, 0.0087,
0.10, 51.0, 117.0, 1.00, 1.50, 0.9795, 0.0005, 0.0028, 0.0087,
0.10, 51.0, 117.0, 1.50, 2.00, 0.9726, 0.0009, 0.0028, 0.0087,
0.10, 51.0, 117.0, 2.00, 2.50, 0.9562, 0.0014, 0.0028, 0.0087,
0.10, 51.0, 117.0, 2.50, 3.00, 0.9282, 0.0036, 0.0028, 0.0087,
0.10, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0028, 0.0087,
0.10, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0028, 0.0087,
0.10, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0028, 0.0087,
0.10, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0028, 0.0087,
0.10, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0025, 0.0355,
0.10, 117.0, 130.0, 0.50, 1.00, 0.9991, 0.0051, 0.0025, 0.0355,
0.10, 117.0, 130.0, 1.00, 1.50, 0.9804, 0.0018, 0.0025, 0.0355,
0.10, 117.0, 130.0, 1.50, 2.00, 0.9957, 0.0023, 0.0025, 0.0355,
0.10, 117.0, 130.0, 2.00, 2.50, 1.0010, 0.0039, 0.0025, 0.0355,
0.10, 117.0, 130.0, 2.50, 3.00, 1.0314, 0.0087, 0.0025, 0.0355,
0.10, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0025, 0.0355,
0.10, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0025, 0.0355,
0.10, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0025, 0.0355,
0.10, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0025, 0.0355,
0.10, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0072, 0.0355,
0.10, 130.0, 145.0, 0.50, 1.00, 0.8676, 0.0053, 0.0072, 0.0355,
0.10, 130.0, 145.0, 1.00, 1.50, 1.0052, 0.0019, 0.0072, 0.0355,
0.10, 130.0, 145.0, 1.50, 2.00, 0.9836, 0.0017, 0.0072, 0.0355,
0.10, 130.0, 145.0, 2.00, 2.50, 0.9739, 0.0026, 0.0072, 0.0355,
0.10, 130.0, 145.0, 2.50, 3.00, 0.9699, 0.0049, 0.0072, 0.0355,
0.10, 130.0, 145.0, 3.00, 3.50, 0.9616, 0.0256, 0.0072, 0.0355,
0.10, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0072, 0.0355,
0.10, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0072, 0.0355,
0.10, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0072, 0.0355,
0.10, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0607, 0.0355,
0.10, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0607, 0.0355,
0.10, 145.0, 150.0, 1.00, 1.50, 0.2987, 0.0143, 0.0607, 0.0355,
0.10, 145.0, 150.0, 1.50, 2.00, 0.5188, 0.0044, 0.0607, 0.0355,
0.10, 145.0, 150.0, 2.00, 2.50, 0.5295, 0.0033, 0.0607, 0.0355,
0.10, 145.0, 150.0, 2.50, 3.00, 0.5416, 0.0050, 0.0607, 0.0355,
0.10, 145.0, 150.0, 3.00, 3.50, 0.5643, 0.0175, 0.0607, 0.0355,
0.10, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0607, 0.0355,
0.10, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0607, 0.0355,
0.10, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0607, 0.0355,
0.80, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0793, 0.0351,
0.80, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0793, 0.0351,
0.80, 17.0, 25.0, 1.00, 1.50, 0.4339, 0.0280, 0.0793, 0.0351,
0.80, 17.0, 25.0, 1.50, 2.00, 0.5016, 0.0056, 0.0793, 0.0351,
0.80, 17.0, 25.0, 2.00, 2.50, 0.3734, 0.0027, 0.0793, 0.0351,
0.80, 17.0, 25.0, 2.50, 3.00, 0.3551, 0.0026, 0.0793, 0.0351,
0.80, 17.0, 25.0, 3.00, 3.50, 0.3257, 0.0029, 0.0793, 0.0351,
0.80, 17.0, 25.0, 3.50, 4.00, 0.3098, 0.0039, 0.0793, 0.0351,
0.80, 17.0, 25.0, 4.00, 4.50, 0.2364, 0.0064, 0.0793, 0.0351,
0.80, 17.0, 25.0, 4.50, 5.00, 0.1388, 0.1672, 0.0793, 0.0351,
0.80, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0172, 0.0351,
0.80, 25.0, 37.0, 0.50, 1.00, 0.8278, 0.0099, 0.0172, 0.0351,
0.80, 25.0, 37.0, 1.00, 1.50, 0.9253, 0.0021, 0.0172, 0.0351,
0.80, 25.0, 37.0, 1.50, 2.00, 0.8911, 0.0015, 0.0172, 0.0351,
0.80, 25.0, 37.0, 2.00, 2.50, 0.8608, 0.0016, 0.0172, 0.0351,
0.80, 25.0, 37.0, 2.50, 3.00, 0.8337, 0.0020, 0.0172, 0.0351,
0.80, 25.0, 37.0, 3.00, 3.50, 0.8157, 0.0028, 0.0172, 0.0351,
0.80, 25.0, 37.0, 3.50, 4.00, 0.7849, 0.0047, 0.0172, 0.0351,
0.80, 25.0, 37.0, 4.00, 4.50, 0.6478, 0.0185, 0.0172, 0.0351,
0.80, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0172, 0.0351,
0.80, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0047, 0.0351,
0.80, 37.0, 51.0, 0.50, 1.00, 0.9607, 0.0064, 0.0047, 0.0351,
0.80, 37.0, 51.0, 1.00, 1.50, 0.9704, 0.0017, 0.0047, 0.0351,
0.80, 37.0, 51.0, 1.50, 2.00, 0.9906, 0.0016, 0.0047, 0.0351,
0.80, 37.0, 51.0, 2.00, 2.50, 0.9907, 0.0027, 0.0047, 0.0351,
0.80, 37.0, 51.0, 2.50, 3.00, 0.9726, 0.0040, 0.0047, 0.0351,
0.80, 37.0, 51.0, 3.00, 3.50, 0.9840, 0.0065, 0.0047, 0.0351,
0.80, 37.0, 51.0, 3.50, 4.00, 0.9860, 0.0417, 0.0047, 0.0351,
0.80, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0047, 0.0351,
0.80, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0047, 0.0351,
0.80, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0026, 0.0058,
0.80, 51.0, 117.0, 0.50, 1.00, 1.0330, 0.0015, 0.0026, 0.0058,
0.80, 51.0, 117.0, 1.00, 1.50, 0.9853, 0.0007, 0.0026, 0.0058,
0.80, 51.0, 117.0, 1.50, 2.00, 0.9738, 0.0010, 0.0026, 0.0058,
0.80, 51.0, 117.0, 2.00, 2.50, 0.9581, 0.0017, 0.0026, 0.0058,
0.80, 51.0, 117.0, 2.50, 3.00, 0.9314, 0.0040, 0.0026, 0.0058,
0.80, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0026, 0.0058,
0.80, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0026, 0.0058,
0.80, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0026, 0.0058,
0.80, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0026, 0.0058,
0.80, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0039, 0.039,
0.80, 117.0, 130.0, 0.50, 1.00, 0.9854, 0.0056, 0.0039, 0.039,
0.80, 117.0, 130.0, 1.00, 1.50, 0.9741, 0.0021, 0.0039, 0.039,
0.80, 117.0, 130.0, 1.50, 2.00, 0.9929, 0.0026, 0.0039, 0.039,
0.80, 117.0, 130.0, 2.00, 2.50, 0.9891, 0.0058, 0.0039, 0.039,
0.80, 117.0, 130.0, 2.50, 3.00, 1.0341, 0.0124, 0.0039, 0.039,
0.80, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0039, 0.039,
0.80, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0039, 0.039,
0.80, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0039, 0.039,
0.80, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0039, 0.039,
0.80, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0088, 0.039,
0.80, 130.0, 145.0, 0.50, 1.00, 0.7791, 0.0059, 0.0088, 0.039,
0.80, 130.0, 145.0, 1.00, 1.50, 0.9725, 0.0022, 0.0088, 0.039,
0.80, 130.0, 145.0, 1.50, 2.00, 0.9671, 0.0018, 0.0088, 0.039,
0.80, 130.0, 145.0, 2.00, 2.50, 0.9559, 0.0027, 0.0088, 0.039,
0.80, 130.0, 145.0, 2.50, 3.00, 0.9495, 0.0049, 0.0088, 0.039,
0.80, 130.0, 145.0, 3.00, 3.50, 0.9258, 0.0271, 0.0088, 0.039,
0.80, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0088, 0.039,
0.80, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0088, 0.039,
0.80, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0088, 0.039,
0.80, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0658, 0.039,
0.80, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0658, 0.039,
0.80, 145.0, 150.0, 1.00, 1.50, 0.2891, 0.0143, 0.0658, 0.039,
0.80, 145.0, 150.0, 1.50, 2.00, 0.4927, 0.0044, 0.0658, 0.039,
0.80, 145.0, 150.0, 2.00, 2.50, 0.4948, 0.0034, 0.0658, 0.039,
0.80, 145.0, 150.0, 2.50, 3.00, 0.5080, 0.0050, 0.0658, 0.039,
0.80, 145.0, 150.0, 3.00, 3.50, 0.5339, 0.0172, 0.0658, 0.039,
0.80, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0658, 0.039,
0.80, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0658, 0.039,
0.80, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0658, 0.039,
0.90, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0815, 0.0403,
0.90, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0815, 0.0403,
0.90, 17.0, 25.0, 1.00, 1.50, 0.4227, 0.0277, 0.0815, 0.0403,
0.90, 17.0, 25.0, 1.50, 2.00, 0.4744, 0.0055, 0.0815, 0.0403,
0.90, 17.0, 25.0, 2.00, 2.50, 0.3588, 0.0027, 0.0815, 0.0403,
0.90, 17.0, 25.0, 2.50, 3.00, 0.3431, 0.0025, 0.0815, 0.0403,
0.90, 17.0, 25.0, 3.00, 3.50, 0.3155, 0.0029, 0.0815, 0.0403,
0.90, 17.0, 25.0, 3.50, 4.00, 0.3011, 0.0039, 0.0815, 0.0403,
0.90, 17.0, 25.0, 4.00, 4.50, 0.2320, 0.0065, 0.0815, 0.0403,
0.90, 17.0, 25.0, 4.50, 5.00, 0.1388, 0.1672, 0.0815, 0.0403,
0.90, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0186, 0.0403,
0.90, 25.0, 37.0, 0.50, 1.00, 0.7907, 0.0104, 0.0186, 0.0403,
0.90, 25.0, 37.0, 1.00, 1.50, 0.8956, 0.0022, 0.0186, 0.0403,
0.90, 25.0, 37.0, 1.50, 2.00, 0.8705, 0.0015, 0.0186, 0.0403,
0.90, 25.0, 37.0, 2.00, 2.50, 0.8429, 0.0016, 0.0186, 0.0403,
0.90, 25.0, 37.0, 2.50, 3.00, 0.8178, 0.0021, 0.0186, 0.0403,
0.90, 25.0, 37.0, 3.00, 3.50, 0.8014, 0.0028, 0.0186, 0.0403,
0.90, 25.0, 37.0, 3.50, 4.00, 0.7696, 0.0048, 0.0186, 0.0403,
0.90, 25.0, 37.0, 4.00, 4.50, 0.6309, 0.0184, 0.0186, 0.0403,
0.90, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0186, 0.0403,
0.90, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0052, 0.0403,
0.90, 37.0, 51.0, 0.50, 1.00, 0.9531, 0.0068, 0.0052, 0.0403,
0.90, 37.0, 51.0, 1.00, 1.50, 0.9625, 0.0019, 0.0052, 0.0403,
0.90, 37.0, 51.0, 1.50, 2.00, 0.9863, 0.0016, 0.0052, 0.0403,
0.90, 37.0, 51.0, 2.00, 2.50, 0.9795, 0.0029, 0.0052, 0.0403,
0.90, 37.0, 51.0, 2.50, 3.00, 0.9604, 0.0044, 0.0052, 0.0403,
0.90, 37.0, 51.0, 3.00, 3.50, 0.9772, 0.0071, 0.0052, 0.0403,
0.90, 37.0, 51.0, 3.50, 4.00, 0.9807, 0.0434, 0.0052, 0.0403,
0.90, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0052, 0.0403,
0.90, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0052, 0.0403,
0.90, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0028, 0.0044,
0.90, 51.0, 117.0, 0.50, 1.00, 1.0343, 0.0016, 0.0028, 0.0044,
0.90, 51.0, 117.0, 1.00, 1.50, 0.9872, 0.0008, 0.0028, 0.0044,
0.90, 51.0, 117.0, 1.50, 2.00, 0.9762, 0.0010, 0.0028, 0.0044,
0.90, 51.0, 117.0, 2.00, 2.50, 0.9612, 0.0019, 0.0028, 0.0044,
0.90, 51.0, 117.0, 2.50, 3.00, 0.9354, 0.0044, 0.0028, 0.0044,
0.90, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0028, 0.0044,
0.90, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0028, 0.0044,
0.90, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0028, 0.0044,
0.90, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0028, 0.0044,
0.90, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0043, 0.0429,
0.90, 117.0, 130.0, 0.50, 1.00, 0.9797, 0.0060, 0.0043, 0.0429,
0.90, 117.0, 130.0, 1.00, 1.50, 0.9662, 0.0023, 0.0043, 0.0429,
0.90, 117.0, 130.0, 1.50, 2.00, 0.9866, 0.0028, 0.0043, 0.0429,
0.90, 117.0, 130.0, 2.00, 2.50, 0.9636, 0.0064, 0.0043, 0.0429,
0.90, 117.0, 130.0, 2.50, 3.00, 0.9923, 0.0130, 0.0043, 0.0429,
0.90, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0043, 0.0429,
0.90, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0043, 0.0429,
0.90, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0043, 0.0429,
0.90, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0043, 0.0429,
0.90, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0104, 0.0429,
0.90, 130.0, 145.0, 0.50, 1.00, 0.7304, 0.0061, 0.0104, 0.0429,
0.90, 130.0, 145.0, 1.00, 1.50, 0.9453, 0.0024, 0.0104, 0.0429,
0.90, 130.0, 145.0, 1.50, 2.00, 0.9539, 0.0019, 0.0104, 0.0429,
0.90, 130.0, 145.0, 2.00, 2.50, 0.9403, 0.0028, 0.0104, 0.0429,
0.90, 130.0, 145.0, 2.50, 3.00, 0.9338, 0.0051, 0.0104, 0.0429,
0.90, 130.0, 145.0, 3.00, 3.50, 0.8999, 0.0282, 0.0104, 0.0429,
0.90, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0104, 0.0429,
0.90, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0104, 0.0429,
0.90, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0104, 0.0429,
0.90, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0701, 0.0429,
0.90, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0701, 0.0429,
0.90, 145.0, 150.0, 1.00, 1.50, 0.2794, 0.0143, 0.0701, 0.0429,
0.90, 145.0, 150.0, 1.50, 2.00, 0.4754, 0.0045, 0.0701, 0.0429,
0.90, 145.0, 150.0, 2.00, 2.50, 0.4724, 0.0034, 0.0701, 0.0429,
0.90, 145.0, 150.0, 2.50, 3.00, 0.4885, 0.0050, 0.0701, 0.0429,
0.90, 145.0, 150.0, 3.00, 3.50, 0.5103, 0.0171, 0.0701, 0.0429,
0.90, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0701, 0.0429,
0.90, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0701, 0.0429,
0.90, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0701, 0.0429,
0.95, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0848, 0.0496,
0.95, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0848, 0.0496,
0.95, 17.0, 25.0, 1.00, 1.50, 0.4286, 0.0294, 0.0848, 0.0496,
0.95, 17.0, 25.0, 1.50, 2.00, 0.4492, 0.0055, 0.0848, 0.0496,
0.95, 17.0, 25.0, 2.00, 2.50, 0.3355, 0.0027, 0.0848, 0.0496,
0.95, 17.0, 25.0, 2.50, 3.00, 0.3215, 0.0025, 0.0848, 0.0496,
0.95, 17.0, 25.0, 3.00, 3.50, 0.2966, 0.0028, 0.0848, 0.0496,
0.95, 17.0, 25.0, 3.50, 4.00, 0.2831, 0.0038, 0.0848, 0.0496,
0.95, 17.0, 25.0, 4.00, 4.50, 0.2163, 0.0062, 0.0848, 0.0496,
0.95, 17.0, 25.0, 4.50, 5.00, 0.1388, 0.1672, 0.0848, 0.0496,
0.95, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0217, 0.0496,
0.95, 25.0, 37.0, 0.50, 1.00, 0.7435, 0.0112, 0.0217, 0.0496,
0.95, 25.0, 37.0, 1.00, 1.50, 0.8553, 0.0024, 0.0217, 0.0496,
0.95, 25.0, 37.0, 1.50, 2.00, 0.8288, 0.0016, 0.0217, 0.0496,
0.95, 25.0, 37.0, 2.00, 2.50, 0.8011, 0.0017, 0.0217, 0.0496,
0.95, 25.0, 37.0, 2.50, 3.00, 0.7775, 0.0021, 0.0217, 0.0496,
0.95, 25.0, 37.0, 3.00, 3.50, 0.7613, 0.0029, 0.0217, 0.0496,
0.95, 25.0, 37.0, 3.50, 4.00, 0.7286, 0.0050, 0.0217, 0.0496,
0.95, 25.0, 37.0, 4.00, 4.50, 0.5967, 0.0197, 0.0217, 0.0496,
0.95, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0217, 0.0496,
0.95, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0056, 0.0496,
0.95, 37.0, 51.0, 0.50, 1.00, 0.9399, 0.0072, 0.0056, 0.0496,
0.95, 37.0, 51.0, 1.00, 1.50, 0.9501, 0.0021, 0.0056, 0.0496,
0.95, 37.0, 51.0, 1.50, 2.00, 0.9783, 0.0017, 0.0056, 0.0496,
0.95, 37.0, 51.0, 2.00, 2.50, 0.9588, 0.0031, 0.0056, 0.0496,
0.95, 37.0, 51.0, 2.50, 3.00, 0.9356, 0.0048, 0.0056, 0.0496,
0.95, 37.0, 51.0, 3.00, 3.50, 0.9580, 0.0079, 0.0056, 0.0496,
0.95, 37.0, 51.0, 3.50, 4.00, 0.9405, 0.0478, 0.0056, 0.0496,
0.95, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0056, 0.0496,
0.95, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0056, 0.0496,
0.95, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0026, 0.002,
0.95, 51.0, 117.0, 0.50, 1.00, 1.0409, 0.0018, 0.0026, 0.002,
0.95, 51.0, 117.0, 1.00, 1.50, 0.9884, 0.0009, 0.0026, 0.002,
0.95, 51.0, 117.0, 1.50, 2.00, 0.9810, 0.0012, 0.0026, 0.002,
0.95, 51.0, 117.0, 2.00, 2.50, 0.9649, 0.0021, 0.0026, 0.002,
0.95, 51.0, 117.0, 2.50, 3.00, 0.9368, 0.0048, 0.0026, 0.002,
0.95, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0026, 0.002,
0.95, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0026, 0.002,
0.95, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0026, 0.002,
0.95, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0026, 0.002,
0.95, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0043, 0.0508,
0.95, 117.0, 130.0, 0.50, 1.00, 0.9708, 0.0066, 0.0043, 0.0508,
0.95, 117.0, 130.0, 1.00, 1.50, 0.9531, 0.0026, 0.0043, 0.0508,
0.95, 117.0, 130.0, 1.50, 2.00, 0.9720, 0.0030, 0.0043, 0.0508,
0.95, 117.0, 130.0, 2.00, 2.50, 0.9201, 0.0068, 0.0043, 0.0508,
0.95, 117.0, 130.0, 2.50, 3.00, 0.9504, 0.0143, 0.0043, 0.0508,
0.95, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0043, 0.0508,
0.95, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0043, 0.0508,
0.95, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0043, 0.0508,
0.95, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0043, 0.0508,
0.95, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0117, 0.0508,
0.95, 130.0, 145.0, 0.50, 1.00, 0.6765, 0.0065, 0.0117, 0.0508,
0.95, 130.0, 145.0, 1.00, 1.50, 0.9060, 0.0026, 0.0117, 0.0508,
0.95, 130.0, 145.0, 1.50, 2.00, 0.9339, 0.0020, 0.0117, 0.0508,
0.95, 130.0, 145.0, 2.00, 2.50, 0.9129, 0.0030, 0.0117, 0.0508,
0.95, 130.0, 145.0, 2.50, 3.00, 0.9082, 0.0053, 0.0117, 0.0508,
0.95, 130.0, 145.0, 3.00, 3.50, 0.8692, 0.0292, 0.0117, 0.0508,
0.95, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0117, 0.0508,
0.95, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0117, 0.0508,
0.95, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0117, 0.0508,
0.95, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0746, 0.0508,
0.95, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0746, 0.0508,
0.95, 145.0, 150.0, 1.00, 1.50, 0.2668, 0.0144, 0.0746, 0.0508,
0.95, 145.0, 150.0, 1.50, 2.00, 0.4567, 0.0046, 0.0746, 0.0508,
0.95, 145.0, 150.0, 2.00, 2.50, 0.4467, 0.0035, 0.0746, 0.0508,
0.95, 145.0, 150.0, 2.50, 3.00, 0.4594, 0.0050, 0.0746, 0.0508,
0.95, 145.0, 150.0, 3.00, 3.50, 0.4741, 0.0166, 0.0746, 0.0508,
0.95, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0746, 0.0508,
0.95, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0746, 0.0508,
0.95, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0746, 0.0508,
0.97, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0900, 0.066,
0.97, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0900, 0.066,
0.97, 17.0, 25.0, 1.00, 1.50, 0.4273, 0.0306, 0.0900, 0.066,
0.97, 17.0, 25.0, 1.50, 2.00, 0.4132, 0.0054, 0.0900, 0.066,
0.97, 17.0, 25.0, 2.00, 2.50, 0.2788, 0.0025, 0.0900, 0.066,
0.97, 17.0, 25.0, 2.50, 3.00, 0.2665, 0.0023, 0.0900, 0.066,
0.97, 17.0, 25.0, 3.00, 3.50, 0.2442, 0.0025, 0.0900, 0.066,
0.97, 17.0, 25.0, 3.50, 4.00, 0.2343, 0.0034, 0.0900, 0.066,
0.97, 17.0, 25.0, 4.00, 4.50, 0.1782, 0.0056, 0.0900, 0.066,
0.97, 17.0, 25.0, 4.50, 5.00, 0.1388, 0.1672, 0.0900, 0.066,
0.97, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0298, 0.066,
0.97, 25.0, 37.0, 0.50, 1.00, 0.7029, 0.0115, 0.0298, 0.066,
0.97, 25.0, 37.0, 1.00, 1.50, 0.8024, 0.0025, 0.0298, 0.066,
0.97, 25.0, 37.0, 1.50, 2.00, 0.7531, 0.0017, 0.0298, 0.066,
0.97, 25.0, 37.0, 2.00, 2.50, 0.7119, 0.0017, 0.0298, 0.066,
0.97, 25.0, 37.0, 2.50, 3.00, 0.6803, 0.0022, 0.0298, 0.066,
0.97, 25.0, 37.0, 3.00, 3.50, 0.6551, 0.0030, 0.0298, 0.066,
0.97, 25.0, 37.0, 3.50, 4.00, 0.6048, 0.0051, 0.0298, 0.066,
0.97, 25.0, 37.0, 4.00, 4.50, 0.4561, 0.0195, 0.0298, 0.066,
0.97, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0298, 0.066,
0.97, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0064, 0.066,
0.97, 37.0, 51.0, 0.50, 1.00, 0.9146, 0.0079, 0.0064, 0.066,
0.97, 37.0, 51.0, 1.00, 1.50, 0.9297, 0.0024, 0.0064, 0.066,
0.97, 37.0, 51.0, 1.50, 2.00, 0.9678, 0.0018, 0.0064, 0.066,
0.97, 37.0, 51.0, 2.00, 2.50, 0.9376, 0.0033, 0.0064, 0.066,
0.97, 37.0, 51.0, 2.50, 3.00, 0.9067, 0.0051, 0.0064, 0.066,
0.97, 37.0, 51.0, 3.00, 3.50, 0.9364, 0.0086, 0.0064, 0.066,
0.97, 37.0, 51.0, 3.50, 4.00, 0.9115, 0.0507, 0.0064, 0.066,
0.97, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0064, 0.066,
0.97, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0064, 0.066,
0.97, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0028, 0.0018,
0.97, 51.0, 117.0, 0.50, 1.00, 1.0334, 0.0021, 0.0028, 0.0018,
0.97, 51.0, 117.0, 1.00, 1.50, 0.9908, 0.0010, 0.0028, 0.0018,
0.97, 51.0, 117.0, 1.50, 2.00, 0.9857, 0.0013, 0.0028, 0.0018,
0.97, 51.0, 117.0, 2.00, 2.50, 0.9713, 0.0024, 0.0028, 0.0018,
0.97, 51.0, 117.0, 2.50, 3.00, 0.9380, 0.0054, 0.0028, 0.0018,
0.97, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0028, 0.0018,
0.97, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0028, 0.0018,
0.97, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0028, 0.0018,
0.97, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0028, 0.0018,
0.97, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0059, 0.0611,
0.97, 117.0, 130.0, 0.50, 1.00, 0.9454, 0.0074, 0.0059, 0.0611,
0.97, 117.0, 130.0, 1.00, 1.50, 0.9402, 0.0029, 0.0059, 0.0611,
0.97, 117.0, 130.0, 1.50, 2.00, 0.9581, 0.0033, 0.0059, 0.0611,
0.97, 117.0, 130.0, 2.00, 2.50, 0.8829, 0.0073, 0.0059, 0.0611,
0.97, 117.0, 130.0, 2.50, 3.00, 0.9024, 0.0153, 0.0059, 0.0611,
0.97, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0059, 0.0611,
0.97, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0059, 0.0611,
0.97, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0059, 0.0611,
0.97, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0059, 0.0611,
0.97, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0150, 0.0611,
0.97, 130.0, 145.0, 0.50, 1.00, 0.6400, 0.0069, 0.0150, 0.0611,
0.97, 130.0, 145.0, 1.00, 1.50, 0.8690, 0.0028, 0.0150, 0.0611,
0.97, 130.0, 145.0, 1.50, 2.00, 0.9044, 0.0022, 0.0150, 0.0611,
0.97, 130.0, 145.0, 2.00, 2.50, 0.8743, 0.0031, 0.0150, 0.0611,
0.97, 130.0, 145.0, 2.50, 3.00, 0.8666, 0.0056, 0.0150, 0.0611,
0.97, 130.0, 145.0, 3.00, 3.50, 0.7880, 0.0312, 0.0150, 0.0611,
0.97, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0150, 0.0611,
0.97, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0150, 0.0611,
0.97, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0150, 0.0611,
0.97, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0832, 0.0611,
0.97, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0832, 0.0611,
0.97, 145.0, 150.0, 1.00, 1.50, 0.2487, 0.0142, 0.0832, 0.0611,
0.97, 145.0, 150.0, 1.50, 2.00, 0.4334, 0.0047, 0.0832, 0.0611,
0.97, 145.0, 150.0, 2.00, 2.50, 0.3983, 0.0034, 0.0832, 0.0611,
0.97, 145.0, 150.0, 2.50, 3.00, 0.4082, 0.0049, 0.0832, 0.0611,
0.97, 145.0, 150.0, 3.00, 3.50, 0.4183, 0.0157, 0.0832, 0.0611,
0.97, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0832, 0.0611,
0.97, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0832, 0.0611,
0.97, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0832, 0.0611
};
// eff(data) / eff(MC) exp 51..65
const float LepEff::fMuonEff51[] = {
0.10, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0487, 0.0279,
0.10, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0487, 0.0279,
0.10, 17.0, 25.0, 1.00, 1.50, 0.5473, 0.0332, 0.0487, 0.0279,
0.10, 17.0, 25.0, 1.50, 2.00, 0.6786, 0.0065, 0.0487, 0.0279,
0.10, 17.0, 25.0, 2.00, 2.50, 0.7053, 0.0040, 0.0487, 0.0279,
0.10, 17.0, 25.0, 2.50, 3.00, 0.6987, 0.0040, 0.0487, 0.0279,
0.10, 17.0, 25.0, 3.00, 3.50, 0.6877, 0.0049, 0.0487, 0.0279,
0.10, 17.0, 25.0, 3.50, 4.00, 0.6619, 0.0068, 0.0487, 0.0279,
0.10, 17.0, 25.0, 4.00, 4.50, 0.5885, 0.0126, 0.0487, 0.0279,
0.10, 17.0, 25.0, 4.50, 5.00, 0.4066, 0.2370, 0.0487, 0.0279,
0.10, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0062, 0.0279,
0.10, 25.0, 37.0, 0.50, 1.00, 0.9640, 0.0100, 0.0062, 0.0279,
0.10, 25.0, 37.0, 1.00, 1.50, 1.0000, 0.0019, 0.0062, 0.0279,
0.10, 25.0, 37.0, 1.50, 2.00, 1.0079, 0.0014, 0.0062, 0.0279,
0.10, 25.0, 37.0, 2.00, 2.50, 1.0030, 0.0016, 0.0062, 0.0279,
0.10, 25.0, 37.0, 2.50, 3.00, 0.9920, 0.0021, 0.0062, 0.0279,
0.10, 25.0, 37.0, 3.00, 3.50, 0.9798, 0.0027, 0.0062, 0.0279,
0.10, 25.0, 37.0, 3.50, 4.00, 0.9777, 0.0047, 0.0062, 0.0279,
0.10, 25.0, 37.0, 4.00, 4.50, 0.9643, 0.0233, 0.0062, 0.0279,
0.10, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0062, 0.0279,
0.10, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0030, 0.0279,
0.10, 37.0, 51.0, 0.50, 1.00, 0.9851, 0.0058, 0.0030, 0.0279,
0.10, 37.0, 51.0, 1.00, 1.50, 0.9870, 0.0013, 0.0030, 0.0279,
0.10, 37.0, 51.0, 1.50, 2.00, 0.9997, 0.0014, 0.0030, 0.0279,
0.10, 37.0, 51.0, 2.00, 2.50, 0.9969, 0.0020, 0.0030, 0.0279,
0.10, 37.0, 51.0, 2.50, 3.00, 0.9876, 0.0030, 0.0030, 0.0279,
0.10, 37.0, 51.0, 3.00, 3.50, 0.9952, 0.0052, 0.0030, 0.0279,
0.10, 37.0, 51.0, 3.50, 4.00, 0.9819, 0.0280, 0.0030, 0.0279,
0.10, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0030, 0.0279,
0.10, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0030, 0.0279,
0.10, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0021, 0.0087,
0.10, 51.0, 117.0, 0.50, 1.00, 1.0330, 0.0012, 0.0021, 0.0087,
0.10, 51.0, 117.0, 1.00, 1.50, 0.9828, 0.0005, 0.0021, 0.0087,
0.10, 51.0, 117.0, 1.50, 2.00, 0.9748, 0.0007, 0.0021, 0.0087,
0.10, 51.0, 117.0, 2.00, 2.50, 0.9608, 0.0011, 0.0021, 0.0087,
0.10, 51.0, 117.0, 2.50, 3.00, 0.9402, 0.0030, 0.0021, 0.0087,
0.10, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0021, 0.0087,
0.10, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0021, 0.0087,
0.10, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0021, 0.0087,
0.10, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0021, 0.0087,
0.10, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0053, 0.0355,
0.10, 117.0, 130.0, 0.50, 1.00, 1.0062, 0.0050, 0.0053, 0.0355,
0.10, 117.0, 130.0, 1.00, 1.50, 0.9883, 0.0018, 0.0053, 0.0355,
0.10, 117.0, 130.0, 1.50, 2.00, 0.9967, 0.0022, 0.0053, 0.0355,
0.10, 117.0, 130.0, 2.00, 2.50, 1.0003, 0.0037, 0.0053, 0.0355,
0.10, 117.0, 130.0, 2.50, 3.00, 1.0356, 0.0087, 0.0053, 0.0355,
0.10, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0053, 0.0355,
0.10, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0053, 0.0355,
0.10, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0053, 0.0355,
0.10, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0053, 0.0355,
0.10, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0082, 0.0355,
0.10, 130.0, 145.0, 0.50, 1.00, 0.9114, 0.0053, 0.0082, 0.0355,
0.10, 130.0, 145.0, 1.00, 1.50, 1.0033, 0.0019, 0.0082, 0.0355,
0.10, 130.0, 145.0, 1.50, 2.00, 0.9914, 0.0016, 0.0082, 0.0355,
0.10, 130.0, 145.0, 2.00, 2.50, 0.9899, 0.0024, 0.0082, 0.0355,
0.10, 130.0, 145.0, 2.50, 3.00, 0.9884, 0.0047, 0.0082, 0.0355,
0.10, 130.0, 145.0, 3.00, 3.50, 0.9630, 0.0238, 0.0082, 0.0355,
0.10, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0082, 0.0355,
0.10, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0082, 0.0355,
0.10, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0082, 0.0355,
0.10, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0564, 0.0355,
0.10, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0564, 0.0355,
0.10, 145.0, 150.0, 1.00, 1.50, 0.2694, 0.0108, 0.0564, 0.0355,
0.10, 145.0, 150.0, 1.50, 2.00, 0.5120, 0.0037, 0.0564, 0.0355,
0.10, 145.0, 150.0, 2.00, 2.50, 0.5937, 0.0029, 0.0564, 0.0355,
0.10, 145.0, 150.0, 2.50, 3.00, 0.6138, 0.0045, 0.0564, 0.0355,
0.10, 145.0, 150.0, 3.00, 3.50, 0.6038, 0.0134, 0.0564, 0.0355,
0.10, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0564, 0.0355,
0.10, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0564, 0.0355,
0.10, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0564, 0.0355,
0.80, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0515, 0.0351,
0.80, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0515, 0.0351,
0.80, 17.0, 25.0, 1.00, 1.50, 0.5351, 0.0328, 0.0515, 0.0351,
0.80, 17.0, 25.0, 1.50, 2.00, 0.6671, 0.0066, 0.0515, 0.0351,
0.80, 17.0, 25.0, 2.00, 2.50, 0.6995, 0.0041, 0.0515, 0.0351,
0.80, 17.0, 25.0, 2.50, 3.00, 0.6910, 0.0040, 0.0515, 0.0351,
0.80, 17.0, 25.0, 3.00, 3.50, 0.6808, 0.0049, 0.0515, 0.0351,
0.80, 17.0, 25.0, 3.50, 4.00, 0.6518, 0.0068, 0.0515, 0.0351,
0.80, 17.0, 25.0, 4.00, 4.50, 0.5794, 0.0126, 0.0515, 0.0351,
0.80, 17.0, 25.0, 4.50, 5.00, 0.3696, 0.2169, 0.0515, 0.0351,
0.80, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0082, 0.0351,
0.80, 25.0, 37.0, 0.50, 1.00, 0.9394, 0.0112, 0.0082, 0.0351,
0.80, 25.0, 37.0, 1.00, 1.50, 0.9881, 0.0021, 0.0082, 0.0351,
0.80, 25.0, 37.0, 1.50, 2.00, 1.0072, 0.0015, 0.0082, 0.0351,
0.80, 25.0, 37.0, 2.00, 2.50, 1.0011, 0.0017, 0.0082, 0.0351,
0.80, 25.0, 37.0, 2.50, 3.00, 0.9884, 0.0022, 0.0082, 0.0351,
0.80, 25.0, 37.0, 3.00, 3.50, 0.9751, 0.0028, 0.0082, 0.0351,
0.80, 25.0, 37.0, 3.50, 4.00, 0.9719, 0.0049, 0.0082, 0.0351,
0.80, 25.0, 37.0, 4.00, 4.50, 0.9673, 0.0250, 0.0082, 0.0351,
0.80, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0082, 0.0351,
0.80, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0051, 0.0351,
0.80, 37.0, 51.0, 0.50, 1.00, 0.9869, 0.0063, 0.0051, 0.0351,
0.80, 37.0, 51.0, 1.00, 1.50, 0.9832, 0.0016, 0.0051, 0.0351,
0.80, 37.0, 51.0, 1.50, 2.00, 0.9997, 0.0015, 0.0051, 0.0351,
0.80, 37.0, 51.0, 2.00, 2.50, 0.9932, 0.0026, 0.0051, 0.0351,
0.80, 37.0, 51.0, 2.50, 3.00, 0.9817, 0.0040, 0.0051, 0.0351,
0.80, 37.0, 51.0, 3.00, 3.50, 1.0192, 0.0070, 0.0051, 0.0351,
0.80, 37.0, 51.0, 3.50, 4.00, 0.9625, 0.0307, 0.0051, 0.0351,
0.80, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0051, 0.0351,
0.80, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0051, 0.0351,
0.80, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0029, 0.0058,
0.80, 51.0, 117.0, 0.50, 1.00, 1.0410, 0.0014, 0.0029, 0.0058,
0.80, 51.0, 117.0, 1.00, 1.50, 0.9876, 0.0006, 0.0029, 0.0058,
0.80, 51.0, 117.0, 1.50, 2.00, 0.9773, 0.0009, 0.0029, 0.0058,
0.80, 51.0, 117.0, 2.00, 2.50, 0.9607, 0.0014, 0.0029, 0.0058,
0.80, 51.0, 117.0, 2.50, 3.00, 0.9413, 0.0035, 0.0029, 0.0058,
0.80, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0029, 0.0058,
0.80, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0029, 0.0058,
0.80, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0029, 0.0058,
0.80, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0029, 0.0058,
0.80, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0056, 0.039,
0.80, 117.0, 130.0, 0.50, 1.00, 1.0023, 0.0056, 0.0056, 0.039,
0.80, 117.0, 130.0, 1.00, 1.50, 0.9816, 0.0020, 0.0056, 0.039,
0.80, 117.0, 130.0, 1.50, 2.00, 0.9943, 0.0026, 0.0056, 0.039,
0.80, 117.0, 130.0, 2.00, 2.50, 0.9744, 0.0054, 0.0056, 0.039,
0.80, 117.0, 130.0, 2.50, 3.00, 1.0371, 0.0124, 0.0056, 0.039,
0.80, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0056, 0.039,
0.80, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0056, 0.039,
0.80, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0056, 0.039,
0.80, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0056, 0.039,
0.80, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0091, 0.039,
0.80, 130.0, 145.0, 0.50, 1.00, 0.8591, 0.0060, 0.0091, 0.039,
0.80, 130.0, 145.0, 1.00, 1.50, 0.9805, 0.0021, 0.0091, 0.039,
0.80, 130.0, 145.0, 1.50, 2.00, 0.9746, 0.0017, 0.0091, 0.039,
0.80, 130.0, 145.0, 2.00, 2.50, 0.9729, 0.0026, 0.0091, 0.039,
0.80, 130.0, 145.0, 2.50, 3.00, 0.9723, 0.0048, 0.0091, 0.039,
0.80, 130.0, 145.0, 3.00, 3.50, 0.9487, 0.0245, 0.0091, 0.039,
0.80, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0091, 0.039,
0.80, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0091, 0.039,
0.80, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0091, 0.039,
0.80, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0619, 0.039,
0.80, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0619, 0.039,
0.80, 145.0, 150.0, 1.00, 1.50, 0.2578, 0.0107, 0.0619, 0.039,
0.80, 145.0, 150.0, 1.50, 2.00, 0.4946, 0.0037, 0.0619, 0.039,
0.80, 145.0, 150.0, 2.00, 2.50, 0.5612, 0.0030, 0.0619, 0.039,
0.80, 145.0, 150.0, 2.50, 3.00, 0.5865, 0.0046, 0.0619, 0.039,
0.80, 145.0, 150.0, 3.00, 3.50, 0.5716, 0.0129, 0.0619, 0.039,
0.80, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0619, 0.039,
0.80, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0619, 0.039,
0.80, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0619, 0.039,
0.90, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0539, 0.0403,
0.90, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0539, 0.0403,
0.90, 17.0, 25.0, 1.00, 1.50, 0.5339, 0.0335, 0.0539, 0.0403,
0.90, 17.0, 25.0, 1.50, 2.00, 0.6515, 0.0066, 0.0539, 0.0403,
0.90, 17.0, 25.0, 2.00, 2.50, 0.6949, 0.0042, 0.0539, 0.0403,
0.90, 17.0, 25.0, 2.50, 3.00, 0.6884, 0.0042, 0.0539, 0.0403,
0.90, 17.0, 25.0, 3.00, 3.50, 0.6772, 0.0050, 0.0539, 0.0403,
0.90, 17.0, 25.0, 3.50, 4.00, 0.6499, 0.0069, 0.0539, 0.0403,
0.90, 17.0, 25.0, 4.00, 4.50, 0.5784, 0.0129, 0.0539, 0.0403,
0.90, 17.0, 25.0, 4.50, 5.00, 0.3696, 0.2169, 0.0539, 0.0403,
0.90, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0097, 0.0403,
0.90, 25.0, 37.0, 0.50, 1.00, 0.9157, 0.0118, 0.0097, 0.0403,
0.90, 25.0, 37.0, 1.00, 1.50, 0.9765, 0.0023, 0.0097, 0.0403,
0.90, 25.0, 37.0, 1.50, 2.00, 1.0065, 0.0016, 0.0097, 0.0403,
0.90, 25.0, 37.0, 2.00, 2.50, 1.0004, 0.0018, 0.0097, 0.0403,
0.90, 25.0, 37.0, 2.50, 3.00, 0.9872, 0.0023, 0.0097, 0.0403,
0.90, 25.0, 37.0, 3.00, 3.50, 0.9722, 0.0029, 0.0097, 0.0403,
0.90, 25.0, 37.0, 3.50, 4.00, 0.9698, 0.0051, 0.0097, 0.0403,
0.90, 25.0, 37.0, 4.00, 4.50, 0.9711, 0.0266, 0.0097, 0.0403,
0.90, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0097, 0.0403,
0.90, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0050, 0.0403,
0.90, 37.0, 51.0, 0.50, 1.00, 0.9782, 0.0065, 0.0050, 0.0403,
0.90, 37.0, 51.0, 1.00, 1.50, 0.9773, 0.0018, 0.0050, 0.0403,
0.90, 37.0, 51.0, 1.50, 2.00, 0.9983, 0.0016, 0.0050, 0.0403,
0.90, 37.0, 51.0, 2.00, 2.50, 0.9842, 0.0027, 0.0050, 0.0403,
0.90, 37.0, 51.0, 2.50, 3.00, 0.9687, 0.0043, 0.0050, 0.0403,
0.90, 37.0, 51.0, 3.00, 3.50, 1.0096, 0.0074, 0.0050, 0.0403,
0.90, 37.0, 51.0, 3.50, 4.00, 0.9823, 0.0334, 0.0050, 0.0403,
0.90, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0050, 0.0403,
0.90, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0050, 0.0403,
0.90, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0034, 0.0044,
0.90, 51.0, 117.0, 0.50, 1.00, 1.0437, 0.0016, 0.0034, 0.0044,
0.90, 51.0, 117.0, 1.00, 1.50, 0.9898, 0.0007, 0.0034, 0.0044,
0.90, 51.0, 117.0, 1.50, 2.00, 0.9806, 0.0009, 0.0034, 0.0044,
0.90, 51.0, 117.0, 2.00, 2.50, 0.9643, 0.0016, 0.0034, 0.0044,
0.90, 51.0, 117.0, 2.50, 3.00, 0.9424, 0.0039, 0.0034, 0.0044,
0.90, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0034, 0.0044,
0.90, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0034, 0.0044,
0.90, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0034, 0.0044,
0.90, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0034, 0.0044,
0.90, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0056, 0.0429,
0.90, 117.0, 130.0, 0.50, 1.00, 0.9979, 0.0060, 0.0056, 0.0429,
0.90, 117.0, 130.0, 1.00, 1.50, 0.9788, 0.0023, 0.0056, 0.0429,
0.90, 117.0, 130.0, 1.50, 2.00, 0.9923, 0.0028, 0.0056, 0.0429,
0.90, 117.0, 130.0, 2.00, 2.50, 0.9511, 0.0059, 0.0056, 0.0429,
0.90, 117.0, 130.0, 2.50, 3.00, 1.0048, 0.0134, 0.0056, 0.0429,
0.90, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0056, 0.0429,
0.90, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0056, 0.0429,
0.90, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0056, 0.0429,
0.90, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0056, 0.0429,
0.90, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0105, 0.0429,
0.90, 130.0, 145.0, 0.50, 1.00, 0.8257, 0.0064, 0.0105, 0.0429,
0.90, 130.0, 145.0, 1.00, 1.50, 0.9611, 0.0023, 0.0105, 0.0429,
0.90, 130.0, 145.0, 1.50, 2.00, 0.9670, 0.0018, 0.0105, 0.0429,
0.90, 130.0, 145.0, 2.00, 2.50, 0.9595, 0.0026, 0.0105, 0.0429,
0.90, 130.0, 145.0, 2.50, 3.00, 0.9620, 0.0050, 0.0105, 0.0429,
0.90, 130.0, 145.0, 3.00, 3.50, 0.9269, 0.0254, 0.0105, 0.0429,
0.90, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0105, 0.0429,
0.90, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0105, 0.0429,
0.90, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0105, 0.0429,
0.90, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0682, 0.0429,
0.90, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0682, 0.0429,
0.90, 145.0, 150.0, 1.00, 1.50, 0.2511, 0.0109, 0.0682, 0.0429,
0.90, 145.0, 150.0, 1.50, 2.00, 0.4802, 0.0038, 0.0682, 0.0429,
0.90, 145.0, 150.0, 2.00, 2.50, 0.5418, 0.0031, 0.0682, 0.0429,
0.90, 145.0, 150.0, 2.50, 3.00, 0.5707, 0.0047, 0.0682, 0.0429,
0.90, 145.0, 150.0, 3.00, 3.50, 0.5605, 0.0131, 0.0682, 0.0429,
0.90, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0682, 0.0429,
0.90, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0682, 0.0429,
0.90, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0682, 0.0429,
0.95, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0579, 0.0496,
0.95, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0579, 0.0496,
0.95, 17.0, 25.0, 1.00, 1.50, 0.5364, 0.0355, 0.0579, 0.0496,
0.95, 17.0, 25.0, 1.50, 2.00, 0.6435, 0.0068, 0.0579, 0.0496,
0.95, 17.0, 25.0, 2.00, 2.50, 0.6871, 0.0044, 0.0579, 0.0496,
0.95, 17.0, 25.0, 2.50, 3.00, 0.6817, 0.0043, 0.0579, 0.0496,
0.95, 17.0, 25.0, 3.00, 3.50, 0.6707, 0.0051, 0.0579, 0.0496,
0.95, 17.0, 25.0, 3.50, 4.00, 0.6469, 0.0071, 0.0579, 0.0496,
0.95, 17.0, 25.0, 4.00, 4.50, 0.5820, 0.0134, 0.0579, 0.0496,
0.95, 17.0, 25.0, 4.50, 5.00, 0.3511, 0.2068, 0.0579, 0.0496,
0.95, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0131, 0.0496,
0.95, 25.0, 37.0, 0.50, 1.00, 0.8763, 0.0126, 0.0131, 0.0496,
0.95, 25.0, 37.0, 1.00, 1.50, 0.9601, 0.0025, 0.0131, 0.0496,
0.95, 25.0, 37.0, 1.50, 2.00, 1.0034, 0.0018, 0.0131, 0.0496,
0.95, 25.0, 37.0, 2.00, 2.50, 0.9993, 0.0019, 0.0131, 0.0496,
0.95, 25.0, 37.0, 2.50, 3.00, 0.9847, 0.0024, 0.0131, 0.0496,
0.95, 25.0, 37.0, 3.00, 3.50, 0.9695, 0.0031, 0.0131, 0.0496,
0.95, 25.0, 37.0, 3.50, 4.00, 0.9712, 0.0056, 0.0131, 0.0496,
0.95, 25.0, 37.0, 4.00, 4.50, 0.9707, 0.0281, 0.0131, 0.0496,
0.95, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0131, 0.0496,
0.95, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0061, 0.0496,
0.95, 37.0, 51.0, 0.50, 1.00, 0.9705, 0.0071, 0.0061, 0.0496,
0.95, 37.0, 51.0, 1.00, 1.50, 0.9689, 0.0021, 0.0061, 0.0496,
0.95, 37.0, 51.0, 1.50, 2.00, 0.9954, 0.0017, 0.0061, 0.0496,
0.95, 37.0, 51.0, 2.00, 2.50, 0.9706, 0.0030, 0.0061, 0.0496,
0.95, 37.0, 51.0, 2.50, 3.00, 0.9521, 0.0047, 0.0061, 0.0496,
0.95, 37.0, 51.0, 3.00, 3.50, 0.9958, 0.0081, 0.0061, 0.0496,
0.95, 37.0, 51.0, 3.50, 4.00, 0.8930, 0.0415, 0.0061, 0.0496,
0.95, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0061, 0.0496,
0.95, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0061, 0.0496,
0.95, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0039, 0.002,
0.95, 51.0, 117.0, 0.50, 1.00, 1.0531, 0.0018, 0.0039, 0.002,
0.95, 51.0, 117.0, 1.00, 1.50, 0.9910, 0.0008, 0.0039, 0.002,
0.95, 51.0, 117.0, 1.50, 2.00, 0.9837, 0.0011, 0.0039, 0.002,
0.95, 51.0, 117.0, 2.00, 2.50, 0.9693, 0.0019, 0.0039, 0.002,
0.95, 51.0, 117.0, 2.50, 3.00, 0.9442, 0.0043, 0.0039, 0.002,
0.95, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0039, 0.002,
0.95, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0039, 0.002,
0.95, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0039, 0.002,
0.95, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0039, 0.002,
0.95, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0067, 0.0508,
0.95, 117.0, 130.0, 0.50, 1.00, 0.9959, 0.0066, 0.0067, 0.0508,
0.95, 117.0, 130.0, 1.00, 1.50, 0.9697, 0.0025, 0.0067, 0.0508,
0.95, 117.0, 130.0, 1.50, 2.00, 0.9842, 0.0030, 0.0067, 0.0508,
0.95, 117.0, 130.0, 2.00, 2.50, 0.9197, 0.0065, 0.0067, 0.0508,
0.95, 117.0, 130.0, 2.50, 3.00, 0.9541, 0.0142, 0.0067, 0.0508,
0.95, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0067, 0.0508,
0.95, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0067, 0.0508,
0.95, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0067, 0.0508,
0.95, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0067, 0.0508,
0.95, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0132, 0.0508,
0.95, 130.0, 145.0, 0.50, 1.00, 0.7778, 0.0069, 0.0132, 0.0508,
0.95, 130.0, 145.0, 1.00, 1.50, 0.9315, 0.0025, 0.0132, 0.0508,
0.95, 130.0, 145.0, 1.50, 2.00, 0.9541, 0.0019, 0.0132, 0.0508,
0.95, 130.0, 145.0, 2.00, 2.50, 0.9400, 0.0028, 0.0132, 0.0508,
0.95, 130.0, 145.0, 2.50, 3.00, 0.9431, 0.0052, 0.0132, 0.0508,
0.95, 130.0, 145.0, 3.00, 3.50, 0.9074, 0.0262, 0.0132, 0.0508,
0.95, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0132, 0.0508,
0.95, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0132, 0.0508,
0.95, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0132, 0.0508,
0.95, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0719, 0.0508,
0.95, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0719, 0.0508,
0.95, 145.0, 150.0, 1.00, 1.50, 0.2335, 0.0107, 0.0719, 0.0508,
0.95, 145.0, 150.0, 1.50, 2.00, 0.4656, 0.0039, 0.0719, 0.0508,
0.95, 145.0, 150.0, 2.00, 2.50, 0.5169, 0.0032, 0.0719, 0.0508,
0.95, 145.0, 150.0, 2.50, 3.00, 0.5467, 0.0048, 0.0719, 0.0508,
0.95, 145.0, 150.0, 3.00, 3.50, 0.5247, 0.0129, 0.0719, 0.0508,
0.95, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0719, 0.0508,
0.95, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0719, 0.0508,
0.95, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0719, 0.0508,
0.97, 17.0, 25.0, 0.00, 0.50, 1.0000, 0.0000, 0.0688, 0.066,
0.97, 17.0, 25.0, 0.50, 1.00, 1.0000, 0.0000, 0.0688, 0.066,
0.97, 17.0, 25.0, 1.00, 1.50, 0.5478, 0.0377, 0.0688, 0.066,
0.97, 17.0, 25.0, 1.50, 2.00, 0.6283, 0.0069, 0.0688, 0.066,
0.97, 17.0, 25.0, 2.00, 2.50, 0.6711, 0.0045, 0.0688, 0.066,
0.97, 17.0, 25.0, 2.50, 3.00, 0.6668, 0.0044, 0.0688, 0.066,
0.97, 17.0, 25.0, 3.00, 3.50, 0.6599, 0.0053, 0.0688, 0.066,
0.97, 17.0, 25.0, 3.50, 4.00, 0.6372, 0.0074, 0.0688, 0.066,
0.97, 17.0, 25.0, 4.00, 4.50, 0.5782, 0.0139, 0.0688, 0.066,
0.97, 17.0, 25.0, 4.50, 5.00, 0.5267, 0.3774, 0.0688, 0.066,
0.97, 25.0, 37.0, 0.00, 0.50, 1.0000, 0.0000, 0.0205, 0.066,
0.97, 25.0, 37.0, 0.50, 1.00, 0.8443, 0.0132, 0.0205, 0.066,
0.97, 25.0, 37.0, 1.00, 1.50, 0.9433, 0.0027, 0.0205, 0.066,
0.97, 25.0, 37.0, 1.50, 2.00, 1.0000, 0.0020, 0.0205, 0.066,
0.97, 25.0, 37.0, 2.00, 2.50, 1.0015, 0.0021, 0.0205, 0.066,
0.97, 25.0, 37.0, 2.50, 3.00, 0.9826, 0.0027, 0.0205, 0.066,
0.97, 25.0, 37.0, 3.00, 3.50, 0.9673, 0.0036, 0.0205, 0.066,
0.97, 25.0, 37.0, 3.50, 4.00, 0.9691, 0.0065, 0.0205, 0.066,
0.97, 25.0, 37.0, 4.00, 4.50, 0.9781, 0.0346, 0.0205, 0.066,
0.97, 25.0, 37.0, 4.50, 5.00, 1.0000, 0.0000, 0.0205, 0.066,
0.97, 37.0, 51.0, 0.00, 0.50, 1.0000, 0.0000, 0.0068, 0.066,
0.97, 37.0, 51.0, 0.50, 1.00, 0.9416, 0.0076, 0.0068, 0.066,
0.97, 37.0, 51.0, 1.00, 1.50, 0.9563, 0.0023, 0.0068, 0.066,
0.97, 37.0, 51.0, 1.50, 2.00, 0.9915, 0.0018, 0.0068, 0.066,
0.97, 37.0, 51.0, 2.00, 2.50, 0.9549, 0.0032, 0.0068, 0.066,
0.97, 37.0, 51.0, 2.50, 3.00, 0.9304, 0.0051, 0.0068, 0.066,
0.97, 37.0, 51.0, 3.00, 3.50, 0.9785, 0.0088, 0.0068, 0.066,
0.97, 37.0, 51.0, 3.50, 4.00, 0.8953, 0.0437, 0.0068, 0.066,
0.97, 37.0, 51.0, 4.00, 4.50, 1.0000, 0.0000, 0.0068, 0.066,
0.97, 37.0, 51.0, 4.50, 5.00, 1.0000, 0.0000, 0.0068, 0.066,
0.97, 51.0, 117.0, 0.00, 0.50, 1.0000, 0.0000, 0.0045, 0.0018,
0.97, 51.0, 117.0, 0.50, 1.00, 1.0436, 0.0021, 0.0045, 0.0018,
0.97, 51.0, 117.0, 1.00, 1.50, 0.9935, 0.0009, 0.0045, 0.0018,
0.97, 51.0, 117.0, 1.50, 2.00, 0.9884, 0.0012, 0.0045, 0.0018,
0.97, 51.0, 117.0, 2.00, 2.50, 0.9752, 0.0022, 0.0045, 0.0018,
0.97, 51.0, 117.0, 2.50, 3.00, 0.9406, 0.0048, 0.0045, 0.0018,
0.97, 51.0, 117.0, 3.00, 3.50, 1.0000, 0.0000, 0.0045, 0.0018,
0.97, 51.0, 117.0, 3.50, 4.00, 1.0000, 0.0000, 0.0045, 0.0018,
0.97, 51.0, 117.0, 4.00, 4.50, 1.0000, 0.0000, 0.0045, 0.0018,
0.97, 51.0, 117.0, 4.50, 5.00, 1.0000, 0.0000, 0.0045, 0.0018,
0.97, 117.0, 130.0, 0.00, 0.50, 1.0000, 0.0000, 0.0091, 0.0611,
0.97, 117.0, 130.0, 0.50, 1.00, 0.9759, 0.0075, 0.0091, 0.0611,
0.97, 117.0, 130.0, 1.00, 1.50, 0.9616, 0.0029, 0.0091, 0.0611,
0.97, 117.0, 130.0, 1.50, 2.00, 0.9747, 0.0033, 0.0091, 0.0611,
0.97, 117.0, 130.0, 2.00, 2.50, 0.8815, 0.0070, 0.0091, 0.0611,
0.97, 117.0, 130.0, 2.50, 3.00, 0.9086, 0.0153, 0.0091, 0.0611,
0.97, 117.0, 130.0, 3.00, 3.50, 1.0000, 0.0000, 0.0091, 0.0611,
0.97, 117.0, 130.0, 3.50, 4.00, 1.0000, 0.0000, 0.0091, 0.0611,
0.97, 117.0, 130.0, 4.00, 4.50, 1.0000, 0.0000, 0.0091, 0.0611,
0.97, 117.0, 130.0, 4.50, 5.00, 1.0000, 0.0000, 0.0091, 0.0611,
0.97, 130.0, 145.0, 0.00, 0.50, 1.0000, 0.0000, 0.0178, 0.0611,
0.97, 130.0, 145.0, 0.50, 1.00, 0.7376, 0.0073, 0.0178, 0.0611,
0.97, 130.0, 145.0, 1.00, 1.50, 0.9009, 0.0027, 0.0178, 0.0611,
0.97, 130.0, 145.0, 1.50, 2.00, 0.9361, 0.0021, 0.0178, 0.0611,
0.97, 130.0, 145.0, 2.00, 2.50, 0.9142, 0.0030, 0.0178, 0.0611,
0.97, 130.0, 145.0, 2.50, 3.00, 0.9191, 0.0055, 0.0178, 0.0611,
0.97, 130.0, 145.0, 3.00, 3.50, 0.8751, 0.0281, 0.0178, 0.0611,
0.97, 130.0, 145.0, 3.50, 4.00, 1.0000, 0.0000, 0.0178, 0.0611,
0.97, 130.0, 145.0, 4.00, 4.50, 1.0000, 0.0000, 0.0178, 0.0611,
0.97, 130.0, 145.0, 4.50, 5.00, 1.0000, 0.0000, 0.0178, 0.0611,
0.97, 145.0, 150.0, 0.00, 0.50, 1.0000, 0.0000, 0.0832, 0.0611,
0.97, 145.0, 150.0, 0.50, 1.00, 1.0000, 0.0000, 0.0832, 0.0611,
0.97, 145.0, 150.0, 1.00, 1.50, 0.2187, 0.0106, 0.0832, 0.0611,
0.97, 145.0, 150.0, 1.50, 2.00, 0.4481, 0.0040, 0.0832, 0.0611,
0.97, 145.0, 150.0, 2.00, 2.50, 0.4776, 0.0032, 0.0832, 0.0611,
0.97, 145.0, 150.0, 2.50, 3.00, 0.5042, 0.0048, 0.0832, 0.0611,
0.97, 145.0, 150.0, 3.00, 3.50, 0.4886, 0.0131, 0.0832, 0.0611,
0.97, 145.0, 150.0, 3.50, 4.00, 1.0000, 0.0000, 0.0832, 0.0611,
0.97, 145.0, 150.0, 4.00, 4.50, 1.0000, 0.0000, 0.0832, 0.0611,
0.97, 145.0, 150.0, 4.50, 5.00, 1.0000, 0.0000, 0.0832, 0.0611
};
